
/*
 *  AI-Trader --- Service Trading using Conceptual Graphs
 *  Copyright (C) 1995 Arno Puder et al.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 aitrader@informatik.uni-frankfurt.de
 */

#include "cgparsetree.h"
#include "util.h"

int Template2IDL::IsBaseType( IDLString type )
{
    return (type == "boolean" ||
            type == "byte" ||
            type == "octet" ||
            type == "char" ||
            type == "error_status_t" ||
            type == "long" ||
            type == "short" ||
            type == "unsigned long" ||
            type == "unsigned short" ||
            type == "float" ||
            type == "double" ||
            type == "small" ||
            type == "hyper" ||
            type == "unsigned small" ||
            type == "unsigned hyper" ||
            type == "ISO_LATIN_1" ||
            type == "ISO_MULTI_LINGUAL" ||
            type == "ISO_UCS" ||
            type == "any" ||
            type == "Object" ||
            type == "TypeCode" ||
            type == "handle_t" ||
            type == "void" ||
            type == "void*");
}


int Template2IDL::IsBaseValue( IDLString value )
{
    return (value == "TRUE" ||
            value == "FALSE" ||
            value == "NULL");
}


int Template2IDL::DetectDeclaratorType( ConceptualGraph *cg )
{
    ConceptualGraph *rgraph;
    ConceptualGraph *cgraph;
    IDLString rname;
    IDLString cname;
    int result;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    rname = "CONSTRUCTED_TYPE";
    rgraph = GetRelation(cg, rname);
    if(rgraph == NULL)
        result = SIMPLE;
    else
    {
        cname = "POINTER";
        cgraph = GetPureConcept(rgraph, cname);
        if(cgraph != NULL)
            result = POINTER;
        else
        {
            cname = "ARRAY";
            cgraph = GetPureConcept(rgraph, cname);
            if(cgraph != NULL)
                result = ARRAY;
            else
            {   
                cname = "FUNCTION_POINTER";
                cgraph = GetPureConcept(rgraph, cname);
                if(cgraph != NULL)
                    result = FUNCTION_POINTER;
                else
                    result = SIMPLE;
            }
        }
    }
    return(result);
}


int Template2IDL::DetectIDL( ConceptualGraph *cg )
{
    ConceptualGraph *cgraph_1;
    ConceptualGraph *cgraph_2;
    ConceptualGraph *rgraph;
    ConceptNode *concept;
    IDLString cname;
    IDLString rname;
    IDLString name;
    IDLString instance;
    int result = 0;

    assert(cg->GetNode()->IsA() == RELATION_NODE);
    cgraph_1 = cg->GetWidth();
    while(cgraph_1 != NULL)
    {
        concept = (ConceptNode*) cgraph_1->GetNode();
        cname = concept->GetName();
        if(cname == "INTERFACE")
        { 
            rname = "ANNOTATIONS";
            rgraph = GetRelation(cgraph_1, rname);
            assert(rgraph != NULL);
            cgraph_2 = rgraph->GetWidth();
            while(cgraph_2 != NULL)
            {
                concept = (ConceptNode*) cgraph_2->GetNode();
                cname = concept->GetName();
                if(cname == "IDL")
                {
                    instance = "CORBA";
                    if(CompareInstances(cgraph_2, instance)) 
                        result = AIT_IDL_CORBA;
                    instance = "DCE";
                    if(CompareInstances(cgraph_2, instance))
                        result = AIT_IDL_DCE;
                }
                cgraph_2 = cgraph_2->GetDepth();
            }
        }
        cgraph_1 = cgraph_1->GetDepth();
    }
    return(result);
}


int Template2IDL::DetectStandardException( ConceptualGraph *cg )
{
    ConceptualGraph *cgraph;
    ConceptualGraph *rgraph;
    ConceptNode *concept;
    SetOfWord instance_list;
    IDLString cname;
    IDLString rname;
    IDLString instance;
    SetOfWord::iterator i;
    int detected = 0;
 
    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    rname = "ANNOTATIONS";
    rgraph = GetRelation(cg, rname);
    if(rgraph == NULL)
        assert(0);
    cgraph = rgraph->GetWidth();
    while(cgraph != NULL)
    {
        concept = (ConceptNode*) cgraph->GetNode();
        cname = concept->GetName();
        if(cname == "EXCEPTION")
        {
            instance_list = concept->GetInstances();
            i = instance_list.begin();
            instance = (*i);
            if(instance == "std")
                detected = 1;
        }
        cgraph = cgraph->GetDepth();
    }
    return(detected);
} 


void Template2IDL::CollectConstants(ConceptualGraph *cg, IDLString tag)
{
    ConceptualGraph *rgraph_1;
    ConceptualGraph *rgraph_2;
    ConceptualGraph *cgraph;
    IDLString rname;
    IDLString value;
    ConceptNode *concept;
    SetOfWord instance_list;
    SetOfWord::iterator i;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    rname = "TYPE";
    rgraph_1 = GetRelation(cg, rname);
    if(rgraph_1 == NULL) 
    {
        rname = "TEMPLATE_TYPE";
        rgraph_1 = GetRelation(cg, rname);
    }
    CollectTypes(rgraph_1, tag);
    rname = "VALUE";
    rgraph_2 = GetRelation(cg, rname);
    assert(rgraph_2 != NULL);
    cgraph = rgraph_2->GetWidth();
    concept = (ConceptNode*) cgraph->GetNode();
    instance_list = concept->GetInstances();
    i = instance_list.begin();
    value = (*i);
    if(idict.ContainsWord(tag))
    {
        if(!IsBaseValue(value))
        {
            if(idict.ContainsWord(value))
                idict.AddRelation(tag, "CONST_DEPEND", value);
            else
            {
                idict.AddWord(value);
                idict.AddRelation(tag, "SYNONYM", value);
                idict.AddRelation(value, "CONST_DEPEND", "*BASE*");
#if 0
                if(idict.AreRelated(tag, "CONST_DEPEND", "*BASE*")) 
                    idict.DeleteRelation(tag, "CONST_DEPEND", "*BASE*");
#endif
            }
        }
        else
        {
            idict.AddWord(tag);
            if(IsBaseValue(value))
                idict.AddRelation(tag, "CONST_DEPEND", "*BASE*");
            else
            {
                if(idict.ContainsWord(value))
                    idict.AddRelation(tag, "CONST_DEPEND", value);
                else 
                {		
                    idict.AddWord(value);
                    idict.AddRelation(tag, "SYNONYM", value);
                    idict.AddRelation(value, "CONST_DEPEND", "*BASE*");
                }
       	    }
        }
    }
}


void Template2IDL::CollectTypes(ConceptualGraph* cg, IDLString tag, int depth)
{
    ConceptualGraph *cgraph_1;
    ConceptualGraph *cgraph_2;
    ConceptualGraph *cgraph_3;
    ConceptualGraph *rgraph_1;
    ConceptualGraph *rgraph_2;
    ConceptualGraph *rgraph_3;
    IDLString rname_1;
    IDLString rname_2;
    IDLString cname;
    IDLString type;
    RelationNode *relation;
    ConceptNode *concept;
    SetOfWord instance_list;
    SetOfWord::iterator i;

    assert(cg->GetNode()->IsA() == RELATION_NODE);
    relation = (RelationNode*) cg->GetNode();
    rname_1 = relation->GetName();
    if(rname_1 == "TYPE" || rname_1 == "DISCRIMINATOR_TYPE") 
    {
        cgraph_1 = cg->GetWidth();
        concept = (ConceptNode*) cgraph_1->GetNode();
        instance_list = concept->GetInstances();
        i = instance_list.begin();
        type = (*i);
        if(!idict.ContainsWord(tag))
            idict.AddWord(tag);
        if(IsBaseType(type))
            idict.AddRelation(tag, "TYPE_DEPEND", "*BASE*");
        else
        {
            if(!idict.ContainsWord(type))
                idict.AddWord(type);
            idict.AddRelation(tag, "TYPE_DEPEND", type);
        }
    }
    if(rname_1 == "CONSTRUCTED_TYPE") 
    {
        cgraph_1 = cg->GetWidth();
        concept = (ConceptNode*) cgraph_1->GetNode();
        cname = concept->GetName();
        if(cname == "ENUM") 
        {
            if(depth == 0) 
            {
                idict.AddWord(tag);
                idict.AddRelation(tag, "TYPE_DEPEND", "*BASE*");
    	    }
        }
        if(cname == "PIPE") 
        {
            rname_2 = "TYPE";
            rgraph_1 = GetRelation(cgraph_1, rname_2);
            CollectTypes(rgraph_1, tag, 1);
        }
        if(cname == "STRUCT") 
        {
            rname_2 = "MEMBERS";
            rgraph_1 = GetRelation(cgraph_1, rname_2);
            if(rgraph_1 != NULL)
            {
                cgraph_2 = rgraph_1->GetWidth();
                while(cgraph_2 != NULL)
                {
                    rgraph_2 = cgraph_2->GetWidth();
                    CollectTypes(rgraph_2, tag, 1);
                    cgraph_2 = cgraph_2->GetDepth();
                }
            } 
        }
        if(cname == "UNION") 
        {
            rname_2 = "CASES";
            rgraph_1 = GetRelation(cgraph_1, rname_2);
            if(rgraph_1 != NULL)
            {
                cgraph_2 = rgraph_1->GetWidth();
                while(cgraph_2 != NULL)
                {
                    rname_2 = "MEMBERS";
                    rgraph_2 = GetRelation(cgraph_2, rname_2);
                    if(rgraph_2 != NULL)
                    {
                        cgraph_3 = rgraph_2->GetWidth();
                        while(cgraph_3 != NULL)
                        {
                            rgraph_3 = cgraph_3->GetWidth();
                            CollectTypes(rgraph_3, tag, 1);
                            cgraph_3 = cgraph_3->GetDepth();
                        }
                    }
                    cgraph_2 = cgraph_2->GetDepth();	 
                }
            } 
        }
        if(cname == "ARRAY" || cname == "FUNCTION_POINTER" ||
           cname == "POINTER")
        {
            rgraph_1 = cgraph_1->GetWidth();
            CollectTypes(rgraph_1, tag, 1);
        }
    }
    if(rname_1 == "TEMPLATE_TYPE") 
    {
        cgraph_1 = cg->GetWidth();
        while(cgraph_1 != NULL)
        { 
            concept = (ConceptNode*) cgraph_1->GetNode();
            cname = concept->GetName();
            if(cname == "ELEMENT_TYPE") 
            {
                instance_list = concept->GetInstances();
                i = instance_list.begin();
	    	 type = (*i);
	    	 if(contains(type, ">") && contains(type, "<")) 
                 {
                     rname_2 = "TEMPLATE_TYPE";
                     rgraph_1 = GetRelation(cgraph_1, rname_2);
                     CollectTypes(rgraph_1, tag);
                 }
                 else
                 {
                     if(IsBaseType(type)) 
                     {
                         if(!idict.ContainsWord(tag))
                             idict.AddWord(tag);
                         idict.AddRelation(tag, "TYPE_DEPEND", "*BASE*");
                     }
                     else
                     {
                         if(!idict.ContainsWord(type))
                             idict.AddWord(type);
	            	 idict.AddWord(tag);
	            	 idict.AddRelation(tag, "TYPE_DEPEND", type);
                     }
                 } 
            }
            if(cname == "TEMPLATE_BOUND")
            {
                instance_list = concept->GetInstances();
                i = instance_list.begin();
                type = (*i);
                idict.AddWord(tag);
                if(!idict.ContainsWord(type))
                    idict.AddWord(type);
                idict.AddRelation(tag, "CONST_DEPEND", type);
            }	 
            cgraph_1 = cgraph_1->GetDepth();
        }
    }
    if(rname_1 == "RECURSIVE_TYPE") 
    {
        /* XXX empty */
    }
}


Template2IDL::Template2IDL(ConceptualGraph* rg)
{
    root = rg;
    idict.RegisterRelation("METHOD_OF",
                           Dict::RelReflexive |
                           Dict::RelTransitive |
                           Dict::RelAcyclic);
    idict.RegisterRelation("MEMBER_OF",
                           Dict::RelReflexive |
                           Dict::RelTransitive |
                           Dict::RelAcyclic);
    idict.RegisterRelation("CONST_DEPEND",
                           Dict::RelReflexive |
                           Dict::RelTransitive |
                           Dict::RelAcyclic);
    idict.RegisterRelation("TYPE_DEPEND",
                           Dict::RelReflexive |
                           Dict::RelTransitive |
                           Dict::RelAcyclic);
    idict.RegisterRelation("SYNONYM",
                           Dict::RelReflexive |
                           Dict::RelTransitive |
                           Dict::RelSymmetric);
    idict.AddWord("*BASE*");
    odict.AddWord("*BASE*");
}


IDLString Template2IDL::ClearScopedName(IDLString name, IDLString scope)
{
    IDLString tname;
    IDLString tscope;
    string tmp;

    tscope = scope;
    tname = name;
    if(!(tscope == ""))
    {
        while(contains(tscope, "::"))
        {
            tmp = before(tscope, "::");
            tmp += "::";
            if(contains(tname, tmp))
                del(tname, tmp);
            del(tscope, tmp);
        }
    }
    return(tname);
}

/********************** CORBA - DEFINITIONEN **********************/


void Template2CORBA::CreateSpecification()
{
    ConceptualGraph *rgraph;
    ConceptualGraph *cgraph;
    IDLString cname;
    IDLString rname;
    IDLString source_name;
    IDLString source_form;
    IDLString scope;
    ConceptNode *concept;

    scope = "";
    rname = "DEFINITIONS";
    rgraph = GetRelation(root, rname);
    assert(rgraph != NULL);
    mode = DetectIDL(rgraph);
    assert(mode);
    switch(mode)
    {
    case AIT_IDL_DCE:
        CollectDeclarations(); 
        cgraph = rgraph->GetWidth();
        while(cgraph != NULL)
        {
            concept = (ConceptNode*) cgraph->GetNode();
            cname = concept->GetName();
            assert(cname == "INTERFACE");
            CreateInterface(cgraph, scope);
            cgraph = cgraph->GetDepth();
        }
        break;
    case AIT_IDL_CORBA:
        source_name = "*BASE*";
        source_form = "SPECIFICATION";
        CollectDeclarations();
        rname = "ANNOTATIONS";
        rgraph = GetRelation(root, rname);
        assert(rgraph != NULL);
        cname = "HIERARCHY";
        cgraph = GetPureConcept(rgraph, cname);
        assert(cgraph != NULL);
        CreateDeclarations(source_name, source_form, scope);
        rname = "MEMBERS";
        rgraph = GetRelation(cgraph, rname);
        assert(rgraph != NULL); 
        RangeMembers(rgraph, source_name, source_form, scope);
        break;
    }
}


void Template2CORBA::CollectDeclarations()
{
    ConceptualGraph *cgraph;
    ConceptualGraph *rgraph;
    ConceptNode *concept;
    IDLString cname;
    IDLString rname;

    rname = "DEFINITIONS";
    rgraph = GetRelation(root, rname);
    assert(rgraph != NULL);
    cgraph = rgraph->GetWidth();
    while(cgraph != NULL)
    {
        concept = (ConceptNode*) cgraph->GetNode();
        cname = concept->GetName();
        assert(cname == "INTERFACE");
        rname = "TYPES";
        rgraph = GetRelation(cgraph, rname);
        if(rgraph != NULL)
            ExamineDeclarations(rgraph->GetWidth());
        rname = "CONSTANTS";
        rgraph = GetRelation(cgraph, rname);
        if(rgraph != NULL)
            ExamineDeclarations(rgraph->GetWidth());
        rname = "OPERATIONS";
        rgraph = GetRelation(cgraph, rname);
        if(rgraph != NULL)
            ExamineDeclarations(rgraph->GetWidth());
        cgraph = cgraph->GetDepth();
    }
}


void Template2CORBA::ExamineDeclarations( ConceptualGraph *cg )
{
    ConceptualGraph *cgraph_1;
    ConceptualGraph *cgraph_2;
    ConceptualGraph *rgraph;
    ConceptNode *concept;
    SetOfWord instance_list;
    IDLString rname;
    IDLString sname;
    IDLString cname_1;
    IDLString cname_2;
    IDLString instance;
    IDLString identifier;
    SetOfWord::iterator i;
    int result = 0;
 
    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    cgraph_1 = cg;
    while(cgraph_1 != NULL)
    {
        rname = "ANNOTATIONS";
        rgraph = GetRelation(cgraph_1, rname);
        assert(rgraph != NULL);
        concept = (ConceptNode*) cgraph_1->GetNode();
        cname_1 = concept->GetName();
        instance_list = concept->GetInstances();
        i = instance_list.begin();
        if(i != instance_list.end())
            identifier = (*i);
        else
	{
            cname_2 = "IDENTIFIER";
            cgraph_2 = GetPureConcept(rgraph, cname_2); 
            assert(cgraph_2 != NULL);
            concept = (ConceptNode*) cgraph_2->GetNode();
            instance_list = concept->GetInstances();
            i = instance_list.begin();
            assert(i != instance_list.end());
            identifier = (*i);
	}
        cname_2 = "SOURCE_NAME";
        cgraph_2 = GetPureConcept(rgraph, cname_2);
        assert(cgraph_2 != NULL);
        concept = (ConceptNode*) cgraph_2->GetNode();
        instance_list = concept->GetInstances();
        i = instance_list.begin();
        assert(i != instance_list.end());
        sname = (*i);
        if(!idict.ContainsWord(identifier))
            idict.AddWord(identifier);
        if(!idict.ContainsWord(sname))
            idict.AddWord(sname);
        if(cname_1 == "OPERATION")
            assert(idict.AddRelation(identifier, "METHOD_OF", sname));	
        else
            assert(idict.AddRelation(identifier, "MEMBER_OF", sname));
        cgraph_2 = rgraph->GetWidth();
        while(cgraph_2 != NULL)
	{
            concept = (ConceptNode*) cgraph_2->GetNode();
            cname_2 = concept->GetName();
            if(cname_2 == "TYPE")
	    {
                instance_list = concept->GetInstances();
                i = instance_list.begin();
                assert(i != instance_list.end());
                instance = (*i);
                rgraph = cgraph_2->GetWidth();
                CollectTypes(rgraph, instance);
                ExamineDeclarations(cgraph_2);
	    }
            if(cname_2 == "CONSTANT")
	    {
                instance_list = concept->GetInstances();
                i = instance_list.begin();
                assert(i != instance_list.end());
                instance = (*i);
                CollectConstants(cgraph_2, instance);
                ExamineDeclarations(cgraph_2);
	    }
            if(cname_2 == "OPERATION")
                ExamineDeclarations(cgraph_2);
            cgraph_2 = cgraph_2->GetDepth();
	}
        cgraph_1 = cgraph_1->GetDepth();
    }
}


int Template2CORBA::CheckDeclaration( ConceptualGraph *cg, IDLString sname,
                                      IDLString sform, IDLString scope )
{
    ConceptualGraph *rgraph;
    IDLString name;
    IDLString rname;
    IDLString cname_1;
    IDLString cname_2;
    IDLString cname_3;
    int result = 0;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    rname = "ANNOTATIONS";
    rgraph = GetRelation(cg, rname);
    assert(rgraph != NULL);
    cname_1 = "SOURCE_NAME";
    cname_2 = "SOURCE_FORM";
    cname_3 = "SCOPE";
    name = ClearScopedName(sname, scope);
    if(scope == "") 
    {
        if(DetectConcept(rgraph, cname_1, name) &&
           DetectConcept(rgraph, cname_2, sform))
            result = 1;
    }
    else if(DetectConcept(rgraph, cname_3, scope) && 
            DetectConcept(rgraph, cname_1, name) &&
            DetectConcept(rgraph, cname_2, sform))
        result = 1;
    return(result);
}


void Template2CORBA::RangeMembers( ConceptualGraph *cg, IDLString sname, 
                                   IDLString sform, IDLString scope )
{
    ConceptualGraph *rgraph;
    ConceptualGraph *cgraph;
    ConceptNode *concept;
    SetOfWord instance_list;
    IDLString name;
    IDLString cname;
    IDLString instance;
    int creation = 1;
    SetOfWord::iterator i;
 
    assert(cg->GetNode()->IsA() == RELATION_NODE);
    cgraph = cg->GetWidth();
    while(cgraph != NULL)
    {
        concept = (ConceptNode*) cgraph->GetNode();
        instance_list = concept->GetInstances();
        i = instance_list.begin();
        assert(i != instance_list.end()); 
        instance = (*i);
        if(!odict.ContainsWord(instance))
        {
            creation = CheckPredecessors(cgraph);
            if(creation)
            {
                cname = concept->GetName();
                if(cname == "MODULE")
                    CreateModule(cgraph, scope);
                if(cname == "INTERFACE")
                    CreateInterface(cgraph, scope);
                CreateDeclarations(sname, sform, scope);
                odict.AddWord(instance);
                RangeMembers(cg, sname, sform, scope);
            }
        }
        cgraph = cgraph->GetDepth();
    }
}


int Template2CORBA::CheckPredecessors( ConceptualGraph *cg )
{
    ConceptualGraph *cgraph;
    ConceptualGraph *rgraph;
    ConceptNode *concept;
    SetOfWord instance_list;
    IDLString instance;
    IDLString rname;
    int result = 1;
    SetOfWord::iterator i;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    rname = "PREDECESSORS";
    rgraph = GetRelation(cg, rname);
    if(rgraph != NULL)
    {
        cgraph = rgraph->GetWidth();
        while(cgraph != NULL)
        {
            concept = (ConceptNode*) cgraph->GetNode();
            instance_list = concept->GetInstances();
            i = instance_list.begin();
     	 assert(i != instance_list.end()); 
     	 instance = (*i);
     	 if(!odict.ContainsWord(instance))
             result = 0;
         cgraph = cgraph->GetDepth();
        }
    }
    return(result); 
}


void Template2CORBA::CreateInterface( ConceptualGraph *cg, IDLString scope )
{
    ConceptualGraph *rgraph; 
    ConceptualGraph *cgraph; 
    ConceptualGraph *ograph; 
    ConceptNode *concept;
    SetOfWord instance_list;
    IDLString name;
    IDLString tname;
    IDLString sname;
    IDLString scoped_name;
    IDLString rname;
    IDLString cname;
    IDLString tscope;
    IDLString interface_list;
    IDLString interface;
    SetOfWord::iterator i;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    concept = (ConceptNode*) cg->GetNode();
    instance_list = concept->GetInstances();
    i = instance_list.begin();
    assert(i != instance_list.end());
    scoped_name = (*i);
    name = ClearScopedName(scoped_name, scope);
    tscope = scope + name + "::";
    rname = "DEPENDANTS";
    rgraph = GetRelation(cg, rname);
    if(rgraph != NULL) 
    {
        cgraph = rgraph->GetWidth();
        while(cgraph != NULL)
    	{
            concept = (ConceptNode*) cgraph->GetNode();
            instance_list = concept->GetInstances();
            i = instance_list.begin();
            assert(i != instance_list.end());
            sname = (*i);
            if(!odict.ContainsWord(sname))
	    {
                tname = ClearScopedName(sname, scope);
                corbaout->EnterNewLevel(forward_level);
                corbaout->PrintIdentifier(tname);
                corbaout->LeaveLevel(forward_level);
	    }
            cgraph = cgraph->GetDepth(); 
    	}
    }
    corbaout->EnterNewLevel(interface_level);
    corbaout->PrintIdentifier(name);
    rname = "SUPERINTERFACES";
    rgraph = GetRelation(cg, rname);
    if(rgraph != NULL)
    {
        corbaout->EnterNewLevel(inheritance_level);
        interface_list = "";
        cgraph = rgraph->GetWidth();
        while(cgraph != NULL)
	{
            concept = (ConceptNode*) cgraph->GetNode();
            instance_list = concept->GetInstances();
            i = instance_list.begin();
            sname = (*i);
            interface = ClearScopedName(sname, scope);
            if(!(interface_list == ""))
                interface_list += ", ";
            interface_list += interface;
            cgraph = cgraph->GetDepth();
	}
        corbaout->PrintIdentifier(interface_list);
        corbaout->LeaveLevel(inheritance_level);
    }
    if(mode == AIT_IDL_CORBA)
    {
        cname = "INTERFACE";
        CreateDeclarations(name, cname, tscope);
        CreateOperations(name, cname, tscope, 1);
        opsdone.erase (opsdone.begin(), opsdone.end());
    }
    else 
    {
        CreateDeclarations(cg, tscope);
    } 
    corbaout->LeaveLevel(interface_level);
    odict.AddWord(scoped_name);
}
    

void Template2CORBA::CreateModule( ConceptualGraph *cg, IDLString scope )
{
    ConceptualGraph *rgraph;
    ConceptNode *concept;
    SetOfWord instance_list;
    IDLString tscope;
    IDLString cname;
    IDLString name;
    IDLString rname;
    IDLString scoped_name;
    SetOfWord::iterator i;
 
    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    concept = (ConceptNode*) cg->GetNode();
    instance_list = concept->GetInstances();
    i = instance_list.begin();
    assert(i != instance_list.end());
    scoped_name = (*i);
    name = ClearScopedName(scoped_name, scope);
    tscope = scope + name + "::";
    corbaout->EnterNewLevel(module_level);
    corbaout->PrintIdentifier(name);
    cname = "MODULE";
    CreateDeclarations(scoped_name, cname, tscope);
    rname = "MEMBERS";
    rgraph = GetRelation(cg, rname);
    assert(rgraph != NULL);
    RangeMembers(rgraph, scoped_name, cname, tscope); 
    corbaout->LeaveLevel(module_level);
}


void Template2CORBA::CreateOperations( IDLString sname, IDLString sform,
                                       IDLString scope, int pos )
{
    ConceptualGraph *cgraph_1;
    ConceptualGraph *cgraph_2;
    ConceptualGraph *rgraph;
    ConceptNode *concept;
    IDLString cname;
    IDLString rname;
    IDLString operation;
    int position;
    Dict::vcwp::const_iterator i;

    Dict::vcwp operation_list = idict.GetRelImmPred(sname, "METHOD_OF");
    i = operation_list.begin();
    while(i != operation_list.end())
    {
        operation = **i;
        if(opsdone.find (operation) == opsdone.end())
        {
            cgraph_1 = GetDeclarationPosition(operation, sname, sform, scope);
            assert(cgraph_1 != NULL);
            rname = "ANNOTATIONS";
            rgraph = GetRelation(cgraph_1, rname);
            assert(rgraph != NULL);
            cname = "OPERATION_POSITION";
            cgraph_2 = GetPureConcept(rgraph, cname);
            if(cgraph_2 == NULL)
            {
                rname = "POSITION";
                rgraph = GetRelation(cgraph_1, rname);
                assert(rgraph != NULL);
                cgraph_2 = rgraph->GetWidth();
            } 
            concept = (ConceptNode*) cgraph_2->GetNode();
            position = concept->GetValue();
            if(position == pos) 
            {
                rname = "ANNOTATIONS";
                rgraph = GetRelation(cgraph_1, rname);
                assert(rgraph != NULL);
                cname = "ATTRIBUTE";
                cgraph_2 = GetPureConcept(rgraph, cname);
                if(cgraph_2 == NULL)
                    CreateOperation(cgraph_1, scope);
                else
                    CreateAttribute(cgraph_1, cgraph_2, scope);
                opsdone.insert (operation);
                pos++;
                CreateOperations(sname, sform, scope, pos);
            }
        }
        ++i;
    }
}


void Template2CORBA::CreateDeclarations( IDLString sname, IDLString sform,
                                         IDLString scope )
{
    ConceptualGraph *cgraph;
    IDLString name;
    IDLString member;
    IDLString type;
    IDLString constant;
    IDLString synonym;
    Dict::vcwp::const_iterator i;
    Dict::vcwp::const_iterator j;
    int creation = 1;

    Dict::vcwp member_list = idict.GetRelImmPred(sname, "MEMBER_OF");
    i = member_list.begin();
    while(i != member_list.end())
    {
        member = **i;
        if(!odict.ContainsWord(member))
        {
            Dict::vcwp type_list = idict.GetRelImmSucc(member, "TYPE_DEPEND");
            j = type_list.begin();
            while(j != type_list.end())
            {
                type = **j;
                if(!odict.ContainsWord(type))
                    creation = 0;
                ++j;
            }
            Dict::vcwp constant_list = idict.GetRelImmSucc(member,
                                                           "CONST_DEPEND");
            j = constant_list.begin();
            while(j != constant_list.end())
            {
                constant = **j;
                if(!odict.ContainsWord(constant))
                    creation = 0;
                ++j;
            }
            if(creation) 
            {
                cgraph = GetDeclarationPosition(member, sname, sform, scope);
                assert(cgraph != NULL);
                CreateDeclaration(cgraph, member, scope);
                odict.AddWord(member);
            }
        }
        ++i;
    }
}


ConceptualGraph*
Template2CORBA::GetDeclarationPosition( IDLString declarator, 
                                        IDLString sname,
                                        IDLString sform, 
                                        IDLString scope )
{
    ConceptualGraph *cgraph;
    ConceptualGraph *rgraph;
    ConceptualGraph *pgraph;
    ConceptNode *concept;
    IDLString cname;
    IDLString rname;
    int detected = 0;

    rname = "DEFINITIONS";
    rgraph = GetRelation(root, rname);
    assert(rgraph != NULL);
    cgraph = rgraph->GetWidth();
    pgraph = NULL;
    while(cgraph != NULL && !detected)
    {
        concept = (ConceptNode*) cgraph->GetNode();
        cname = concept->GetName();
        assert(cname == "INTERFACE");
        rname = "TYPES";
        rgraph = GetRelation(cgraph, rname);
        if(rgraph != NULL)
            pgraph = GetPosition(rgraph->GetWidth(), declarator, sname,
                                 sform, scope);
        if(pgraph != NULL)
            detected = 1;
        else
        {
            rname = "CONSTANTS";
            rgraph = GetRelation(cgraph, rname);
            if(rgraph != NULL)
                pgraph = GetPosition(rgraph->GetWidth(), declarator, sname,
                                     sform, scope);
            if(pgraph != NULL)
                detected = 1;
            else
            {
                rname = "OPERATIONS";
                rgraph = GetRelation(cgraph, rname);
                if(rgraph != NULL)
                    pgraph = GetPosition(rgraph->GetWidth(), declarator,
                                         sname, sform, scope);
                if(pgraph != NULL)
                    detected = 1;
            }
        }
        cgraph = cgraph->GetDepth();
    }
    return(pgraph);
}


ConceptualGraph* Template2CORBA::GetPosition( ConceptualGraph *cg, 
                                              IDLString declarator,
                                              IDLString sname,
                                              IDLString sform,
                                              IDLString scope )
{
    ConceptualGraph *rgraph;
    ConceptualGraph *cgraph_1;
    ConceptualGraph *cgraph_2;
    ConceptualGraph *pgraph;
    ConceptNode *concept;
    IDLString cname;
    IDLString rname;
    IDLString instance;
    IDLString identifier;
    SetOfWord instance_list;
    SetOfWord::iterator i;
    int result = 0;
    int detected = 0;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    cgraph_1 = cg;
    pgraph = NULL;
    while(cgraph_1 != NULL && !detected)
    {
        rname = "ANNOTATIONS";
        rgraph = GetRelation(cgraph_1, rname);
        assert(rgraph != NULL);
        concept = (ConceptNode*) cgraph_1->GetNode();
        instance_list = concept->GetInstances();
        i = instance_list.begin();
        if(i != instance_list.end())
            identifier = (*i);
        else
        {
            cname = "IDENTIFIER";
            cgraph_2 = GetPureConcept(rgraph, cname);
            assert(cgraph_2 != NULL);
            concept = (ConceptNode*) cgraph_2->GetNode();
            instance_list = concept->GetInstances();
            i = instance_list.begin();
            assert(i != instance_list.end());
            identifier = (*i);
        }
        cgraph_2 = rgraph->GetWidth();
        while(cgraph_2 != NULL)
        {
            concept = (ConceptNode*) cgraph_2->GetNode();
            cname = concept->GetName();
            if(cname == "TYPE" || cname == "CONSTANT" || cname == "OPERATION")
            {
                instance_list = concept->GetInstances();
                i = instance_list.begin();
                assert(i != instance_list.end());
                instance = (*i);
                result = CheckDeclaration(cgraph_2, sname, sform, scope);
                if(result && declarator == instance)
                {
                    pgraph = cgraph_2;
                    detected = 1;
                }
            }
            cgraph_2 = cgraph_2->GetDepth();
        }
        result = CheckDeclaration(cgraph_1, sname, sform, scope);
        if(result && declarator == identifier) 
        {
            pgraph = cgraph_1;
            detected = 1;
        }
        cgraph_1 = cgraph_1->GetDepth();
    }
    return(pgraph);
}


void
Template2CORBA::CreateDeclaration( ConceptualGraph *cg, IDLString declarator,
                                   IDLString scope )
{
    ConceptualGraph *rgraph;
    ConceptualGraph *cgraph;
    ConceptNode *concept;
    SetOfWord instance_list;
    IDLString instance;
    IDLString rname;
    IDLString cname;
    SetOfWord::iterator i;
    int creation = 1;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    concept = (ConceptNode*) cg->GetNode();
    cname = concept->GetName();
    if(cname == "TYPE")
    { 
        rname = "ANNOTATIONS";
        rgraph = GetRelation(cg, rname);
        assert(rgraph != NULL);
        cgraph = rgraph->GetWidth();
        while(cgraph != NULL)
    	{
            concept = (ConceptNode*) cgraph->GetNode();
            cname = concept->GetName();
            if(cname == "ALIAS" || cname == "RAISES_EXPRESSION")
                creation = 0;
            if(cname == "EXCEPTION")
            {
                instance_list = concept->GetInstances();
                i = instance_list.begin();
                assert(i != instance_list.end());
                instance = (*i);
                if(!(string(instance) == string("std")))
                    CreateException(cg, scope, declarator);
                creation = 0;
            } 
            cgraph = cgraph->GetDepth();
    	}
        if(creation)
            CreateTypeDeclaration(cg, scope, declarator);
    }
    else
        CreateConstantDeclaration(cg, scope, declarator);
}


void Template2CORBA::CreateDeclarations( ConceptualGraph *cg, IDLString scope )
{
    ConceptualGraph *cgraph_1;
    ConceptualGraph *cgraph_2;
    ConceptualGraph *cgraph_3;
    ConceptualGraph *rgraph;
    ConceptualGraph *rgraph_2;
    ConceptNode *concept;
    SetOfWord instance_list;
    SetOfWord::iterator i, j;
    IDLString rname;
    IDLString cname;
    IDLString tname;
    int counter = 0;
    int pos;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    rname = "CONSTANTS";
    rgraph = GetRelation(cg, rname);
    if(rgraph != NULL)
        cgraph_2 = rgraph->GetWidth();
    else
        cgraph_2 = NULL;
    rname = "TYPES";
    rgraph = GetRelation(cg, rname);
    if(rgraph != NULL)
        cgraph_1 = rgraph->GetWidth();
    else
        cgraph_1 = NULL;
    while(cgraph_1 != NULL)
    {
        concept = (ConceptNode*) cgraph_1->GetNode();
        instance_list = concept->GetInstances();
        i = instance_list.begin();
        if(i == instance_list.end())
	{
            rname = "ANNOTATIONS";
            rgraph = GetRelation(cgraph_1, rname);
            assert(rgraph != NULL);
            cgraph_3 = rgraph->GetWidth();
            while(cgraph_3 != NULL)
	    {
                concept = (ConceptNode*) cgraph_3->GetNode();
                cname = concept->GetName();
                if(cname == "IDENTIFIER")
		{
                    instance_list = concept->GetInstances();
                    j = instance_list.begin();
                    tname = (*j);
		}
                cgraph_3 = cgraph_3->GetDepth();
	    }
	}
        else
            tname = (*i);
        RangeTypeDeclarations(cgraph_1, cgraph_2, scope, tname);
        cgraph_1 = cgraph_1->GetDepth();
    }
    while(cgraph_2 != NULL)
    {
        concept = (ConceptNode*) cgraph_2->GetNode();
        instance_list = concept->GetInstances();
        i = instance_list.begin();
        tname = (*i);
        RangeConstantDeclarations(cgraph_1, cgraph_2, scope, tname);
        cgraph_2 = cgraph_2->GetDepth();
    }
    rname = "OPERATIONS";
    rgraph = GetRelation(cg, rname);
    assert(rgraph != NULL);
    cgraph_1 = rgraph->GetWidth();
    while(cgraph_1 != NULL)
    {
        counter++;
        cgraph_1 = cgraph_1->GetDepth();
    }
    for(int k = 1; k <= counter; k++)
    {
        cgraph_1 = rgraph->GetWidth();
        while(cgraph_1 != NULL)
 	{
            rname = "POSITION";
            rgraph_2 = GetRelation(cgraph_1, rname);
            assert(rgraph_2 != NULL);
            cgraph_2 = rgraph_2->GetWidth();
            concept = (ConceptNode*) cgraph_2->GetNode();
            pos = concept->GetValue();
            if(pos == k)
                CreateOperation(cgraph_1, scope);
            cgraph_1 = cgraph_1->GetDepth();
	}   
    } 
}


void
Template2CORBA::RangeTypeDeclarations( ConceptualGraph *cg_1,
                                       ConceptualGraph *cg_2, 
                                       IDLString scope, IDLString type )
{
    IDLString tname;
    Dict::vcwp::const_iterator i;

    Dict::vcwp type_list = idict.GetRelImmSucc(type, "TYPE_DEPEND");
    i = type_list.begin();
    while(i != type_list.end())
    {
        tname = **i;
        if(!odict.ContainsWord(tname))
            RangeTypeDeclarations(cg_1, cg_2, scope, tname);
        ++i;
    }
    Dict::vcwp const_list = idict.GetRelImmSucc(type, "CONST_DEPEND");
    i = const_list.begin();
    while(i != type_list.end())
    {
        tname = **i;
        if(!odict.ContainsWord(tname))
            RangeConstantDeclarations(cg_1, cg_2, scope, tname);
        ++i;
    } 
    if(!odict.ContainsWord(type))
    {
        CreateTypeDeclaration(cg_1, scope, type);
        odict.AddWord(type);
    }
}


void
Template2CORBA::RangeConstantDeclarations( ConceptualGraph *cg_1,
                                           ConceptualGraph *cg_2, 
                                           IDLString scope, IDLString type )
{
    IDLString cname;
    Dict::vcwp::const_iterator i;

    Dict::vcwp const1_list = idict.GetRelImmSucc(type, "CONST_DEPEND");
    i = const1_list.begin();
    if(i == const1_list.end())
    {
        Dict::vcwp const2_list = idict.GetRelImmSucc(type, "SYNONYM");
        i = const2_list.begin();
        while(i != const2_list.end())
    	{
            cname = **i;
            if(!odict.ContainsWord(cname))
                RangeConstantDeclarations(cg_1, cg_2, scope, cname);
            ++i;
    	}
    }
    else
    { 
        while(i != const1_list.end())
    	{
            cname = **i;
            if(!odict.ContainsWord(cname))
                RangeConstantDeclarations(cg_1, cg_2, scope, cname);
            ++i;
    	}
    }
    Dict::vcwp type_list = idict.GetRelImmSucc(type, "TYPE_DEPEND");
    i = type_list.begin();
    while(i != type_list.end())
    {
        cname = **i;
        if(!odict.ContainsWord(cname))
            RangeTypeDeclarations(cg_1, cg_2, scope, cname);
        ++i;
    } 
    if(!odict.ContainsWord(type))
    {
        CreateConstantDeclaration(cg_2, scope, type);
        odict.AddWord(type);
    }
}


void
Template2CORBA::CreateAttribute( ConceptualGraph *cg_1, ConceptualGraph *cg_2,
                                 IDLString scope )
{
    ConceptualGraph *rgraph;
    ConceptNode *concept;
    SetOfWord instance_list;
    IDLString instance;
    IDLString name;
    IDLString rname;
    IDLString attribute;
    SetOfWord::iterator i;

    assert(cg_1->GetNode()->IsA() == CONCEPT_NODE);
    assert(cg_2->GetNode()->IsA() == CONCEPT_NODE);
    concept = (ConceptNode*) cg_2->GetNode();
    instance_list = concept->GetInstances();
    i = instance_list.begin();
    assert(i != instance_list.end());
    name = (*i);
    concept = (ConceptNode*) cg_1->GetNode();
    instance_list = concept->GetInstances();
    i = instance_list.begin();
    assert(i != instance_list.end());
    instance = (*i);
    del(instance, "_get_");
    if(instance == name)
    {
        corbaout->EnterNewLevel(attribute_level);
        rname = "FEATURES";
        rgraph = GetRelation(cg_2, rname);
        if(rgraph != NULL)
    	{
            attribute = "readonly";
            corbaout->PrintAttribute(attribute);
	}
        CreateTypes(cg_1, scope, name);
        corbaout->PrintIdentifier(name);
        corbaout->LeaveLevel(attribute_level);
    }
}


void Template2CORBA::CheckForRaisesExpression( ConceptualGraph *cg,
                                               IDLString name )
{
    ConceptualGraph *rgraph;
    ConceptualGraph *cgraph;
    ConceptNode *concept;
    SetOfWord instance_list;
    IDLString cname;
    IDLString rname;
    IDLString sname;
    IDLString sform;
    IDLString scope;
    IDLString declarator;
    SetOfWord::iterator i;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    scope = "";
    sname = "";
    sform = "";
    rname = "ANNOTATIONS";
    rgraph = GetRelation(cg, rname);
    assert(rgraph != NULL);
    cgraph = rgraph->GetWidth();
    while(cgraph != NULL)
    {
        concept = (ConceptNode*) cgraph->GetNode();
        cname = concept->GetName();
        if(cname == "SOURCE_NAME" || cname == "SCOPE" ||
           cname == "SOURCE_FORM")
        { 
            instance_list = concept->GetInstances();
            i = instance_list.begin();
            assert(i != instance_list.end());
            if(cname == "SOURCE_NAME")
                sname = (*i);
            if(cname == "SOURCE_FORM")
                sform = (*i);
            if(cname == "SCOPE")
                scope = (*i);
	}
        cgraph = cgraph->GetDepth();
    }
    assert(!(scope == "") && !(sname == "") && !(sform == ""));
    declarator = name;
    declarator += "_Except";
    cgraph = GetDeclarationPosition(declarator, sname, sform, scope);
    if(cgraph != NULL)
        CreateRaisesExpression(cgraph, name); 
}


void Template2CORBA::CheckForContextExpression( ConceptualGraph *cg )
{
    ConceptualGraph *rgraph;
    ConceptualGraph *cgraph;
    ConceptNode *concept;
    IDLString rname;
    IDLString cname;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    rname = "ANNOTATIONS";
    rgraph = GetRelation(cg, rname);
    assert(rgraph != NULL);
    cgraph = rgraph->GetWidth();
    while(cgraph != NULL)
    {
        concept = (ConceptNode*) cgraph->GetNode();
        cname = concept->GetName();
        if(cname == "CONTEXT")
            CreateContextExpression(cgraph);
        cgraph = cgraph->GetDepth();
    }
}


void Template2CORBA::CreateOperation( ConceptualGraph *cg, IDLString scope )
{
    ConceptualGraph *graph;
    ConceptualGraph *hgraph;
    ConceptualGraph *sgraph;
    ConceptualGraph *tgraph;
    ConceptualGraph *rgraph;
    ConceptualGraph *cgraph;
    ConceptNode *concept;
    RelationNode *relation;
    SetOfWord instance_list;
    SetOfWord::iterator i;
    IDLString name;
    IDLString scoped_name;
    IDLString attribute;
    IDLString type;
    IDLString rname;
    int attribute_detected = 0;
    int outparameter_detected;
    IDLString tscope;
    int k;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    corbaout->EnterNewLevel(operation_level);
    concept = (ConceptNode*) cg->GetNode();
    instance_list = concept->GetInstances(); 
    i = instance_list.begin();
    name = (*i);
    tscope = scope + name + "::";
    rname = "SIGNATURE";
    rgraph = GetRelation(cg, rname);
    assert(rgraph != NULL);
    sgraph = rgraph;
    outparameter_detected = LookForOutputParameters(sgraph);
    rname = "ATTRIBUTES";
    rgraph = GetRelation(cg, rname);
    if(rgraph != NULL)
    {
        cgraph = rgraph->GetWidth();
        concept = (ConceptNode*) cgraph->GetNode();
        instance_list = concept->GetInstances();
        i = instance_list.begin();
        while( i != instance_list.end())
        {
            attribute = (*i);
            if(attribute == "maybe")
                attribute_detected = 1;
            ++i;
        }
    }
    rname = "TYPE";
    rgraph = GetRelation(cg, rname);
    if(rgraph != NULL)
    {
        cgraph = rgraph->GetWidth();
        concept = (ConceptNode*) cgraph->GetNode();
        instance_list = concept->GetInstances();
        i = instance_list.begin();
        type = (*i);
        if(type == "void") 
        {
            if(attribute_detected && outparameter_detected)
            {
                attribute = "oneway";
                corbaout->PrintAttribute(attribute);
            }
        }
        corbaout->PrintType(type);
    }
    else
    {
        rname = "TEMPLATE_TYPE";
        rgraph = GetRelation(cg, rname);
        assert(rgraph != NULL);
        CreateTypes(cg, tscope, name);
    }
    corbaout->PrintIdentifier(name);
    CreateSignature(sgraph, scope);
    if(mode == AIT_IDL_CORBA)
    {
        CheckForRaisesExpression(cg, name);
        CheckForContextExpression(cg);
    }
    corbaout->LeaveLevel(operation_level);
}


int Template2CORBA::LookForOutputParameters( ConceptualGraph *cg )
{
    ConceptualGraph *cgraph_1;
    ConceptualGraph *cgraph_2;
    ConceptualGraph *rgraph;
    ConceptNode *concept;
    SetOfWord instance_list;
    SetOfWord::iterator i;
    IDLString rname;
    IDLString attribute;
    int result = 0;

    assert(cg->GetNode()->IsA() == RELATION_NODE);
    cgraph_1 = cg->GetWidth();
    rname = "ATTRIBUTES";
    while(cgraph_1 != NULL)
    {
        rgraph = GetRelation(cgraph_1, rname);
        assert (rgraph != NULL);
        cgraph_2 = rgraph->GetWidth();
        concept = (ConceptNode*) cgraph_2->GetNode();
        instance_list = concept->GetInstances();
        i = instance_list.begin();
        while(i != instance_list.end())
	{
            attribute = (*i);
            if(attribute == "out" || attribute == "inout")
                result = 1;
            ++i;
	}
        cgraph_1 = cgraph_1->GetDepth();
    }
    return(result);
}


void Template2CORBA::CreateRaisesExpression( ConceptualGraph *cg,
                                             IDLString name ) 
{
    ConceptualGraph *rgraph_1;
    ConceptualGraph *rgraph_2;
    ConceptualGraph *rgraph_3;
    ConceptualGraph *cgraph_1;
    ConceptualGraph *cgraph_2;
    ConceptualGraph *cgraph_3;
    ConceptNode *concept;
    SetOfWord instance_list;
    SetOfWord::iterator i;
    IDLString exception;
    IDLString exception_list;
    IDLString rname;
    IDLString cname;
    int j;
    int counter = 0;
    IDLString label;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    rname = "CONSTRUCTED_TYPE";
    rgraph_2 = GetRelation(cg, rname);
    assert(rgraph_2 != NULL);
    cgraph_2 = rgraph_2->GetWidth();
    concept = (ConceptNode*) cgraph_2->GetNode();
    cname = concept->GetName();
    assert(cname == "UNION");
    rname = "CASES";
    rgraph_1 = GetRelation(cgraph_2, rname);
    assert(rgraph_1 != NULL);
    cgraph_1 = rgraph_1->GetWidth();
    while(cgraph_1 != NULL)
    {
        counter++;
        cgraph_1 = cgraph_1->GetDepth();
    }
    exception_list = "";
    for(j = 2; j <= counter; j++)
    {
        cgraph_1 = rgraph_1->GetWidth();
        while(cgraph_1 != NULL)
	{
            concept = (ConceptNode*) cgraph_1->GetNode();
            instance_list = concept->GetInstances();
            i = instance_list.begin();
            assert(i != instance_list.end());
            label = (*i);
            if(dec(j) == label)
	    {
                rname = "MEMBERS";
                rgraph_2 = GetRelation(cgraph_1, rname);
                assert(rgraph_2 != NULL);
                cgraph_2 = rgraph_2->GetWidth();
                rname = "TYPE";
                rgraph_3 = GetRelation(cgraph_2, rname);
                assert(rgraph_3 != NULL);
                cgraph_3 = rgraph_3->GetWidth();
                concept = (ConceptNode*) cgraph_3->GetNode();
                instance_list = concept->GetInstances();
                i = instance_list.begin();
	     assert(i != instance_list.end());
	     exception = (*i);
	     del(exception, "_Except");
	     del(exception, name);
	     if(!(exception_list == ""))
                 exception_list += ", ";
	     exception_list += exception;
	    }
            cgraph_1 = cgraph_1->GetDepth();
	}
    }
    corbaout->EnterNewLevel(raises_level);
    corbaout->PrintIdentifier(exception_list);
    corbaout->LeaveLevel(raises_level);
}


void Template2CORBA::CreateContextExpression( ConceptualGraph *cg )
{
    ConceptualGraph *rgraph;
    ConceptualGraph *cgraph;
    ConceptNode *concept;
    SetOfWord instance_list;
    SetOfWord::iterator i;
    IDLString rname;
    IDLString literal;
    IDLString literal_list;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    rname = "MEMBERS";
    rgraph = GetRelation(cg, rname);
    assert(rgraph != NULL);
    literal_list == "";
    cgraph = rgraph->GetWidth();
    while(cgraph != NULL)
    { 
        concept = (ConceptNode*) cgraph->GetNode();
        instance_list = concept->GetInstances();
        i = instance_list.begin();
        literal = (*i);
        literal = "\"" + literal + "\"";
        if(!(literal_list == ""))
            literal_list += ", ";
        literal_list += literal;
        cgraph = cgraph->GetDepth();
    }
    corbaout->EnterNewLevel(context_level);
    corbaout->PrintIdentifier(literal_list);
    corbaout->LeaveLevel(context_level);
}


void Template2CORBA::CreateParameter( ConceptualGraph *cg, IDLString scope )
{
    ConceptualGraph *rgraph;
    ConceptualGraph *cgraph;
    ConceptNode *concept;
    SetOfWord instance_list;
    IDLString instance;
    IDLString attribute;
    IDLString rname;
    IDLString name;
    SetOfWord::iterator i;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    corbaout->EnterNewLevel(parameter_level);
    concept = (ConceptNode*) cg->GetNode();
    instance_list = concept->GetInstances();
    i = instance_list.begin();
    assert(i != instance_list.end());
    name = (*i);
    rname = "ATTRIBUTES";
    rgraph = GetRelation(cg, rname);
    assert(rgraph != NULL);
    cgraph = rgraph->GetWidth();
    concept = (ConceptNode*) cgraph->GetNode();
    instance_list = concept->GetInstances();
    i = instance_list.begin();
    while(i != instance_list.end())
    {
        instance = (*i);
        if(instance == "in" || instance == "out" || instance == "inout")
	{
            attribute = instance;
	}
        ++i;
    }
    corbaout->PrintAttribute(attribute);
    CheckDeclarator(cg, scope, name ); 
    corbaout->LeaveLevel(parameter_level);
}


void Template2CORBA::CreateSignature( ConceptualGraph *cg, IDLString scope )
{
    ConceptualGraph *cgraph_1;
    ConceptualGraph *cgraph_2;
    ConceptualGraph *rgraph;
    ConceptNode *concept;
    int j;
    int pos;
    int counter = 0;
    IDLString rname;

    assert(cg->GetNode()->IsA() == RELATION_NODE); 
    corbaout->EnterNewLevel(signature_level);
    cgraph_1 = cg->GetWidth();
    while(cgraph_1 != NULL)
    {
        counter++;
        cgraph_1 = cgraph_1->GetDepth();
    }
    for(j = 1; j < counter; j++)
    {
        cgraph_1 = cg->GetWidth();
        while(cgraph_1 != NULL)
	{
            rname = "POSITION";
            rgraph = GetRelation(cgraph_1, rname);
            assert(rgraph != NULL);
            cgraph_2 = rgraph->GetWidth();
            concept = (ConceptNode*) cgraph_2->GetNode();
            pos = concept->GetValue();
            if(pos == j)
                CreateParameter(cgraph_1, scope);
            cgraph_1 = cgraph_1->GetDepth();
	}
    }
    corbaout->LeaveLevel(signature_level);
}


void Template2CORBA::CreateTypeDeclaration( ConceptualGraph *cg,
                                            IDLString scope, 
                                            IDLString type )
{
    ConceptualGraph *cgraph_1;
    ConceptualGraph *cgraph_2;
    ConceptualGraph *rgraph;
    IDLString declarator;
    IDLString cname;
    IDLString rname;
    ConceptNode *concept;
    SetOfWord instance_list;
    SetOfWord::iterator i, j;
    int type_dcl_detected = 1;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    cgraph_1 = cg;
    while(cgraph_1 != NULL)
    {
        concept = (ConceptNode*) cgraph_1->GetNode();
        instance_list = concept->GetInstances();
        i = instance_list.begin();
        if(i == instance_list.end())
	{
            rname = "ANNOTATIONS";
            rgraph = GetRelation(cgraph_1, rname);
            assert(rgraph != NULL);
            cgraph_2 = rgraph->GetWidth();
            while(cgraph_2 != NULL)
	    {
                concept = (ConceptNode*) cgraph_2->GetNode();
                cname = concept->GetName();
                if(cname == "IDENTIFIER")
		{
                    instance_list = concept->GetInstances();
                    j = instance_list.begin();
                    declarator = (*j);
		}
                cgraph_2 = cgraph_2->GetDepth();
	    }
            type_dcl_detected = 0;
	}
        else
            declarator = (*i);
        if(declarator == type && !DetectStandardException(cgraph_1))
	{
            corbaout->EnterNewLevel(type_level);
            if(!type_dcl_detected)
                CreateTypes(cgraph_1, scope, type);
            else
	    {
                corbaout->EnterNewLevel(type_dcl_level);
                CheckDeclarator(cgraph_1, scope, type);
                corbaout->LeaveLevel(type_dcl_level);
	    }
            corbaout->LeaveLevel(type_level);
	}
        cgraph_1 = cgraph_1->GetDepth();
    }
}


void Template2CORBA::CheckDeclarator( ConceptualGraph *cg, IDLString scope, 
                                      IDLString name )
{
    ConceptualGraph *rgraph_1;
    ConceptualGraph *rgraph_2;
    ConceptualGraph *cgraph_1;
    ConceptualGraph *cgraph_2;
    ConceptualGraph *cgraph_3;
    ConceptNode *concept;
    SetOfWord instance_list;
    SetOfWord::iterator i;
    IDLString rname;
    IDLString cname;
    IDLString bound;
    IDLString dtype;
    IDLString declarator;
    int pos;
    int type;
    int unconvertable;
    int dim;
    int j;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    type = DetectDeclaratorType(cg);
    switch(type)
    {
    case SIMPLE:
	CreateDeclarator(cg, scope);
	break;
    case POINTER:
	unconvertable = 0;
	rname = "CONSTRUCTED_TYPE";
	rgraph_1 = GetRelation(cg, rname);
	assert(rgraph_1 != NULL);
	cname = "POINTER";
	cgraph_1 = GetPureConcept(rgraph_1, cname);
	assert(cgraph_1 != NULL);
	rname = "TYPE";
	rgraph_1 = GetRelation(cg, rname);
	if(rgraph_1 != NULL)
        {
	    cgraph_2 = rgraph_1->GetWidth();
	    assert(cgraph_2 != NULL);
	    concept = (ConceptNode*) cgraph_2->GetNode();
	    instance_list = concept->GetInstances();
	    i = instance_list.begin();
	    assert(i != instance_list.end());
	    dtype = (*i);
	    if(contains(dtype, "_Except"))
            {
		corbaout->EnterNewLevel(sequence_level);
		corbaout->PrintType(dtype);
		corbaout->LeaveLevel(sequence_level);
		corbaout->PrintIdentifier(name);
            }
	    else
                unconvertable = 1;
        }
	else
            unconvertable = 1;
	if(unconvertable)
        {
	    declarator = name;
	    concept = (ConceptNode*) cgraph_1->GetNode();
	    dim = concept->GetValue();
	    for(j = 1; j <= dim; j++)
                declarator = "*" + declarator;
	    dtype = " /* ???? ";
            declarator = dtype + declarator;
	    dtype = " Konvertierung nicht moeglich !!! */";
	    declarator += dtype;
	    dtype = "octet";
	    corbaout->PrintType(dtype);
            declarator = name + declarator;
	    corbaout->PrintIdentifier(declarator); 	
        }
	break;
    case ARRAY:
        declarator = name;
	rname = "CONSTRUCTED_TYPE";
	rgraph_1 = GetRelation(cg, rname);
	assert(rgraph_1 != NULL);
	cname = "ARRAY";
	cgraph_1 = GetPureConcept(rgraph_1, cname);
	assert(cgraph_1 != NULL);
	concept = (ConceptNode*) cgraph_1->GetNode();
	dim = concept->GetValue();
	rname = "DIMENSIONS";
	rgraph_1 = GetRelation(cgraph_1, rname);
	assert(rgraph_1 != NULL);
	rname = "POSITION";
	for(j = 1; j <= dim; j++)
        {
	    cgraph_2 = rgraph_1->GetWidth();
	    while(cgraph_2 != NULL)
            {
		rgraph_2 = GetRelation(cgraph_2, rname);
		assert(rgraph_2 != NULL);
		cgraph_3 = rgraph_2->GetWidth();
		assert(cgraph_3 != NULL);
		concept = (ConceptNode*) cgraph_3->GetNode();
		pos = concept->GetValue();
		if(pos == j)
                {
		    concept = (ConceptNode*) cgraph_2->GetNode();
		    instance_list = concept->GetInstances();
		    i = instance_list.begin();
		    assert(i != instance_list.end());
                    bound = string("[") + string(*i) + "]";
		    declarator += bound;
                }
		cgraph_2 = cgraph_2->GetDepth();
            }
        }
	CreateTypes(cgraph_1, scope, name);
	corbaout->PrintIdentifier(declarator);
	break;
    case FUNCTION_POINTER:
	rname = "CONSTRUCTED_TYPE";
	rgraph_1 = GetRelation(cg, rname);
	assert(rgraph_1 != NULL);
	cname = "FUNCTION_POINTER";
	cgraph_1 = GetPureConcept(rgraph_1, cname);
	assert(cgraph_1 != NULL);
	rname = "ANNOTATIONS";
	rgraph_1 = GetRelation(cg, rname);
	if(rgraph_1 == NULL)
        {
	    rgraph_1 = GetRelation(cgraph_1, rname);
	    assert(rgraph_1 != NULL);
        }
	cgraph_2 = GetPureConcept(rgraph_1, cname);
	assert(cgraph_2 != NULL);
	concept = (ConceptNode*) cgraph_2->GetNode();
	instance_list = concept->GetInstances();
	i = instance_list.begin();
	assert(i != instance_list.end());
	declarator = (*i);
	prepend(declarator, " ");
	declarator += " Konvertierung nicht moeglich !!! */";
	rname = "TYPE";
	rgraph_1 = GetRelation(cgraph_1, rname);
	assert(rgraph_1 != NULL);
	cgraph_3 = rgraph_1->GetWidth();
	assert(cgraph_3 != NULL);
	concept = (ConceptNode*) cgraph_3->GetNode();
	instance_list = concept->GetInstances();
	i = instance_list.begin();
	assert(i != instance_list.end());
	dtype = (*i);
	prepend(declarator, dtype);
	dtype = "octet";
	corbaout->PrintType(dtype);
	name += " /* ";	
	prepend(declarator, name);
	corbaout->PrintIdentifier(declarator);
	break;
    }
}


void Template2CORBA::CreateDeclarator( ConceptualGraph *cg, IDLString scope )
{
    ConceptualGraph *rgraph;
    ConceptNode *concept;
    SetOfWord instance_list;
    SetOfWord::iterator i;
    IDLString declarator;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    concept = (ConceptNode*) cg->GetNode();
    instance_list = concept->GetInstances();
    i = instance_list.begin();
    assert(i != instance_list.end());
    declarator = (*i);
    CreateTypes(cg, scope, declarator);
    corbaout->PrintIdentifier(declarator);
}


void Template2CORBA::CreateConstantDeclaration( ConceptualGraph *cg,
                                                IDLString scope, 
                                                IDLString type )
{
    ConceptualGraph *cgraph;
    ConceptualGraph *rgraph;
    ConceptNode *concept;
    IDLString declarator;
    SetOfWord instance_list;
    SetOfWord::iterator i;
    int k;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE); 
    cgraph = cg;
    while(cgraph != NULL)
    {
        concept = (ConceptNode*) cgraph->GetNode();
        instance_list = concept->GetInstances();
        i = instance_list.begin();
        assert(i != instance_list.end());
        declarator = (*i);
        if(declarator == type)
    	{
            rgraph = cgraph->GetWidth();
            corbaout->EnterNewLevel(constant_level);
            CreateTypes(cgraph, scope, type);
            corbaout->PrintIdentifier(declarator);
            CreateConstExpression(rgraph);
            corbaout->LeaveLevel(constant_level);
	}
        cgraph = cgraph->GetDepth();
    }
}


void Template2CORBA::CreateCases( ConceptualGraph *cg, IDLString scope )
{
    ConceptualGraph *cgraph;
    ConceptualGraph *rgraph;
    ConceptualGraph *dgraph;
    ConceptNode *concept;
    SetOfWord instance_list;
    SetOfWord::iterator i;
    IDLString rname;
    IDLString dname;
    IDLString label;
 
    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    rname = "MEMBERS";
    dname = "";
    cgraph = cg;
    while(cgraph != NULL)
    {
        concept = (ConceptNode*) cgraph->GetNode();
        instance_list = concept->GetInstances();
        i = instance_list.begin();
        while(i != instance_list.end())
	{
            label = (*i);
            if(label == "default") 
	    {
                dname = label;
                dgraph = cgraph;
	    }
            else 
	    {
                corbaout->EnterNewLevel(case_level);
                corbaout->PrintExpression(label);
                corbaout->LeaveLevel(case_level);
	    }
            ++i;
	}
        rgraph = GetRelation(cgraph, rname);
        assert(rgraph != NULL);
        if(!(string(dname) == string("default")))
            CreateMembers(rgraph, scope);
        cgraph = cgraph->GetDepth();
    }
    if(dname == "default")
    {
        corbaout->EnterNewLevel(case_level);
        corbaout->PrintExpression(dname);
        corbaout->LeaveLevel(case_level);
        rgraph = GetRelation(dgraph, rname);
        assert(rgraph != NULL);
        CreateMembers(rgraph, scope);
    }
}


void Template2CORBA::CreateMembers( ConceptualGraph *cg, IDLString scope )
{
    ConceptualGraph *cgraph_1;
    ConceptualGraph *cgraph_2;
    ConceptualGraph *rgraph;
    ConceptNode *concept;
    SetOfWord instance_list;
    SetOfWord::iterator i;
    IDLString member;
    IDLString rname;
    int j;
    int counter = 0;
    long position;

    assert(cg->GetNode()->IsA() == RELATION_NODE);
    cgraph_1 = cg->GetWidth();
    while(cgraph_1 != NULL)
    {
        counter++;
        cgraph_1 = cgraph_1->GetDepth();
    }
    rname = "POSITION";
    for(j = 1; j <= counter; j++)
    {
        cgraph_1 = cg->GetWidth();
        while(cgraph_1 != NULL)
        {
            concept = (ConceptNode*) cgraph_1->GetNode();
            instance_list = concept->GetInstances();
            i = instance_list.begin();
            member = (*i);
            rgraph = GetRelation(cgraph_1, rname);
            assert(rgraph != NULL);
            cgraph_2 = rgraph->GetWidth();
            concept = (ConceptNode*) cgraph_2->GetNode();
            position = concept->GetValue();
            if(position == j) 
	    {
                corbaout->EnterNewLevel(member_level);
                CheckDeclarator(cgraph_1, scope, member);
                corbaout->LeaveLevel(member_level);
	    }
            cgraph_1 = cgraph_1->GetDepth();
        }
    }
}


void Template2CORBA::CreateConstructedType( ConceptualGraph *cg,
                                            IDLString scope, 
                                            IDLString tag )
{
    ConceptualGraph *cgraph_1;
    ConceptualGraph *cgraph_2;
    ConceptualGraph *rgraph_1;
    ConceptualGraph *rgraph_2;
    IDLString cname;
    IDLString rname;
    IDLString dname;
    IDLString name;
    IDLString name_list;
    IDLString tscope;
    ConceptNode *concept;
    SetOfWord instance_list;
    SetOfWord::iterator i;
    int pos;
    int j;
    int counter = 0;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    concept = (ConceptNode*) cg->GetNode();
    cname = concept->GetName();
    if(cname == "STRUCT")
    {
        corbaout->EnterNewLevel(struct_level);
        instance_list = concept->GetInstances();
        i = instance_list.begin();
        if(i == instance_list.end())
            name = tag;
        else
            name = (*i);
        corbaout->PrintIdentifier(name);
        tscope = scope + name + "::";
        rname = "MEMBERS";
        rgraph_1 = GetRelation(cg, rname);		
        CreateMembers(rgraph_1, tscope);
        corbaout->LeaveLevel(struct_level);
    }
    if(cname == "UNION")
    {
        corbaout->EnterNewLevel(union_level);
        instance_list = concept->GetInstances();
        i = instance_list.begin();
        if(i == instance_list.end())
            name = tag;
        else
            name = (*i);
        corbaout->PrintIdentifier(name);
        tscope = scope + name + "::";
        corbaout->EnterNewLevel(switch_level);
        dname = "switch_type"; /* Bezeichner, falls 'enum' ohne Identifier */
        rname = "DISCRIMINATOR_TYPE";
        rgraph_1 = GetRelation(cg, rname);
        if(rgraph_1 != NULL)
            CreateTypes(cg, tscope, dname);
        else
	{
            rname = "CONSTRUCTED_TYPE";
            rgraph_1 = GetRelation(cg, rname);
            if(rgraph_1 != NULL)
                CreateTypes(cg, tscope, dname);
            else 
	    {
                /*
                 * XXX
                 * was if (mode = AIT_IDL_DCE) { ... } which looks like a bug.
                 */
                assert (mode == AIT_IDL_DCE);
                /*
                 * Nicht-diskriminiertes Union durch
                 * Typ 'any' ersetzt !!!
                 */
                dname = "any";
                corbaout->PrintType(dname);
	    } 
	}
        corbaout->LeaveLevel(switch_level);
        rname = "CASES";
        rgraph_1 = GetRelation(cg, rname);
        assert(rgraph_1 != NULL);	
        cgraph_1 = rgraph_1->GetWidth();
        CreateCases(cgraph_1, tscope); 
        corbaout->LeaveLevel(union_level);
    }  
    if(cname == "ENUM")
    {
        name_list = "";
        corbaout->EnterNewLevel(enum_level);
        instance_list = concept->GetInstances();
        i = instance_list.begin();
        if(i == instance_list.end())
            name = tag;
        else
            name = (*i);
        corbaout->PrintIdentifier(name);
        rname = "MEMBERS";
        rgraph_1 = GetRelation(cg, rname);
        assert(rgraph_1 != NULL);
        cgraph_1 = rgraph_1->GetWidth();
        while(cgraph_1 != NULL)
	{
            counter++;
            cgraph_1 = cgraph_1->GetDepth();
	}
        for(j = 1; j <= counter; j++)
	{
            cgraph_1 = rgraph_1->GetWidth();
            while(cgraph_1 != NULL)
	    {
                rname = "POSITION";
                rgraph_2 = GetRelation(cgraph_1, rname);
                assert(rgraph_2 != NULL);
                cgraph_2 = rgraph_2->GetWidth();
                concept = (ConceptNode*) cgraph_2->GetNode();
                pos = concept->GetValue();
                if(pos == j)
	    	{
                    concept = (ConceptNode*) cgraph_1->GetNode();
                    instance_list = concept->GetInstances();
                    i = instance_list.begin();
                    name = (*i);
                    if(!(name_list == ""))
                        name_list += ", ";
                    name_list += name;
		}
                cgraph_1 = cgraph_1->GetDepth();
	    }
	}
        corbaout->PrintIdentifier(name_list);
        corbaout->LeaveLevel(enum_level);
    }
    if(cname == "PIPE")
    {
        /* Pipe-Type momentan durch 'octet' ersetzt !!! */
        dname = "octet";
        corbaout->PrintType(dname);
    }
}


void Template2CORBA::CreateTemplateType( ConceptualGraph *cg, IDLString scope )
{
    ConceptualGraph *rgraph;
    ConceptualGraph *cgraph_1;
    ConceptualGraph *cgraph_2;
    ConceptualGraph *tgraph;
    ConceptNode *concept;
    IDLString cname;
    IDLString rname;
    IDLString type;
    IDLString bound;
    SetOfWord instance_list;
    IDLLevelType level;
    SetOfWord::iterator i;
    int interface_type_detected = 0;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    cgraph_1 = cg;
    type = "";
    bound = "";
    while(cgraph_1 != NULL)
    {
        concept = (ConceptNode*) cgraph_1->GetNode();
        cname = concept->GetName();
        if(cname == "SEQUENCE")
            level = sequence_level;
        if(cname == "STRING")
            level = string_level;
        if(cname == "ELEMENT_TYPE")
        {
            rname = "ANNOTATIONS";
            rgraph = GetRelation(cgraph_1, rname);
            if(rgraph != NULL)
	    {
                cname = "INTERFACE_TYPE";
                cgraph_2 = GetPureConcept(rgraph, cname);
                if(cgraph_2 != NULL)
                    interface_type_detected = 1;
	    }
            tgraph = cgraph_1;
            instance_list = concept->GetInstances();
            i = instance_list.begin();
            type = (*i);
        }
        if(cname == "TEMPLATE_BOUND")
        {
            instance_list = concept->GetInstances();
            i = instance_list.begin();
            bound = (*i);
        }
        cgraph_1 = cgraph_1->GetDepth();
    }
    corbaout->EnterNewLevel(level);
    if(interface_type_detected)
    {
        del(type, "_interface_type"); 
        corbaout->PrintType(type);
    }
    else
    {
        if(!(type == ""))
    	{
            if(contains(type, "<") && contains(type, ">"))
                CreateTemplateType(tgraph, scope);
            else
                CreateBaseType(tgraph, scope);
    	}
        if(!(bound == ""))
            corbaout->PrintExpression(bound);
    }
    corbaout->LeaveLevel(level);
}


void Template2CORBA::CreateBaseType( ConceptualGraph *cg, IDLString scope )
{
    IDLString type;
    IDLString ctype;
    ConceptNode *concept;
    SetOfWord instance_list;
    SetOfWord::iterator i;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    concept = (ConceptNode*) cg->GetNode();
    instance_list = concept->GetInstances();
    i = instance_list.begin();
    type = (*i);
    if(idict.ContainsWord(type))
    {
        ctype = ClearScopedName(type, scope);	
    }
    else
    {
        if(IsBaseType(type))
            ctype = type;
        else
            ctype = "none";
        if(type == "small" || type == "unsigned small")
            ctype = "octet";
        if(type == "hyper" || type == "unsigned hyper")
            ctype = "octet";
        if(type == "byte")
            ctype = "octet";
        if(type == "void")
            ctype = "void /* : Typ 'void' existiert nur als "
                "Operation-Return-Type!!!*/ ";
        if(type == "handle_t" || type == "void*")
            ctype = "sequence<octet>";
        if(type == "error_status_t")
            ctype = "unsigned long";
        if(type == "ISO_LATIN_1" || type == "ISO_MULTI_LINGUAL" ||
           type == "ISO_UCS")
            ctype = "char";
    }
    corbaout->PrintType(ctype);
}


void Template2CORBA::CreateConstExpression( ConceptualGraph *cg )
{
    ConceptualGraph *rgraph;
    ConceptualGraph *cgraph;
    RelationNode *relation;
    ConceptNode *concept;
    SetOfWord instance_list;
    SetOfWord::iterator i;
    IDLString rname;
    IDLString expression;
    int string_detected = 0;

    assert(cg->GetNode()->IsA() == RELATION_NODE);
    rgraph = cg;
    while(rgraph != NULL)
    {
        relation = (RelationNode*) rgraph->GetNode();
        rname = relation->GetName();
        if(rname == "VALUE")
            cgraph = rgraph->GetWidth();
        if(rname == "TEMPLATE_TYPE")
            string_detected = 1;
        rgraph = rgraph->GetDepth();
    }
    concept = (ConceptNode*) cgraph->GetNode();
    instance_list = concept->GetInstances();
    i = instance_list.begin();
    expression = (*i);
    if(string_detected)
    {
        prepend(expression, "\"");
        expression += "\"";
    }
    corbaout->PrintExpression(expression);
}


void Template2CORBA::CreateTypes( ConceptualGraph *cg,
                                  IDLString scope, IDLString tag )
{
    ConceptualGraph *rgraph_1;
    ConceptualGraph *rgraph_2;
    ConceptualGraph *cgraph_1;
    ConceptualGraph *cgraph_2;
    IDLString rname;
    IDLString cname;
    IDLString name;
    RelationNode *relation;
    ConceptNode *concept;
    SetOfWord instance_list;
    SetOfWord::iterator i;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    rgraph_1 = cg->GetWidth();
    while(rgraph_1 != NULL)
    {
        cgraph_1 = rgraph_1->GetWidth();
        relation = (RelationNode*) rgraph_1->GetNode();
        rname = relation->GetName();
        if(rname == "CONSTRUCTED_TYPE")
            CreateConstructedType(cgraph_1, scope, tag);
        if(rname == "TYPE")
            CreateBaseType(cgraph_1, scope);
        if(rname == "DISCRIMINATOR_TYPE")
            CreateBaseType(cgraph_1, scope);
        if(rname == "TEMPLATE_TYPE") 
	{
            rname = "ANNOTATIONS";
            rgraph_2 = GetRelation(cg, rname);
            if(rgraph_2 != NULL)
	    {
                cname = "INTERFACE_TYPE";
                cgraph_2 = GetPureConcept(rgraph_2, cname);
                if(cgraph_2 != NULL)
		{
                    cname = "ELEMENT_TYPE";
                    cgraph_2 = GetPureConcept(rgraph_1, cname);
                    assert(cgraph_2 != NULL);
                    concept = (ConceptNode*) cgraph_2->GetNode();
                    instance_list = concept->GetInstances();
                    i = instance_list.begin();
                    assert(i != instance_list.end());
                    name = (*i);
                    del(name, "_interface_type");
                    corbaout->PrintType(name);
		} 
                else
                    CreateTemplateType(cgraph_1, scope);
	    }
            else
                CreateTemplateType(cgraph_1, scope);
	}
        if(rname == "RECURSIVE_TYPE")
        {
            corbaout->EnterNewLevel(sequence_level);
            concept = (ConceptNode*) cgraph_1->GetNode();
            instance_list = concept->GetInstances();
            i = instance_list.begin();
            name = (*i);
            corbaout->PrintType(name);
            corbaout->LeaveLevel(sequence_level);
        }
        rgraph_1 = rgraph_1->GetDepth();	
    }
}


void Template2CORBA::CreateException( ConceptualGraph *cg, IDLString scope, 
                                      IDLString name )
{
    ConceptualGraph *cgraph;
    ConceptualGraph *rgraph;
    ConceptNode *concept;
    SetOfWord instance_list;
    SetOfWord::iterator i;
    IDLString rname;
    IDLString cname;
    IDLString tscope;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    corbaout->EnterNewLevel(exception_level);
    del(name, "_Except");
    corbaout->PrintIdentifier(name);
    tscope = scope + name + "::";
    rname = "CONSTRUCTED_TYPE";
    rgraph = GetRelation(cg, rname);
    if(rgraph != NULL)
    { 
        cgraph = rgraph->GetWidth();
        concept = (ConceptNode*) cgraph->GetNode();
        cname = concept->GetName();
        assert(cname == "STRUCT");
        rname = "MEMBERS";
        rgraph = GetRelation(cgraph, rname);
        CreateMembers(rgraph, tscope);
    }
    corbaout->LeaveLevel(exception_level);
}





/*********************** CG_DCE_Conversion ***************************/


void Template2DCE::CreateSpecification()
{
    ConceptualGraph *cgraph;
    ConceptualGraph *rgraph;
    IDLString scope;
    IDLString cname;
    IDLString rname;
    ConceptNode *concept;

    scope = "";
    rname = "DEFINITIONS";
    rgraph = GetRelation(root, rname);
    mode = DetectIDL(rgraph);
    assert(mode);
    cgraph = rgraph->GetWidth();
    while(cgraph != NULL)
    {
        concept = (ConceptNode*) cgraph->GetNode();
        cname = concept->GetName();
        if(cname == "INTERFACE")
            CreateInterface(cgraph, scope); 
        cgraph = cgraph->GetDepth();
    }
}
 

void Template2DCE::CreateInterface( ConceptualGraph *cg, IDLString scope )
{
    ConceptualGraph *cgraph_1; 
    ConceptualGraph *cgraph_2; 
    ConceptualGraph *cgraph_3; 
    ConceptualGraph *rgraph_1; 
    ConceptualGraph *rgraph_2; 
    ConceptNode *concept;
    SetOfWord instance_list;
    IDLString interface;
    IDLString tscope;
    IDLString rname;
    IDLString cname;
    IDLString tag_1;
    IDLString tag_2;
    SetOfWord::iterator i, j;
    int counter;
    int pos;
    int k;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    rname = "CONSTANTS";
    rgraph_1 = GetRelation(cg, rname);
    if(rgraph_1 != NULL) 
    {
        cgraph_1 = rgraph_1->GetWidth();
        while(cgraph_1 != NULL)
    	{
            concept = (ConceptNode*) cgraph_1->GetNode();
            instance_list = concept->GetInstances();
            i = instance_list.begin();
            assert(i != instance_list.end());
            tag_1 = (*i);
            CollectConstants(cgraph_1, tag_1);
            cgraph_1 = cgraph_1->GetDepth();
    	}
    }
    rname = "TYPES";
    rgraph_1 = GetRelation(cg, rname);
    if(rgraph_1 != NULL) 
    {
        cgraph_1 = rgraph_1->GetWidth();
        while(cgraph_1 != NULL)
    	{
            concept = (ConceptNode*) cgraph_1->GetNode();
            instance_list = concept->GetInstances();
            i = instance_list.begin();
            if(i == instance_list.end())
	    {
                rname = "ANNOTATIONS";
                rgraph_2 = GetRelation(cgraph_1, rname);
                assert(rgraph_2 != NULL);
                cgraph_2 = rgraph_2->GetWidth();
                while(cgraph_2 != NULL)
	    	{
                    concept = (ConceptNode*) cgraph_2->GetNode();
                    cname = concept->GetName();
                    if(cname == "IDENTIFIER")
	    	    {
                        instance_list = concept->GetInstances();
                        j = instance_list.begin();
                        tag_1 = (*j);
	    	    }
                    cgraph_2 = cgraph_2->GetDepth();
	        }
	    }	
            else
                tag_1 = (*i);
            rgraph_2 = cgraph_1->GetWidth();
            while(rgraph_2 != NULL)
    	    {
                CollectTypes(rgraph_2, tag_1);
                rgraph_2 = rgraph_2->GetDepth();
    	    }
            cgraph_1 = cgraph_1->GetDepth();
        }
    }
    concept = (ConceptNode*) cg->GetNode();
    instance_list = concept->GetInstances();
    i = instance_list.begin();
    interface = (*i);
    tscope = scope + interface + "::";
    dceout->EnterNewLevel(interface_level);
    rname = "ATTRIBUTES";
    rgraph_1 = GetRelation(cg, rname);
    if(rgraph_1 != NULL)
        cgraph_1 = rgraph_1;
    else
        cgraph_1 = NULL;
    CreateInterfaceAttributes(cgraph_1);
    dceout->PrintIdentifier(interface);
    CreateDeclarations(cg, tscope); 
    rname = "OPERATIONS";
    rgraph_1 = GetRelation(cg, rname);
    assert(rgraph_1 != NULL); 
    cgraph_1 = rgraph_1->GetWidth();
    counter = 0;
    cgraph_2 = cgraph_1;
    while(cgraph_2 != NULL)
    {
        counter++;
        cgraph_2 = cgraph_2->GetDepth();
    }
    for(k = 1; k <= counter; k++) 
    {
        cgraph_2 = cgraph_1;
        while(cgraph_2 != NULL)
	{
            rname = "POSITION";
            rgraph_2 = GetRelation(cgraph_2, rname);
            assert(rgraph_2 != NULL);
            cgraph_3 = rgraph_2->GetWidth();
            concept = (ConceptNode*) cgraph_3->GetNode();
            pos = concept->GetValue();
            if(pos == k)
                CreateOperation(cgraph_2, tscope);
            cgraph_2 = cgraph_2->GetDepth();
	}
    }
    dceout->LeaveLevel(interface_level);
}
    

void Template2DCE::CreateInterfaceAttributes( ConceptualGraph *cg )
{
    ConceptualGraph *cgraph;
    ConceptNode *concept;
    SetOfWord instance_list;
    SetOfWord::iterator i;
    IDLString attribute;
    IDLString attribute_list;
    IDLString all_endpoints;
    IDLString cname;

    if(cg == NULL)
    {
        assert(mode == AIT_IDL_CORBA); 
        attribute_list = "local";
    }
    else 
    {
        assert(cg->GetNode()->IsA() == RELATION_NODE);
        attribute_list = "";
        cgraph = cg->GetWidth();
        while(cgraph != NULL)
    	{
            concept = (ConceptNode*) cgraph->GetNode();
            cname = concept->GetName();
            if(cname == "INTERFACE_ATTRIBUTE")
	    {
                if(!(attribute_list == ""))
                    attribute_list += ", ";
                attribute_list += "local";
	    } 
            if(cname == "INTERFACE_UUID")
	    {
                instance_list = concept->GetInstances();
                i = instance_list.begin();
                attribute = (*i);
                if(!(attribute_list == ""))
                    attribute_list += ", ";
                attribute_list += attribute;
	    }
            if(cname == "INTERFACE_VERSION")
	    {
                instance_list = concept->GetInstances();
                i = instance_list.begin();
                attribute = (*i);
                if(!(attribute_list == ""))
                    attribute_list += ", ";
                attribute_list += attribute;
	    }
            if(cname == "INTERFACE_PTR_SEMANTIC")
	    {
                instance_list = concept->GetInstances();
                i = instance_list.begin();
                attribute = (*i);
                if(!(attribute_list == ""))
                    attribute_list += ", ";
                attribute_list += attribute;
	    } 
            if(cname == "INTERFACE_SERVER_ENDPOINTS")
	    {
                instance_list = concept->GetInstances();
                i = instance_list.begin();
                all_endpoints = "";
                while(i != instance_list.end())
	        {
                    attribute = (*i);
                    if(!(all_endpoints == ""))
                        all_endpoints = all_endpoints + ", ";
                    all_endpoints = all_endpoints + "\"" + attribute + "\"";
                    ++i;
	    	}
                if(!(attribute_list == ""))
                    attribute_list += ", ";
                attribute_list += all_endpoints;
	    }
            cgraph = cgraph->GetDepth();
        }
    }
    dceout->PrintAttribute(attribute_list);
}


void Template2DCE::CreateDeclarations( ConceptualGraph *cg, IDLString scope )
{
    ConceptualGraph *cgraph_1;
    ConceptualGraph *cgraph_2;
    ConceptualGraph *cgraph_3;
    ConceptualGraph *rgraph;
    ConceptNode *concept;
    SetOfWord instance_list;
    SetOfWord::iterator i, j;
    IDLString rname;
    IDLString cname;
    IDLString tname;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    rname = "CONSTANTS";
    rgraph = GetRelation(cg, rname);
    if(rgraph != NULL)
        cgraph_2 = rgraph->GetWidth();
    else
        cgraph_2 = NULL;
    rname = "TYPES";
    rgraph = GetRelation(cg, rname);
    if(rgraph != NULL)
        cgraph_1 = rgraph->GetWidth();
    else
        cgraph_1 = NULL;
    while(cgraph_1 != NULL)
    {
        concept = (ConceptNode*) cgraph_1->GetNode();
        instance_list = concept->GetInstances();
        i = instance_list.begin();
        if(i == instance_list.end())
	{
            rname = "ANNOTATIONS";
            rgraph = GetRelation(cgraph_1, rname);
            assert (rgraph != NULL);
            cgraph_3 = rgraph->GetWidth();
            while(cgraph_3 != NULL)
	    {
                concept = (ConceptNode*) cgraph_3->GetNode();
                cname = concept->GetName();
                if(cname == "IDENTIFIER")
		{
                    instance_list = concept->GetInstances();
                    j = instance_list.begin();
                    tname = (*j);
		}
                cgraph_3 = cgraph_3->GetDepth();
	    }
	}
        else
            tname = (*i);
        RangeTypeDeclarations(cgraph_1, cgraph_2, scope, tname);
        cgraph_1 = cgraph_1->GetDepth();
    }
    while(cgraph_2 != NULL)
    {
        concept = (ConceptNode*) cgraph_2->GetNode();
        instance_list = concept->GetInstances();
        i = instance_list.begin();
        tname = (*i);
        RangeConstantDeclarations(cgraph_1, cgraph_2, scope, tname);
        cgraph_2 = cgraph_2->GetDepth();
    }
}


void Template2DCE::RangeTypeDeclarations( ConceptualGraph *cg_1,
                                          ConceptualGraph *cg_2,
                                          IDLString scope, IDLString type )
{
    IDLString tname;
    Dict::vcwp::const_iterator i;

    Dict::vcwp type_list = idict.GetRelImmSucc(type, "TYPE_DEPEND");
    i = type_list.begin();
    while(i != type_list.end())
    {
        tname = **i;
        if(!odict.ContainsWord(tname))
            RangeTypeDeclarations(cg_1, cg_2, scope, tname);
        ++i;
    }
    Dict::vcwp const_list = idict.GetRelImmSucc(type, "CONST_DEPEND");
    i = const_list.begin();
    while(i != const_list.end())
    {
        tname = **i;
        if(!odict.ContainsWord(tname))
            RangeConstantDeclarations(cg_1, cg_2, scope, tname);
        ++i;
    } 
    if(!odict.ContainsWord(type))
    {
        CreateTypeDeclaration(cg_1, scope, type);
        odict.AddWord(type);
    }
}


void Template2DCE::RangeConstantDeclarations( ConceptualGraph *cg_1,
                                              ConceptualGraph *cg_2, 
                                              IDLString scope, IDLString type )
{
    IDLString cname;
    Dict::vcwp::const_iterator i;

    Dict::vcwp const1_list = idict.GetRelImmSucc(type, "CONST_DEPEND");
    i = const1_list.begin();
    if(i == const1_list.end())
    {
        Dict::vcwp const2_list = idict.GetRelImmSucc(type, "SYNONYM");
        i = const2_list.begin();
        while(i != const2_list.end())
    	{
            cname = **i;
            if(!odict.ContainsWord(cname))
                RangeConstantDeclarations(cg_1, cg_2, scope, cname);
            ++i;
    	}
    }
    else
    { 
        while(i != const1_list.end())
    	{
            cname = **i;
            if(!odict.ContainsWord(cname))
                RangeConstantDeclarations(cg_1, cg_2, scope, cname);
            ++i;
    	}
    }
    Dict::vcwp type_list = idict.GetRelImmSucc(type, "TYPE_DEPEND");
    i = type_list.begin();
    while(i != type_list.begin())
    {
        cname = **i;
        if(!odict.ContainsWord(cname))
            RangeTypeDeclarations(cg_1, cg_2, scope, cname);
        ++i;
    } 
    if(!odict.ContainsWord(type))
    {
        CreateConstantDeclaration(cg_2, scope, type);
        odict.AddWord(type);
    }
}


void Template2DCE::CreateOperation( ConceptualGraph *cg, IDLString scope )
{
    ConceptualGraph *cgraph;
    ConceptualGraph *rgraph;
    ConceptNode *concept;
    SetOfWord instance_list;
    SetOfWord::iterator i;
    IDLString identifier;
    IDLString tscope;
    IDLString attribute;
    IDLString attribute_list;
    IDLString rname;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    concept = (ConceptNode*) cg->GetNode();
    instance_list = concept->GetInstances(); 
    i = instance_list.begin();
    identifier = (*i);
    tscope = scope + identifier + "::";
    dceout->EnterNewLevel(operation_level);
    rname = "ATTRIBUTES";
    rgraph = GetRelation(cg, rname);
    if(rgraph != NULL)
    {
        attribute_list = "";
        cgraph = rgraph->GetWidth();
        while(cgraph != NULL)
	{
            concept = (ConceptNode*) cgraph->GetNode();
            instance_list = concept->GetInstances();
            i = instance_list.begin();
            attribute = (*i);
            if(!(attribute_list == ""))
                attribute_list += ", ";
            attribute_list += attribute;
            cgraph = cgraph->GetDepth();
	} 
        dceout->PrintAttribute(attribute_list);
    }
    rname = "TYPE";
    rgraph = GetRelation(cg, rname);
    if(rgraph != NULL)
        CreateTypes(cg, tscope, identifier);
    rname = "TEMPLATE_TYPE";
    rgraph = GetRelation(cg, rname);
    if(rgraph != NULL)
    {
        /* A GOOD IDEA ???? */
    }
    dceout->PrintIdentifier(identifier);
    rname = "SIGNATURE";
    rgraph = GetRelation(cg, rname);
    if(rgraph != NULL)
        CreateSignature(rgraph, tscope);
    dceout->LeaveLevel(operation_level); 
}


void Template2DCE::CreateSignature( ConceptualGraph* cg, IDLString scope )
{
    ConceptualGraph *cgraph_1;
    ConceptualGraph *cgraph_2;
    ConceptualGraph *rgraph;
    ConceptNode *concept;
    SetOfWord instance_list;
    SetOfWord::iterator i;
    int j;
    int pos;
    int counter = 0;
    IDLString pname;
    IDLString rname;

    assert(cg->GetNode()->IsA() == RELATION_NODE); 
    cgraph_1 = cg->GetWidth();
    while(cgraph_1 != NULL)
    {
        counter++;
        cgraph_1 = cgraph_1->GetDepth();
    }
    if(counter == 1 && mode == AIT_IDL_DCE)
    { 
        pname = "void";
        dceout->PrintExpression(pname);
    }
    else
    {
        if(mode == AIT_IDL_DCE)
            counter--; 
        for(j = 1; j <= counter; j++)
    	{
            cgraph_1 = cg->GetWidth();
            while(cgraph_1 != NULL)
	    {
                rname = "POSITION";
                rgraph = GetRelation(cgraph_1, rname);
                assert(rgraph != NULL);
                cgraph_2 = rgraph->GetWidth();
                concept = (ConceptNode*) cgraph_2->GetNode();
                pos = concept->GetValue();
                if(pos == j)
		{
                    concept = (ConceptNode*) cgraph_1->GetNode();
                    instance_list = concept->GetInstances();
                    i = instance_list.begin();
                    pname = (*i);
                    dceout->EnterNewLevel(parameter_level);
                    CreateDeclaration(cgraph_1, scope, pname);
                    dceout->LeaveLevel(parameter_level);
		}
                cgraph_1 = cgraph_1->GetDepth();
	    }
	}
    }
}


void Template2DCE::CreateTypeDeclaration( ConceptualGraph* cg,
                                          IDLString scope, 
                                          IDLString type )
{
    ConceptualGraph *cgraph_1;
    ConceptualGraph *cgraph_2;
    ConceptualGraph *rgraph;
    IDLString rname;
    IDLString cname;
    IDLString declarator;
    ConceptNode *concept;
    SetOfWord instance_list;
    SetOfWord::iterator i, j;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    cgraph_1 = cg;
    while(cgraph_1 != NULL)
    {
        concept = (ConceptNode*) cgraph_1->GetNode();
        instance_list = concept->GetInstances();
        i = instance_list.begin();
        if(i == instance_list.end())
	{
            rname = "ANNOTATIONS";
            rgraph = GetRelation(cgraph_1, rname);
            assert(rgraph != NULL);
            cgraph_2 = rgraph->GetWidth();
            while(cgraph_2 != NULL)
	    {
                concept = (ConceptNode*) cgraph_2->GetNode();
                cname = concept->GetName();
                if(cname == "IDENTIFIER")
		{
                    instance_list = concept->GetInstances();
                    j = instance_list.begin();
                    assert(j != instance_list.end());
                    declarator = (*j);
		}
                cgraph_2 = cgraph_2->GetDepth();
	    }
	}
        else
            declarator = (*i);
        if(declarator == type)
	{
            if(mode != AIT_IDL_DCE || !DetectStandardException(cgraph_1))
	    {
                dceout->EnterNewLevel(type_level);
                CreateDeclaration(cgraph_1, scope, type);
                dceout->LeaveLevel(type_level);
	    }
	}
        cgraph_1 = cgraph_1->GetDepth();
    }
}


void Template2DCE::CreateDeclaration( ConceptualGraph *cg, IDLString scope, 
                                      IDLString name )
{
    ConceptualGraph *cgraph_1;
    ConceptualGraph *cgraph_2;
    ConceptualGraph *cgraph_3;
    ConceptualGraph *rgraph;
    ConceptualGraph *tgraph;
    ConceptNode *concept;
    SetOfWord instance_list, attribute_list;
    SetOfWord::iterator i;
    IDLString attribute;
    IDLString attributes;
    IDLString declarator;
    IDLString rname;
    IDLString cname;
    IDLString tname;
    IDLString bound;
    IDLString form;
    IDLString type;
    int interface_type_detected = 0;
    int recursion_detected = 0;
    int out_detected = 0;
    int dtype;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    rname = "ATTRIBUTES";
    rgraph = GetRelation(cg, rname);
    if(rgraph != NULL)
    {
        cgraph_1 = rgraph->GetWidth();
        concept = (ConceptNode*) cgraph_1->GetNode();
        attribute_list = concept->GetInstances();
        attributes = "";
        i = attribute_list.begin();
        assert(i != attribute_list.end());
        while(i != attribute_list.end())
    	{
            attribute = (*i);
            if(!(attributes == ""))
                attributes += ", ";
            if(attribute == "inout") 
	    {
                attributes += "in, out";
                out_detected = 1;
	    }
            else
	    {
                if(attribute == "out")
                    out_detected = 1;
                attributes += attribute;
	    }
            ++i;
	}
    }
    dtype = DetectDeclaratorType(cg);
    switch(dtype)
    {
    case POINTER:
	CreatePointerDeclaration(cg, attributes, scope, name);
	break;
    case ARRAY:
	CreateArrayDeclaration(cg, attributes, scope, name);
	break;
    case FUNCTION_POINTER:
	CreateFunctionPointerDeclaration(cg, attributes, scope, name);
	break;
    case SIMPLE:
	rname = "ANNOTATIONS";
	rgraph = GetRelation(cg, rname);
	if(rgraph != NULL)
        {
	    cname = "INTERFACE_TYPE";
	    cgraph_1 = GetPureConcept(rgraph, cname);
	    if(cgraph_1 != NULL)
                interface_type_detected = 1;
        }
 	concept = (ConceptNode*) cg->GetNode();
 	instance_list = concept->GetInstances();
 	i = instance_list.begin();
 	if(i != instance_list.end())
            declarator = (*i);
 	else
            declarator = name;
 	rname = "TEMPLATE_TYPE";
 	rgraph = GetRelation(cg, rname);
 	if(rgraph != NULL)
        {
     	    cgraph_1 = rgraph->GetWidth();
     	    bound = "";
     	    while(cgraph_1 != NULL)
            {
		concept = (ConceptNode*) cgraph_1->GetNode();
	 	cname = concept->GetName();
	 	if(cname == "STRING")
                    form = "string";
	 	if(cname == "SEQUENCE")
                    form = "sequence";
	 	if(cname == "TEMPLATE_BOUND")
                {
	 	    instance_list = concept->GetInstances();
	 	    i = instance_list.begin();
	 	    bound = (*i);
	 	    gsub(bound, "::", "_");
                }
	 	if(cname == "ELEMENT_TYPE")
                {
	 	    instance_list = concept->GetInstances();
	 	    i = instance_list.begin();
		    assert(i != instance_list.end());
	 	    type = (*i);
	 	    if(contains(type, "sequence") ||
                       contains(type, "string"))
                        recursion_detected = 1;
	 	    tgraph = cgraph_1->GetWidth();
		    rname = "ANNOTATIONS";
		    rgraph = GetRelation(cgraph_1, rname);
		    if(rgraph != NULL)
                    {
			cgraph_2 = rgraph->GetWidth();
			while(cgraph_2 != NULL)
                        {
			    concept = (ConceptNode*) cgraph_2->GetNode();
			    cname = concept->GetName();
			    if(cname == "INTERFACE_TYPE")
                                interface_type_detected = 1;
			    if(cname == "STRUCT")
                                prepend(type, "struct ");
			    if(cname == "UNION")
                                prepend(type, "union ");
			    cgraph_2 = cgraph_2->GetDepth();
                        }
                    }
                }
	 	cgraph_1 = cgraph_1->GetDepth();
            }
     	    if(form == "string")
            {
	 	if(out_detected)
                    prepend(declarator, "*");
	 	if(bound == "")
                {
	 	    if(contains(attributes, "ptr"))
                    {
			if(!contains(attributes, "string"))
                        {
			    if(!(attributes == ""))
                                attributes += ", ";
 			    attributes += "string";
                        }
                    }
	 	    else 
                    {
			if(!(attributes == ""))
                            attributes += ", ";
			if(!contains(attributes, "string"))
                            attributes += "string, ptr";
			else
                            attributes += "ptr";
                    }
	 	    dceout->PrintAttribute(attributes);
	 	    tname = "char";
		    prepend(declarator, "*");
	 	    dceout->PrintType(tname);
                }
	 	else
                {
	 	    attributes += "string";
	 	    dceout->PrintAttribute(attributes);
	 	    tname = "char";
	 	    dceout->PrintType(tname);
	 	    declarator = declarator + "[" + bound + "]";
                }
	 	dceout->PrintIdentifier(declarator);
            }	
     	    else 
            {
		if(interface_type_detected) 
                {
	     	    prepend(declarator, "*");
		    if(out_detected)
                        prepend(declarator, "*");
	  	    if(!(attributes == ""))
                        dceout->PrintAttribute(attributes);
	 	    dceout->PrintType(type);
                }
	 	else
                { 
	 	    if(out_detected)
                        prepend(declarator, "*");
	 	    if(!(attributes == ""))
                        dceout->PrintAttribute(attributes); 
	 	    if(!(bound == "")) 
                    {
	 	    	if(recursion_detected)
                            tname = ConvertElementType(tgraph, scope);
	    	 	else
                            tname = CompareType(type);
	    	 	dceout->PrintType(tname); 
	    	 	declarator = declarator + "[" + bound + "]";
	    	 	dceout->PrintIdentifier(declarator);
                    }
	     	    else
                    {
	    		if(recursion_detected)
                            tname = ConvertElementType(tgraph, scope);
	    	 	else
                            tname = CompareType(type);

	         	dceout->EnterNewLevel(struct_level);
	     	 	dceout->PrintIdentifier(declarator);

	     	 	dceout->EnterNewLevel(member_level);
                        dceout->PrintType(string("long"));
	     	 	dceout->PrintIdentifier(string("size"));
	    	 	dceout->LeaveLevel(member_level);

	    	 	dceout->EnterNewLevel(member_level);
	    	 	attribute = "size_is(size)";
	    	 	dceout->PrintAttribute(attribute);

		 	if(contains(tname, "&"))
		 	    attribute = "*" + tname;
		 	else
                            attribute = tname;
                        del (tname, "&");
                        if (IsUnconvertableType (tname))
                            attribute += " /* NOT YET CONVERTED */";

                        dceout->PrintType(attribute);
	    	 	tname = declarator + "_arr[*]";
	    	 	dceout->PrintIdentifier(tname);
	    	 	dceout->LeaveLevel(member_level);
	    	 	dceout->LeaveLevel(struct_level);
	    	 	dceout->PrintIdentifier(declarator);
                    }
                }
            }
        }
 	else 
        {
	    if(!(attributes == ""))
                dceout->PrintAttribute(attributes);
            CreateTypes(cg, scope, name);
            if(out_detected && !contains(declarator, "["))
                prepend(declarator, "*");
     	    dceout->PrintIdentifier(declarator);
        }
	break;
    }
}


void Template2DCE::CreatePointerDeclaration( ConceptualGraph *cg,
                                             IDLString attributes,
                                             IDLString scope,
                                             IDLString name )
{
    ConceptualGraph *rgraph;
    ConceptualGraph *cgraph_1;
    ConceptualGraph *cgraph_2;
    ConceptNode *concept;
    SetOfWord instance_list;
    IDLString rname;
    IDLString cname;
    IDLString ptype;
    IDLString type;
    IDLString declarator;
    SetOfWord::iterator i;
    int dim;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    if(!(attributes == ""))
        dceout->PrintAttribute(attributes);
    concept = (ConceptNode*) cg->GetNode();
    instance_list = concept->GetInstances();
    i = instance_list.begin();
    assert(i != instance_list.end());
    declarator = (*i);
    rname = "CONSTRUCTED_TYPE";
    rgraph = GetRelation(cg, rname);
    assert(rgraph != NULL);
    cgraph_1 = rgraph->GetWidth();
    rname = "ANNOTATIONS";
    rgraph = GetRelation(cg, rname);
    if(rgraph == NULL)
    {
        rgraph = GetRelation(cgraph_1, rname);
        assert(rgraph != NULL);
    }
    cname = "POINTER";
    cgraph_2 = GetPureConcept(rgraph, cname);
    assert(cgraph_2 != NULL);
    concept = (ConceptNode*) cgraph_2->GetNode();
    instance_list = concept->GetInstances();
    i = instance_list.begin();
    if(i != instance_list.end())
        ptype = (*i);
    else
        ptype = "unknown";
    concept = (ConceptNode*) cgraph_1->GetNode();
    dim = concept->GetValue();
    if(dim == 2)
        prepend(declarator, "**");
    else
        prepend(declarator, "*");
    if(ptype == "ptr")
    {
        type = ConvertElementType(cgraph_1->GetWidth(), scope);
        dceout->PrintType(type);
    }
    else
        CreateTypes(cgraph_1, scope, name);
    dceout->PrintIdentifier(declarator);
}


void Template2DCE::CreateArrayDeclaration( ConceptualGraph *cg,
                                           IDLString attributes,
                                           IDLString scope,
                                           IDLString name )
{
    ConceptualGraph *rgraph;
    ConceptualGraph *cgraph_1;
    ConceptualGraph *cgraph_2;
    ConceptualGraph *cgraph_3;
    ConceptNode *concept;
    SetOfWord instance_list;
    IDLString rname;
    IDLString cname;
    IDLString tname;
    IDLString declarator;
    SetOfWord::iterator i;
    int pos;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    if(!(attributes == ""))
        dceout->PrintAttribute(attributes);
    rname = "CONSTRUCTED_TYPE";
    rgraph = GetRelation(cg, rname);
    assert(rgraph != NULL);
    cgraph_1 = rgraph->GetWidth();
    rname = "ANNOTATIONS";
    rgraph = GetRelation(cg, rname);
    if(rgraph == NULL)
    {
        rgraph = GetRelation(cgraph_1, rname);
        assert(rgraph != NULL);
    }
    cgraph_2 = rgraph->GetWidth();
    while(cgraph_2 != NULL)
    {
        concept = (ConceptNode*) cgraph_2->GetNode();
        cname = concept->GetName();
        if(cname == "FIXED_ARRAY" || cname == "VARYING_ARRAY" ||
           cname == "CONFORMANT_ARRAY")
        {
            instance_list = concept->GetInstances();
            i = instance_list.begin();
            assert(i != instance_list.end());
            declarator = (*i);
            if(cname == "FIXED_ARRAY")
                CreateTypes(cgraph_1, scope, name);
            if(cname == "CONFORMANT_ARRAY") 
	    {
                tname = ConvertElementType(cgraph_1->GetWidth(), scope);
                dceout->PrintType(tname);
	    }
            if(cname == "VARYING_ARRAY")
	    {
                rname = "CONSTRUCTED_TYPE";
                rgraph = GetRelation(cgraph_1, rname);
                assert(rgraph != NULL);
                cgraph_2 = rgraph->GetWidth();
                rname = "MEMBERS";
                rgraph = GetRelation(cgraph_2, rname);
                cgraph_2 = rgraph->GetWidth();
                while(cgraph_2 != NULL)
	        {
                    rname = "POSITION";
                    rgraph = GetRelation(cgraph_2, rname);
                    assert(rgraph != NULL);
                    cgraph_3 = rgraph->GetWidth();
                    concept = (ConceptNode*) cgraph_3->GetNode();
                    pos = concept->GetValue();
                    if(pos == 4) CreateTypes(cgraph_2, scope, name);
                    cgraph_2 = cgraph_2->GetDepth();
	        }
	    }
        }
        cgraph_2 = cgraph_2->GetDepth();
    }
    dceout->PrintIdentifier(declarator);
}


void Template2DCE::CreateFunctionPointerDeclaration( ConceptualGraph *cg,
                                                     IDLString attributes,
                                                     IDLString scope,
                                                     IDLString name )
{
    ConceptualGraph *rgraph;
    ConceptualGraph *cgraph_1;
    ConceptualGraph *cgraph_2;
    ConceptNode *concept;
    SetOfWord instance_list;
    IDLString rname;
    IDLString cname;
    IDLString declarator;
    SetOfWord::iterator i;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    if(!(attributes == ""))
        dceout->PrintAttribute(attributes);
    rname = "CONSTRUCTED_TYPE";
    rgraph = GetRelation(cg, rname);
    assert(rgraph != NULL);
    cgraph_1 = rgraph->GetWidth();
    rname = "ANNOTATIONS";
    rgraph = GetRelation(cg, rname);
    if(rgraph == NULL)
    {
        rgraph = GetRelation(cgraph_1, rname);
        assert(rgraph != NULL);
    }
    cname = "FUNCTION_POINTER";
    cgraph_2 = GetPureConcept(rgraph, cname);
    assert(cgraph_2 != NULL);
    concept = (ConceptNode*) cgraph_2->GetNode();
    instance_list = concept->GetInstances();
    i = instance_list.begin();
    assert(i != instance_list.end());
    declarator = (*i);
    CreateTypes(cgraph_1, scope, name);
    dceout->PrintIdentifier(declarator);
}


IDLString Template2DCE::CompareType( IDLString type )
{
    return(type);
}


IDLString Template2DCE::ConvertElementType( ConceptualGraph *cg,
                                            IDLString scope )
{
    ConceptualGraph *cgraph_1;
    ConceptualGraph *cgraph_2;
    ConceptualGraph *rgraph;
    ConceptualGraph *tgraph;
    ConceptNode *concept;
    IDLString cname;
    IDLString rname;
    IDLString form;
    IDLString bound;
    IDLString type;
    SetOfWord instance_list;
    SetOfWord::iterator i;
    int interface_type_detected = 0;
    int recursion_detected = 0;

    assert(cg->GetNode()->IsA() == RELATION_NODE);
    cgraph_1 = cg->GetWidth();
    while(cgraph_1 != NULL)
    {
        concept = (ConceptNode*) cgraph_1->GetNode();
        cname = concept->GetName();
        if(cname == "STRING")
            form = "string";
        if(cname == "SEQUENCE")
            form = "sequence";
        if(cname == "TEMPLATE_BOUND")
	{
            instance_list = concept->GetInstances();
            i = instance_list.begin();
            bound = (*i);
	}
        if(cname == "ELEMENT_TYPE")
	{
            instance_list = concept->GetInstances();
            i = instance_list.begin();
            type = (*i);
            if(contains(type, "sequence") || contains(type, "string"))
                recursion_detected = 1;
            tgraph = cgraph_1->GetWidth();
            rname = "ANNOTATIONS";
            rgraph = GetRelation(cgraph_1, rname);
            if(rgraph != NULL)
	    {
                cgraph_2 = rgraph->GetWidth();
                while(cgraph_2 != NULL)
		{
                    concept = (ConceptNode*) cgraph_2->GetNode();
                    cname = concept->GetName();
                    if(cname == "INTERFACE_TYPE")
                        interface_type_detected = 1;
                    if(cname == "STRUCT")
                        prepend(type, "struct ");
                    if(cname == "UNION")
                        prepend(type, "union ");
                    cgraph_2 = cgraph_2->GetDepth();
		}
	    }
	}
        cgraph_1 = cgraph_1->GetDepth();
    }
    if(form == "string")
        type = "char";
    if(interface_type_detected)
        prepend(type, "&");
    return(type);
    /*
     * A GOOD IDEA ????
     */
}


void Template2DCE::CreateConstantDeclaration( ConceptualGraph *cg,
                                              IDLString scope,
                                              IDLString type )
{
    ConceptualGraph *cgraph;
    ConceptualGraph *rgraph;
    ConceptNode *concept;
    IDLString declarator;
    SetOfWord instance_list;
    SetOfWord::iterator i;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE); 
    cgraph = cg;
    while(cgraph != NULL)
    {
        concept = (ConceptNode*) cgraph->GetNode();
        instance_list = concept->GetInstances();
        i = instance_list.begin();
        declarator = (*i);
        if(declarator == type)
    	{
            rgraph = cgraph->GetWidth();
            dceout->EnterNewLevel(constant_level);
            CreateTypes(cgraph, scope, type);
            dceout->PrintIdentifier(declarator);
            CreateConstExpression(rgraph, scope);
            dceout->LeaveLevel(constant_level);
	}
        cgraph = cgraph->GetDepth();
    }
}


void Template2DCE::CreateCases( ConceptualGraph *cg, IDLString scope,
                                int union_n_e )
{
    ConceptualGraph *cgraph_1;
    ConceptualGraph *cgraph_2;
    ConceptualGraph *rgraph;
    ConceptualGraph *dgraph;
    ConceptNode *concept;
    SetOfWord instance_list;
    SetOfWord::iterator i;
    IDLString label;
    IDLString label_list;
    IDLString rname;
    int default_case = 0;
 
    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    cgraph_1 = cg;
    while(cgraph_1 != NULL)
    {
        concept = (ConceptNode*) cgraph_1->GetNode();
        instance_list = concept->GetInstances();
        i = instance_list.begin();
        assert(i != instance_list.end());
        label = (*i);
        if(label == "default") 
	{
            default_case = 1;
            rname = "MEMBERS";
            dgraph = GetRelation(cgraph_1, rname);
            assert(dgraph != NULL);
	}
        else 
	{
            dceout->EnterNewLevel(case_level);
            if(union_n_e)
	    {
                label_list = label;
                ++i;
                while(i != instance_list.end())
		{
		 label = (*i);
		 prepend(label, ", ");
		 label_list += label;
                 ++i;
		}
                dceout->PrintExpression(label_list);
	    }
            else
	    {
                dceout->PrintExpression(label);
                ++i;
                while(i != instance_list.end())
	    	{
                    label = (*i);
                    dceout->PrintExpression(label);
                    ++i;
	    	}
	    }
            dceout->LeaveLevel(case_level);
            rname = "MEMBERS";
            rgraph = GetRelation(cgraph_1, rname);
            assert(rgraph != NULL);
            cgraph_2 = rgraph->GetWidth();
            CreateMembers(cgraph_2, scope);
	}
        cgraph_1 = cgraph_1->GetDepth();
    }
    if(default_case)
    {
        dceout->EnterNewLevel(case_level);
        label = "default";
        dceout->PrintExpression(label);
        dceout->LeaveLevel(case_level);
        cgraph_2 = dgraph->GetWidth();
        CreateMembers(cgraph_2, scope);
    }
}


void Template2DCE::CreateMembers( ConceptualGraph* cg, IDLString scope )
{
    ConceptualGraph *cgraph_1;
    ConceptualGraph *cgraph_2;
    ConceptualGraph *rgraph;
    RelationNode *relation;
    ConceptNode *concept;
    SetOfWord instance_list;
    SetOfWord::iterator i;
    IDLString tname;
    IDLString member;
    IDLString rname;
    int j;
    int counter = 1;
    long position;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    cgraph_1 = cg;
    while(cgraph_1 != NULL)
    {
        cgraph_1 = cgraph_1->GetDepth();
        counter++;
    }
    for(j = 1; j <= counter; j++)
    {
        cgraph_1 = cg;
        while(cgraph_1 != NULL)
        {
            concept = (ConceptNode*) cgraph_1->GetNode();
            instance_list = concept->GetInstances();
            i = instance_list.begin();
            member = (*i);
            rname = "POSITION";
            rgraph = GetRelation(cgraph_1, rname);
            assert(rgraph != NULL);
            cgraph_2 = rgraph->GetWidth();
            concept = (ConceptNode*) cgraph_2->GetNode();
            position = concept->GetValue();
            if(position == j)
	    {
                dceout->EnterNewLevel(member_level);
                CreateDeclaration(cgraph_1, scope, member);
                dceout->LeaveLevel(member_level);
	    }
            cgraph_1 = cgraph_1->GetDepth();
        }
    }
}


void Template2DCE::CreateConstructedType( ConceptualGraph *cg,
                                          IDLString scope, 
                                          IDLString tag )
{
    ConceptualGraph *cgraph_1;
    ConceptualGraph *cgraph_2;
    ConceptualGraph *rgraph_1;
    ConceptualGraph *rgraph_2;
    IDLString cname;
    IDLString rname;
    IDLString dname;
    IDLString tname;
    IDLString name;
    IDLString name_list;
    IDLString tscope;
    IDLString ttag;
    int pos;
    ConceptNode *concept;
    SetOfWord instance_list;
    SetOfWord::iterator i;
    int j;
    int union_n_e = 0;
    int counter = 0;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    concept = (ConceptNode*) cg->GetNode();
    cname = concept->GetName();
    if(cname == "STRUCT")
    {
        dceout->EnterNewLevel(struct_level);
        instance_list = concept->GetInstances();
        i = instance_list.begin();
        if(i != instance_list.end())
	{
            name = (*i);
            dceout->PrintIdentifier(name);
            tscope = scope + name + "::";
	}
        else 
	{
            if(!(tag == ""))
	    {
                dceout->PrintIdentifier(tag);
                tscope = scope + tag  +"::";
	    }
	}
        rname = "MEMBERS";
        rgraph_1 = GetRelation(cg, rname);
        if(rgraph_1 != NULL)
	{
            cgraph_1 = rgraph_1->GetWidth();     
            CreateMembers(cgraph_1, tscope);
	}
        dceout->LeaveLevel(struct_level);
    }
    if(cname == "UNION")
    {
        dceout->EnterNewLevel(union_level);
        instance_list = concept->GetInstances();
        i = instance_list.begin();
        if(i != instance_list.end())
	{
            name = (*i);
            tscope = scope + name + "::";
            ++i;
            if(i != instance_list.end())
                ttag = (*i);
            else
                ttag = "";
	}
        else 
	{
            name = "";
            if(!(tag == ""))
	    {
                dceout->PrintIdentifier(tag);
                tscope = scope + tag  +"::";
	    }
	}
        rname = "ANNOTATIONS";
        rgraph_1 = GetRelation(cg, rname);
        if(rgraph_1 != NULL)
	{
            cgraph_1 = rgraph_1->GetWidth();
            while(cgraph_1 != NULL)
	    {
                concept = (ConceptNode*) cgraph_1->GetNode();
                cname = concept->GetName();
                if(cname == "UNION_DISCRIMINATOR_NAME")
		{
                    instance_list = concept->GetInstances();
                    i = instance_list.begin();
                    dname = (*i);
		}
                cgraph_1 = cgraph_1->GetDepth();
	    }
	}
        else if(mode == AIT_IDL_CORBA)
            dname = name + "_disc";
        rname = "DISCRIMINATOR_TYPE";
        rgraph_1 = GetRelation(cg, rname);
        if(rgraph_1 == NULL)
	{
            rname = "CONSTRUCTED_TYPE";
            rgraph_1 = GetRelation(cg, rname);
            if(rgraph_1 == NULL)
                union_n_e = 1;
	}
        if(!(ttag == ""))
            dceout->PrintIdentifier(ttag);
        if(!union_n_e)
	{
            dceout->EnterNewLevel(switch_level);
            CreateTypes(cg, tscope, tag);
            dceout->PrintIdentifier(dname);
            dceout->LeaveLevel(switch_level);
            if(!(name == ""))
                dceout->PrintIdentifier(name);
	}
        else
            if(!(name == ""))
                dceout->PrintIdentifier(name);
        rname = "CASES";
        rgraph_1 = GetRelation(cg, rname);
        if(rgraph_1 != NULL)
	{
            cgraph_2 = rgraph_1->GetWidth();
            CreateCases(cgraph_2, scope, union_n_e);
	}
        dceout->LeaveLevel(union_level);
    }
    if(cname == "ENUM")
    {
        name_list = "";
        dceout->EnterNewLevel(enum_level);
        rname = "MEMBERS";
        rgraph_1 = GetRelation(cg, rname);
        assert (rgraph_1 != NULL);
        cgraph_1 = rgraph_1->GetWidth();
        while(cgraph_1 != NULL)		
	{
            counter++;
            cgraph_1 = cgraph_1->GetDepth();
	}
        for(j = 1; j <= counter; j++)
	{
            cgraph_1 = rgraph_1->GetWidth();
            while(cgraph_1 != NULL)
	    {
                rname = "POSITION";
                rgraph_2 = GetRelation(cgraph_1, rname);
                cgraph_2 = rgraph_2->GetWidth();
                concept = (ConceptNode*) cgraph_2->GetNode();
                pos = concept->GetValue();
                if(pos == j)
	    	{
                    concept = (ConceptNode*) cgraph_1->GetNode();
                    instance_list = concept->GetInstances();
                    i = instance_list.begin();
                    name = (*i);
                    if(!(name_list == ""))
                        name_list += ", ";
                    name_list += name;
		}
                cgraph_1 = cgraph_1->GetDepth();
	    }
	}
        dceout->PrintIdentifier(name_list);
        dceout->LeaveLevel(enum_level);
    }
    if(cname == "PIPE")
    {
        name_list = "";
        dceout->EnterNewLevel(pipe_level);
        CreateTypes(cg, scope, tag);
        instance_list = concept->GetInstances();
        i = instance_list.begin();
        while(i != instance_list.end())
	{
            name = (*i);
            if(!(name_list == ""))
                name_list += ", ";
            name_list += name;
            ++i;
	}
        if(!(name_list == ""))
            dceout->PrintIdentifier(name_list);
        else
            assert(0);
        dceout->LeaveLevel(pipe_level);	
    }
}


void Template2DCE::CreateBaseType( ConceptualGraph *cg, IDLString scope )
{
    IDLString ctype;
    ConceptNode *concept;
    SetOfWord instance_list;
    SetOfWord::iterator i;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    concept = (ConceptNode*) cg->GetNode();
    instance_list = concept->GetInstances();
    i = instance_list.begin();
    ctype = (*i);
    if(idict.ContainsWord(ctype)) 
    {
        dceout->PrintType(ctype);
    }
    else 
    {
        if (IsUnconvertableType (ctype))
            ctype += " /* NOT YET CONVERTED */";
        dceout->PrintType(ctype);
        /* A GOOD IDEA ???? */
    }
}


void Template2DCE::CreateConstExpression( ConceptualGraph *cg,
                                          IDLString scope )
{
    ConceptualGraph *rgraph;
    ConceptualGraph *cgraph;
    RelationNode *relation;
    ConceptNode *concept;
    SetOfWord instance_list;
    SetOfWord::iterator i;
    IDLString rname;
    IDLString expression;
    int string_detected = 0;

    assert(cg->GetNode()->IsA() == RELATION_NODE);
    rgraph = cg;
    while(rgraph != NULL)
    {
        relation = (RelationNode*) rgraph->GetNode();
        rname = relation->GetName();
        if(rname == "VALUE")
            cgraph = rgraph->GetWidth();
        if(rname == "TEMPLATE_TYPE")
            string_detected = 1;
        rgraph = rgraph->GetDepth();
    }
    concept = (ConceptNode*) cgraph->GetNode();
    instance_list = concept->GetInstances();
    i = instance_list.begin();
    expression = (*i);
    if(string_detected)
    {
        prepend(expression, "\"");
        expression += "\"";
    }
    dceout->PrintExpression(expression);
}


void Template2DCE::CreateTypes( ConceptualGraph *cg, IDLString scope,
                                IDLString tag )
{
    ConceptualGraph *rgraph;
    ConceptualGraph *cgraph;
    IDLString rname;
    IDLString tname;
    IDLString cname;
    IDLString tmp;
    RelationNode *relation;
    ConceptNode *concept;
    SetOfWord instance_list;
    SetOfWord::iterator i;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    rgraph = cg->GetWidth();
    while(rgraph != NULL)
    {
        cgraph = rgraph->GetWidth();
        relation = (RelationNode*) rgraph->GetNode();
        rname = relation->GetName();
        if(rname == "CONSTRUCTED_TYPE")
            CreateConstructedType(cgraph, scope, tag);
        if(rname == "TYPE")
            CreateBaseType(cgraph, scope);
        if(rname == "DISCRIMINATOR_TYPE")
            CreateBaseType(cgraph, scope);
        if(rname == "TEMPLATE_TYPE") 
	{
            tname = ConvertElementType(rgraph, scope);
            dceout->PrintType(tname);
	}
        if(rname == "RECURSIVE_TYPE")
        {
            concept = (ConceptNode*) cg->GetNode();
            cname = concept->GetName();
            if(cname == "STRUCT_MEMBER")
                tmp = "struct ";
            else
                tmp = "union ";
            concept = (ConceptNode*) cgraph->GetNode();
            instance_list = concept->GetInstances();
            i = instance_list.begin();
            tname = (*i);
            prepend(tname, tmp);
            dceout->PrintType(tname); 
        }
        rgraph = rgraph->GetDepth();	
    }
}

bool
Template2DCE::IsUnconvertableType (IDLString type)
{
    return type == "any" || type == "Object" || type == "TypeCode";
}

