
/*
 *  AI-Trader --- Service Trading using Conceptual Graphs
 *  Copyright (C) 1995 Arno Puder et al.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 aitrader@informatik.uni-frankfurt.de
 */



#include <dict.h>
#include <fstream.h>
#include <stream.h>
#include <assert.h>
#include <cg.h>
#include <stdlib.h>
#include "idlstring.h"
#include "printidl.h"
#include "cgtemplate.h"

#define AIT_IDL_DCE   1 
#define AIT_IDL_CORBA 2

#define SIMPLE 0
#define POINTER 1
#define ARRAY 2
#define FUNCTION_POINTER 3


class Template2IDL : public CGTemplate
{
protected:
    ConceptualGraph *root;
    Dict	idict;
    Dict	odict;
    set<string, less<string> > opsdone;

public:
    Template2IDL(ConceptualGraph* rg);
    ~Template2IDL(){};
    void CollectConstants(ConceptualGraph* cg, IDLString tag);
    void CollectTypes(ConceptualGraph* cg, IDLString tag, int depth = 0);
    int IsBaseType(IDLString type);
    int IsBaseValue(IDLString value);
    IDLString ClearScopedName(IDLString name, IDLString scope);
    int DetectIDL( ConceptualGraph *cg );
    int DetectDeclaratorType( ConceptualGraph *cg );
    int DetectStandardException( ConceptualGraph *cg );
};


class Template2CORBA: public Template2IDL
{
private:
    CORBA2Stream *corbaout;
    int mode;
    void CreateInterface( ConceptualGraph *cg, IDLString scope );
    void CreateModule( ConceptualGraph *cg, IDLString scope );
    void CreateDeclarations( ConceptualGraph *cg, IDLString scope );
    void CreateOperation( ConceptualGraph *cg, IDLString scope );
    void CreateAttribute( ConceptualGraph *cg_1, ConceptualGraph *cg_2,
                          IDLString scope );
    int LookForOutputParameters( ConceptualGraph *cg );
    void CreateRaisesExpression( ConceptualGraph *cg, IDLString name );
    void CreateContextExpression( ConceptualGraph *cg );
    void CreateTypes( ConceptualGraph *cg, IDLString scope, IDLString tag );
    void CreateSignature( ConceptualGraph *cg, IDLString scope );
    void CreateParameter( ConceptualGraph *cg, IDLString scope );
    void RangeTypeDeclarations( ConceptualGraph *cg_1, ConceptualGraph *cg_2, 
                                IDLString scope, IDLString type );
    void RangeConstantDeclarations( ConceptualGraph *cg_1,
                                    ConceptualGraph *cg_2, 
                                    IDLString scope, IDLString type );
    void CreateTypeDeclaration( ConceptualGraph *cg, IDLString scope, 
                                IDLString type );
    void CreateConstantDeclaration( ConceptualGraph *cg, IDLString scope, 
                                    IDLString type );
    void CreateDeclarator( ConceptualGraph *cg, IDLString scope );
    void CheckDeclarator( ConceptualGraph *cg, IDLString scope,
                          IDLString name );
    void CreateMembers( ConceptualGraph *cg, IDLString scope );
    void CreateCases( ConceptualGraph *cg, IDLString scope );
    void CreateException( ConceptualGraph *cg, IDLString scope,
                          IDLString name );
    void CreateConstructedType( ConceptualGraph *cg, IDLString scope, 
                                IDLString tag );
    void CreateBaseType( ConceptualGraph *cg, IDLString scope );
    void CreateTemplateType( ConceptualGraph *cg, IDLString scope );
    void CreateConstExpression( ConceptualGraph *cg );
    void CheckForRaisesExpression( ConceptualGraph *cg, IDLString name );
    void CheckForContextExpression( ConceptualGraph *cg );
    void RangeMembers( ConceptualGraph *cg, IDLString sname, IDLString sform,
                       IDLString scope );
    void CheckDependencies( ConceptualGraph *cg, IDLString scope,
                            IDLString name );
    void CollectDeclarations();
    void ExamineDeclarations( ConceptualGraph *cg );
    int CheckDeclaration( ConceptualGraph *cg, IDLString sname,
                          IDLString sform, IDLString scope );
    int CheckPredecessors( ConceptualGraph *cg );
    void CreateOperations( IDLString sname, IDLString sform, IDLString scope,
                           int pos );
    void CreateDeclarations( IDLString sname, IDLString sform,
                             IDLString scope );
    ConceptualGraph* GetDeclarationPosition( IDLString declarator,
                                             IDLString sname, 
                                             IDLString sform,
                                             IDLString scope );
    ConceptualGraph* GetPosition( ConceptualGraph *cg, IDLString declarator, 
                                  IDLString sname, IDLString sform,
                                  IDLString scope );
    void CreateDeclaration( ConceptualGraph *cg, IDLString declarator,
                            IDLString scope );
    
public:
    Template2CORBA( ConceptualGraph *cg, ostream& o ) : Template2IDL(cg) 
    {
        corbaout = new CORBA2Stream(o, 1);
        mode = 0;
    }
    ~Template2CORBA() { delete corbaout; }
    void CreateSpecification();
};


class Template2DCE: public Template2IDL
{
private:
    DCE2Stream *dceout;
    int mode;
public:
    Template2DCE( ConceptualGraph *cg, ostream& o ) : Template2IDL(cg) 
    { 
        dceout = new DCE2Stream(o, 1);
        mode = 0;
    }
    ~Template2DCE() { delete dceout; }
    void CreateSpecification();
    void CreateInterface( ConceptualGraph *cg, IDLString scope );
    void CreateDeclarations( ConceptualGraph *cg, IDLString scope );
    void CreateOperation( ConceptualGraph *cg, IDLString scope );
    void CreateTypes( ConceptualGraph *cg, IDLString scope, IDLString tag );
    void CreateSignature( ConceptualGraph *cg, IDLString scope );
    void RangeTypeDeclarations( ConceptualGraph *cg_1, ConceptualGraph *cg_2, 
                                IDLString scope, IDLString type );
    void RangeConstantDeclarations( ConceptualGraph *cg_1,
                                    ConceptualGraph *cg_2, 
                                    IDLString scope, IDLString type );
    void CreateTypeDeclaration( ConceptualGraph *cg, IDLString scope,
                                IDLString type );
    void CreateConstantDeclaration( ConceptualGraph *cg, IDLString scope,
                                    IDLString type );
    void CreateDeclaration( ConceptualGraph *cg, IDLString scope,
                            IDLString type );
    void CreatePointerDeclaration( ConceptualGraph *cg, IDLString attributes, 
                                   IDLString scope, IDLString name );
    void CreateArrayDeclaration( ConceptualGraph *cg, IDLString attributes, 
                                 IDLString scope, IDLString name );
    void CreateFunctionPointerDeclaration( ConceptualGraph *cg,
                                           IDLString attributes, 
                                           IDLString scope, IDLString name );
    void CreateCases( ConceptualGraph *cg, IDLString scope, int union_n_e );
    void CreateMembers( ConceptualGraph *cg, IDLString scope );
    void CreateConstructedType( ConceptualGraph *cg, IDLString scope,
                                IDLString tag );
    void CreateBaseType( ConceptualGraph *cg, IDLString scope );
    void CreateInterfaceAttributes( ConceptualGraph *cg );
    IDLString ConvertElementType( ConceptualGraph *cg, IDLString scope );
    void CreateConstExpression( ConceptualGraph *cg, IDLString scope );
    IDLString CompareType( IDLString type );
    bool IsUnconvertableType (IDLString type);
};
