/*
 *  AI-Trader --- Service Trading using Conceptual Graphs
 *  Copyright (C) 1995 Arno Puder et al.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 aitrader@informatik.uni-frankfurt.de
 */


#include <assert.h>
#include <stream.h>
#include "cgtransform.h"
#include "util.h"

/*
 * XXX in diesem File gibts an einigen Stellen noch Probleme mit dem
 * ``Umhaengen'' von Teilbauemen. Dabei wird u.U. beim SetWidth() bzw.
 * SetDepth() ein Teilbaum geloescht, auf den spaeter noch zugegriffen
 * wird.
 */

IDLString
ConceptualGraph2Template::EncodeIfaceName (IDLString name, int pos)
{
    assert (freq (name, "#") == 0);
    return name + "#" + dec (pos);
}

IDLString
ConceptualGraph2Template::DecodeIfaceName (IDLString name, int &pos)
{
    assert (freq (name, "#") == 1);
    IDLString s = before (name, "#");
    pos = atoi (after(name, "#").c_str());
    return s;
}

vector<string>
ConceptualGraph2Template::SortIfaceNames (const Dict::vcwp &iflist)
#ifdef __GNUG__
    return sorted;
#endif
{
#ifndef __GNUG__
    vector<string> sorted;
#endif
    Dict::vcwp::const_iterator j;

    sorted.insert (sorted.end(), iflist.size(), string());
    for (j = iflist.begin(); j != iflist.end(); ++j) {
        int pos;
        string iface = DecodeIfaceName (**j, pos);
        assert (pos < sorted.size());
        sorted[pos] = iface;
    }
#ifndef __GNUG__
    return sorted;
#endif
}

ConceptualGraph2Template::ConceptualGraph2Template( ConceptualGraph *cg )
{
    ConceptNode *concept;
    IDLString name;
    IDLString scope;
   
    root = cg;
    idict.RegisterRelation("MEMBER_OF",
                           Dict::RelReflexive |
                           Dict::RelTransitive |
                           Dict::RelAcyclic);
    idict.RegisterRelation("INHERITS_FROM",
                           Dict::RelReflexive |
                           Dict::RelTransitive |
                           Dict::RelAcyclic);
    idict.RegisterRelation("PREDECESSOR_OF",
                           Dict::RelReflexive |
                           Dict::RelTransitive |
                           Dict::RelAcyclic);
    idict.RegisterRelation("DEPENDS_ON",
                           Dict::RelReflexive |
                           Dict::RelTransitive);
    ddict.RegisterRelation("INHERITS_FROM",
                           Dict::RelReflexive |
                           Dict::RelTransitive |
                           Dict::RelAcyclic);
    idict.AddWord("*BASE*");
    odict.AddWord("*BASE*");
    ddict.AddWord("*BASE*");
    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    concept = (ConceptNode*) cg->GetNode();
    assert(string(concept->GetName()) == string("SPECIFICATION"));
    name = "*BASE*";
    scope = "";
    LookUpModuleSpecification(cg, scope, name);
}


ConceptualGraph2Template::~ConceptualGraph2Template()
{
}


void ConceptualGraph2Template::Print( ostream &ostr )
{
    root->Print(ostr);
}


void ConceptualGraph2Template::LookUpModuleSpecification( ConceptualGraph *cg, 
                                                          IDLString scope, 
                                                          IDLString name )
{
    ConceptualGraph *rgraph;
    ConceptualGraph *graph_1;
    ConceptualGraph *graph_2;
    ConceptNode *concept;
    IDLString instance;
    IDLString sinstance;
    IDLString tscope;
    IDLString scoped_name;
    IDLString cname;
    IDLString rname;
    IDLString sname[MAXN];
    IDLString sscope;
    SetOfWord instance_list;
    SetOfWord::iterator i;
    int maxnr;

    scoped_name = GetScopedName(scope, name);
    tscope = ActualizeScope(scope, name);
    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    if(name == "*BASE*")
        rname = "DEFINITIONS";
    else
        rname = "SCOPE";
    rgraph = GetRelation(cg, rname);
    assert(rgraph != NULL);
    graph_1 = rgraph;
    graph_2 = graph_1->GetWidth();
    while(graph_2 != NULL)
    {
        concept = (ConceptNode*) graph_2->GetNode();
        cname = concept->GetName();
        if(cname == "INTERFACE")
        {
            instance_list = concept->GetInstances();
            i = instance_list.begin();
            instance = (*i);
            MaintainNameScopeDictionary(tscope, scoped_name, instance);
            if(contains(tscope, "::"))
            {
                maxnr = split(tscope, sname, MAXN, "::");
                rname = "DEFINITIONS";
                rgraph = GetRelation(root, rname);
                assert(rgraph != NULL);
                sinstance = GetScopedInterfaceName(rgraph, sname, 0,
                                                   maxnr - 1, instance);
                assert(!(sinstance == ""));
            }
            else
                sinstance = instance;
            DetectGlobalInterfaceTypeUsage(cg, instance, sinstance,
                                           name, scope);
            sscope = "";
            rname = "*BASE*";	
            DetectGlobalInterfaceTypeUsage(root, sinstance, sinstance,
                                           rname, sscope);
            LookUpInterfaceSpecification(graph_2, tscope, instance);
        }
        if(cname == "MODULE")
        {
            instance_list = concept->GetInstances();
            i = instance_list.begin();
            instance = (*i);
            MaintainNameScopeDictionary(tscope, scoped_name, instance);
            LookUpModuleSpecification(graph_2, tscope, instance);
        }
        if(cname == "TYPE" || cname == "CONSTANT")
        {
            CollectScopedNames(cg, graph_2);
        }
        graph_1 = graph_2;
        graph_2 = graph_2->GetDepth();
    }
}


void
ConceptualGraph2Template::DetectGlobalInterfaceTypeUsage( ConceptualGraph *cg, 
                                                          IDLString interface,
                                                          IDLString sinterface,
                                                          IDLString name,
                                                          IDLString scope )
{
    ConceptualGraph *rgraph;
    ConceptualGraph *graph_1;
    ConceptualGraph *graph_2;
    ConceptNode *concept;
    SetOfWord instance_list;
    IDLString instance;
    IDLString cname;
    IDLString rname;
    IDLString tscope;
    IDLString scoped_name;
    SetOfWord::iterator i;
    int detected = 0;
    int det = 0;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    tscope = ActualizeScope(scope, name);
    if(name == "*BASE*")
        rname = "DEFINITIONS";
    else
        rname = "SCOPE";
    rgraph = GetRelation(cg, rname);
    assert(rgraph != NULL);
    graph_1 = rgraph->GetWidth();
    while(graph_1 != NULL)
    {
        concept = (ConceptNode*) graph_1->GetNode();
        cname = concept->GetName();
        if(cname == "INTERFACE")
	{
            instance_list = concept->GetInstances();
            i = instance_list.begin();
            assert(i != instance_list.end());
            instance = (*i);
            if(!(instance == interface))
                det = DetectLocalInterfaceTypeUsage(graph_1, interface,
                                                    instance);
            if(det) 
	    {
                detected = 1;
                MaintainDependencyDictionary(tscope, interface, instance);
	    }
	}
        if(cname == "MODULE")
	{
            instance_list = concept->GetInstances();
            i = instance_list.begin();
            assert(i != instance_list.end());
            instance = (*i);
            DetectGlobalInterfaceTypeUsage(graph_1, interface, sinterface,
                                           instance, tscope);
	}
        if(cname == "TYPE" || cname == "CONSTANT")
	{
            det = SubstituteInterfaceTypeUsage(graph_1, interface);
            if(det)
                detected = 1;
	}
        graph_1 = graph_1->GetDepth();
    }
    if(detected)
        DefineInterfaceType(rgraph, interface, name, tscope);
} 


void ConceptualGraph2Template::DefineInterfaceType( ConceptualGraph *cg,
                                                    IDLString interface,
                                                    IDLString name,
                                                    IDLString scope )
{
    ConceptualGraph *graph;
    ConceptualGraph *cgraph;
    ConceptualGraph *cgraph_1;
    ConceptualGraph *cgraph_2;
    ConceptualGraph *cgraph_3;
    ConceptualGraph *cgraph_4;
    ConceptualGraph *igraph;
    ConceptualGraph *rgraph_1;
    ConceptualGraph *rgraph_2;
    IDLString instance;
    IDLString iname;
    IDLString rname;
    IDLString cname;

    assert(cg->GetNode()->IsA() == RELATION_NODE);
    igraph = cg;
    graph = cg->GetWidth();
    while(graph != NULL)
    {
        igraph = graph;
        graph = graph->GetDepth();
    }
    iname = interface;
    gsub(iname, "::", "_");
    iname += "_interface_type";
    cname = "TYPE";
    cgraph = CreateNewConceptNode(cname, iname);
    assert(cgraph != NULL);
    rname = "ANNOTATIONS"; 
    rgraph_1 = CreateNewRelationNode(rname);
    assert(rgraph_1 != NULL);
    cname = "SCOPE";
    cgraph_1 = CreateNewConceptNode(cname, scope);
    assert(cgraph_1 != NULL);
    cname = "SOURCE_NAME";
    cgraph_2 = CreateNewConceptNode(cname, name);
    assert(cgraph_2 != NULL);
    cname = "SOURCE_FORM";
    instance = "MODULE";
    cgraph_3 = CreateNewConceptNode(cname, instance);
    assert(cgraph_3 != NULL);
    cname = "ALIAS";
    cgraph_4 = CreateNewConceptNode(cname, interface);
    assert(cgraph_4 != NULL);  
    cgraph_3->SetDepth(cgraph_4);
    cgraph_2->SetDepth(cgraph_3);
    cgraph_1->SetDepth(cgraph_2);
    rgraph_1->SetWidth(cgraph_1);
    rname = "TYPE";
    rgraph_2 = CreateNewRelationNode(rname);
    assert(rgraph_2 != NULL);
    cname = "TYPE_TYPE";
    instance = "byte";
    cgraph_4 = CreateNewConceptNode(cname, instance);
    assert(cgraph_4 != NULL);
    rgraph_2->SetWidth(cgraph_4);
    rgraph_1->SetDepth(rgraph_2);
    cgraph->SetWidth(rgraph_1);
    if(igraph->GetNode()->IsA() == RELATION_NODE)
        igraph->SetWidth(cgraph);
    else
        igraph->SetDepth(cgraph);
}


void ConceptualGraph2Template::CreateDependencyGraph()
{
    ConceptualGraph *graph_1;
    ConceptualGraph *graph_2;
    ConceptualGraph *cgraph;
    ConceptualGraph *rgraph;
    IDLString rname;
    IDLString cname;
    IDLString instance;

    rname = "ANNOTATIONS";
    rgraph = GetRelation(root, rname);
    assert(rgraph == NULL); 
    graph_1 = root;
    rgraph = root->GetWidth();
    while(rgraph != NULL)
    {
        graph_1 = rgraph;
        rgraph = rgraph->GetDepth();
    }
    rgraph = CreateNewRelationNode(rname);
    assert(rgraph != NULL);
    if(graph_1->GetNode()->IsA() == CONCEPT_NODE)
        graph_1->SetWidth(rgraph);
    else
        graph_1->SetDepth(rgraph);
    cname = "HIERARCHY";
    instance = "*NONAME*";
    cgraph = CreateNewConceptNode(cname, instance);
    graph_2 = rgraph->GetWidth();
    if(graph_2 == NULL)
        graph_1 = rgraph; 
    else 
    {
        while(graph_2 != NULL)
	{
            graph_1 = graph_2;
            graph_2 = graph_2->GetDepth();
	}
    }  
    if(graph_1->GetNode()->IsA() == RELATION_NODE)
        graph_1->SetWidth(cgraph);
    else
        graph_1->SetDepth(cgraph);
    cname = "*BASE*";
    CreateDependencySubGraph(cgraph, cname);
}


void ConceptualGraph2Template::CreateDependencySubGraph( ConceptualGraph *cg, 
                                                         IDLString base )
{
    ConceptualGraph *graph;
    ConceptualGraph *graph_1;
    ConceptualGraph *rgraph_4;
    ConceptualGraph *rgraph_3;
    ConceptualGraph *rgraph_2;
    ConceptualGraph *rgraph_1;
    ConceptualGraph *cgraph;
    IDLString cname;
    IDLString rname;
    IDLString member;
    IDLString interface;
    IDLString predecessor;
    IDLString dependant;
    Dict::vcwp::const_iterator i;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    graph = cg;
    Dict::vcwp member_list = idict.GetRelImmPred(base, "MEMBER_OF");
    i = member_list.begin();
    if(i != member_list.end())
    {
        rname = "MEMBERS";
        graph_1 = graph;
        rgraph_1 = CreateNewRelationNode(rname);
        graph_1->SetWidth(rgraph_1);
        graph_1 = rgraph_1;
        graph = graph_1;
    }
    while(i != member_list.end())
    {
        member = **i;
        cname = GetMemberForm(member);
        cgraph = CreateNewConceptNode(cname, member);
        assert(cgraph != NULL);
        CreateDependencySubGraph(cgraph, member); 
        if(graph_1->GetNode()->IsA() == RELATION_NODE)
            graph_1->SetWidth(cgraph);
        else
            graph_1->SetDepth(cgraph);
        graph_1 = cgraph;
        ++i;
    }

    Dict::vcwp interface_list = idict.GetRelImmSucc(base, "INHERITS_FROM");
    vector<string> sorted = SortIfaceNames (interface_list);
    vector<string>::const_iterator j;

    j = sorted.begin();
    if(j != sorted.end())
    {
        rname = "SUPERINTERFACES";
        graph_1 = graph;
        rgraph_2 = CreateNewRelationNode(rname);
        if(graph_1->GetNode()->IsA() == CONCEPT_NODE)
            graph_1->SetWidth(rgraph_2);
        else
            graph_1->SetDepth(rgraph_2);
        graph_1 = rgraph_2;
        graph = graph_1;
        cname = "INTERFACE";
    }
    while(j != sorted.end())
    {
        interface = *j;
        cgraph = CreateNewConceptNode(cname, interface);
        if(graph_1->GetNode()->IsA() == RELATION_NODE)
            graph_1->SetWidth(cgraph);
        else
            graph_1->SetDepth(cgraph);
        graph_1 = cgraph;
        ++j;
    }
    Dict::vcwp predecessor_list = idict.GetRelAllPred(base, "PREDECESSOR_OF");
    i = predecessor_list.begin();
    if(i != predecessor_list.end())
    {
        rname = "PREDECESSORS";
        graph_1 = graph;
        rgraph_3 = CreateNewRelationNode(rname);
        if(graph_1->GetNode()->IsA() == CONCEPT_NODE)
            graph_1->SetWidth(rgraph_3);
        else
            graph_1->SetDepth(rgraph_3);
        graph_1 = rgraph_3;
        graph = graph_1;
    }
    while(i != predecessor_list.end())
    {
        predecessor = **i;
        cname = GetMemberForm(predecessor);
        cgraph = CreateNewConceptNode(cname, predecessor);
        if(graph_1->GetNode()->IsA() == RELATION_NODE)
            graph_1->SetWidth(cgraph);
        else
            graph_1->SetDepth(cgraph);
        graph_1 = cgraph;
        ++i;
    }
    Dict::vcwp dependant_list = idict.GetRelImmSucc(base, "DEPENDS_ON");
    i = dependant_list.begin();
    if(i != dependant_list.end())
    {
        rname = "DEPENDANTS";
        graph_1 = graph;
        rgraph_4 = CreateNewRelationNode(rname);
        if(graph_1->GetNode()->IsA() == CONCEPT_NODE)
            graph_1->SetWidth(rgraph_4);
        else
            graph_1->SetDepth(rgraph_4);
        graph_1 = rgraph_4;
        graph = graph_1;
        cname = "INTERFACE";
    }
    while(i != dependant_list.end())
    {
        dependant = **i;
        cgraph = CreateNewConceptNode(cname, dependant);
        if(graph_1->GetNode()->IsA() == RELATION_NODE)
            graph_1->SetWidth(cgraph);
        else
            graph_1->SetDepth(cgraph);
        graph_1 = cgraph;
        ++i;
    }
}


IDLString ConceptualGraph2Template::GetMemberForm( IDLString name )
{
    ConceptualGraph *rgraph;
    ConceptualGraph *cgraph;
    ConceptualGraph *graph;
    ConceptNode *concept;
    IDLString cname;
    IDLString rname;
    IDLString scope[MAXN];
    int maxnr;

    maxnr = split(name, scope, MAXN, "::");
    rname = "DEFINITIONS";
    rgraph = GetRelation(root, rname);
    assert(rgraph != NULL);
    graph = GetInstance(rgraph, scope, 0, maxnr -1);
    if(graph->GetNode()->IsA() == RELATION_NODE)
        cgraph = graph->GetWidth();
    else
        cgraph = graph->GetDepth();
    concept = (ConceptNode*) cgraph->GetNode();
    cname = concept->GetName();
    return(cname);
}


int
ConceptualGraph2Template::SubstituteInterfaceTypeUsage(ConceptualGraph *cg,
                                                       IDLString name )
{
    ConceptualGraph *graph_1;
    ConceptualGraph *graph_2;
    ConceptualGraph *graph_3;
    ConceptualGraph *graph_4;
    ConceptNode *concept;
    RelationNode *relation;
    SetOfWord instance_list;
    IDLString instance;
    IDLString rname;
    IDLString cname;
    SetOfWord::iterator i;
    int detected = 0;
    int det = 0;
    int modified = 0;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    graph_1 = cg->GetWidth();
    while(graph_1 != NULL && !modified)
    {
        relation = (RelationNode*) graph_1->GetNode();
        rname = relation->GetName();
        if(rname == "TYPE" || rname == "DISCRIMINATOR_TYPE")
	{
            graph_2 = graph_1->GetWidth();
            concept = (ConceptNode*) graph_2->GetNode();
            instance_list = concept->GetInstances();
            i = instance_list.begin();
            instance = (*i);
            if(instance == name) 
	    {
                ExchangeInterfaceType(cg, name);
                detected = 1;
                modified = 1;
                /* ZZZ */
	    }
	}
        if(rname == "CONSTRUCTED_TYPE")
	{
            graph_2 = graph_1->GetWidth();
            concept = (ConceptNode*) graph_2->GetNode();
            cname = concept->GetName();
            if(cname == "UNION")
	    {
                det = SubstituteInterfaceTypeUsage(graph_2, name);
                if(det)
                    detected = 1;
                rname = "CASES";
                graph_3 = GetRelation(graph_2, rname);
                assert(graph_3 != NULL); 
                graph_3 = graph_3->GetWidth();
                while(graph_3 != NULL)
		{
                    rname = "MEMBERS";
                    graph_4 = GetRelation(graph_3, rname);
                    assert(graph_4 != NULL);
                    graph_4 = graph_4->GetWidth();
                    while(graph_4 != NULL)
		    {
                        det = SubstituteInterfaceTypeUsage(graph_4, name);
                        if(det)
                            detected = 1;
                        graph_4 = graph_4->GetDepth();
		    }
                    graph_3 = graph_3->GetDepth();
		}
	    }
            if(cname == "STRUCT")
	    {
                rname = "MEMBERS";
                graph_3 = GetRelation(graph_2, rname);
                if(graph_3 != NULL)
                    graph_3 = graph_3->GetWidth();
                while(graph_3 != NULL)
		{
                    det = SubstituteInterfaceTypeUsage(graph_3, name);
                    if(det)
                        detected = 1;
                    graph_3 = graph_3->GetDepth();
		}
	    }
            if(cname == "ARRAY")
	    {
                det = SubstituteInterfaceTypeUsage(graph_2, name);
                if(det)
                    detected = 1;
	    }
            if(cname == "POINTER")
	    {
                det = SubstituteInterfaceTypeUsage(graph_2, name);
                if(det)
                    /* Keine Modifikation, da nur Pointer
                     * auf _Except-Typen auftreten!!!
                     */ 
                    detected = 1;
	    }
	}
        if(rname == "TEMPLATE_TYPE")
	{
            graph_2 = graph_1->GetWidth();
            while(graph_2 != NULL)
	    {
                concept = (ConceptNode*) graph_2->GetNode();
                cname = concept->GetName();
                if(cname == "ELEMENT_TYPE")
		{
                    if(graph_2->GetWidth() != NULL)
		    {
                        det = SubstituteInterfaceTypeUsage(graph_2, name);
                        if(det)
                            detected = 1;
		    }
                    else
		    {
                        instance_list = concept->GetInstances();
                        i = instance_list.begin();
                        instance = (*i);
                        if(instance == name) 
			{
                            ModifyAndExchangeInterfaceType(graph_2,
                                                           &instance_list,
                                                           name);
                            /*
                             * XXX
                             * instance_list ist jetzt kein Pointer auf
                             * die Instanzliste des cg mehr! Muss jetzt
                             * von Hand neu gesetzt werden.
                             */
                            concept->SetInstances (instance_list);
                            detected = 1;
                            modified = 1;
			}
		    }
		}
                graph_2 = graph_2->GetDepth();
	    }
	}
        /* ZZZ */
        if (!modified)
            graph_1 = graph_1->GetDepth();
    }
    return(detected);
}


void ConceptualGraph2Template::ExchangeInterfaceType( ConceptualGraph *cg, 
                                                      IDLString name )
{
    ConceptualGraph *rgraph_1;
    ConceptualGraph *rgraph_2;
    ConceptualGraph *rgraph_3;
    ConceptualGraph *cgraph_1;
    ConceptualGraph *cgraph_2;
    ConceptualGraph *graph_1;
    ConceptualGraph *graph_2;
    RelationNode *relation;
    IDLString rname;
    IDLString cname;
    IDLString instance;
    int detected;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    rname = "ANNOTATIONS";
    graph_1 = GetRelation(cg, rname);
    if(graph_1 != NULL)
    {
        cgraph_1 = graph_1->GetWidth();
        while(cgraph_1 != NULL)
	{
            graph_1 = cgraph_1;
            cgraph_1 = cgraph_1->GetDepth();
	}
    }
    else
        graph_1 = CreateNewRelationNode(rname);
    cname = "INTERFACE_TYPE";
    cgraph_1 = CreateNewConceptNode(cname, name);
    if(graph_1->GetNode()->IsA() == RELATION_NODE)
        graph_1->SetWidth(cgraph_1);
    else
        graph_1->SetDepth(cgraph_1);
    rname = "TEMPLATE_TYPE";
    rgraph_1 = CreateNewRelationNode(rname);
    cname = "SEQUENCE";
    instance = "*NONAME*";
    cgraph_1 = CreateNewConceptNode(cname, instance);
    cname = "ELEMENT_TYPE";
    instance = name;
    gsub(instance, "::","_");
    instance += "_interface_type";
    cgraph_2 = CreateNewConceptNode(cname, instance);
    cgraph_1->SetDepth(cgraph_2);
    rgraph_1->SetWidth(cgraph_1);
    rgraph_2 = cg->GetWidth();
    assert(rgraph_2 != NULL);
    graph_2 = cg;
    detected = 0;
    while(!detected)
    {
        relation = (RelationNode*) rgraph_2->GetNode();
        rname = relation->GetName();
        if(rname == "TYPE" || rname == "DISCRIMINATOR_TYPE")
	{
            rgraph_3 = CopyGraph(rgraph_2->GetDepth());
            rgraph_1->SetDepth(rgraph_3);
            if(graph_2->GetNode()->IsA() == CONCEPT_NODE)
                graph_2->SetWidth(rgraph_1);
            else
                graph_2->SetDepth(rgraph_1);
            detected = 1; 
	}
        if(!detected)
	{
            graph_2 = rgraph_2;
            rgraph_2 = rgraph_2->GetDepth();
	}
    }
    if(graph_1->GetNode()->IsA() == RELATION_NODE)
    { 
        graph_2 = cg;
        rgraph_2 = cg->GetWidth();
        while(rgraph_2 != NULL)
	{
            graph_2 = rgraph_2;
            rgraph_2 = rgraph_2->GetDepth();
	}
        if(graph_2->GetNode()->IsA() == RELATION_NODE)
            graph_2->SetDepth(graph_1);
        else
            graph_2->SetWidth(graph_1);
    }
}


void
ConceptualGraph2Template::ModifyAndExchangeInterfaceType( ConceptualGraph *cg,
                                                          SetOfWord *inst,
                                                          IDLString name )
{
    ConceptualGraph *rgraph_1;
    ConceptualGraph *rgraph_2;
    ConceptualGraph *cgraph_1;
    ConceptualGraph *cgraph_2;
    IDLString rname;
    IDLString cname;
    IDLString instance;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    assert(cg->GetWidth() == NULL);
    inst->erase(inst->begin(), inst->end());
    instance = name;
    gsub(instance, "::", "_");
    instance += "_interface_type";
    instance = "sequence" + instance;
    instance += ">";
    inst->insert(Word (instance, false));
    rname = "ANNOTATIONS";
    rgraph_1 = CreateNewRelationNode(rname);
    cname = "INTERFACE_TYPE";
    cgraph_1 = CreateNewConceptNode(cname, name);
    rgraph_1->SetWidth(cgraph_1);
    rname = "TEMPLATE_TYPE";
    rgraph_2 = CreateNewRelationNode(rname);
    cname = "SEQUENCE";
    instance = "*NONAME*";
    cgraph_1 = CreateNewConceptNode(cname, instance);
    cname = "ELEMENT_TYPE";
    instance = name;
    gsub(instance, "::","_");
    instance += "_interface_type";
    cgraph_2 = CreateNewConceptNode(cname, instance);
    cgraph_1->SetDepth(cgraph_2);
    rgraph_2->SetWidth(cgraph_1);
    rgraph_1->SetDepth(rgraph_2);
    cg->SetWidth(rgraph_1);
}


void
ConceptualGraph2Template::LookUpInterfaceSpecification( ConceptualGraph *cg,
                                                        IDLString scope, 
                                                        IDLString name )
{
    ConceptualGraph *graph_1;
    ConceptualGraph *graph_2;
    ConceptualGraph *graph_3;
    ConceptualGraph *graph_4;
    ConceptualGraph *graph_5;
    ConceptualGraph *sgraph;
    ConceptNode *concept;
    RelationNode *relation;
    IDLString instance;
    IDLString tscope;
    IDLString scoped_name;
    IDLString rname_1;
    IDLString rname_2;
    SetOfWord instance_list;
    SetOfWord::iterator i;

    scoped_name = GetScopedName(scope, name);
    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    graph_1 = cg;
    graph_2 = graph_1->GetWidth();
    while(graph_2 != NULL)
    {
        relation = (RelationNode*) graph_2->GetNode();
        rname_1 = relation->GetName();
        if(rname_1 == "SUPERINTERFACES")
	{
            graph_3 = graph_2->GetWidth();
            int pos = 0;
            while(graph_3 != NULL)
 	    {
                concept = (ConceptNode*) graph_3->GetNode();
                instance_list = concept->GetInstances();
                i = instance_list.begin();
                assert (i != instance_list.end());
                instance = (*i);
                MaintainInheritanceDictionary(scope, scoped_name,
                                              instance, pos++);
                graph_3 = graph_3->GetDepth();
   	    }
            sgraph = CopyGraph(graph_2->GetDepth());
            if(graph_1->GetNode()->IsA() == CONCEPT_NODE) {
                graph_1->SetWidth(sgraph);
                /* XXX */
                graph_2 = graph_1->GetWidth();
            } else {
                graph_1->SetDepth(sgraph);
                /*
                 * XXX wegen graph_1->GetDepth() == graph_2
                 * zeigt nach dem graph_1->SetDepth()
                 * graph_2 auf freigegebenen Speicher!
                 */
                graph_2 = graph_1->GetDepth();
            }
            /* XXX */
            continue;
	}
        if(rname_1 == "TYPES" || rname_1 == "CONSTANTS")
	{
            graph_3 = graph_2->GetWidth();
            while(graph_3 != NULL)
 	    {
                CollectScopedNames(graph_1, graph_3);
                graph_3 = graph_3->GetDepth();
   	    }
	}
        if(rname_1 == "OPERATIONS")
	{
            graph_3 = graph_2->GetWidth();
            while(graph_3 != NULL)
 	    {
                CollectScopedNames(graph_1, graph_3);
                rname_2 = "SIGNATURE";
                graph_4 = GetRelation(graph_3, rname_2);
                if(graph_4 != NULL)
                    graph_5 = graph_4->GetWidth();
                while(graph_5 != NULL)
		{
                    CollectScopedNames(graph_4, graph_5);
                    graph_4 = graph_5;
                    graph_5 = graph_5->GetDepth();
		}
                graph_3 = graph_3->GetDepth();
   	    }
	}
        graph_1 = graph_2;
        graph_2 = graph_2->GetDepth();
    }
}


IDLString ConceptualGraph2Template::GetScopedName(IDLString scope, 
                                                  IDLString name ) 
{
    IDLString base;

    base = name;
    if(!(base == "*BASE*") && !(scope == ""))
        base = scope + base;
    return(base);
}


IDLString ConceptualGraph2Template::ActualizeScope( IDLString scope,
                                                    IDLString name )
{
    IDLString base;

    if(!(name == "*BASE*"))
        base = name + "::";
    else
        base = "";
    if(!(scope == ""))
        base = scope + base;
    return(base);
}


void ConceptualGraph2Template::MaintainNameScopeDictionary( IDLString scope,
                                                            IDLString module,
                                                            IDLString member )
{
    IDLString name;
    IDLString predecessor;
    int success;
    Dict::vcwp::const_iterator i;
  
    name = GetScopedName(scope, member); 
    if(!idict.ContainsWord(name))
        idict.AddWord(name);
    success = idict.AddRelation(name, "MEMBER_OF", module);
    Dict::vcwp predecessor_list = idict.GetRelImmPred(module, "MEMBER_OF");
    i = predecessor_list.begin();
    while(i != predecessor_list.end())
    {
        predecessor = **i;
        if(!(name == predecessor))
            success = idict.AddRelation(predecessor, "PREDECESSOR_OF", name);
        assert(success);
        ++i;
    }
    assert(success);
}


void
ConceptualGraph2Template::MaintainInheritanceDictionary(
    IDLString scope,
    IDLString interface,
    IDLString superinterface,
    int position)
{
    ConceptualGraph *graph;
    IDLString module[MAXN];
    IDLString name;
    IDLString rname;
    int maxnr;
    int success;

    if(!contains(superinterface, "::"))
    {
        if(contains(scope, "::"))
	{
            maxnr = split(scope, module, MAXN, "::");
            rname = "DEFINITIONS";
            graph = GetRelation(root, rname);
            name = GetScopedInterfaceName(graph, module, 0, maxnr - 1,
                                          superinterface);
            assert(!(name == ""));
        }
        else
            name = superinterface;
    }
    else
        name = superinterface;
    if (!ddict.ContainsWord (interface))
        ddict.AddWord (interface);
    if (!ddict.ContainsWord(name))
        ddict.AddWord(name);
    success = ddict.AddRelation(interface, "INHERITS_FROM", name);
    assert(success);


    name = EncodeIfaceName (name, position);

    if (!idict.ContainsWord(name))
        idict.AddWord(name);
    success = idict.AddRelation(interface, "INHERITS_FROM", name);
    assert(success);
}


void
ConceptualGraph2Template::MaintainDependencyDictionary( IDLString scope,
                                                        IDLString interface,
                                                        IDLString dependant )
{
    ConceptualGraph *graph;
    IDLString module[MAXN];
    IDLString name;
    IDLString rname;
    int maxnr;
    int success;

    if(!contains(dependant, "::"))
    {
        if(contains(scope, "::"))
	{
            maxnr = split(scope, module, MAXN, "::");
            rname = "DEFINITIONS";
            graph = GetRelation(root, rname);
            name = GetScopedInterfaceName(graph, module, 0, maxnr - 1,
                                          dependant);
            assert (!(name == ""));
        }
        else
            name = dependant;
    }
    else
        name = dependant;
    if(!idict.ContainsWord(name))
        idict.AddWord(name);
    if(!idict.ContainsWord(interface))
        idict.AddWord(interface);
    success = idict.AddRelation(name, "DEPENDS_ON", interface);
    assert(success);
}


IDLString
ConceptualGraph2Template::GetScopedInterfaceName( ConceptualGraph *cg,
                                                  IDLString scope[],
                                                  int nr, int maxnr,
                                                  IDLString interface )	
{
    ConceptualGraph *graph_1;
    ConceptualGraph *graph_2;
    ConceptualGraph *rgraph;
    IDLString name;
    IDLString cname;
    IDLString rname;
    IDLString result;
 
    assert(cg->GetNode()->IsA() == RELATION_NODE);
    result = "";
    name = scope[nr];
    gsub(name, "::", "");
    cname = "MODULE";
    graph_1 = GetConcept(cg, cname , name);
    assert(graph_1 != NULL);
    rname = "SCOPE";
    graph_2 = GetRelation(graph_1, rname);
    assert(graph_2 != NULL);
    if(nr < maxnr) 
    {
        result = GetScopedInterfaceName(graph_2, scope, ++nr, maxnr,
                                        interface);
        if(result == "")
        {
            cname = "INTERFACE";
            rgraph = GetConcept(graph_2, cname, interface);
            if(rgraph != NULL)
                result = interface;
        } 
    }
    else
    {
        cname = "INTERFACE";
        rgraph = GetConcept(graph_2, cname, interface);
        if(rgraph != NULL)
            result = interface;
    }
    if(!(result == string("")))
    {
        name += "::";
        result = name + result;
    }
    return(result);
}


void ConceptualGraph2Template::CreateLeanStructure()
{
    IDLString base;

    base = "*BASE*";
    RangeMembershipStructure(base);
}


void ConceptualGraph2Template::RangeMembershipStructure( IDLString base )
{
    IDLString member;
    Dict::vcwp::const_iterator i;

    Dict::vcwp member_list = idict.GetRelImmPred(base, "MEMBER_OF");
    i = member_list.begin();
    while(i != member_list.end())
    {
        member = **i;
        if(!odict.ContainsWord(member))
            RangeInheritanceStructure(member);
        ++i;
    }  
}


void ConceptualGraph2Template::RangeInheritanceStructure( IDLString base )
{
    IDLString member;
    IDLString interface;
    Dict::vcwp::const_iterator j;
    vector<string>::const_iterator i;

    Dict::vcwp interface_list = idict.GetRelImmSucc(base, "INHERITS_FROM");
    vector<string> sorted = SortIfaceNames (interface_list);

    i = sorted.begin();
    while(i != sorted.end())
    {
        interface = *i;
        if(!odict.ContainsWord(interface))
            RangeInheritanceStructure(interface);
        ++i;
    }
    Dict::vcwp member_list = idict.GetRelImmSucc(base, "MEMBER_OF");
    j = member_list.begin();
    assert(j != member_list.end());
    member = **j;
    if(!odict.ContainsWord(member))
        RangeInheritanceStructure(member);
    if(!odict.ContainsWord(base))
        ResolveHierarchy(base);
}


void ConceptualGraph2Template::ResolveHierarchy( IDLString sname )
{
    ConceptualGraph *graph_1;
    ConceptualGraph *graph_2;
    ConceptualGraph *snode;
    ConceptualGraph *inode;
    ConceptualGraph *tnode;
    IDLString scope[MAXN];
    IDLString tscope[MAXN];
    IDLString rname;
    IDLString member;
    IDLString interface;
    int nr;
    int maxnr;
    Dict::vcwp::const_iterator j;
    vector<string>::const_iterator i;

    maxnr = split(sname, scope, MAXN, "::");
    rname = "DEFINITIONS";
    snode = GetRelation(root, rname);
    assert(snode != NULL);
    inode = GetInstance(snode, scope, 0, maxnr - 1);
    if(inode->GetNode()->IsA() == RELATION_NODE)
        graph_1 = inode->GetWidth();
    else
        graph_1 = inode->GetDepth();

    Dict::vcwp interface_list = idict.GetRelImmSucc(sname, "INHERITS_FROM");
    vector<string> sorted = SortIfaceNames (interface_list);

    i = sorted.begin();
    while(i != sorted.end())
    {
        interface = *i;
        maxnr = split(interface, tscope, MAXN, "::");
        tnode= GetInstance(snode, tscope, 0, maxnr - 1);
        if(tnode->GetNode()->IsA() == RELATION_NODE)
            graph_2 = tnode->GetWidth();
        else
            graph_2 = tnode->GetDepth();
        CollectInheritedDefinitions(graph_1, graph_2);
        RemoveInterface(tnode, interface, sname);
        inode = GetInstance(snode, scope, 0, maxnr - 1);
        if(inode->GetNode()->IsA() == RELATION_NODE)
            graph_1 = inode->GetWidth();
        else
            graph_1 = inode->GetDepth();
        ++i;
    }
    Dict::vcwp member_list = idict.GetRelImmSucc(sname, "MEMBER_OF");
    j = member_list.begin();
    assert(j != member_list.end());
    member = **j;
    if(!(member == "*BASE*"))
    {
        maxnr = split(member, tscope, MAXN, "::");
        tnode = GetInstance(snode, tscope, 0, maxnr - 1);
        if(tnode->GetNode()->IsA() == RELATION_NODE)
            graph_2 = tnode->GetWidth();
        else
            graph_2 = tnode->GetDepth();
    }
    else 
    {
        graph_2 = root;
        tnode = snode;
    }
    CollectGlobalDefinitions(graph_1, graph_2, member);
    RemoveModule(tnode, member, sname);
    odict.AddWord(sname);
    RangeMembershipStructure(sname);
}


int
ConceptualGraph2Template::DetectLocalInterfaceTypeUsage( ConceptualGraph *cg,
                                                         IDLString interface, 
                                                         IDLString name )
{
    ConceptualGraph *graph;
    ConceptualGraph *rgraph;
    ConceptualGraph *tgraph;
    RelationNode *relation;
    IDLString rname;
    IDLString tname;
    int detected = 0;
    int det = 0;
 
    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    rgraph = cg->GetWidth();
    while(rgraph != NULL)
    {
        relation = (RelationNode*) rgraph->GetNode();
        rname = relation->GetName();
        if(rname == "TYPES" || rname == "CONSTANTS")
	{
            graph = rgraph->GetWidth();
            while(graph != NULL)
	    {
                det = SubstituteInterfaceTypeUsage(graph, interface);
                if(det)
                    detected = 1;
                graph = graph->GetDepth();
	    }
	}
        if(rname == "OPERATIONS")
	{
            graph = rgraph->GetWidth();
            while(graph != NULL)
	    {
                det = SubstituteInterfaceTypeUsage(graph, interface);
                if(det)
                    detected = 1;
                tname = "SIGNATURE";
                tgraph = GetRelation(graph, tname);
                if(tgraph != NULL)
                    tgraph = tgraph->GetWidth();
                while(tgraph != NULL)
		{
                    det = SubstituteInterfaceTypeUsage(tgraph, interface);
                    if(det)
                        detected = 1;
                    tgraph = tgraph->GetDepth();
		}
                graph = graph->GetDepth();
	    }
	}
        rgraph = rgraph->GetDepth();
    }
    return(detected);
}


void ConceptualGraph2Template::InsertScopedType( ConceptualGraph *target,
                                                 IDLString type )
{
    ConceptualGraph *rgraph_1;
    ConceptualGraph *rgraph_2;
    ConceptualGraph *graph_1;
    ConceptualGraph *graph_2;
    ConceptualGraph *cgraph_1;
    ConceptualGraph *cgraph_2;
    ConceptualGraph *cgraph_3;
    ConceptNode *concept;
    IDLString cname;
    IDLString rname;
    IDLString name;
    IDLString scope[MAXN];
    int maxelem;

    maxelem = split(type, scope, MAXN, "::");
    rname = "DEFINITIONS";
    rgraph_1 = GetRelation(root, rname);
    assert(rgraph_1 != NULL);
    cgraph_1 = GetType(rgraph_1, scope, 0, maxelem - 1);
    if(cgraph_1 != NULL)
    {
        rgraph_2 = CopyGraph(cgraph_1->GetWidth());
        rname = "TYPES";
        graph_1 = GetInsertPosition(target, rname);
        if(graph_1 == NULL)
            graph_1 = InsertNewRelation(target, rname);
        assert(graph_1 != NULL);
        concept = (ConceptNode*) cgraph_1->GetNode();
        cname = concept->GetName();
        name = type;
        gsub(name, "::", "_");
        cgraph_2 = CreateNewConceptNode(cname, name);
        assert(cgraph_2 != NULL);
        cgraph_2->SetWidth(rgraph_2);
        if(graph_1->GetNode()->IsA() == RELATION_NODE)
            graph_1->SetWidth(cgraph_2);
        else
            graph_1->SetDepth(cgraph_2);
        rname = "ANNOTATIONS";
        graph_2 = GetInsertPosition(cgraph_2, rname);
        if(graph_2 == NULL)
            graph_2 = InsertNewRelation(cgraph_2, rname);
        assert(graph_2 != NULL);
        cname = "ALIAS";
        cgraph_3 = CreateNewConceptNode(cname, type);
        assert(cgraph_3 != NULL);
        if(graph_2->GetNode()->IsA() == RELATION_NODE)
            graph_2->SetWidth(cgraph_3);
        else
            graph_2->SetDepth(cgraph_3);
    } 
}


void ConceptualGraph2Template::CollectScopedNames( ConceptualGraph* target, 
                                                   ConceptualGraph *cg )
{
    ConceptualGraph *graph_1;
    ConceptualGraph *graph_2;
    ConceptualGraph *graph_3;
    ConceptualGraph *graph_4;
    ConceptualGraph *tgraph;
    ConceptNode *concept;
    RelationNode *relation;
    SetOfWord instance_list;
    IDLString rname;
    IDLString cname;
    IDLString tname;
    IDLString name;
    SetOfWord::iterator i;

    assert(cg->GetNode()->IsA() == CONCEPT_NODE);
    graph_1 = cg->GetWidth();
    while(graph_1 != NULL)
    {
        relation = (RelationNode*) graph_1->GetNode();
        rname = relation->GetName();
        if(rname == "TYPE" || rname == "DISCRIMINATOR_TYPE") 
    	{
            graph_2 = graph_1->GetWidth();
            concept = (ConceptNode*) graph_2->GetNode();
            instance_list = concept->GetInstances();
            i = instance_list.begin();
            tname = (*i);
            if(contains(tname, "::"))
	    {
                InsertScopedType(target, tname);
                /*
                 * XXX war:
                 *   instance_list->clear();
                 * das loescht die Instanzliste im Konzeptgraphen. Soll
                 * das so sein?
                 */
                instance_list.erase (instance_list.begin(),
                                     instance_list.end());
                concept->SetInstances (instance_list);

                gsub(tname, "::", "_");
                cname = concept->GetName();
                tgraph = CreateNewConceptNode(cname, tname);
                tgraph->SetWidth(CopyGraph(graph_2->GetWidth()));
                tgraph->SetDepth(CopyGraph(graph_2->GetDepth()));
                graph_1->SetWidth(tgraph);
	    }
    	}
        if(rname == "CONSTRUCTED_TYPE") 
    	{
            graph_2 = graph_1->GetWidth();
            concept = (ConceptNode*) graph_2->GetNode();
            cname = concept->GetName();
            if(cname == "STRUCT")
	    {
                name = "MEMBERS";
                graph_3 = GetRelation(graph_2, name);
                if(graph_3 != NULL)
                    graph_3 = graph_3->GetWidth();
                while(graph_3 != NULL)
		{
                    CollectScopedNames(target, graph_3);
                    graph_3 = graph_3->GetDepth();
		}
	    }
            if(cname == "UNION")
	    {
                name = "CASES";
                graph_3 = GetRelation(graph_2, name);
                if(graph_3 != NULL) graph_3 = graph_3->GetWidth();
                while(graph_3 != NULL)
		{
                    name = "MEMBERS";
                    graph_4 = GetRelation(graph_3, name);
                    if(graph_4 != NULL)
                        graph_4 = graph_4->GetWidth();
                    while(graph_4 != NULL)
		    {
                        CollectScopedNames(target, graph_4);
                        graph_4 = graph_4->GetDepth();
		    }
                    graph_3 = graph_3->GetDepth();
		}
	    }
            if(cname == "ARRAY")
                CollectScopedNames(target, graph_2);
	}
        if(rname == "TEMPLATE_TYPE")
	{
            graph_2 = graph_1->GetWidth();
            while(graph_2 != NULL)
	    {
                concept = (ConceptNode*) graph_2->GetNode();
                cname = concept->GetName();
                if(cname == "ELEMENT_TYPE") 
		{
                    if(graph_2->GetWidth() != NULL)
                        CollectScopedNames(target, graph_2);
                    else
		    {
                        instance_list = concept->GetInstances();
                        i = instance_list.begin();
                        tname = (*i);
                        if(contains(tname, "::"))
	    		{
                            InsertScopedType(target, tname);
                            /*
                             * XXX war:
                             *   instance_list->clear();
                             * das loescht die Instanzliste im Konzeptgraphen.
                             * Soll das so sein?
                             */
                            instance_list.erase (instance_list.begin(),
                                                 instance_list.end());
                            concept->SetInstances (instance_list);
                            gsub(tname, "::", "_");
                            tgraph = CreateNewConceptNode(cname, tname);
                            tgraph->SetWidth(CopyGraph(graph_2->GetWidth()));
                            tgraph->SetDepth(CopyGraph(graph_2->GetDepth()));
                            /*
                             * XXX Achtung, das loescht den Graph,
                             * worauf graph_2 zeigt
                             */
                            graph_1->SetWidth(tgraph);
                            graph_2 = tgraph;
	    		}
		    }
		}
                graph_2 = graph_2->GetDepth();
	    }
	}
        graph_1 = graph_1->GetDepth();
    }
}


void
ConceptualGraph2Template::CollectGlobalDefinitions( ConceptualGraph *target, 
                                                    ConceptualGraph *source, 
                                                    IDLString sname )
{
    ConceptualGraph *cgraph;
    ConceptualGraph *rgraph;
    ConceptNode *concept;
    IDLString cname;
    IDLString rname;

    assert(target->GetNode()->IsA() == CONCEPT_NODE);
    assert(source->GetNode()->IsA() == CONCEPT_NODE);
    if(sname == "*BASE*")
        rname = "DEFINITIONS";
    else
        rname = "SCOPE";
    rgraph = GetRelation(source, rname);
    assert(rgraph != NULL);
    cgraph = rgraph->GetWidth();
    while(cgraph != NULL)
    {
        concept = (ConceptNode*) cgraph->GetNode();
        cname = concept->GetName();
        if(cname == "TYPE")
    	{
            rname = "TYPES";
            CheckAndInsertDefinition(target, cgraph, rname);
    	} 
        if(cname == "CONSTANT")
    	{
            rname = "CONSTANTS";
            CheckAndInsertDefinition(target, cgraph, rname);
    	} 
        cgraph = cgraph->GetDepth();
    }
}


void
ConceptualGraph2Template::CollectInheritedDefinitions(
    ConceptualGraph *target,
    ConceptualGraph *source ) 
{
    ConceptualGraph *cgraph;
    ConceptualGraph *rgraph;
    IDLString cname;
    IDLString rname;

    assert(target->GetNode()->IsA() == CONCEPT_NODE);
    assert(source->GetNode()->IsA() == CONCEPT_NODE);
    rname = "TYPES";
    rgraph = GetRelation(source, rname);
    if(rgraph != NULL)
    {
        cgraph = rgraph->GetWidth();
        while(cgraph != NULL)
	{
            CheckAndInsertDefinition(target, cgraph, rname);
            cgraph = cgraph->GetDepth();
	}
    }
    rname = "CONSTANTS";
    rgraph = GetRelation(source, rname);
    if(rgraph != NULL)
    {  
        cgraph = rgraph->GetWidth();
        while(cgraph != NULL)
	{
            CheckAndInsertDefinition(target, cgraph, rname);
            cgraph = cgraph->GetDepth();
	}
    }
    rname = "OPERATIONS";
    rgraph = GetRelation(source, rname);
    if(rgraph != NULL)
    {  
        cgraph = rgraph->GetWidth();
        while(cgraph != NULL)
	{
            CheckAndInsertDefinition(target, cgraph, rname);
            cgraph = cgraph->GetDepth();
	}
    }
}


void
ConceptualGraph2Template::CheckAndInsertDefinition( ConceptualGraph *target,
                                                    ConceptualGraph *source,
                                                    IDLString rname )
{
    ConceptualGraph *cgraph;
    ConceptualGraph *rgraph;
    ConceptualGraph *tgraph;
    ConceptNode *concept;
    SetOfWord instance_list;
    SetOfWord id_list;
    IDLString instance;
    IDLString name;
    IDLString cname;
    SetOfWord::iterator i;

    assert(target->GetNode()->IsA() == CONCEPT_NODE);
    assert(source->GetNode()->IsA() == CONCEPT_NODE);
    concept = (ConceptNode*) source->GetNode();
    instance_list = concept->GetInstances();
    i = instance_list.begin();
    if(i == instance_list.end()) 
    {
        name = "ANNOTATIONS";
        rgraph = GetRelation(source, name);
        assert(rgraph != NULL);
        cgraph = rgraph->GetWidth();
        while(cgraph != NULL)
	{
            concept = (ConceptNode*) cgraph->GetNode();
            cname = concept->GetName();
            if(cname == "IDENTIFIER")
 	    {
                id_list = concept->GetInstances();
                i = id_list.begin();
                while(i != id_list.end())
		{
                    instance = (*i);
                    tgraph = CheckDefinition(target, source, instance, rname);
                    if(tgraph != NULL)
                        InsertDefinition(tgraph, source, rname);
                    else
                        ++i;
		}
	    }
            cgraph = cgraph->GetDepth();
   	}
    }
    else
    {
        while(i != instance_list.end())
	{
            instance = (*i);
            tgraph = CheckDefinition(target, source, instance, rname);
            if(tgraph != NULL)
                InsertDefinition(tgraph, source, rname);
            ++i;
	}
    }
}


void ConceptualGraph2Template::InsertDefinition( ConceptualGraph* target, 
                                                 ConceptualGraph* source, 
                                                 IDLString rname )
{
    ConceptualGraph *sgraph;
    ConceptualGraph *rgraph;
    ConceptualGraph *graph;
    ConceptNode *concept;
    IDLString name;
    int tpos;
    int spos;
    int flag = 0;

    assert(source->GetNode()->IsA() == CONCEPT_NODE);
    sgraph = CopyGraph(source);
    sgraph->SetDepth(NULL);
    if(target->GetNode()->IsA() == CONCEPT_NODE)
        flag = 1;
    name = "POSITION";
    if(rname == "OPERATIONS")
    {
        if(flag) 
	{
            graph = GetRelation(target, name);
            assert(graph != NULL);
            graph = graph->GetWidth();
            concept = (ConceptNode*) graph->GetNode();
            tpos = concept->GetValue();
	}
        else
            tpos = 0;
        graph = GetRelation(sgraph, name);
        assert(graph != NULL); 
        graph = graph->GetWidth();
        concept = (ConceptNode*) graph->GetNode();
        spos = concept->GetValue();
        concept->SetValue(tpos + 1);
        name = "ANNOTATIONS";
        graph = GetRelation(sgraph, name);

        bool have_operation_pos = false;
        if(graph != NULL)
    	{
            graph = graph->GetWidth();
            rgraph = graph;
            while(graph != NULL) 
	    {
                if (string(graph->GetNode()->GetName()) ==
		    string("OPERATION_POSITION"))
                    have_operation_pos = true;
                if(graph->GetDepth() == NULL)
                    rgraph = graph;
                graph = graph->GetDepth();
	    }
	}
        else
            rgraph = InsertNewRelation(sgraph, name);
        name = "OPERATION_POSITION";
        if (!have_operation_pos)
            InsertNewConcept(rgraph, name, spos);
    }
    if(flag)
        target->SetDepth(sgraph);
    else
        target->SetWidth(sgraph);
}


void ConceptualGraph2Template::RemoveInterface( ConceptualGraph* node,
                                                IDLString dname, 
                                                IDLString sname )
{
    ConceptualGraph *graph_1;
    ConceptualGraph *graph_2;
    ConceptualGraph *graph_3;
    ConceptNode *concept;
    IDLString cname;
    IDLString interface;
    Dict::vcwp::const_iterator i;
    int deletion = 1;

    Dict::vcwp interface_list = ddict.GetRelImmPred(dname, "INHERITS_FROM");
    i = interface_list.begin();
    while(i != interface_list.end())
    {
        interface = **i;
        if(!(sname == interface) && !odict.ContainsWord(interface))
            deletion = 0;
        ++i;
    }
    if(!odict.ContainsWord(dname))
        deletion = 0;
    if(deletion)
    { 
        graph_1 = node;
        if(graph_1->GetNode()->IsA() == CONCEPT_NODE)
            graph_2 = graph_1->GetDepth();
        else
            graph_2 = graph_1->GetWidth();
        assert(graph_2 != NULL);
        if(graph_2->GetDepth() != NULL)
            graph_3 = CopyGraph(graph_2->GetDepth());
        else
            graph_3 = NULL;
        if(graph_1->GetNode()->IsA() == CONCEPT_NODE)
            graph_1->SetDepth(graph_3);
        else
        {
            assert(graph_3 != NULL);
            graph_1->SetWidth(graph_3);
        }
    }
}


void ConceptualGraph2Template::RemoveModule( ConceptualGraph* node,
                                             IDLString dname, 
                                             IDLString sname )
{
    ConceptualGraph *graph_1;
    ConceptualGraph *graph_2;
    ConceptualGraph *graph_3;
    ConceptualGraph *graph_4;
    ConceptNode *concept;
    IDLString cname;
    IDLString rname;
    IDLString member;
    Dict::vcwp::const_iterator i;
    int deletion = 1;

    Dict::vcwp member_list = idict.GetRelImmPred(dname, "MEMBER_OF");
    i = member_list.begin();
    while(i != member_list.end())
    {
        member = **i;
        if(!(sname == member) && !odict.ContainsWord(member))
            deletion = 0;
        ++i;
    }
    if(!odict.ContainsWord(dname))
        deletion = 0;
    if(deletion)
    {
        if(dname == "*BASE*")
	{
            graph_1 = node;
            graph_2 = graph_1->GetWidth();
	}
        else
	{
            if(node->GetNode()->IsA() == CONCEPT_NODE)
                graph_1 = node->GetDepth();
            else
                graph_1 = node->GetWidth();
            rname = "SCOPE";
            graph_2 = GetRelation(graph_1, rname);
            assert(graph_2 != NULL);
            graph_1 = graph_2;
            graph_2 = graph_2->GetWidth();
	} 
        while(graph_2 != NULL)
	{ 
            concept = (ConceptNode*) graph_2->GetNode();
            cname = concept->GetName();
            if(cname == "TYPE" || cname == "CONSTANT")
            {
                if(graph_2->GetDepth() != NULL)
                    graph_3 = CopyGraph(graph_2->GetDepth());
                else
                    graph_3 = NULL;
                if(graph_1->GetNode()->IsA() == CONCEPT_NODE)
                    graph_1->SetDepth(graph_3);
                else
                {
                    assert(graph_3 != NULL); 
                    graph_1->SetWidth(graph_3);
                }
                graph_2 = graph_3;
            }
            else
            {
                graph_1 = graph_2;
                graph_2 = graph_2->GetDepth();
            }
        }
        if(!(dname == "*BASE*"))
	{
            if(node->GetNode()->IsA() == CONCEPT_NODE)
                graph_1 = node->GetDepth();
            else
                graph_1 = node->GetWidth();
            rname = "SCOPE";
            graph_2 = GetRelation(graph_1, rname);
            assert(graph_2 != NULL);
            if(graph_2->GetWidth() != NULL)
                graph_3 = CopyGraph(graph_2->GetWidth());
            else
                graph_3 = NULL;
            if(graph_3 != NULL)
	    {
                graph_4 = graph_3;
                while(graph_4 != NULL)
		{
                    graph_2 = graph_4;
                    graph_4 = graph_4->GetDepth();
		}
                if(graph_1->GetDepth() != NULL)
                    graph_4 = CopyGraph(graph_1->GetDepth());
                else
                    graph_4 = NULL;
                if(node->GetNode()->IsA() == CONCEPT_NODE)
                    node->SetDepth(graph_3);
                else
                    node->SetWidth(graph_3);
                graph_2->SetDepth(graph_4); 
	    } 
            else
	    {
                graph_4 = CopyGraph (graph_1->GetDepth());
                if(node->GetNode()->IsA() == CONCEPT_NODE)
                    node->SetDepth(graph_4);
                else 
		{
                    assert(graph_1->GetDepth() != NULL);
                    node->SetWidth(graph_4);
		}
	    }
	}
    }
}
