
/*
 *  AI-Trader --- Service Trading using Conceptual Graphs
 *  Copyright (C) 1995 Arno Puder et al.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 aitrader@informatik.uni-frankfurt.de
 */

#include "printidl.h"


DCE2Stream::DCE2Stream( ostream &o, int d )
{
    strout = new String2Stream(o);
    level = root_level;
    state = INITIAL;
    next = NULL;
    depth = d;
    synonym_counter = 0;
}


DCE2Stream::DCE2Stream( String2Stream *s, IDLLevelType l, int d )
{
    IDLString keyword;
    IDLString ch;

    strout = s;
    level = l;
    state = INITIAL;
    next = NULL;
    depth = d;
    synonym_counter = 0;
    switch(level)
    {
    case operation_level:
	strout->PrintWhiteSpace(depth);
	break;
    case type_level:
	strout->PrintWhiteSpace(depth);
	keyword = "typedef";
	strout->PrintWord(keyword);
	break;
    case constant_level:
	strout->PrintWhiteSpace(depth);
	keyword = "const";
	strout->PrintWord(keyword);
	break;
    case struct_level:
	strout->PrintWhiteSpace(1);
	keyword = "struct";
	strout->PrintWord(keyword);
	break;
    case union_level:
	strout->PrintWhiteSpace(1);
	keyword = "union";
	strout->PrintWord(keyword);
	break;
    case switch_level:
	strout->PrintWhiteSpace(1);
	keyword = "switch";
	strout->PrintWord(keyword);
	ch = "(";
	strout->PrintWord(ch);
	break;
    case enum_level:
	strout->PrintWhiteSpace(1);
	keyword = "enum";
	strout->PrintWord(keyword);
	strout->PrintWhiteSpace(1);
	ch = "{";
	strout->PrintWord(ch);
	break;
    case pipe_level:	
	strout->PrintWhiteSpace(1);
	keyword = "pipe";
	strout->PrintWord(keyword);
	break;
    case case_level:
    case case_n_e_level:
	state = EXPRESSION;
	break;
    }	
}


DCE2Stream::~DCE2Stream()
{
    IDLString ch;

    if(next != NULL)
        delete next;
    switch(level)
    {
    case union_level:
	assert(state == CASE || state == MEMBER);
	if(state == CASE)
            strout->PrintWhiteSpace(depth);
	ch = "}";
	strout->PrintWord(ch);
	break;
    case struct_level:
 	assert(state == FINAL);
	strout->PrintWhiteSpace(depth);
	ch = "}";
	strout->PrintWord(ch);
	break;
    case interface_level:
 	assert(state == FINAL);
	strout->PrintWhiteLine(1);
	strout->PrintWhiteSpace(depth);
	ch = "}";
	strout->PrintLine(ch);
	strout->PrintWhiteLine(1);
	break;
    case member_level:
    case constant_level:
 	assert(state == FINAL);
	ch = ";";
	strout->PrintLine(ch);
	break; 
    case operation_level:
 	assert(state == FINAL);
	strout->PrintWhiteLine(1);
	strout->PrintWhiteSpace(depth);
	ch = ")";
	strout->PrintWord(ch);
	ch = ";";
	strout->PrintLine(ch);
	break;
    case enum_level:
 	assert(state == FINAL);
	strout->PrintWhiteSpace(1);
	ch = "}";
	strout->PrintWord(ch);
	break;
    case switch_level:
 	assert(state == FINAL);
	ch = ")";
	strout->PrintWord(ch);
	break;
    case root_level:
	delete strout;
	break;
    case type_level:
 	assert(state == FINAL);
	strout->PrintWhiteLine(1);
	break;
    }
		
}


void DCE2Stream::EnterNewLevel( IDLLevelType newlevel )
{
    IDLString ch;
    IDLString keyword;

    if(next != NULL) 
    {
        next->EnterNewLevel(newlevel);
        if(level == newlevel)
            synonym_counter++;
    }
    else
    {
        switch(level)
	{
        case interface_level:
	    assert(state == FINAL);
	    strout->PrintWhiteLine(1);
	    next = new DCE2Stream(strout, newlevel, depth + 1);
	    break;
        case parameter_level:
        case switch_level:
        case pipe_level:	    
	    assert(state == TYPE);
	    next = new DCE2Stream(strout, newlevel, depth + 1);
	    break;	    
        case type_level:
	    assert(state == TYPE || state == INITIAL);
	    next = new DCE2Stream(strout, newlevel, depth + 1);
	    break;	    
        case operation_level:
	    assert(state == PARAMETER || state == FINAL);
            if(state == PARAMETER) ch = "(";
	    else ch = ",";
	    strout->PrintLine(ch);
	    strout->PrintWhiteSpace(depth + 1);
	    next = new DCE2Stream(strout, newlevel, depth + 1);
	    state = FINAL;
	    break;	    
        case struct_level:
	    assert(newlevel == member_level);
	    if(state == TAG)
            {
		ch = "{";
		strout->PrintLine(ch);
            }
	    strout->PrintWhiteSpace(depth + 1);
	    next = new DCE2Stream(strout, newlevel, depth + 1);
	    state = FINAL;
	    break;	    
        case union_level:
	    switch(state)
            {
            case INITIAL:
            case TAG:
                if(newlevel == case_level)
                {
                    strout->PrintWhiteSpace(1);
                    ch = "{";
                    strout->PrintLine(ch);
                    next = new DCE2Stream(strout, case_n_e_level, depth + 1);
                    state = FINAL;
                }
                else 
                {
                    assert(newlevel == switch_level);
                    next = new DCE2Stream(strout, newlevel, depth + 1);
                    state = SWITCH;
                }
                break;
            case IDENTIFIER:
            case CASE:
                assert(newlevel == case_level); 
                strout->PrintWhiteSpace(1);
                next = new DCE2Stream(strout, newlevel, depth + 1);
                state = MEMBER;
                break;
            case MEMBER:
                assert(newlevel == member_level); 
                next = new DCE2Stream(strout, newlevel, depth + 1);
                break;
            default:
                assert(0);
            }
	    break;
        case constant_level:
        case member_level:
	    assert(state == INITIAL || state == TYPE);
	    next = new DCE2Stream(strout, newlevel, depth + 1);
	    break;
        case root_level:
	    if(state == FINAL)
            {
		ch = "/"; ch += "*";
		strout->PrintLine(ch);
            }
	    next = new DCE2Stream(strout, newlevel, depth + 1);
	    break;
        default:
	    assert(0);
	}
    }
}

void DCE2Stream::LeaveLevel( IDLLevelType deletelevel )
{
    IDLString sign;
    IDLLevelType nextlevel;
    int counter;

    assert(next != NULL); 
    nextlevel = next->GetLevel();
    counter = next->GetSynonymCounter();
    if(nextlevel == case_n_e_level && deletelevel == case_level &&
       level == union_level && counter == 0)
        nextlevel = case_level; 

    if(nextlevel == deletelevel)
    {
        if(!counter)
	{
            delete next;
            next = NULL;
            switch(level)
	    {
            case switch_level:
            case pipe_level:
            case parameter_level:
            case member_level:
            case type_level:
	        state = IDENTIFIER;
	        break;
            case union_level:
		if(deletelevel == member_level) 
                {
		    state = CASE;
		    strout->PrintWhiteSpace(depth + 1);
                }
		if(deletelevel == switch_level)
                    state = IDENTIFIER;
		if(deletelevel == case_level)
                    state = MEMBER;
		break;
            case operation_level:
		state = FINAL;
		break;
            case root_level:
		if(state == FINAL)
                {
		    sign = "*/";
		    strout->PrintLine(sign);
                }
		state = FINAL;
		break;
	    }
	}
        else 
	{
            next->LeaveLevel(deletelevel);
            next->SetSynonymCounter(counter - 1);
	}
    }
    else
        next->LeaveLevel(deletelevel);
}


void DCE2Stream::PrintAttribute( IDLString attribute )
{
    IDLString ch;

    if(next != NULL)
        next->PrintAttribute(attribute);
    else
    {
        assert(state == INITIAL);
        switch(level)
    	{
        case interface_level:
   	    strout->PrintWhiteSpace(depth);
   	    ch = "[";
   	    strout->PrintWord(ch);
   	    strout->PrintWord(attribute);
   	    ch = "]";
   	    strout->PrintWord(ch);
   	    state = IDENTIFIER;
	    break;
        case type_level:
   	    strout->PrintWhiteSpace(1);
        case operation_level:
        case parameter_level:
        case member_level:
   	    ch = "[";
   	    strout->PrintWord(ch);
   	    strout->PrintWord(attribute);
   	    ch = "]";
   	    strout->PrintWord(ch);
   	    strout->PrintWhiteSpace(1);
   	    state = TYPE;
	    break;
        default:
	    assert(0);
    	}
    }
}


void DCE2Stream::PrintIdentifier( IDLString identifier )
{
    IDLString keyword;
    IDLString ch;
    
    if(next != NULL)
        next->PrintIdentifier(identifier);
    else
    {
        assert(state == IDENTIFIER || state == INITIAL);
        switch(level)
	{
        case interface_level:
	    strout->PrintWhiteSpace(1);
	    keyword = "interface";
	    strout->PrintWord(keyword);
	    strout->PrintWhiteSpace(1);
	    strout->PrintLine(identifier);
	    ch = "{";
	    strout->PrintWhiteSpace(level + 1);
	    strout->PrintLine(ch);
	    state = FINAL;
	    break;
        case type_level:
	    strout->PrintWhiteSpace(1);
	    strout->PrintWord(identifier);
	    ch = ";";
	    strout->PrintLine(ch);
	    state = FINAL;
	    break;
        case constant_level:
	    strout->PrintWhiteSpace(1);
	    strout->PrintWord(identifier);
	    strout->PrintWhiteSpace(1);
	    ch = "=";
	    strout->PrintWord(ch);
	    state = EXPRESSION;
	    break;    
        case operation_level:
	    strout->PrintWhiteSpace(1);
	    strout->PrintWord(identifier);
	    state = PARAMETER;
	    break;
        case enum_level:
        case parameter_level:
        case pipe_level:
        case switch_level:
        case member_level:
	    strout->PrintWhiteSpace(1);
	    strout->PrintWord(identifier);
	    state = FINAL;
	    break; 
        case struct_level:
	    strout->PrintWhiteSpace(1);
	    strout->PrintWord(identifier);
	    state = TAG;
	    break;
        case union_level:
	    if(state == INITIAL)
            {
	        strout->PrintWhiteSpace(1);
		strout->PrintWord(identifier);
		state = TAG;
            }
	    if(state == IDENTIFIER)
            {
		strout->PrintWhiteSpace(1);
		strout->PrintLine(identifier);
	    	ch = "{";
	    	strout->PrintWhiteSpace(depth);
	    	strout->PrintLine(ch);
		strout->PrintWhiteSpace(depth + 1);
		state = CASE;
            }
	    break;
        default:
	    assert(0);
	}
    }
}


void DCE2Stream::PrintExpression( IDLString expression )
{
    IDLString keyword;
    IDLString ch;

    if(next != NULL)
        next->PrintExpression(expression);
    else
    {
        assert(state == EXPRESSION || state == PARAMETER);
        switch(level)
	{
        case constant_level:
	    strout->PrintWhiteSpace(1);
	    strout->PrintWord(expression);
	    state = FINAL;
	    break;
        case case_level:
	    if(expression != IDLString ("default"))
            {
		strout->PrintWhiteSpace(1);
		keyword = "case";
		strout->PrintWord(keyword);
            }
	    strout->PrintWhiteSpace(1);
	    strout->PrintWord(expression);
	    strout->PrintWhiteSpace(1);
	    ch = ":";
	    strout->PrintWord(ch);
	    break;
        case case_n_e_level:
	    if(expression != IDLString ("default"))
            {
		strout->PrintWhiteSpace(1);
		ch = "[";
		strout->PrintWord(ch);
		keyword = "case";
		strout->PrintWord(keyword);
		strout->PrintWhiteSpace(1);
		ch = "(";
		strout->PrintWord(ch);
		strout->PrintWord(expression);
		ch = ")";
		strout->PrintWord(ch);
		ch = "]";
		strout->PrintWord(ch);
		state = MEMBER;
            }
	    else
            {
		strout->PrintWhiteSpace(1);
		ch = "[";
		strout->PrintWord(ch);
		strout->PrintWord(expression);
		ch = "]";
		strout->PrintWord(ch);
		state = MEMBER;
            }
	    break;
        case operation_level:
	    assert(expression == "void");
	    ch = "(";
	    strout->PrintWord(ch);
	    strout->PrintWord(expression);
	    state = FINAL;
	    break;
        default:
	    assert(0);
	}
    }
}


void DCE2Stream::PrintType( IDLString type )
{
    IDLString keyword;
    IDLString ch;

    if(next != NULL)
        next->PrintType(type);
    else
    {
        assert(state == TYPE || state == INITIAL);
        switch(level) 
        {
        case pipe_level:
        case type_level: 
        case constant_level:
        case parameter_level:
            strout->PrintWhiteSpace(1);
        case operation_level:
        case member_level:
        case switch_level:
            strout->PrintWord(type);
            state = IDENTIFIER;
            break;
        default:
            assert(0);
        }
    }
}


CORBA2Stream::CORBA2Stream( ostream &o, int d )
{
    strout = new String2Stream(o);
    level = root_level;
    state = FINAL;
    next = NULL;
    depth = d;
    synonym_counter = 0;
}


CORBA2Stream::CORBA2Stream( String2Stream *s, IDLLevelType l, int d )
{
    IDLString keyword;
    IDLString ch;
    
    strout = s;
    level = l;
    next = NULL;
    depth = d;
    synonym_counter = 0;
    switch(level)
    {
    case type_level:
    case attribute_level:
    case operation_level:
    case case_level:
	state = INITIAL;
	break;
    case module_level:
	strout->PrintWhiteSpace(depth);
	keyword = "module";
	strout->PrintWord(keyword);
	state = IDENTIFIER;
	break;
    case interface_level:
        strout->PrintWhiteSpace(depth);
	keyword = "interface";
	strout->PrintWord(keyword);
	state = IDENTIFIER;
	break;
    case constant_level:
	strout->PrintWhiteSpace(depth);
	keyword = "const";
	strout->PrintWord(keyword);
	state = TYPE;
	break;
    case exception_level:
	strout->PrintWhiteSpace(depth);
	keyword = "exception";
	strout->PrintWord(keyword);
	state = IDENTIFIER;
	break;
    case struct_level:
	strout->PrintWhiteSpace(1);
	keyword = "struct";
	strout->PrintWord(keyword);
	state = IDENTIFIER;
	break;
    case union_level:
	strout->PrintWhiteSpace(1);
	keyword = "union";
	strout->PrintWord(keyword);
	state = IDENTIFIER;
	break;
    case switch_level:
	strout->PrintWhiteSpace(1);
	keyword = "switch";
	strout->PrintWord(keyword);
	ch = "(";
	strout->PrintWord(ch);
	state = TYPE;
	break;
    case enum_level:
	strout->PrintWhiteSpace(1);
	keyword = "enum";
	strout->PrintWord(keyword);
	state = INITIAL;
	break;
    case sequence_level:	
	strout->PrintWhiteSpace(1);
	keyword = "sequence";
	strout->PrintWord(keyword);
	ch = "<";
	strout->PrintWord(ch);
	state = TYPE;
	break;
    case string_level:
	strout->PrintWhiteSpace(1);
	keyword = "string";
	strout->PrintWord(keyword);
	state = INITIAL;
	break;
    case raises_level:
	strout->PrintWhiteSpace(depth);
	keyword = "raises";
	strout->PrintWord(keyword);
	ch = "(";
	strout->PrintWord(ch);
	state = IDENTIFIER;
	break;
    case context_level:
	strout->PrintWhiteSpace(depth);
	keyword = "context";
	strout->PrintWord(keyword);
	ch = "(";
	strout->PrintWord(ch);
	state = IDENTIFIER;
	break;
    case member_level:
	state = TYPE;
	break;
    case parameter_level:
	strout->PrintWhiteSpace(depth);
	state = ATTRIBUTE;
	break;
    case signature_level:
	ch = "(";
	strout->PrintWord(ch);
	state = PARAMETER;
	break;
    case inheritance_level:
	strout->PrintWhiteSpace(1);
	ch = ":";
	strout->PrintWord(ch);
	state = IDENTIFIER;
	break;
    case forward_level:
	state = IDENTIFIER;
	break;
    case type_dcl_level:
	strout->PrintWhiteSpace(depth);
	keyword = "typedef";
	strout->PrintWord(keyword);
	state = TYPE;
	break;
    default:
	assert(0);
    }
}


CORBA2Stream::~CORBA2Stream()
{
    IDLString ch;
    
    if(next != NULL)
        delete next;
    switch(level)
    {
    case exception_level:
	if(state == MEMBER)
        {
	    ch = "};";
	    strout->PrintLine(ch);
	    break;
        }
	else
        {
	    assert(state == FINAL);
	    strout->PrintWhiteSpace(depth);
	    ch = "};";
	    strout->PrintLine(ch);
        }
	break;
    case module_level:
    case interface_level:
 	assert(state == FINAL);
	strout->PrintWhiteLine(1);
	strout->PrintWhiteSpace(depth);
	ch = "};";
	strout->PrintLine(ch);
	break;
    case attribute_level:
    case operation_level:
    case member_level:
    case constant_level:
    case forward_level:
    case type_level:
 	assert(state == FINAL);
	ch = ";";
	strout->PrintLine(ch);
	break; 
    case struct_level:
    case union_level:
 	assert(state == FINAL);
	strout->PrintWhiteSpace(depth);
	ch = "}";
	strout->PrintWord(ch);
	break;
    case enum_level:
 	assert(state == FINAL);
	ch = "}";
	strout->PrintWord(ch);
	break;
    case signature_level:
	ch = ")";
	if(state == PARAMETER)
            strout->PrintWord(ch);
	else 
        {
	    assert(state == FINAL); 
	    strout->PrintWhiteSpace(depth);
	    strout->PrintWord(ch);
        }
	break;   
    case raises_level:
    case context_level:
    case switch_level:
 	assert(state == FINAL);
	ch = ")";
	strout->PrintWord(ch);
	break;
    case sequence_level:
 	assert(state == FINAL);
	ch = ">";
	strout->PrintWord(ch);
	break;
    case type_dcl_level: 
    case parameter_level:
 	assert(state == FINAL);
	break;
    case string_level:
	assert(state == FINAL || state == INITIAL);
	break;	
    case case_level:
    case inheritance_level:
        break;
    case root_level:
	delete strout;
	break;
    default:
	assert(0);
    }
}


void CORBA2Stream::EnterNewLevel( IDLLevelType newlevel )
{
    IDLString ch;
    IDLString keyword;

    if(next != NULL) 
    {
        next->EnterNewLevel(newlevel);
        if(level == newlevel)
            synonym_counter++;
    }
    else
    {
        switch(level)
	{
        case module_level:
 	    assert(state == FINAL);
	    strout->PrintWhiteLine(1);
	    next = new CORBA2Stream(strout, newlevel, depth + 1);
	    break;
        case interface_level:
	    if(state == INHERITANCE)
            {
		if(newlevel != inheritance_level)
                {
		    ch = "";
		    strout->PrintLine(ch);
		    strout->PrintWhiteSpace(depth);
		    ch = "{";
		    strout->PrintLine(ch);
		    strout->PrintWhiteLine(1);
		    state = FINAL;
                }
            }	    
	    else 
            {
		assert(state == FINAL);
		strout->PrintWhiteLine(1);
            }
            next = new CORBA2Stream(strout, newlevel, depth + 1);
	    break;
        case operation_level:
	    switch(state)
            {
            case FINAL:
                assert(newlevel == raises_level || newlevel == context_level);
                ch = "";
                strout->PrintLine(ch);
                next = new CORBA2Stream(strout, newlevel, depth);
                break;
            case PARAMETER:
                assert(newlevel == signature_level);
                next = new CORBA2Stream(strout, newlevel, depth);
                state = FINAL;
                break;
            case INITIAL:
            case TYPE:
                assert(newlevel == string_level || newlevel == sequence_level);
                if(state = INITIAL) strout->PrintWhiteSpace(depth);
                next = new CORBA2Stream(strout, newlevel, depth);
                state = IDENTIFIER;
                break;
            default:
                assert(0);
                break;
            } 
	    break;
        case signature_level:
	    assert(newlevel == parameter_level);
	    switch(state)
            {
            case PARAMETER:
                ch = "";
                strout->PrintLine(ch);
                state = FINAL;
                break;
            case FINAL:
                ch = ",";
                strout->PrintLine(ch);
                break;
            default:
                assert(0);
            }
	    next = new CORBA2Stream(strout, newlevel, depth + 1);
	    break;
        case parameter_level:
	    assert(state == TYPE);
	    assert(newlevel == string_level || newlevel == sequence_level);
	    next = new CORBA2Stream(strout, newlevel, depth + 1);
	    break;	
        case attribute_level:
	    assert(newlevel == sequence_level || newlevel == string_level);
	    if(state == INITIAL)
            {
		strout->PrintWhiteSpace(depth);
		keyword = "attribute";
	  	strout->PrintWord(keyword);
		state = TYPE;
            }
	    else
                assert(state == TYPE);
	    break; 	
        case exception_level:
	    assert(newlevel == member_level);
	    switch(state)
            {
            case MEMBER:
                strout->PrintWhiteLine(1);
                strout->PrintWhiteSpace(depth);		    
                state = FINAL;
                next = new CORBA2Stream(strout, newlevel, depth + 1);
                break;
            case FINAL:
                strout->PrintWhiteSpace(depth);		    
                next = new CORBA2Stream(strout, newlevel, depth + 1);
                break;
            default:
                assert(0);
            }
	    break;
        case struct_level:
	    assert(newlevel == member_level);
	    if(state == MEMBER)
            {
		strout->PrintWhiteSpace(1);
		ch = "{";
		strout->PrintLine(ch);
		strout->PrintWhiteSpace(depth + 1);
		next = new CORBA2Stream(strout, newlevel, depth + 1);
		state = FINAL;
            }
	    else
            { 
		assert(state == FINAL);
		strout->PrintWhiteSpace(depth + 1);
		next = new CORBA2Stream(strout, newlevel, depth + 1);
            }
	    break;	    
        case type_level:
            if(newlevel == struct_level ||
               newlevel == union_level ||
               newlevel == enum_level)
                strout->PrintWhiteSpace(depth);
            next = new CORBA2Stream(strout, newlevel, depth);
            break;
        case type_dcl_level:
	    assert(state == TYPE);
	    next = new CORBA2Stream(strout, newlevel, depth + 1);
	    state = IDENTIFIER;
	    break;	    
        case switch_level:
	    assert(state == TYPE);
	    assert(newlevel == enum_level);
	    next = new CORBA2Stream(strout, newlevel, depth + 1);
	    state = FINAL;
	    break;	    
        case union_level:
	    switch(state)
            {
            case SWITCH:
                assert(newlevel == switch_level);
                strout->PrintWhiteSpace(1);
                next = new CORBA2Stream(strout, newlevel, depth + 1);
                break;
            case CASE:
                assert(newlevel == case_level);
                strout->PrintWhiteSpace(1);
                ch = "{";
                strout->PrintLine(ch);
                strout->PrintWhiteSpace(depth + 1);
                next = new CORBA2Stream(strout, newlevel, depth);
                break;
            case MEMBER:
                assert(newlevel == case_level || newlevel == member_level); 
                if(newlevel == case_level)
                    next = new CORBA2Stream(strout, newlevel, depth);
                if(newlevel == member_level) 
                    next = new CORBA2Stream(strout, newlevel, depth + 1);
                break;
            case FINAL:
                assert(newlevel == case_level);
                strout->PrintWhiteSpace(depth + 1);
                next = new CORBA2Stream(strout, newlevel, depth);
                break;
            default:
                assert(0);
            }
	    break;
        case member_level:
	    assert(state == TYPE);
	    next = new CORBA2Stream(strout, newlevel, depth + 1);
	    state = IDENTIFIER;
	    break;
        case constant_level:
	    assert(state == TYPE);
	    assert(newlevel == string_level);
	    next = new CORBA2Stream(strout, newlevel, depth + 1);
	    break;
        case root_level:
	    next = new CORBA2Stream(strout, newlevel, depth + 1);
	    break;
        default:
	    assert(0);
	}
    }
}

void CORBA2Stream::LeaveLevel( IDLLevelType deletelevel )
{
    IDLString ch;
    IDLLevelType nextlevel;
    int counter;

    assert(next != NULL); 
    nextlevel = next->GetLevel();
    counter = next->GetSynonymCounter();
    if(nextlevel == deletelevel)
    {
        if(!counter)
	{
            delete next;
            next = NULL;
            switch(level)
	    {
            case interface_level:
		if(deletelevel == inheritance_level) 
                {
		    strout->PrintWhiteLine(1);
		    strout->PrintWhiteSpace(depth);
		    ch = "{";
		    strout->PrintLine(ch);
		    state = FINAL;
                }
		break;
            case operation_level:
		assert(deletelevel == sequence_level ||  
		       deletelevel == string_level ||
		       deletelevel == signature_level ||
		       deletelevel == raises_level ||
		       deletelevel == context_level);
		break;	
            case signature_level:
		assert(deletelevel == parameter_level);
		break;
            case parameter_level:
            case attribute_level:
		assert(deletelevel == sequence_level || 
		       deletelevel == string_level); 
		state = IDENTIFIER;
		break;
            case type_level:
		assert(deletelevel == type_dcl_level ||
		       deletelevel == struct_level ||
		       deletelevel == union_level ||
		       deletelevel == enum_level);
		state = FINAL;
		break;
            case union_level:
		if(deletelevel == switch_level)
                    state = CASE;
		if(deletelevel == member_level) state = FINAL;
		if(deletelevel == case_level) state = MEMBER;
		break;
            case constant_level:
		assert(deletelevel == string_level);
		state = IDENTIFIER;
		break;
            case root_level:
		strout->PrintWhiteLine(2);
		break;
	    }
	} 
        else 
        {
            next->LeaveLevel(deletelevel);
            next->SetSynonymCounter(counter - 1);
	}
    }
    else
        next->LeaveLevel(deletelevel);
}


void CORBA2Stream::PrintAttribute( IDLString attribute )
{
    IDLString ch;
    IDLString keyword;

    if(next != NULL)
        next->PrintAttribute(attribute);
    else
    {
        switch(level)
    	{
        case attribute_level:
	    assert(state == INITIAL);
	    strout->PrintWhiteSpace(depth);
	    strout->PrintWord(attribute);
	    strout->PrintWhiteSpace(1);
	    keyword = "attribute";
	    strout->PrintWord(keyword);
	    state = TYPE;
	    break;
        case operation_level:
	    assert(state == INITIAL);
	    strout->PrintWhiteSpace(depth);
	    strout->PrintWord(attribute);
	    state = TYPE;
	    break;
        case parameter_level:
     	    assert(state == ATTRIBUTE);
 	    strout->PrintWord(attribute);
	    state = TYPE;
	    break;	    
        default:
	    assert(0);
    	}
    }
}


void CORBA2Stream::PrintIdentifier( IDLString identifier )
{
    IDLString keyword;
    IDLString ch;

    if(next != NULL)
        next->PrintIdentifier(identifier);
    else
    {
        switch(level)
	{
        case forward_level:
	    strout->PrintWhiteSpace(1);
	    keyword = "interface";
	    strout->PrintWord(keyword);
	    strout->PrintWhiteSpace(1);
	    strout->PrintWord(identifier);
	    state = FINAL;
	    break;
        case module_level:
	    assert(state == IDENTIFIER);
	    strout->PrintWhiteSpace(1);
	    strout->PrintWord(identifier);
	    ch = "{";
	    strout->PrintLine(ch);
	    strout->PrintWhiteLine(1);
	    state = FINAL;
	    break;
        case interface_level:
     	    assert(state == IDENTIFIER);
	    strout->PrintWhiteSpace(1);
	    strout->PrintWord(identifier);
	    state = INHERITANCE;
	    break;
        case constant_level:
     	    assert(state == IDENTIFIER);
	    strout->PrintWhiteSpace(1);
	    strout->PrintWord(identifier);
	    strout->PrintWhiteSpace(1);
	    ch = "=";
	    strout->PrintWord(ch);
	    state = EXPRESSION;
	    break;    
        case operation_level:
     	    assert(state == IDENTIFIER);
	    strout->PrintWhiteSpace(1);
	    strout->PrintWord(identifier);
	    state = PARAMETER;
	    break;
        case type_dcl_level:
        case parameter_level:
     	    assert(state == IDENTIFIER);
	    strout->PrintWhiteSpace(1);
	    strout->PrintWord(identifier);
	    state = FINAL;
	    break; 
        case enum_level:
	    if(state == INITIAL)
            {
	    	strout->PrintWhiteSpace(1);
	    	strout->PrintWord(identifier);
	    	strout->PrintWhiteSpace(1);
		ch = "{";
	    	strout->PrintWord(ch);
	    	state = IDENTIFIER;
            }
	    else 
            {
     	        assert(state == IDENTIFIER);
		strout->PrintWord(identifier);
		state = FINAL;
            }
	    break;
        case exception_level:
	    strout->PrintWhiteSpace(1);
	    strout->PrintWord(identifier);
	    ch = "{";
	    strout->PrintWhiteSpace(1);
	    strout->PrintWord(ch);
	    state = MEMBER;
	    break;
        case inheritance_level:
        case member_level:
     	    assert(state == IDENTIFIER);
	    strout->PrintWhiteSpace(1);
	    strout->PrintWord(identifier);
	    state = FINAL;
	    break; 
        case struct_level:
     	    assert(state == IDENTIFIER);
	    strout->PrintWhiteSpace(1);
	    strout->PrintWord(identifier);
	    state = MEMBER;
	    break;
        case union_level:
     	    assert(state == IDENTIFIER);
	    strout->PrintWhiteSpace(1);
	    strout->PrintWord(identifier);
	    state = SWITCH;
	    break;
        case context_level:
        case raises_level:
        case attribute_level:
     	    assert(state == IDENTIFIER);
	    strout->PrintWhiteSpace(1);
	    strout->PrintWord(identifier);
	    state = FINAL;
	    break;
        default:
            assert(0);
	}
    }
}


void CORBA2Stream::PrintExpression( IDLString expression )
{
    IDLString keyword;
    IDLString ch;

    if(next != NULL)
        next->PrintExpression(expression);
    else
    {
        switch(level)
	{
        case sequence_level:
	    assert(state == FINAL);
	    ch = ",";
	    strout->PrintWord(ch);
	    strout->PrintWhiteSpace(1);
	    strout->PrintWord(expression);
	    break;
        case string_level:
	    assert(state == INITIAL);
	    ch = "<";
	    strout->PrintWord(ch);
	    strout->PrintWord(expression);
	    ch = ">";
	    strout->PrintWord(ch);
	    state = FINAL;
	    break;
        case constant_level:
     	    assert(state == EXPRESSION);
	    strout->PrintWhiteSpace(1);
	    strout->PrintWord(expression);
	    state = FINAL;
	    break;
        case case_level:
	    assert(state == INITIAL);
	    if(expression != IDLString ("default"))
            {
		strout->PrintWhiteSpace(1);
		keyword = "case";
		strout->PrintWord(keyword);
            }
	    strout->PrintWhiteSpace(1);
	    strout->PrintWord(expression);
	    strout->PrintWhiteSpace(1);
	    ch = ":";
	    strout->PrintWord(ch);
	    break;
        default:
	    assert(0);
	}
    }
}


void CORBA2Stream::PrintType( IDLString type )
{
    IDLString keyword;
    IDLString ch;

    if(next != NULL)
        next->PrintType(type);
    else
    {
        switch(level) 
	{
        case attribute_level:
	    if(state == INITIAL)
            {
		strout->PrintWhiteSpace(depth);
		keyword = "attribute";
		strout->PrintWord(keyword);
		strout->PrintWhiteSpace(1);
		strout->PrintWord(type);
		state = IDENTIFIER;
            }
	    else 
            {
		assert(state == TYPE);
		strout->PrintWhiteSpace(1);
		strout->PrintWord(type);
		state = IDENTIFIER;
            }
	    break;
        case operation_level:
            if(state == INITIAL)
                strout->PrintWhiteSpace(depth);
	    else 
            {	
		assert(state == TYPE);
		strout->PrintWhiteSpace(1);
            }
	    strout->PrintWord(type);
	    state = IDENTIFIER;
	    break;
        case parameter_level:
	    assert(state == TYPE);
	    strout->PrintWhiteSpace(1);
	    strout->PrintWord(type);
	    state = IDENTIFIER;
	    break;
        case sequence_level:
	    assert(state == TYPE);
	    strout->PrintWord(type);
	    state = FINAL;
	    break;
        case constant_level:
	    assert(state == TYPE);
	    strout->PrintWhiteSpace(1);
	    strout->PrintWord(type);
	    state = IDENTIFIER;
	    break;
        case switch_level:
	    assert(state == TYPE);
	    strout->PrintWord(type);
	    state = FINAL;
	    break;
        case type_dcl_level:
	    assert(state == TYPE);
	    strout->PrintWhiteSpace(1);
	    strout->PrintWord(type);
	    state = IDENTIFIER;
	    break;
        case member_level:
	    assert(state == TYPE);
	    strout->PrintWord(type);
	    state = IDENTIFIER;
	    break;
        default:
	    assert(0);
	}
    }
}
