/*
 *  AI-Trader --- Service Trading using Conceptual Graphs
 *  Copyright (C) 1995 Arno Puder et al.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 aitrader@informatik.uni-frankfurt.de
 */

import java.lang.*;
import java.util.*;
import java.awt.*;
import java.applet.Applet;

import AlertBox;
import EntryBox;
import CanvasObject;
import CanvasBoxText;
import CanvasArrow;
import MyCanvas;
import ConceptualGraph;
import CGParser;
import View;
import Popup;

class CGArrow extends CanvasArrow {
public CGArrow (int x0, int y0, int x1, int y1)
    {
        super (x0, y0, x1, y1, Color.black);
    }
}


class CGNode extends CanvasBoxText {
private CGNode parent;
private CGArrow arrow;
private Vector childs;
private Vector arrows;
private boolean isconcept;
public int layout_y;

public CGNode (int x, int y, String text, FontMetrics fm, boolean iscon)
    {
	super (x, y, text, fm,
	       iscon ? Color.black : Color.white,
	       iscon ? Color.white : Color.black,
	       iscon ? Color.black : Color.white);
        super.Move (-Bbox().width/2, -Bbox().height/2);
        parent = null;
        childs = new Vector();
        arrows = new Vector();
        arrow = null;
	isconcept = iscon;
	layout_y = 0;
    }

public void Move (int xinc, int yinc)
    {
        AutoUpdateOff ();
        super.Move (xinc, yinc);
        for (int i = arrows.size()-1; i >= 0; --i) {
            ((CGArrow)arrows.elementAt(i)).MoveTail (xinc, yinc);
        }
        if (arrow != null)
            arrow.MoveHead (xinc, yinc);
        AutoUpdateOn ();
    }

public void MoveTo (int x, int y)
    {
        Point p = GetPos ();
	Move (x - p.x, y - p.y);
    }

public void AddChild (CGNode n, CGArrow a)
    {
        childs.addElement (n);
        arrows.addElement (a);
        Rectangle pos = Bbox();
        a.MoveTailTo (pos.x + pos.width - 1, pos.y + pos.height/2);
    }

public void DelChild (CGNode n)
        throws NoSuchElementException
    {
        int idx = childs.indexOf (n);
        if (idx == -1)
            throw new NoSuchElementException ();
        else {
            childs.removeElementAt (idx);
            arrows.removeElementAt (idx);
        }
    }
    
public void SetParent (CGNode n, CGArrow a)
    {
        parent = n;
        arrow = a;
        if (arrow != null) {
            Rectangle pos = Bbox();
            arrow.MoveHeadTo (pos.x, pos.y + pos.height/2);
        }
    }

public CGNode GetParent ()
    {
	return parent;
    }

public CGArrow GetParentArrow ()
    {
	return arrow;
    }

public void SetText (String text)
    {
        int width = Bbox().width;

        AutoUpdateOff ();
        super.SetText (text);
        width = Bbox().width - width;
        for (int i = arrows.size()-1; i >= 0; --i)
            ((CGArrow)arrows.elementAt(i)).MoveTail (width, 0);
        AutoUpdateOn ();
    }

public int NumChilds ()
    {
	return childs.size();
    }

public CGNode ChildAt (int idx)
        throws ArrayIndexOutOfBoundsException
    {
	return (CGNode)childs.elementAt (idx);
    }

public CGArrow ArrowAt (int idx)
        throws ArrayIndexOutOfBoundsException
    {
	return (CGArrow)arrows.elementAt (idx);
    }

public boolean IsConcept ()
    {
	return isconcept;
    }
}


public class CGEditor extends MyCanvas
    implements PopupHandler, EntryBoxHandler {
private CGNode topnode;

private CGNode curnode;
private CGArrow curarrow;
private int dragx, dragy;

private boolean readonly;

private Vector viewers;
private Hashtable templates;

private final static int stateIdle = 0;
private final static int stateDrag = 1;
private final static int stateCreate = 2;
private final static int stateMenu = 3;
private final static int stateEdit = 4;
private int state;

private final int layoutXDist = 20;
private final int layoutYDist = 20;


public CGEditor (Dimension prefsize, boolean editable)
    {
	super (prefsize);
	readonly = !editable;
        state = stateIdle;
	topnode = null;
	curnode = null;
	curarrow = null;
        viewers = new Vector();
        templates = new Hashtable();
    }

public boolean handleEvent (Event e)
    {
	if (readonly)
	    return false;

        switch (e.id) {
        case Event.LOST_FOCUS:
            ShowStatus (-1, -1);
            break;

        case Event.MOUSE_DOWN:
	    Pick (e);
            return true;

        case Event.MOUSE_MOVE:
            ShowStatus (e.x, e.y);
            return true;

        case Event.MOUSE_DRAG:
	    Drag (e);
            return true;

        case Event.MOUSE_UP:
	    Drop (e);
            return true;

	case Event.KEY_PRESS:
	    if (e.key == '\033')
		Abort ();
	    else if (e.key == 'l')
		Layout (null);
	    return true;
        }
        return false;
    }

private void ShowStatus (int x, int y)
    {
        String s = "";
        CGNode node = null;

        if (x >= 0 && y >= 0) {
            try {
                CanvasObject co = super.Find (x, y);
                if (co instanceof CGNode) {
                    node = (CGNode)co;
                }
            } catch (NoSuchElementException ex) {}

            if (node == null) {
                if (topnode == null)
                    s = "Button1: create";
            } else {
                s = "Button1: create, Ctrl-Button1: drag, Shift-Button1: edit";
                if (node.IsConcept()) {
                    CGNode parent = node.GetParent();
                    if (parent == null || parent.NumChilds() > 1)
                        s += ", Shift-Ctrl-Button1: remove";
                    if (templates.size() > 0)
                        s += ", Button3: menu";
                } else {
                    s += ", Shift-Ctrl-Button1: remove";
                }
            }
        }
        GlobalVars.SetStatus (s);
    }

private void MenuPopup (Event ev, CGNode node)
    {
        if (templates.size() == 0)
            return;

        int ex = ev.x, ey = ev.y;
        for (Component p = this; p != null; p = p.getParent()) {
            ex += p.location().x;
            ey += p.location().y;
        }

        Vector ents = new Vector();
        Enumeration e = templates.keys();
        while (e.hasMoreElements()) {
            String key = (String)e.nextElement();
            ents.addElement (key);
        }
        Popup popup = new Popup (GlobalVars.frame, ex, ey, ents, "Relations",
                                 this, null);
        popup.pack();
        popup.show();

        curnode = node;
        state = stateMenu;
    }

public void popupHandler (Popup popup, Object arg)
    {
        state = stateIdle;

        String s = popup.GetSelectionString();
        if (s == null)
            return;

        ConceptualGraph cg = (ConceptualGraph)templates.get(s);
        AutoUpdateOff ();
        CGNode top = ConvertInput (cg);
        CGArrow arrow = new CGArrow (0, 0, 0, 0);
        Add (arrow);
        ToBottom (arrow);
        curnode.AddChild (top, arrow);
        top.SetParent (curnode, arrow);
        Layout (curnode);
        AutoUpdateOn ();

        NotifyViewers ();
    }

private void Pick (Event ev)
    {
	if (state != stateIdle)
	    return;

	CGNode node = null;
	try {
	    CanvasObject co = super.Find (ev.x, ev.y);
	    if (co instanceof CGNode) {
		node = (CGNode)co;
	    }
	} catch (NoSuchElementException ex) {}

        if (node != null && node.IsConcept() &&
            (ev.modifiers & Event.META_MASK) != 0) {
            // template menu

            MenuPopup (ev, node);
        } else {
            switch (ev.modifiers & (Event.SHIFT_MASK | Event.CTRL_MASK)) {
            case Event.SHIFT_MASK:
                // edit
                if (node != null) {
                    String s = node.GetText();
                    if (s != null) {
                        curnode = node;
                        Edit (s);
                    }
                }
                break;

            case Event.CTRL_MASK:
                // drag
                if (node != null) {
                    state = stateDrag;
                    ToTop (node);
                    curnode = node;
                    dragx = ev.x;
                    dragy = ev.y;
                }
                break;

            case Event.SHIFT_MASK|Event.CTRL_MASK:
                // delete subtree
                if (node != null) {
                    CGNode parent = node.GetParent ();
                    if (parent == null || parent.IsConcept() ||
                        parent.NumChilds() > 1) {
                        
                        AutoUpdateOff ();
                        if (parent != null) {
                            parent.DelChild (node);
                            Remove (node.GetParentArrow());
                        }
                        RemoveFromCanvas (node);
                        AutoUpdateOn ();
                    }
                    if (node == topnode)
                        topnode = null;
                    NotifyViewers();
                }
                break;
                
            default:
                // create node(s)
                if (node == null && topnode == null) {
                    topnode = new CGNode (ev.x, ev.y, "something",
                                          getFontMetrics (getFont()),
                                          true);
                    Add (topnode);
                    NotifyViewers ();
                } else if (node != null && topnode != null) {
                    state = stateCreate;
                    curnode = node;
                    
                    Rectangle pos = node.Bbox();
                    curarrow = new CGArrow (pos.x+pos.width-1,
                                            pos.y+pos.height/2,
                                            ev.x, ev.y);
                    Add (curarrow);
                }
                break;
            }
        }
    }

private void Drag (Event ev)
    {
	switch (state) {
	case stateDrag:
            curnode.Move (ev.x - dragx, ev.y - dragy);
            dragx = ev.x;
            dragy = ev.y;
	    break;

	case stateCreate:
	    curarrow.MoveHeadTo (ev.x, ev.y);
	    break;
        }
    }

private void Drop (Event ev)
    {
	switch (state) {
	case stateDrag:
	    state = stateIdle;
            CheckResize ();
	    break;

	case stateCreate:
	    CGNode node = new CGNode (ev.x, ev.y, "something",
				      getFontMetrics (getFont()),
				      !curnode.IsConcept());
            curnode.AddChild (node, curarrow);
            node.SetParent (curnode, curarrow);
            ToBottom (curarrow);
            Add (node);

	    if (!node.IsConcept()) {
		int x = ev.x + node.Bbox().width + layoutXDist;
		CGNode node2 = new CGNode (x, ev.y, "something",
					   getFontMetrics (getFont()),
					   true);
		CGArrow arrow2 = new CGArrow (0, 0, 0, 0);
		node.AddChild (node2, arrow2);
		node2.SetParent (node, arrow2);
		Add (node2);
		Add (arrow2);
                ToBottom (arrow2);
	    }
            state = stateIdle;
            NotifyViewers();
            break;
	}
        ShowStatus (ev.x, ev.y);
    }

private void Abort ()
    {
	switch (state) {
	case stateCreate:
	    Remove (curarrow);
	    state = stateIdle;
	    break;
	}
        state = stateIdle;
    }

public void entryBoxHandler (EntryBox e, Object arg)
    {
        if (state != stateEdit)
            return;
        state = stateIdle;
        if (e.GetButton() == 0) {
            curnode.SetText (e.GetText());
            NotifyViewers();
        }
    }

protected void Edit (String txt)
    {
        state = stateEdit;
        EntryBox editor = new EntryBox ("Edit Node", "Enter node:", txt,
                                        this, null);
        editor.pack();
        editor.show();
    }

private void RemoveFromCanvas (CGNode root)
    {
	super.Remove (root);
	for (int i = root.NumChilds()-1; i >= 0; --i) {
	    super.Remove (root.ArrowAt(i));
	    RemoveFromCanvas (root.ChildAt(i));
	}
    }

private int LayoutPass1 (CGNode root, int y, Vector vdist, int level)
    {
	Rectangle bbox = root.Bbox();
	root.layout_y = y;

	if (level >= vdist.size()) {
	    vdist.addElement (new Integer (bbox.width));
	} else {
	    int width = ((Integer)vdist.elementAt(level)).intValue();
	    if (width < bbox.width) {
		vdist.setElementAt (new Integer (bbox.width), level);
	    }
	}
	int nchilds = root.NumChilds(), y2 = y;
	for (int i = 0; i < nchilds; ++i) {
	    y2 = LayoutPass1 (root.ChildAt(i), y2, vdist, level+1);
	}
	return Math.max (y + bbox.height + layoutYDist, y2);
    }

private void LayoutPass2 (CGNode root, int x, Vector vdist, int level)
    {
	Rectangle bbox = root.Bbox();
	int width = ((Integer)vdist.elementAt(level)).intValue();
	root.MoveTo (x + (width - bbox.width)/2, root.layout_y);

	int nchilds = root.NumChilds();
	for (int i = 0; i < nchilds; ++i) {
	    LayoutPass2 (root.ChildAt(i), x+width+layoutXDist, vdist, level+1);
	}
    }

public void Layout (CGNode root)
    {
	int x, y;
	if (root == null) {
            if ((root = topnode) == null)
                return;
	    x = layoutXDist;
	    y = layoutYDist;
	} else {
	    x = root.Bbox().x;
	    y = root.Bbox().y;
	}
	AutoUpdateOff ();
	Vector vdist = new Vector();
	LayoutPass1 (root, y, vdist, 0);
	LayoutPass2 (root, x, vdist, 0);
	CheckResize ();
	AutoUpdateOn ();
    }

private CGNode ConvertInput (ConceptualGraph cg)
    {
        StringBuffer ostr = new StringBuffer();
        cg.getNode().print (ostr);
        String s = ostr.toString();
        s = s.substring (1, s.length()-1);

        CGNode node = new CGNode (0, 0, s, getFontMetrics(getFont()),
                                  cg.getNode() instanceof ConceptNode);
        Add (node);

        ConceptualGraph cg2;
        for (cg2 = cg.getWidth(); cg2 != null; cg2 = cg2.getDepth()) {
            CGNode node2 = ConvertInput (cg2);
            CGArrow arrow2 = new CGArrow (0, 0, 0, 0);
            Add (arrow2);
            ToBottom (arrow2);

            node.AddChild (node2, arrow2);
            node2.SetParent (node, arrow2);
        }
        return node;
    }

private void ConvertOutput (StringBuffer ostr, CGNode node)
    {
        String s = node.GetText();

        if (node.IsConcept()) {
            ostr.append ("[" + s + "]");
        } else {
            ostr.append ("(" + s + ")");
        }

        int nchilds = node.NumChilds();
        if (nchilds == 1) {
            ostr.append ("->");
            ConvertOutput (ostr, node.ChildAt (0));
        } else if (nchilds > 1) {
            ostr.append ("-");
            for (int i = 0; i < nchilds; ++i) {
                ostr.append ("->");
                ConvertOutput (ostr, node.ChildAt (i));
                if (i+1 != nchilds)
                    ostr.append (",");
            }
            ostr.append (".");
        }
    }

public void SetCG (ConceptualGraph cg)
    {
	AutoUpdateOff ();
	if (topnode != null) {
	    RemoveFromCanvas (topnode);
	    topnode = null;
	}
        if (cg != null) {
            topnode = ConvertInput (cg);
            Layout (null);
        }
	AutoUpdateOn ();
        Abort();
        NotifyViewers();
    }

public ConceptualGraph GetCG ()
    {
        if (topnode == null)
            return null;

        StringBuffer ostr = new StringBuffer();
        ConvertOutput (ostr, topnode);

        CGParser p = new CGParser ();
        try {
            return p.Parse (ostr.toString());
        } catch (Exception e) {
            AlertBox a = new AlertBox ("Parse Error", "Cannot parse graph");
            a.pack();
            a.show();
            return null;
        }
    }

public void AddView (View v)
    {
        viewers.addElement (v);
    }

public void DelView (View v)
    {
        viewers.removeElement (v);
    }

private void NotifyViewers ()
    {
        for (int i = 0; i < viewers.size(); ++i)
            ((View)viewers.elementAt(i)).UpdateView (this);
    }

public void AddTemplate (String name, ConceptualGraph cg)
    {
        templates.put (name, cg);
    }

public void DelTemplate (String name)
        throws NoSuchElementException
    {
        templates.remove (name);
    }
}
