/*
 *  AI-Trader --- Service Trading using Conceptual Graphs
 *  Copyright (C) 1995 Arno Puder et al.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 aitrader@informatik.uni-frankfurt.de
 */

import java.lang.*;
import java.util.*;
import java.awt.*;

import BorderPanel;
import GlobalVars;

class ListCanvas extends Canvas {
private ListBox lb;

public ListCanvas (ListBox lb)
    {
        this.lb = lb;
    }

public void paint (Graphics g)
    {
        lb.Redraw (g);
    }
}


public class ListBox extends Panel {
private final static int HDIST = 1;
private Canvas cv;
private Scrollbar sb;
private Vector ents, sels;
private int yoffs, height, width;
private int xoffs, dragx;
private boolean multisel, dirty;

public ListBox ()
    {
        setLayout (new GridBagLayout ());

        BorderPanel p = new BorderPanel (BorderPanel.SUNKEN);
        GlobalVars.constrain (this, p, 0, 0, 1, 1,
                              GridBagConstraints.BOTH, 1.0, 1.0);

        p.setLayout (new GridBagLayout());
        cv = new ListCanvas (this);
        GlobalVars.constrain (p, cv, 0, 0, 1, 1,
                              GridBagConstraints.BOTH, 1.0, 1.0);

        sb = new Scrollbar (Scrollbar.VERTICAL);
        GlobalVars.constrain (this, sb, 1, 0, 1, 1,
                              GridBagConstraints.BOTH, 0.0, 1.0);

        ents = new Vector();
        sels = new Vector();
        multisel = false;
        yoffs = 0;
        height = 0;
        width = 0;
        dirty = true;
        xoffs = 0;
        ConfigScroll();
    }

public ListBox (int rows, boolean multi)
    {
        this ();
        multisel = multi;
    }
    
public void addItem (String item)
    {
        ents.addElement (item);
        dirty = true;
        Redraw (cv.getGraphics());
        ConfigScroll();
    }

public void delItem (int pos)
    {
        ents.removeElementAt (pos);
        for (int i = 0; i < sels.size(); ++i) {
            if (((Integer)sels.elementAt(i)).intValue() == pos) {
                sels.removeElementAt (i);
            }
        }
        dirty = true;
        Redraw (cv.getGraphics());
        ConfigScroll();
    }

public int countItems ()
    {
        return ents.size ();
    }
    
public void setItems (Vector v)
    {
        ents.removeAllElements ();
        for (int i = 0; i < v.size(); ++i) {
            ents.addElement ((String)v.elementAt (i));
        }
        sels.removeAllElements ();
        yoffs = 0;
        xoffs = 0;
        dirty = true;
        Redraw (cv.getGraphics());
        ConfigScroll ();
    }

public void makeVisible (int idx)
    {
        int oldyoffs = yoffs;

        if (idx < 0)
            idx = 0;
        if (idx >= ents.size())
            idx = ents.size()-1;

        if (idx < yoffs)
            yoffs = idx;
        else if (yoffs + VisibleEntries() <= idx)
            yoffs = idx - VisibleEntries() + 1;

        if (oldyoffs != yoffs)
            Redraw (cv.getGraphics());
    }
    
public void clear ()
    {
        ents.removeAllElements ();
        sels.removeAllElements ();
        yoffs = 0;
        xoffs = 0;
        dirty = true;
        Redraw (cv.getGraphics());
        ConfigScroll();
    }

public String getItem (int idx)
    {
        if (idx >= ents.size())
            return null;
        return (String)ents.elementAt (idx);
    }

public int getSelectedIndex ()
    {
        return sels.size() == 1
            ? ((Integer)sels.firstElement()).intValue()
            : -1;
    }

public int[] getSelectedIndexes ()
    {
        int idxs[] = new int[sels.size()];
        
        for (int i = 0; i < sels.size(); ++i) {
            idxs[i] = ((Integer)sels.elementAt(i)).intValue();
        }
        return idxs;
    }

public String getSelectedItem ()
    {
        return sels.size() == 1
            ? (String)ents.elementAt(((Integer)sels.firstElement()).intValue())
            : null;
    }

public String[] getSelectedItems ()
    {
        String items[] = new String[sels.size()];
        
        for (int i = 0; i < sels.size(); ++i) {
            int idx = ((Integer)sels.elementAt(i)).intValue();
            items[i] = (String)ents.elementAt (idx);
        }
        return items;
    }

public void select (int idx)
    {
        if (idx >= ents.size())
            idx = ents.size()-1;
        else if (idx < 0)
            idx = 0;
        if (ents.elementAt (idx) == null)
            return;
        if (!isSelected (idx)) {
            if (!multisel)
                sels.removeAllElements ();
            sels.addElement (new Integer (idx));
            Redraw (cv.getGraphics());
        }
    }

public void deselect (int idx)
    {
        for (int i = 0; i < sels.size(); ++i) {
            if (((Integer)sels.elementAt(i)).intValue() == idx) {
                sels.removeElementAt (i);
                Redraw (cv.getGraphics());
                break;
            }
        }
    }

public boolean isSelected (int idx)
    {
        for (int i = 0; i < sels.size(); ++i) {
            if (((Integer)sels.elementAt(i)).intValue() == idx)
                return true;
        }
        return false;
    }

public boolean allowsMultipleSelections ()
    {
        return multisel;
    }

public void setMultipleSelections (boolean multi)
    {
        multisel = multi;
        if (sels.size() > 1 && !multi) {
            select (((Integer)sels.elementAt(0)).intValue());
        }
    }

public boolean handleEvent (Event ev)
    {
        switch (ev.id) {
        case Event.SCROLL_LINE_UP:
        case Event.SCROLL_LINE_DOWN:
        case Event.SCROLL_PAGE_UP:
        case Event.SCROLL_PAGE_DOWN:
        case Event.SCROLL_ABSOLUTE:
            yoffs = ((Integer)ev.arg).intValue();
            Redraw (cv.getGraphics());
            return true;

        case Event.MOUSE_DOWN:
            dragx = ev.x;
            int s = (ev.y - cv.bounds().y) / EntryHeight();
            if (yoffs+s < ents.size() && ents.elementAt (yoffs+s) != null) {

                boolean issel = isSelected (yoffs + s);
                if (issel && multisel) {
                    deselect (yoffs + s);
                    getParent().postEvent (
                        new Event (this, Event.LIST_DESELECT,
                                   new Integer (yoffs+s)));
                } else if (!issel) {
                    select (yoffs + s);
                    getParent().postEvent (
                        new Event (this, Event.LIST_SELECT,
                                   new Integer (yoffs+s)));
                }
            }
            return true;

        case Event.MOUSE_DRAG:
            if ((ev.modifiers & Event.ALT_MASK) != 0) {
                int invisible = EntryWidth() - cv.size().width;
                int oxoffs = xoffs;
                xoffs -= ev.x - dragx;
                if (xoffs > 0)
                    xoffs = 0;
                else if (xoffs < -invisible)
                    xoffs = -invisible;
                if (oxoffs != xoffs) {
                    dragx = ev.x;
                    Redraw (cv.getGraphics());
                }
            }
            return true;
        }
        return false;
    }

private void CheckResize ()
    {
        int nvis = VisibleEntries ();
        if (nvis != height) {
            if (yoffs + nvis > ents.size()) {
                yoffs = ents.size() - nvis;
                if (yoffs < 0)
                    yoffs = 0;
            }
            ConfigScroll ();
        }
    }

public void Redraw (Graphics g)
    {
        if (isValid()) {
            CheckResize();

            g.setColor (getBackground ());
            g.fillRect (0, 0, cv.size().width, cv.size().height);
            g.setColor (Color.black);

            int eheight = EntryHeight ();
            int nvis = VisibleEntries ();
            if (nvis > ents.size() - yoffs)
                nvis = ents.size() - yoffs;
            int wd = EntryWidth();
            for (int i = 0; i < nvis; ++i) {
                String s = (String)ents.elementAt (yoffs+i);
                if (s != null) {
                    g.drawString ((String)ents.elementAt (yoffs+i), xoffs,
                                  i*eheight + HDIST + Ascent());
                } else {
                    g.drawLine (xoffs, i*eheight + eheight/2,
                                xoffs+wd-1, i*eheight + eheight/2);
                }
            }
            g.setColor (getBackground ());
            g.setXORMode (Color.black);
            for (int i = 0; i < sels.size(); ++i) {
                int selection = ((Integer)sels.elementAt(i)).intValue();
                if (selection >= yoffs && selection - yoffs < nvis) {
                    g.fillRect (xoffs, (selection - yoffs)*eheight,
                                wd, eheight);
                }
            }
        }
    }

private int EntryWidth ()
    {
        if (dirty && isValid()) {
            width = 0;
            FontMetrics fm = cv.getFontMetrics (cv.getFont());
            for (int i = 0; i < ents.size(); ++i) {
                String s = (String)ents.elementAt(i);
                if (s != null)
                    width = Math.max (width, fm.stringWidth (s));
            }
            dirty = false;
        }
        return Math.max (width, cv.size().width);
    }

private int EntryHeight ()
    {
        return isValid()
            ? cv.getFontMetrics (cv.getFont()).getHeight() + 2*HDIST
            : 10;
    }

private int VisibleEntries ()
    {
        return isValid()
            ? cv.size().height / EntryHeight()
            : 10;
    }

private int Ascent ()
    {
        return isValid()
            ? cv.getFontMetrics (cv.getFont()).getAscent()
            : 5;
    }

private void ConfigScroll ()
    {
        int nvis = VisibleEntries ();
        int sz = ents.size() - nvis;
        if (sz < 0)
            sz = 0;

        sb.setValues (yoffs, nvis, 0, sz);
        sb.setPageIncrement (nvis);
        sb.setLineIncrement (1);
    }

public Dimension preferredSize ()
    {
        return new Dimension (150, 200);
    }

public Dimension minimumSize ()
    {
        return preferredSize();
    }
}
