/*
 *  AI-Trader --- Service Trading using Conceptual Graphs
 *  Copyright (C) 1995 Arno Puder et al.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 aitrader@informatik.uni-frankfurt.de
 */

import java.lang.*;
import java.util.*;
import java.awt.*;

import PopupHandler;


class PopLabel extends Canvas {
protected String label;

public PopLabel (String l)
    {
        label = l;
    }

public Dimension preferredSize ()
    {
        Dimension sz = new Dimension (0, 0);
        sz.width = getFontMetrics (getFont()).stringWidth (label) + 8;
        sz.height= getFontMetrics (getFont()).getHeight() + 6;
        return sz;
    }

public Dimension minimumSize ()
    {
        return preferredSize ();
    }

public void update (Graphics g)
    {
        paint (g);
    }

public void paint (Graphics g)
    {
        g.setColor (getBackground());
        g.fillRect (0, 0, size().width, size().height);
        g.setColor (Color.black);
        g.drawString (label, 2, 4 + getFontMetrics(getFont()).getAscent());
    }
}


class PopString extends PopLabel {
private boolean inside;

public PopString (String l)
    {
        super (l);
        inside = false;
    }

public boolean handleEvent (Event ev)
    {
        switch (ev.id) {
        case Event.MOUSE_ENTER:
            inside = true;
            repaint();
            break;

        case Event.MOUSE_EXIT:
            inside = false;
            repaint();
            break;
        }
        return false;
    }

public void paint (Graphics g)
    {
        super.paint (g);
        if (inside) {
            Popup.Draw3DBorder (this, g);
        }
    }
}


class PopSeparator extends Canvas {
public PopSeparator ()
    {
    }

public Dimension preferredSize ()
    {
        return new Dimension (1, 2);
    }

public Dimension minimumSize ()
    {
        return preferredSize ();
    }

public void update (Graphics g)
    {
        paint (g);
    }

public void paint (Graphics g)
    {
        g.setColor (Popup.darkGray);
        g.drawLine (0, 0, size().width-1, 0);
        g.setColor (Popup.lightGray);
        g.drawLine (0, 1, size().width-1, 1);
    }
}


public class Popup extends Window {
private PopupHandler callback;
private Object callbackArg;
private Vector entries;
private Vector buttons;
private int selection;
private int xpos, ypos;
private GridLayout layout;
private static Insets insets = new Insets (2, 2, 2, 2);
public static final Color lightGray = new Color (230, 230, 230);
public static final Color darkGray = new Color (100, 100, 100);

public Popup (Frame f, int x, int y, Vector ents, String title)
    {
	this (f, x, y, ents, title, null, null);
    }

public Popup (Frame f, int x, int y, Vector ents, String title,
	      PopupHandler cb, Object cbArg)
    {
        super (f);

	callback = cb;
	callbackArg = cbArg;

        buttons = new Vector ();
        selection = -1;
        xpos = x;
        ypos = y;
        entries = ents;

	setLayout (new GridBagLayout());
        Component c = new PopLabel (title);
        GlobalVars.constrain (this, c, 0, 0, 1, 1,
                              GridBagConstraints.BOTH, 1.0, 0.0);

        c = new PopSeparator ();
        GlobalVars.constrain (this, c, 0, 1, 1, 1,
                              GridBagConstraints.BOTH, 1.0, 0.0);

        int n = ents.size();
        for (int i = 0; i < n; ++i) {
            c = new PopString ((String)ents.elementAt (i));
            GlobalVars.constrain (this, c, 0, 2+i, 1, 1,
                                  GridBagConstraints.BOTH, 1.0, 0.0);
            buttons.addElement (c);
        }
    }

public int GetSelection ()
    {
        return selection;
    }

public String GetSelectionString ()
    {
        if (selection < 0)
            return null;
        return (String)entries.elementAt(selection);
    }

public boolean handleEvent (Event ev)
    {
        switch (ev.id) {
        case Event.MOUSE_UP:
            selection = buttons.indexOf (ev.target);
            hide ();
            if (callback != null)
		callback.popupHandler (this, callbackArg);
            return true;
        }
        return false;
    }

public void pack ()
    {
        // super.pack() will force the window to not appear when
        // this.show() is called.
        Dimension d = preferredSize();
	// hack for appletviewer...
	if (d.width < 150)
	    d.width = 150;
        resize (d);
        move (xpos - d.width/2, ypos - d.height/2);
        validate ();
    }

public Insets insets ()
    {
        return insets;
    }

public void paint (Graphics g)
    {
        g.setColor (Color.black);
        Draw3DBorder (this, g);
    }

public static void Draw3DBorder (Component c, Graphics g)
    {
        int wd = c.size().width;
        int ht = c.size().height;
        g.setColor (lightGray);
        g.drawLine (0, 0, wd-1, 0);
        g.drawLine (0, 1, wd-2, 1);
        g.drawLine (0, 2, 0, ht-1);
        g.drawLine (1, 2, 1, ht-2);
        g.setColor (darkGray);
        g.drawLine (2, ht-2, wd-1, ht-2);
        g.drawLine (1, ht-1, wd-1, ht-1);
        g.drawLine (wd-2, 2, wd-2, ht-2);
        g.drawLine (wd-1, 1, wd-1, ht-2);
    }
}
