
/*
 *  AI-Trader --- Service Trading using Conceptual Graphs
 *  Copyright (C) 1995 Arno Puder et al.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 aitrader@informatik.uni-frankfurt.de
 */


import java.util.*;
import ConceptualGraph;
import TransportLayer;
import parser;
import AITException;


public class Proxy {
	public static final int RelNone = 0;
	public static final int RelSymmetric = 1;
	public static final int RelReflexive = 2;
	public static final int RelTransitive = 4;
	public static final int RelAcyclic = 8;

	private TransportLayer transportLayer;
	private parser theParser;

	public Proxy( TransportLayer theConnection )
	{
		transportLayer = theConnection;
	}

	private void executeCommand( String command )
	{
		String result;

		result = transportLayer.transmitString( command );
                try {
                    /* allocate a parser object */
                    theParser = new parser();

                    theParser.setToBeParsed(result);
                    theParser.parse();

                } catch (java.lang.Exception ex) {
                    System.err.println("Exception: " + ex.getMessage());
                    ex.printStackTrace();
                    System.exit(-1);
                }
	}

	private void checkForError()
		throws AITException
	{
		// This function should check whether the last
		// command caused an error
                int errnum = theParser.getErrNum();
		if (errnum != 0) {
                    String errmsg = theParser.getErrMsg();
                    theParser.setError (0, "");
                    // Error occured. Throw an exception
                    throw (new AITException(errnum, errmsg));
		}
	}

	public String getVersion()
		throws AITException
	{
		executeCommand( "@version" );
		checkForError();
		return theParser.getAnswerString();
	}

	public void setVerbose( boolean flag )
		throws AITException
	{
		if( flag )
			executeCommand( "@verbose-on" );
		else
			executeCommand( "@verbose-off" );
		checkForError();
	}

	public void saveTraderDB( String name )
		throws AITException
	{
		executeCommand( "@save '" + name + "'" );
		checkForError();
	}

	public void defineMatchingRule( String name, String rule )
		throws AITException
	{
		executeCommand( "@define-matching-rule '" + name + "\n" + 
						rule + "\n@end" );
		checkForError();
	}

	public void deleteMatchingRule( String name )
		throws AITException
	{
		executeCommand( "@delete-matching-rule '" + name + "'" );
		checkForError();
	}

	public Vector getMatchingRuleNames()
		throws AITException
	{
		executeCommand( "@show-matching-rules" );
		checkForError();

                Vector v = new Vector();
                String s = theParser.getAnswerString();
                StringTokenizer st = new StringTokenizer (s, "\n");
                while (st.hasMoreTokens())
                    v.addElement (st.nextToken());
		return v;
	}

	public String getMatchingRule( String name )
		throws AITException
	{
		executeCommand( "@show-matching-rule '" + name + "'" );
		checkForError();
		return theParser.getAnswerString();
	}

	public void registerRelation( String rel, int prop )
		throws AITException
	{
		String command;
	
		command = "@register-relation '" + rel + "'";
		if( prop != RelNone ) {
			command += " : ";
		}
		boolean flag = false;
		if( (prop & RelSymmetric) == RelSymmetric ) {
			command += "@symmetric";
			flag = true;
		}
		if( (prop & RelReflexive) == RelReflexive ) {
			if( flag )
				command += ", ";
			command += "@reflexive";
			flag = true;
		}
		if( (prop & RelTransitive) == RelTransitive ) {
			if( flag )
				command += ", ";
			command += "@transitive";
			flag = true;
		}
		if( (prop & RelAcyclic) == RelAcyclic ) {
			if( flag )
				command += ", ";
			command += "@acyclic";
		}
		executeCommand( command );
		checkForError();
	}

	public Vector getAllRelations()
		throws AITException
	{
		executeCommand( "@show-all-relations" );
		checkForError();
		return theParser.getAnswerVector();
	}

        public int getRelationProperties (String rel)
        	throws AITException
	{
            executeCommand ("@get-relation-prop '" + rel + "'");
            checkForError ();
            // XXX
            return 0;
        }

	public void deleteWord( String word )
		throws AITException
	{
		executeCommand( "@delete-word \"" + word + '"' );
		checkForError();
	}

	public void addWord( String word )
		throws AITException
	{
		executeCommand( "@add-word \"" + word + '"' );
		checkForError();
	}

	public boolean containsWord( String word )
		throws AITException
	{
		executeCommand( "@contains-word \"" + word + '"' );
		checkForError();
		return theParser.getAnswerBoolean();
	}

	public void addRelation( String word1, String rel, String word2 )
		throws AITException
	{
		executeCommand( "@add-relation \"" + word1 + "\" : '" +
						rel + "' : \"" + word2 + '"' );
		checkForError();
	}

	public Vector getRelationPred( String word, String rel )
		throws AITException
	{
		executeCommand( "@get-relation-pred \"" + word + "\" : '" +
						rel + "'" );
		checkForError();
		return theParser.getAnswerVector();
	}

	public Vector getRelationAllPred( String word, String rel )
		throws AITException
	{
		executeCommand( "@get-relation-all-pred \"" + word + "\" : '" +
						rel + "'" );
		checkForError();
		return theParser.getAnswerVector();
	}

        public Vector getRelationSucc( String word, String rel )
		throws AITException
	{
		executeCommand( "@get-relation-succ \"" + word + "\" : '" +
						rel + "'" );
		checkForError();
		return theParser.getAnswerVector();
	}

        public Vector getRelationAllSucc( String word, String rel )
		throws AITException
	{
		executeCommand( "@get-relation-all-succ \"" + word + "\" : '" +
						rel + "'" );
		checkForError();
		return theParser.getAnswerVector();
	}

	public boolean areRelated( String word1, Vector rels, String word2 )
		throws AITException
	{
		String command;
		boolean flag = false;
	
		command = new String();
		command = "@are-related \"" + word1 + "\" : ";
		for( Enumeration e = rels.elements(); e.hasMoreElements(); ) {
			if( flag )
				command += ", ";
			command += "'" + e.nextElement() + "'";
			flag = true;
		}
		command += " : \"" + word2 + '"';
		executeCommand( command );
		checkForError();
		return theParser.getAnswerBoolean();
	}

	public String add( ConceptualGraph cg )
		throws AITException
	{
		StringBuffer ostr;

		ostr = new StringBuffer();
		ostr.append( "@add " );
		cg.print( ostr );
		executeCommand( ostr.toString() );
		checkForError();
		return theParser.getAnswerString();
	}

	public void replace( String aKey, ConceptualGraph cg )
		throws AITException
	{
		StringBuffer ostr;
	
		ostr = new StringBuffer();
                if (cg != null) {
                    ostr.append( "@add " + aKey + " : " );
                    cg.print( ostr );
                } else {
                    ostr.append ("@remove" + aKey);
                }
		executeCommand( ostr.toString() );
		checkForError();
	}

	public void join( String aKey, ConceptualGraph cg )
		throws AITException
	{
		StringBuffer ostr;

		ostr = new StringBuffer();
		ostr.append( "@join " + aKey + " : " );
		cg.print( ostr );
		executeCommand( ostr.toString() );
		checkForError();
	}

	public Vector match( ConceptualGraph query, Vector selecRules,
                             String profilekey )
		throws AITException
	{
		StringBuffer ostr;

		ostr = new StringBuffer();
		ostr.append( "@match " + profilekey + " : ");
		boolean first = true;
		for( Enumeration e = selecRules.elements(); e.hasMoreElements(); ) {
			if( !first )
				ostr.append( ", " );
			first = false;
			ostr.append( "'" +e.nextElement() + "'" );
		}
		ostr.append( " : " );
		query.print( ostr );
		executeCommand( ostr.toString() );
		checkForError();
		return theParser.getAnswerVector();
	}

	public ConceptualGraph getGraph( String aKey )
		throws AITException
	{
		executeCommand( "@get-graph " + aKey );
		checkForError();
		return theParser.getAnswerConceptualGraph();
	}

	public int getNumberOfVariants( String conType )
		throws AITException
	{
		String result;
	
		executeCommand( "@number-of-variants '" + conType + "'" );
		checkForError();
		return theParser.getAnswerInt();
	}

	public String getNthVariant( String conType, int num )
		throws AITException
	{
		executeCommand( "@get-graph '" + conType + "' : " + num );
		checkForError();
		return theParser.getAnswerString();
	}

	public int getNumberOfGraphs ()
		throws AITException
	{
            // doesnt work
            executeCommand ("@number-of-graphs");
            checkForError();
            return theParser.getAnswerInt();
	}


	public String getNthGraph (int num)
		throws AITException
	{
            // doesnt work
            checkForError();
            return "";
	}


	public void addProvider( String aKey, String theMaintainer,
                                 String anOID )
		throws AITException
	{
		executeCommand( "@add-provider " + aKey + ": '" +
                                theMaintainer +
                                "', '" + anOID + "'" );
		checkForError();
	}

	public void removeProvider( String aKey, String anOID )
		throws AITException
	{
		//assert( 0 );
		checkForError();
	}

	public Vector getProviders( String aKey )
		throws AITException
	{
		executeCommand( "@get-provider " + aKey );
		checkForError();
		return theParser.getAnswerVector();
	}

        public String createProfile ()
		throws AITException
        {
		executeCommand ("@create-profile");
		checkForError ();
		return theParser.getAnswerString ();
        }

        public void removeProfile (String key)
		throws AITException
        {
		executeCommand ("@remove-profile " + key);
		checkForError ();
        }

        public void setMinMatchval (String key, int mmv)
		throws AITException
        {
		executeCommand ("@set-minmatchval " + key + " : " + mmv);
		checkForError ();
        }

        public void setTimeout (String key, int tmo)
		throws AITException
        {
		executeCommand ("@set-timeout " + key + " : " + tmo);
		checkForError ();
        }

        public int getMinMatchval (String key)
		throws AITException
        {
		executeCommand ("@get-minmatchval " + key);
		checkForError ();
                return theParser.getAnswerInt ();
        }

	public int getTimeout (String key)
		throws AITException
        {
		executeCommand ("@get-timeout " + key);
		checkForError ();
                return theParser.getAnswerInt ();
        }
}
