/*
 *  AI-Trader --- Service Trading using Conceptual Graphs
 *  Copyright (C) 1995 Arno Puder et al.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 aitrader@informatik.uni-frankfurt.de
 */

import java.lang.*;
import java.util.*;
import java.awt.*;

class MyPanel extends Panel {
private Viewport vp;
public MyPanel (Viewport vp)
    {
        this.vp = vp;
    }
public void paint (Graphics g)
    {
        vp.repaint();
    }
}

public class Viewport extends Panel {
private static Insets insets = new Insets (2, 2, 2, 2);
private Scrollbar vert, horz;
private Component child;
private Panel frame;
    
public Viewport (Component child, Dimension size)
    {
        this.child = child;

        horz = new Scrollbar (Scrollbar.HORIZONTAL);
        vert = new Scrollbar (Scrollbar.VERTICAL);
        frame = new MyPanel (this);
        frame.resize (size);

        frame.setLayout (null);
        frame.add (child);

        setLayout (new BorderLayout());
        add ("Center", frame);
        add ("East", vert);
        add ("South", horz);

        validate ();

        resizeHorz ();
        resizeVert ();
    }

public Insets insets ()
    {
        return insets;
    }

public boolean handleEvent(Event evt)
    {
        switch (evt.id) {
          case Event.SCROLL_LINE_UP:
          case Event.SCROLL_LINE_DOWN:
          case Event.SCROLL_PAGE_UP:
          case Event.SCROLL_PAGE_DOWN:
          case Event.SCROLL_ABSOLUTE:
            if (evt.target == vert) {
                child.move (child.location().x,
                            -((Integer)evt.arg).intValue());
            }
            if (evt.target == horz) {
                child.move (-((Integer)evt.arg).intValue(),
                            child.location().y);
            }
            break;
        }
        return super.handleEvent(evt);
    }

private void resizeHorz()
    {
        int canvasWidth = frame.size().width;
        int childWidth = child.size().width;
        int xpos = - child.location().x;

        if (canvasWidth <= 0) {
            System.out.println("Canvas has no width; can't resize scrollbar");
            return;
        }

        //Shift everything to the right if we're displaying empty space
        //on the right side.
        if ((xpos + canvasWidth) > childWidth) {
            int newtx = childWidth - canvasWidth;
            if (newtx < 0) {
                newtx = 0;
            }
            child.move (-newtx, child.location().y);
            xpos = newtx;
        }

        horz.setValues(//draw the part of the image that starts at this x:
                       xpos, 
                       //amount to scroll for a "page":
                       (int)(canvasWidth * 0.9), 
                       //minimum image x to specify:
                       0,
                       //maximum image x to specify:
                       childWidth - canvasWidth);
        //"visible" arg to setValues() has no effect if scrollbar is visible.
        horz.setPageIncrement((int)(canvasWidth * 0.9));
        return;
    }

private void resizeVert() {
        int canvasHeight = frame.size().height;
        int childHeight = child.size().height;
        int ypos = - child.location().y;

        if (canvasHeight <= 0) {
            System.out.println("Canvas has no height; can't resize scrollbar");
            return;
        }

        //Shift everything downward if we're displaying empty space
        //on the bottom.
        if ((ypos + canvasHeight) > childHeight) {
            int newty = childHeight - canvasHeight;
            if (newty < 0) {
                newty = 0;
            }
            child.move (child.location().x, -newty);
            ypos = newty;
        }

        vert.setValues(//initially draw part of image starting at this y:
                       ypos, 
                       //visible arg--amount to scroll for a "page":
                       (int)(canvasHeight * 0.9), 
                       //minimum image y to specify:
                       0,
                       //maximum image y to specify:
                       childHeight - canvasHeight);

        //"visible" arg to setValues() has no effect if scrollbar is visible.
	vert.setPageIncrement((int)(canvasHeight * 0.9));
        return;
    }

public void paint(Graphics g)
    {
        //This method probably was called due to applet or child being resized.
	Dimension fsz = frame.size();
	Dimension csz = child.size();
	if (csz.width < fsz.width || csz.height < fsz.height) {
	    csz.width = Math.max (csz.width, fsz.width);
	    csz.height = Math.max (csz.height, fsz.height);
	    child.resize (csz);
	}
        resizeHorz();
        resizeVert();
    }
}
