
/*
 *  AI-Trader --- Service Trading using Conceptual Graphs
 *  Copyright (C) 1995 Arno Puder et al.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 aitrader@informatik.uni-frankfurt.de
 */

#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#ifdef _AIX
#include <strings.h>
#endif
#include <signal.h>
#include <time.h>
#include <netdb.h>

#include <stream.h>
#include <fstream.h>
#include <strstream.h>
#include <rpc/rpc.h>
#include <Trader.h>
#include <rpc_client.h>

#include "idlparser.h"
#include "cgparsetree.h"

#include "stl.h"
#include "access.h"


/*
 * IDL-Toolset functions
 *-------------------------------------------------------------------
 * The following commands are processed here:
 *   @idl-tool-conv-to-cg <idl>
 *   @idl-tool-conv-to-corba <cg>
 *   @idl-tool-conv-to-dce <cg>
 *   @idl-tool-test-subtype <idl1> '\0' <idl2>
 */


char *initial_kb;
char *mrule_name_dce;
char *mrule_name_corba;
char *mrule_name_odp;

Server *toolset_server = NULL;
KeyCG toolset_db_key;

static void log (const string &);

void idl_tool_conv_to_cg( char *param, char **result )
{
    ostrstream ostr1;
    ostr1 << param << ends;
    istrstream istr1( ostr1.str() );
    ostr1.rdbuf()->freeze( 0 );

    ostrstream ostr2;
	
    ostr2 << "@answer-graph=";
    IDLParser idlparser( istr1, "IDL-Toolset", ostr2 );
    idlparser.Parse();

    ostr2 << ends;
    istrstream istr2( ostr2.str() );
    ostr2.rdbuf()->freeze( 0 );
    Server server;
    Parser parser( &server, istr2 );
    parser.Parse();
    ConceptualGraph *cg;
    server.GetAnswer( &cg );

    ostrstream ostr3;
    if( cg != NULL ) {
        cg->PrettyPrint( ostr3, 2 );
        delete cg;
    }

    ostr3 << ends;
    *result = strdup( ostr3.str() );
    ostr3.rdbuf()->freeze( 0 );
}



void idl_tool_conv_to_corba( char *param, char **result )
{
    ostrstream ostr;
    ostr << "@answer-graph=" << param << ends;
    istrstream istr( ostr.str() );
    ostr.rdbuf()->freeze( 0 );
    Server server;
    Parser parser( &server, istr );
    parser.Parse();
    ConceptualGraph *cg;
    server.GetAnswer( &cg );

    ostrstream ostr2;
    if( cg != NULL ) {
        Template2CORBA CORBA_con( cg, ostr2 );
        CORBA_con.CreateSpecification();
    }
    ostr2 << ends;
    *result = strdup( ostr2.str() );
    ostr2.rdbuf()->freeze( 0 );
}



void idl_tool_conv_to_dce( char *param, char **result )
{
    ostrstream ostr;
    ostr << "@answer-graph=" << param << ends;
    istrstream istr( ostr.str() );
    ostr.rdbuf()->freeze( 0 );
    Server server;
    Parser parser( &server, istr );
    parser.Parse();
    ConceptualGraph *cg;
    server.GetAnswer( &cg );

    ostrstream ostr2;
    if( cg != NULL ) {
        Template2DCE DCE_con( cg, ostr2 );
        DCE_con.CreateSpecification();
    }
    ostr2 << ends;
    *result = strdup( ostr2.str() );
    ostr2.rdbuf()->freeze( 0 );
}



bool idl_toolset_match( char *mrule_name, ConceptualGraph *cg )
{
    SetOfString selecRules;
    SetOfMatchResult matches;
	
    selecRules.insert( mrule_name );
    toolset_server->Match( cg, selecRules, matches );
    // If we have any matches with our toolset server,
    // then the subtyping rule succeeded
    return matches.begin() != matches.end();
}



void idl_tool_test_subtype( char *param, char **result )
{
    Server server;
    IDLParser *idlparser;
    Parser *parser;
	
    ostrstream ostr_idl1_1;
    ostr_idl1_1 << param << ends;
    istrstream istr_idl1_1( ostr_idl1_1.str() );
    ostr_idl1_1.rdbuf()->freeze( 0 );

    ostrstream ostr_idl1_2;
	
    ostr_idl1_2 << "@answer-graph=";
    idlparser = new IDLParser( istr_idl1_1, "IDL-Toolset", ostr_idl1_2 );
    idlparser->Parse();
    delete idlparser;

    ostr_idl1_2 << ends;
    istrstream istr_idl1_2( ostr_idl1_2.str() );
    ostr_idl1_2.rdbuf()->freeze( 0 );
    parser = new Parser( &server, istr_idl1_2 );
    parser->Parse();
    delete parser;
    ConceptualGraph *cg1;
    server.GetAnswer( &cg1 );
    if( cg1 == NULL ) {
        *result = strdup( "000" );
        return;
    }

    ostrstream ostr_idl2_1;
    ostr_idl2_1 << (param + strlen( param ) + 1) << ends;
    istrstream istr_idl2_1( ostr_idl2_1.str() );
    ostr_idl2_1.rdbuf()->freeze( 0 );

    ostrstream ostr_idl2_2;
	
    ostr_idl2_2 << "@answer-graph=";
    idlparser = new IDLParser( istr_idl2_1, "IDL-Toolset", ostr_idl2_2 );
    idlparser->Parse();
    delete idlparser;

    ostr_idl2_2 << ends;
    istrstream istr_idl2_2( ostr_idl2_2.str() );
    ostr_idl2_2.rdbuf()->freeze( 0 );
    parser = new Parser( &server, istr_idl2_2 );
    parser->Parse();
    delete parser;
    ConceptualGraph *cg2;
    server.GetAnswer( &cg2 );
    if( cg2 == NULL ) {
        *result = strdup( "000" );
        return;
    }

    toolset_server->Replace( toolset_db_key, cg1 );
    char matches[ 4 ];
    cout << "-------------------------------------###" << endl;
    
    matches[ 0 ] = idl_toolset_match( mrule_name_dce, cg2 ) ? '1' : '0';
    matches[ 1 ] = idl_toolset_match( mrule_name_corba, cg2 ) ? '1' : '0';
    matches[ 2 ] = idl_toolset_match( mrule_name_odp, cg2 ) ? '1' : '0';
    matches[ 3 ] = '\0';
    *result = strdup( matches );
}



void idl_toolset( char *command, char **result )
{
    if( strncmp( command, "@idl-tool-conv-to-cg", 20 ) == 0 )
        idl_tool_conv_to_cg( command + 20, result );
    if( strncmp( command, "@idl-tool-conv-to-corba", 23 ) == 0 )
        idl_tool_conv_to_corba( command + 23, result );
    if( strncmp( command, "@idl-tool-conv-to-dce", 21 ) == 0 )
        idl_tool_conv_to_dce( command + 21, result );
    if( strncmp( command, "@idl-tool-test-subtype", 22 ) == 0 )
        idl_tool_test_subtype( command + 22, result );
}



void idl_toolset_init()
{
    ifstream file( initial_kb );

    toolset_server = new Server();
	
    if( !file ) {
        log(string("/* couldn't open file '") + initial_kb + "': " +
            strerror (errno) + " */");
        exit( 1 );
    }
    Parser *parser = new Parser( toolset_server, file, initial_kb );
    parser->Parse();
    if (parser->ErrorOccured ()) {
        cerr << parser->GetErrorFile() << ':'
             << parser->GetErrorLine() << ':'
             << parser->GetErrorMsg() << endl;
        delete parser;
        exit( 1 );
    }
    delete parser;
	
    // Create a small dummy CG
    SetOfWord insts;
    ConceptNode *cn = new ConceptNode( "dummy", insts );
    ConceptualGraph *cg = new ConceptualGraph( cn, NULL, NULL );
    // Export dummy CG to get a new service database key
    toolset_server->Add( cg, toolset_db_key );
}




/*---End of Toolset section-------------------------------------------*/

static string thePeerAddr = "<server>";
static int theUserCount = 0;
static const int maxUserCount = 10;


bool
denied (const string &s)
{
    if (s.find ("@import") < s.length ())
        return true;

    ostrstream ostr;
    ostr << s << ends;
    istrstream istr (ostr.str());
    ostr.rdbuf()->freeze (0);
    
    AccessChecker ac;
    Parser parser(&ac, istr, NULL);
    parser.Parse ();

    return !ac.IsAllowed();
}


void
setpeer (const sockaddr_in &sin)
{
    struct hostent *hent = gethostbyaddr ((char *)&sin.sin_addr.s_addr,
                                          sizeof (sin.sin_addr.s_addr),
                                          sin.sin_family);
    if (hent)
        thePeerAddr = hent->h_name;
    else
        thePeerAddr = inet_ntoa (sin.sin_addr);
}


void
log (const string &s)
{
    time_t t = time (NULL);
    string timestr = ctime (&t);
    //AP timestr.remove (timestr.size()-1, 1);
    timestr.replace (timestr.size()-1, 1, "");

    cout << "/* " << thePeerAddr << " [" << timestr << "]: */ " << s << endl;
}


void cleanup( int )
{
    int stat, pid;

    while ((pid = waitpid (-1, &stat, WNOHANG)) > 0) {
        cout << "/* child " << pid << " exited */" << endl;
        --theUserCount;
    }
    signal( SIGCLD, cleanup );
}



int net_csend( int sock, char *s )
{
	unsigned long size;
	int bytes_to_send;
	int len;

	bytes_to_send = strlen( s ) + 1;
	size = htonl( bytes_to_send );
	len = write( sock, &size, sizeof( unsigned long ) );
	if( len == -1 ) return -1; /* return any error */

	if( len < sizeof( unsigned long ) ) { /* partial write? */
		errno = EIO;
		return -1;
	}

	len = write( sock, s, bytes_to_send );
	if( len == -1 ) return -1; /* return any error */

	if( len < bytes_to_send ) { /* partial write? */
		errno = EIO;
		return -1;
	}

	return 0; /* success */
}



int net_read( int sock, char *buf, int buflen )
{
	int left;
	int len;
	char *p;

	left = buflen;
	p = buf;
	while( left > 0 ) {
		len = read( sock, p, left );
		if( len == -1 ) return -1; /* error occured */
		if( len == 0 ) break; /* can't read any more */

		left -= len;
		p += len;
	}

	return buflen - left;
}



int net_crecv( int sock, char **buf )
{
	unsigned long size;
	int len;
	unsigned long buflen;

	/* read the response header */
	len = net_read( sock, (char *) &size, sizeof( unsigned long ) );
	if( len == -1 ) return -1; /* oops */
	if( len < sizeof( unsigned long ) ) {
		errno = EIO;
		return -1;
	}

	/* read buffer */
	buflen = ntohl( size );
	*buf = (char *) malloc( buflen );
	len = net_read( sock, (char *) *buf, buflen );
	if( len == -1 ) return -1;
	if( len < buflen ) {
		errno = EIO;
		return -1;
	}

	return buflen;
}


int net_slisten( int port )
{
	int sock, new_sock;
	struct sockaddr_in sin, from;
	socklen_t len;

	memset ( (char *) &sin, 0, sizeof(sin) );
	sin.sin_family = AF_INET;
	sin.sin_port = htons (port);

	sock = socket( AF_INET, SOCK_STREAM, 0 );
	if( sock < 0 ) return -1;

        int on = 1;
        setsockopt (sock, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof (on));

	if( bind( sock, (sockaddr *) &sin, sizeof(sin) ) < 0 ) return -1;

	listen( sock, 1 );
	while( 1 ) {
		len = sizeof( from );
		new_sock = accept( sock, (sockaddr *) &from, &len );
		if( new_sock < 0 ) {
			if( errno == EINTR ) continue;
			return -1;
		}
                if (theUserCount > maxUserCount) {
                    net_csend (new_sock, "too many users. try again later");
                    close (new_sock);
                } else {
                    ++theUserCount;
                    switch(fork ()) {
                    case -1: /* error? */
			return -1;

                    case 0: /* child */
                        setpeer (from);
                        (void) close( sock );
                        idl_toolset_init();
                        return new_sock;

                    default: /* parent */
			signal( SIGCLD, cleanup );
                        close( new_sock );
                        break;
                    }
		}
	}
	/* NOTREACHED */
}



int net_copen( char *host, int port )
{
	int sock;
	struct sockaddr_in sin;
	struct hostent *hp;

	hp = gethostbyname( host );
	if( !hp ) return -1;

	memset ( (char *) &sin, 0, sizeof(sin) );
	memcpy ( (char *) &(sin.sin_addr), hp->h_addr, hp->h_length );
	sin.sin_family = hp->h_addrtype;
	sin.sin_port = htons (port);

	sock = socket( AF_INET, SOCK_STREAM, 0 );
	if( sock < 0 ) return -1;

	if( connect( sock, (sockaddr *) &sin, sizeof (sin) ) < 0 ) return -1;

	return sock;
}



void net_server( int sock, char *aitserver_host )
{
    RPCClient         *rpc_proxy;

    char *java_host;
    int len, wlen, i;
    char *buf;
    char response[ 200 ];
    int rc;

    /* First string denotes the hostname of the Java client */
    len = net_crecv( sock, &java_host );
    if( len == -1 ) {
        log (string ("/* read on socket failed: ") +
             strerror (errno) + " */");
        exit( 1 );
    }

    rpc_proxy = new RPCClient( aitserver_host, 0x20000, 60 );
    if( rpc_proxy->GetErrorCode() == -1 ) {
        log ("/* cannot connect to AIT server */");
        exit( 1 );
    }

    rpc_proxy->Connect( java_host, &rc );
    if( rc == 0 ) {
        log ("/* AIT server refused connection */");
        exit( 1 );
    }

    while( 1 ) { /* until an error occurs */
        /* read a request */
        len = net_crecv( sock, &buf );
        
        if( len == -1 ) {
            rpc_proxy->Disconnect( java_host, &rc );
            log (string("/* read on socket failed: ") +
                 strerror (errno) + " */");
            exit( 1 );
        }

        if (denied (buf)) {
            len = net_csend (sock, "@error 20 : 'Operation not permitted'");
            log (string(buf) + " /* -> denied */");
        } else {
            char *answer;
			if( strncmp( buf, "@idl-tool", 9 ) == 0 ) {
				idl_toolset( buf, &answer );
			} else {
				rpc_proxy->ExecRequest( buf, &answer);
			}
        
            len = net_csend( sock, answer );
            log (string(buf) + " /* -> " + answer + " */");
#ifdef dont_do_that
            /* XXX muss hier noch ein free(answer) hin? */
            free(answer);
#endif dont_do_that
        }
        free( buf );
                
        if( len == -1 ) {
            rpc_proxy->Disconnect( java_host, &rc );
            log (string("/* write on socket failed: ") +
                 strerror (errno) +
                 " */");
            exit( 1 );
        }
    }
}



void main( int argc, char *argv[] )
{
    int sock;

    if( argc != 6 ) {
        cerr<<form ("Usage: %s <aitserver-host> <initial.kb> <dce-rule-name> <corba-rule-name> <odp-rule-name>\n", argv[ 0 ]);
        exit( 1 );
    }
	
    initial_kb = argv[ 2 ];
    mrule_name_dce = argv[ 3 ];
    mrule_name_corba = argv[ 4 ];
    mrule_name_odp = argv[ 5 ];

    sock = net_slisten( 3388 );
    net_server( sock, argv[ 1 ] );
}

