
/*
 *  AI-Trader --- Service Trading using Conceptual Graphs
 *  Copyright (C) 1995 Arno Puder et al.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 aitrader@informatik.uni-frankfurt.de
 */


#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <fstream.h>
#include <strstream.h>
#include <string.h>

#include <rpc_server.h>

#include <Trader.h>
#include "version.h"


class AITraderServer : public RPCServer {
private:
	Server   server;
	char	*ResultString;
	bool     shuttingDown;
	int      numClients;
	int	 verbosity;
public:
	AITraderServer( u_long ProgNr, int verbosity = 0 );
	~AITraderServer();
	virtual int Connect( char *User );
	virtual int Disconnect( char *user );
	virtual void ExecRequest( char *Request, char **Result );
	virtual int InitShutdown( char *Passwd );
	virtual int ForceShutdown( char *Passwd );
	bool LoadInitialKB( char *fn );
};


AITraderServer::AITraderServer( u_long ProgNr, int verb )
	: RPCServer( ProgNr )
{
	ResultString = NULL;
	shuttingDown = false;
	numClients = 0;
	verbosity = verb;
}



AITraderServer::~AITraderServer()
{
	if( ResultString ) {
		delete[] ResultString;
	}
}



int AITraderServer::Connect( char *User )
{
	if( shuttingDown ) {
		cerr << "Rejecting client " << User << endl;
		return 0;
	}
	cout << "Client " << User << " connected" << endl;
	numClients++;
	return 1;
}



int AITraderServer::Disconnect( char *User )
{
	cout << "Client " << User << " disconnected" << endl;
	numClients--;
	if( shuttingDown && numClients == 0 ) {
		cout << "Exiting..." << endl;
		exit( 0 );
	}
	return 1;
}



int AITraderServer::InitShutdown( char *PassWord )
{
	cout << "Initiated shutdown" << endl;
	shuttingDown = true;
	return 0;
}



int AITraderServer::ForceShutdown( char *PassWord )
{
	printf( "Client forced shutdown: passwd = \"%s\" ...\n", PassWord );
	return 0;
}


void AITraderServer::ExecRequest( char *request, char **result )
{
	int		l, i;

	/* Free old result string if necessary */
	if( ResultString ) {
		delete[] ResultString;
	}

	ostrstream ostr;
	ostr << request << ends;
	istrstream istr( ostr.str() );
	ostr.rdbuf()->freeze( 0 );
	Parser *parser = new Parser( &server, istr );
	parser->Parse();
	delete parser;
	string answer = server.GetAnswerString();
	
	/* Allocate new result string */
	l = answer.length();
	ResultString = (char*) new char[ l + 1 ];

	/* Copy string to result string */
	for( i = 0; i < l; i++ ) {
		ResultString[ i ] = answer[ i ];
	}
	ResultString[ i ] = '\0';

	/* Return result */
	*result = ResultString;

	/* Print message */
	if (verbosity > 0) {
		cout << request << endl;
		cout << ResultString << endl << endl;
	}
}



bool AITraderServer::LoadInitialKB( char *fn )
{
	ifstream file( fn );
	
	if( !file ) {
		printf( "Couldn't open file '%s'\n", fn );
		exit( 1 );
	}
	Parser *parser = new Parser( &server, file, fn );
	parser->Parse();
	if (parser->ErrorOccured ()) {
            cerr << parser->GetErrorFile() << ':'
                 << parser->GetErrorLine() << ':'
                 << parser->GetErrorMsg() << endl;
            delete parser;
            return false;
	}
	delete parser;
        return true;
}

static void
usage (char *appname)
{
//	cerr.form ( "Usage: %s [-verbose] <file>\n", appname);
	exit( 1 );
}
	
void main( int argc, char *argv[] )
{
	AITraderServer	*rpc_server;
	int verbosity = 0, i = 1;

	if( argc < 2 ) {
		usage (argv[0]);
	}
	if (!strncmp ("-v", argv[i], 2)) {
		++i;
		verbosity = 1;
	}
	if (i >= argc) {
		usage (argv[0]);
	}
	rpc_server = new AITraderServer( 0x20000, verbosity );

	/* Another server is running */
	if( rpc_server->GetErrorCode() == -1 ) {
		delete rpc_server;
		cerr << "Another AI-Trader Server is running on this system" << endl;
		exit( 1 );
	}
	if (!rpc_server->LoadInitialKB( argv[i] ))
            exit (1);

        if (verbosity > 0) {
            cout << "AI-Trader server version " << TRADER_VERSION
                 << " ready." << endl;
        }
	rpc_server->RunServer();
}
