
/*
 *  AI-Trader --- Service Trading using Conceptual Graphs
 *  Copyright (C) 1995 Arno Puder et al.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 aitrader@informatik.uni-frankfurt.de
 */


// $Id: aitsh.cc,v 1.12 1998/02/06 11:05:40 roemer Exp $


#include <stdlib.h>
#include <stdio.h>
#include <fstream.h>
#include <strstream.h>
#include <sys/types.h>
#include <rpc/rpc.h>
#include <Trader.h>
#include <rpc_client.h>

#include "stl.h"

extern "C" {
    int gethostname (char *name, size_t len);
}

#ifdef HAVE_LIBREADLINE
/*
 * readline/readline.h has no prototypes, so we declare the functions we
 * need ourselves
 */
extern "C" {
  char *readline (const char *);
  int rl_bind_key (int, int (*)(void));
  int rl_insert (void);
  void add_history (const char *);
}
#endif

class Application {
private:
    int                argc;
    char             **argv;
    AITraderInterface *server;
    RPCClient         *rpc_proxy;
    char               hostName[ 256 ];
    bool               isLocal;
    bool               LoadFile( char *Name );
    void               InputLoop();
    void               PrintUsage();
    string	       GetLine() const;
public:
    Application( int an_argc, char *an_argv[] );
	~Application();
    void Run();
};


Application::Application( int an_argc, char *an_argv[] )
{
	argc = an_argc;
	argv = an_argv;
	server = NULL;
	rpc_proxy = NULL;
	gethostname( hostName, 255 );
}


Application::~Application()
{
	if( server )
		delete server;
	if( rpc_proxy )
		delete rpc_proxy;
}


bool Application::LoadFile( char *Name )
{
	ifstream file( Name );
	
	if( !file ) {
		printf( "Couldn't open file '%s'\n", Name );
		exit( -1 );
	}
	Parser *parser = new Parser( server, file, Name );
	parser->Parse();
	if (parser->ErrorOccured ()) {
            cerr << parser->GetErrorFile() << ':'
                 << parser->GetErrorLine() << ':'
                 << parser->GetErrorMsg() << endl;
            delete parser;
            return false;
	}
	delete parser;
        return true;
}


void Application::PrintUsage()
{
	printf( "Usage: %s [<file>]\n", argv[ 0 ] );
	printf( "       %s -remote <host>\n", argv[ 0 ] );
	exit( 1 );
}


void Application::Run()
{
	if( argc > 3 )
		PrintUsage();
	if( argc == 3 ) {
		// Open connection to a remote server
		if( strcmp( argv[ 1 ], "-remote" ) != 0 )
			PrintUsage();
		rpc_proxy = new RPCClient( argv[ 2 ], 0x20000, 60 );
		if( rpc_proxy->GetErrorCode() == -1 ) {
			cout << "No connection to remote server" << endl;
			exit( 1 );
		}
		int rc;
		
		rpc_proxy->Connect( hostName, &rc );
		if( rc == 0 ) {
			cout << "Connection to remote server not allowed" << endl;
			exit( 1 );
		}
		isLocal = false;
		InputLoop();
		return;
	}
	// Create local server
	server = new Server();
	if( argc == 2 ) {
		if (!LoadFile( argv[ 1 ] ))
                    exit (1);
        }
	isLocal = true;
	InputLoop();
}


string
Application::GetLine () const
{
#ifdef HAVE_LIBREADLINE
	static bool inited = false;

	if (!inited) {
		/*
	 	 * we don't want TAB to perform filename completion
	 	 */
		rl_bind_key ('\t', rl_insert);
		inited = true;
	}
	char *command = readline ("aitsh> ");
	if (!command)
		return "@exit";
	string result (command);
	if (!(result == ""))
		add_history (command);
	free (command);
	return result;
#else
	cout << "aitsh> ";
    char command[1024], c;
    cin.get (command, 1024, '\n');
    if (cin.get (c) && c != '\n') {
        cerr << endl;
        cerr << "input line too long" << endl;
        return string ("");
    }
	return string (command);
#endif
}

void Application::InputLoop()
{
	cout << "AI-Trader shell" << endl << endl;

	while( true ) {
		string command = GetLine ();
		if (command == "@exit") {
			int rc;

			if (!isLocal)
				rpc_proxy->Disconnect( hostName, &rc );
			return;
		}
		if( isLocal ) {
			ostrstream ostr;
			ostr << command << ends;
			istrstream istr( ostr.str() );
			ostr.rdbuf()->freeze( 0 );
			Parser *parser = new Parser( server, istr );
			parser->Parse();
			delete parser;
			cout << server->GetAnswerString() << endl;
		} else {
			char *answer;
			rpc_proxy->ExecRequest ((char *)command.c_str(), &answer);
			cout << answer << endl;
		}
	}
}


int main( int argc, char *argv[] )
{
	Application app( argc, argv );

	app.Run();
}
