
/*
 *  AI-Trader --- Service Trading using Conceptual Graphs
 *  Copyright (C) 1995 Arno Puder et al.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 aitrader@informatik.uni-frankfurt.de
 */


// $Id: cgtcl.cc,v 1.30 1997/07/14 15:22:50 roemer Exp $


#include <stdlib.h>
#include <stdio.h>
#include <stream.h>
#include <iostream.h>
#include <fstream.h>
#include <strstream.h>
#include <ctype.h>
#include <sys/types.h>
#include <rpc/rpc.h>
#include <assert.h>

extern "C" {
#include <tcl.h>
}

extern "C" {
    int gethostname (char *name, size_t len);
}

#include <rpc_client.h>

#define HAVE_IDL
#ifdef HAVE_IDL
#include "idlparser.h"
#include "cgparsetree.h"
#endif

#include "Trader.h"
#include "key.h"
#include "proxy.h"
#include "cgtcl.h"
#include "util.h"

#ifdef HAVE_CORBA
#include "orbixClient.h"
#endif

#define STR2CHAR(s)	((char *)(s).c_str())

//--------------------------------------------------------------------------

static Tcl_Interp *CG_Interp = NULL;
static string templateGraph;

//--------------------------------------------------------------------------

// Proxy for local knowledgebase


//-------------------------------------------------------------------------

//static LocalTclClient *TclClient;
static AITraderInterface *TclClient = NULL;

//-------------------------------------------------------------------

#ifdef HAVE_CORBA
class orbixTransportLayer : public TransportLayer 
{
private:
	orbixClient *orbix;
	int error;
public:
	orbixTransportLayer( char *host );
	~orbixTransportLayer();
	int GetError();
	virtual void TransmitString( string &in, string &out );
};



orbixTransportLayer::orbixTransportLayer( char *host )
{
	orbix = new orbixClient( host );
	error = 0;
	if( orbix->GetErrorCode() == -1 ) {
		error = -1;
		return;
	}
}


orbixTransportLayer::~orbixTransportLayer()
{
}


int orbixTransportLayer::GetError()
{
	return error;
}



void orbixTransportLayer::TransmitString( string &in, string &out )
{
	char *answer;
	
	orbix->sendRequest( STR2CHAR(in), &answer );
	out = answer;
}

#endif /* HAVE_CORBA */

//-------------------------------------------------------------------


class RPCTransportLayer : public TransportLayer 
{
private:
	RPCClient rpc_proxy;
	int error;
public:
	RPCTransportLayer( char *host );
	~RPCTransportLayer();
	int GetError();
	virtual void TransmitString( string &in, string &out );
};



RPCTransportLayer::RPCTransportLayer( char *host )
	: rpc_proxy( host, 0x20000, 120 )
{
	error = 0;
	if( rpc_proxy.GetErrorCode() == -1 ) {
		error = -1;
		return;
	}
	
	char hostName[ 256 ];
	int rc;

	gethostname( hostName, 255 );
	rpc_proxy.Connect( hostName, &rc );
	if( rc == 0 )
		error = -2;
}


RPCTransportLayer::~RPCTransportLayer()
{
	char hostName[ 256 ];
	int rc;
	
	gethostname( hostName, 255 );
	rpc_proxy.Disconnect( hostName, &rc );
}


int RPCTransportLayer::GetError()
{
	return error;
}



void RPCTransportLayer::TransmitString( string &in, string &out )
{
	char *answer;
	
	rpc_proxy.ExecRequest( STR2CHAR(in), &answer );
	out = answer;
}

static TransportLayer *transportLayer = NULL;

//-------------------------------------------------------------------

static string CGLastError;


int CG_Error( Tcl_Interp *interp, int errorNum )
{
	static char *Msg[] = {
		"Second parameter must be -c or -r",		/* 0 */
		"Wrong number of arguments",			/* 1 */
		"Already connected",				/* 2 */
		"Timeout",					/* 3 */
		"Not yet connected to any server",		/* 4 */
		"Can not open file",				/* 5 */
		"First parameter must be -local or -remote",	/* 6 */

	};

	CGLastError = string( "CG: " ) + Msg[ errorNum ];
	Tcl_SetResult( interp, STR2CHAR (CGLastError), TCL_VOLATILE );
	return TCL_ERROR;
}


int CG_CheckError( Tcl_Interp *interp )
{
#if 0
	if( ErrorHasOccured ) {
		Tcl_SetResult( interp, CGLastError, TCL_VOLATILE );
		return TCL_ERROR;
	}
#endif
	return TCL_OK;
}



int CG_CompareArg( char *name, char *arg )
{
	int i = 0;
	while( *name == *arg && *name && *arg ) {
		name++;
		arg++;
		i++;
	}
	if( i < 2 || *arg )
		return 0;
	else
		return 1;
}

int CG_GetLattice( Tcl_Interp *interp, char *arg )
{
	if( CG_CompareArg( "-concept", arg ) ) {
		// Simply do nothing
		//TclClient->SwitchToConcepttypelist();
		return 1;
	}
	if( CG_CompareArg( "-relation", arg ) ) {
		// We don't support this option any more!
		assert( 0 );
		
		//TclClient->SwitchToRelationtypelist();
		return 1;
	}
	CG_Error( interp, 0 );
	return 0;
}


static KeyCG
string2key (const char *str)
{
    return atol (str);
}

static char *
key2string (KeyCG key)
{
    return (dec (key));
}

static ConceptualGraph *
ConvertStringToCG (Tcl_Interp *interp, char *str)
{
	LocalTclClient client;
	ostrstream ostr;
	ostr << "@answer-graph=" << str << ends;
	istrstream istr( ostr.str() );
	ostr.rdbuf()->freeze( 0 );
	
	Parser *parser = new Parser( &client, istr, NULL );
	parser->Parse();
	if (parser->ErrorOccured ()) {
		ostrstream ostr;
		ostr << parser->GetErrorMsg() << " in line "
                     << parser->GetErrorLine() << ends;
		Tcl_SetResult (interp, ostr.str (), TCL_VOLATILE);
		ostr.rdbuf()->freeze (0);
		delete parser;
		return NULL;
	}
	delete parser;
	
	ConceptualGraph *cg;
	client.GetAnswer( &cg );
	if (!cg) {
		Tcl_SetResult (interp, "CG: cannot parse graph", TCL_STATIC);
		return NULL;
	}
	/*
	 * client.GetAnswer (&cg) arranges to not destroy cg when
	 * client is destroyed. Thus we don't need to copy the graph
	 *
	 * the caller is responsible for freeing cg when no longer needed.
	 */
	return cg;
}

static ConceptualGraph *
GetGraphByKey (Tcl_Interp *interp, char *key_string)
{
	ConceptualGraph *cg;
	KeyCG key = string2key (key_string);
	TclClient->GetGraph (key, &cg);
	if (!cg) {
		ostrstream ostr;
		ostr << "CG: no graph for key " << key << ends;
		Tcl_SetResult (interp, ostr.str (), TCL_VOLATILE);
		ostr.rdbuf()->freeze (0);
		return NULL;
	}
	/*
	 * caller is responsible for freeing cg when no longer needed
	 */
	return cg;
}

static void
StripBlanks( string &word )
{
    int i;
    for (i = 0; i < word.length() && word[i] == ' '; ++i)
        ;
    if (i > 0)
        word.replace (0, i, "");

    for (i = word.length() - 1; i >= 0 && word[i] == ' '; --i)
        ;
    if (i < word.length() - 1)
        word.replace (i+1, word.length()-1 - i, "");
}

static void
ReadTemplateFile (void)
{
    ifstream ifile ("templates.kb");
    if (ifile) {
        char c;
        ifile >> c;
        while (!ifile.eof()) {
            templateGraph += c;
            ifile >> c;
        }
    }
}

//-------------------------------------------------------------------------


static bool localConnection;


int CG_Connect( ClientData clientData, Tcl_Interp *interp,
	       int argc, char *argv[] )
{
	if( argc < 2 ) {
		return CG_Error( interp, 1 );
	}
	if( TclClient != NULL ) {
		return CG_Error( interp, 2 );
	}
    
	if( CG_CompareArg( "-local", argv[ 1 ] ) ) {
    		if (argc > 3)
			return CG_Error (interp, 1);
		localConnection = true;
		TclClient = new LocalTclClient();
		ErrorString err;

                string name = (argc == 2)
                    ? LocateInitialKB ()
                    : string (argv[2]);
		err = ((LocalTclClient*)TclClient)->Connect(name);
		if (!err) {
			Tcl_SetResult( interp, STR2CHAR (err), TCL_VOLATILE );
			delete TclClient;
			TclClient = NULL;
			return TCL_ERROR;
		}
	} else if( CG_CompareArg( "-remote", argv[ 1 ] ) ||
                   CG_CompareArg( "-corba", argv[ 1 ] )) {
		if (argc != 3)
			return CG_Error (interp, 1);
		localConnection = false;

		int lerror;
		if (CG_CompareArg( "-corba", argv[ 1 ]) ){
#ifdef HAVE_CORBA
			orbixTransportLayer *t = new orbixTransportLayer (argv[2]);
			lerror = t->GetError();
			transportLayer=t;
#else
                        return CG_Error( interp, 6 );
#endif
		}
		else{
			RPCTransportLayer *t = new RPCTransportLayer (argv[2]);
			lerror = t->GetError();
			transportLayer=t;
		}


		if( lerror == -1 ) {
			interp->result = "Cannot connect to remote server";
			return TCL_ERROR;
		}
		if( lerror == -2 ) {
			interp->result = "Connection to remote server not allowed";
			return TCL_ERROR;
		}
		TclClient = new ClientProxy( transportLayer );
	} else
		return CG_Error( interp, 6 );

	string ServerVersion;

	TclClient->GetVersion( ServerVersion );
	Tcl_SetResult( interp, STR2CHAR (ServerVersion), TCL_VOLATILE );
	return TCL_OK;
}



int CG_Disconnect( ClientData clientData, Tcl_Interp *interp,
				   int argc, char *argv[] )
{
	CG_CHECK( interp, argc, 1 );
	if( localConnection ) {
		string name( "checkpoint.kb" );
		TclClient->SaveTraderDB( &name );
	} else {
		delete transportLayer;
		transportLayer = NULL;
	}
	delete TclClient;
	TclClient = NULL;
	return TCL_OK;
}


int CG_SaveDataBase( ClientData, Tcl_Interp *interp, int argc, char *argv[] )
{
	CG_CHECK( interp, argc, 2 );
	string name( argv[ 1 ] );
	TclClient->SaveTraderDB( &name );
	return TCL_OK;
}



int CG_Shutdown( ClientData, Tcl_Interp *interp, int argc, char *argv[] )
{
#if 0
	CG_CHECK( interp, argc, 2 );
	if( TclClient->ShutdownTrader( argv[ 1 ] ) == CG_ERROR ) {
		interp->result = "CG: Unable to shutdown server";
		return TCL_ERROR;
	}
#endif
	return TCL_OK;
}



int CG_ElemExists( ClientData, Tcl_Interp *interp, int argc, char *argv[] )
{
	CG_CHECK( interp, argc, 3 );
#if 0
	if( !CG_GetLattice( interp, argv[ 1 ] ) )
		return TCL_ERROR;
#endif

	bool yes;

	string word( argv[ 2 ] );
	StripBlanks( word );
	TclClient->ContainsWord( word, yes );
	interp->result = yes ? "1" : "0";
	return TCL_OK;
}



int CG_GetVersion( ClientData, Tcl_Interp *interp, int argc, char *argv[] )
{
	CG_CHECK( interp, argc, 1 );
	string Result;
	TclClient->GetVersion( Result );
	Tcl_SetResult( interp, STR2CHAR (Result), TCL_VOLATILE );
	return TCL_OK;
}



int CG_ContainsWord( ClientData, Tcl_Interp *interp, int argc, char *argv[] )
{
	CG_CHECK( interp, argc, 2 );
	bool flag;
        string name (argv[1]);
	TclClient->ContainsWord( name, flag );
	interp->result = flag ? "1" : "0";
	return TCL_OK;
}



int CG_AddWord( ClientData, Tcl_Interp *interp, int argc, char *argv[] )
{
	CG_CHECK( interp, argc, 2 );
	string word( argv[ 1 ] );
	StripBlanks( word );
	TclClient->AddWord( word );
	return TCL_OK;
}



int CG_GetRelationPred( ClientData, Tcl_Interp *interp,
		       int argc, char *argv[] )
{
	CG_CHECK( interp, argc, 3 );
        SetOfWord pred;
        string name (argv[1]), rel (argv[2]);
	TclClient->GetRelationPred( name, rel, pred );
        SetOfWord::const_iterator i;
	for (i = pred.begin(); i != pred.end(); ++i ) {
            Tcl_AppendElement( interp, STR2CHAR (*i) );
	}
	return TCL_OK;
}



int CG_GetRelationSucc( ClientData, Tcl_Interp *interp,
		       int argc, char *argv[] )
{
	CG_CHECK( interp, argc, 3 );
        SetOfWord succ;
        string name (argv[1]), rel (argv[2]);
	TclClient->GetRelationSucc( name, rel, succ );
        SetOfWord::const_iterator i;
	for (i = succ.begin(); i != succ.end(); ++i) {
		Tcl_AppendElement( interp, STR2CHAR (*i) );
	}
	return TCL_OK;
}




int CG_GetAllRelations( ClientData, Tcl_Interp *interp,
		       int argc, char *argv[] )
{
	CG_CHECK( interp, argc, 1 );
        SetOfString rels;
	TclClient->GetAllRelations( rels );
        SetOfString::const_iterator i;
	for (i = rels.begin(); i != rels.end(); ++i) {
		Tcl_AppendElement( interp, STR2CHAR (*i) );
	}
	return TCL_OK;
}




int CG_AddRelation( ClientData, Tcl_Interp *interp, int argc, char *argv[] )
{
	CG_CHECK( interp, argc, 4 );
	string w1( argv[ 1 ] );
	StripBlanks( w1 );
	string rel( argv[ 2 ] );
	StripBlanks( rel );
	string w2( argv[ 3 ] );
	StripBlanks( w2 );
	TclClient->AddRelation( w1, rel, w2 );
	return TCL_OK;
}


/*
 * cg_add_graph (ConcetualGraph) -> KeyCG
 */
int CG_AddGraph (ClientData, Tcl_Interp *interp,
		 int argc, char *argv[])
{
	CG_CHECK (interp, argc, 2);
	
	ConceptualGraph *cg = ConvertStringToCG (interp, argv[1]);
	if (!cg)
		return TCL_ERROR;

	KeyCG key;
	TclClient->Add (cg, key);
	Tcl_SetResult (interp, key2string (key), TCL_VOLATILE);
	delete cg;
	return TCL_OK;
}

/*
 * cg_replace_graph (KeyCG, ConcetualGraph)
 */
int CG_ReplaceGraph (ClientData, Tcl_Interp *interp,
		     int argc, char *argv[])
{
	CG_CHECK (interp, argc, 3);

	ConceptualGraph *cg = ConvertStringToCG (interp, argv[2]);
	if (!cg)
		return TCL_ERROR;

	KeyCG key = string2key (argv[1]);
	TclClient->Replace (key, cg);
	delete cg;
	return TCL_OK;
}

/*
 * cg_join_graph (KeyCG, ConceptualGraph)
 */
int CG_JoinGraph (ClientData, Tcl_Interp *interp, int argc, char *argv[])
{
	CG_CHECK (interp, argc, 3);

	ConceptualGraph *cg = ConvertStringToCG (interp, argv[2]);
	if (!cg)
		return TCL_ERROR;

	KeyCG key = string2key (argv[1]);
	TclClient->Join (key, cg);
	delete cg;
	return TCL_OK;
}


/*
 * cg_get_graph (keyCG) -> ConceptualGraph
 */
int CG_GetGraph (ClientData, Tcl_Interp *interp, int argc, char *argv[])
{
	CG_CHECK (interp, argc, 2);

	ConceptualGraph *cg = GetGraphByKey (interp, argv[1]);
	if (!cg)
		return TCL_ERROR;

	ostrstream ostr;
	cg->Print (ostr);
	ostr << ends;
	Tcl_SetResult (interp, ostr.str(), TCL_VOLATILE);
	ostr.rdbuf()->freeze (0);
	delete cg;
	return TCL_OK;
}

/*
 * cg_get_root_concept (keyCG) -> "[concept:{instance1, instance2, ...}]"
 */
int CG_GetRootConcept (ClientData, Tcl_Interp *interp, int argc, char *argv[])
{
	CG_CHECK (interp, argc, 2);

	ConceptualGraph *cg = GetGraphByKey (interp, argv[1]);
	if (!cg)
		return TCL_ERROR;

	cg->SetWidth (NULL);
	ostrstream ostr;
	cg->Print (ostr);
	ostr << ends;
	Tcl_SetResult (interp, ostr.str(), TCL_VOLATILE);
	ostr.rdbuf()->freeze (0);
	delete cg;
	return TCL_OK;
}

/*
 * cg_get_providers (KeyCG) -> {Provider Provider ...}
 */
int CG_GetProviders (ClientData, Tcl_Interp *interp, int argc, char *argv[])
{
	CG_CHECK (interp, argc, 2);

	SetOfServiceEntry entrySet;
	KeyCG key = string2key (argv[1]);
	TclClient->GetProviders (key, entrySet);

	SetOfServiceEntry::iterator it;
	for (it = entrySet.begin (); it != entrySet.end (); it++) {
		Tcl_AppendElement (interp, (char *)(*it).GetOID().c_str());
	}
	return TCL_OK;
}

/*
 * cg_add_provider (KeyCG, Provider)
 */
int CG_AddProvider (ClientData, Tcl_Interp *interp, int argc, char *argv[])
{
	CG_CHECK (interp, argc, 3);

	KeyCG key = string2key (argv[1]);
        string maintainer ("(no maintainer)");
        string oid (argv[2]);
	TclClient->AddProvider (key, maintainer, oid);
	return TCL_OK;
}


/*
 * cg_get_number_of_instances (KeyCG) -> number_of_instances
 */
int CG_GetNumberOfInstances (ClientData, Tcl_Interp *interp,
			     int argc, char *argv[])
{
	CG_CHECK (interp, argc, 2);

	ConceptualGraph *cg = GetGraphByKey (interp, argv[1]);
	if (!cg)
		return TCL_ERROR;

	int number = ((ConceptNode *)cg->GetNode())->GetInstances().size();
	char buf[20];
	sprintf(buf, "%d", number);
	Tcl_SetResult (interp, buf, TCL_VOLATILE);
	delete cg;
	return TCL_OK;
}


/*
 * cg_get_instance (KeyCG, int num) -> instance_name
 */
int CG_GetInstance (ClientData, Tcl_Interp *interp,
		    int argc, char *argv[])
{
	CG_CHECK (interp, argc, 3);
	
	ConceptualGraph *cg = GetGraphByKey (interp, argv[1]);
	if (!cg)
		return TCL_ERROR;

	int nth = atoi (argv[2]);
	SetOfWord &inst = ((ConceptNode *)cg->GetNode())->GetInstances();
        SetOfWord::const_iterator i;
	for (i = inst.begin (); i != inst.end () && nth > 0; ++i, --nth)
		;
	if (i == inst.end()) {
		ostrstream ostr;
		ostr << "CG: requested instance #" << argv[2]
			<< " but there are only " << inst.size()
			<< " instances";
		Tcl_SetResult (interp, ostr.str (), TCL_VOLATILE);
		ostr.rdbuf()->freeze (0);
		delete cg;
		return TCL_ERROR;
	}
	Tcl_SetResult (interp, STR2CHAR (*i), TCL_VOLATILE);
	delete cg;
	return TCL_OK;
}


/*
 * cg_get_number_of_variants (String type_name) -> number_of_variants
 */
int CG_GetNumberOfVariants (ClientData, Tcl_Interp *interp,
			    int argc, char *argv[])
{
	CG_CHECK (interp, argc, 2);

	string type (argv[1]);
	int number = 0;
	TclClient->GetNumberOfVariants (type, number);

	char buf[20];
	sprintf (buf, "%d", number);
	Tcl_SetResult (interp, buf, TCL_VOLATILE);
	return TCL_OK;
}


/*
 * cg_get_variant (String type_name, int num) -> KeyCG
 */
int CG_GetVariant (ClientData, Tcl_Interp *interp, int argc, char *argv[])
{
	CG_CHECK (interp, argc, 3);

	string type (argv[1]);
	KeyCG key;
	TclClient->GetNthVariant (type, atoi (argv[2]), key);
	Tcl_SetResult (interp, key2string (key), TCL_VOLATILE);
	return TCL_OK;
}


/*
 * cg_free_command( String command ) -> String answer
 */
int CG_FreeCommand( ClientData, Tcl_Interp *interp, int argc, char *argv[] )
{
	CG_CHECK( interp, argc, 2 );

	ostrstream ostr;
	ostr << argv[ 1 ] << ends;
	istrstream istr( ostr.str() );
	ostr.rdbuf()->freeze( 0 );
	Parser *parser = new Parser( TclClient, istr, NULL );
	parser->Parse();
	if( parser->ErrorOccured() ) {
		Tcl_SetResult( interp, parser->GetErrorMsg(), TCL_VOLATILE );
		delete parser;
		return TCL_ERROR;
	}
	Tcl_SetResult( interp, STR2CHAR( TclClient->GetAnswerString() ),
				   TCL_VOLATILE );
	delete parser;
	return TCL_OK;
}


/*
 * cg_query_graph (ConceptualGraph query, KeyCG, StringList select_rules)
 *  -> {{KeyCG Akin-Index} {KeyCG Akin-Index} ...}
 */
int CG_QueryGraph (ClientData, Tcl_Interp *interp, int argc, char *argv[])
{
	CG_CHECK (interp, argc, 4);

	if (*argv[1] == '\0') {
		interp->result = "CG: No graph specified for query";
		return TCL_ERROR;
	}

	ConceptualGraph *cg = ConvertStringToCG (interp, argv[1]);
	if (!cg)
		return TCL_ERROR;

	int listArgc;
	char **listArgv;
	if (Tcl_SplitList( interp, argv[3], &listArgc, &listArgv) != TCL_OK)
		return TCL_ERROR;

	SetOfString selecRules;
	for (int k = 0; k < listArgc; k++) {
		string str (listArgv[k]);
		selecRules.insert (str);
	}
	free ((char *)listArgv);

	SetOfMatchResult resultSet;
	TclClient->Match (cg, selecRules, resultSet, string2key (argv[2]));
	SetOfMatchResult::iterator it;
	for (it = resultSet.begin (); it != resultSet.end (); it++) {
                string resString = key2string ((*it).GetKeyCG());
                resString += string(" ") + dec ((*it).GetMatchVal());
                Tcl_AppendElement (interp, STR2CHAR (resString));
	}
	delete cg;
	return TCL_OK;
}



int CG_ExtractTypeFromGraph( ClientData, Tcl_Interp *interp,
			    int argc, char *argv[] )
{
	CG_CHECK( interp, argc, 2 );
	
	if( *argv[ 1 ] == '\0' ) {
		return TCL_ERROR;
	}
	ConceptualGraph *cg = ConvertStringToCG (interp, argv[1]);
	if (!cg)
		return TCL_ERROR;
	
	Word typeName = cg->GetNode()->GetName();
	Tcl_SetResult( interp, STR2CHAR (typeName), TCL_VOLATILE );
	delete cg;
	return TCL_OK;
}



int CG_ExtractInstanceFromGraph( ClientData, Tcl_Interp *interp,
				int argc, char *argv[] )
{
	CG_CHECK( interp, argc, 2 );

	if( *argv[ 1 ] == '\0' )
		return TCL_ERROR;

	ConceptualGraph *cg = ConvertStringToCG (interp, argv[1]);
	if (!cg)
		return TCL_ERROR;

        SetOfWord &inst = ((ConceptNode *) cg->GetNode())->GetInstances();
	Word Result ("");
	if (!inst.empty()) {
            Result = *(inst.begin());
	}
	Tcl_SetResult( interp, STR2CHAR (Result), TCL_VOLATILE );
	delete cg;
	return TCL_OK;
}



int CG_GetVerbose( ClientData, Tcl_Interp *interp, int argc, char *argv[] )
{
#if 0
	CG_CHECK( interp, argc, 1 );
	interp->result = TclClient->CGRc->GetVerbose() ? "1" : "0";
#endif
	return TCL_OK;
}



int CG_GetCommandLogFile( ClientData, Tcl_Interp *interp,
			 int argc, char *argv[] )
{
#if 0
	CG_CHECK( interp, argc, 1 );
	string Result = TclClient->CGRc->GetClientCommandLogfile();
	Tcl_SetResult( interp, Result, TCL_VOLATILE );
#endif
	return TCL_OK;
}



int CG_GetOutputLogFile( ClientData, Tcl_Interp *interp, int argc, char *argv[] )
{
#if 0
	CG_CHECK( interp, argc, 1 );
	string Result = TclClient->CGRc->GetClientOutputLogfile();
	Tcl_SetResult( interp, Result, TCL_VOLATILE );
#endif
	return TCL_OK;
}



int CG_GetLastErrorText( ClientData, Tcl_Interp *interp, int argc, char *argv[] )
{
#if 0
	CG_CHECK( interp, argc, 1 );
#if 0
	string Result = TclClient->CGRc->GetLastError();
	Tcl_SetResult( interp, Result, TCL_VOLATILE );
#endif
	Tcl_SetResult( interp, CGLastError, TCL_VOLATILE );
#endif
	return TCL_OK;
}



void CG_PrintSpaces( ostream &ostr, int &indent, int incr )
{
	if( indent == -1 )
		return;
	indent += incr;
	int num = indent;
	ostr << endl;
	while( num-- )
		ostr << " ";
}



int CG_DumpCG( Tcl_Interp *interp, ostream &ostr, char *cg,
	      int indent, int incr )
{
	int listArgc;
	char **listArgv;

	if( Tcl_SplitList( interp, cg, &listArgc, &listArgv ) != TCL_OK )
		return TCL_ERROR;
	if( listArgc == 0 ) {
		free( (char *) listArgv );
		return TCL_OK;
	}

	ostr << "[" << listArgv[ 0 ] << "] ";
	int count = listArgc - 1;
	if( count == 0 ) {
		free( (char *) listArgv );
		return TCL_OK;
	}
	if( count != 1 ) {
		ostr << "- ";
		CG_PrintSpaces( ostr, indent, incr );
	}
    
	int i = 1;
	while( i < listArgc ) {
		int listArgc1;
		char **listArgv1;

		if( Tcl_SplitList( interp, listArgv[ i ], &listArgc1,
				  &listArgv1 ) != TCL_OK ) {
			free( (char *) listArgv );
			return TCL_ERROR;
		}
		if( listArgc1 <= 1 ) {
			free( (char *) listArgv );
			free( (char *) listArgv1 );
			return TCL_ERROR;
		}
		ostr << "-> (";
		ostr << listArgv1[ 0 ];
		ostr << ") ";
		if( listArgc1 > 2 ) {
			ostr << "- ";
			CG_PrintSpaces( ostr, indent, incr );
		}
	
		int i1 = 1;
		while( i1 < listArgc1 ) {
			ostr << "-> ";
	    
			if( CG_DumpCG( interp, ostr, listArgv1[ i1 ],
				      indent, incr ) != TCL_OK ) {
				free( (char *) listArgv );
				free( (char *) listArgv1 );
				return TCL_ERROR;
			}
			i1++;
			if( i1 < listArgc1 ) {
				ostr << ", ";
				CG_PrintSpaces( ostr, indent, 0 );
			}
		}
		if( listArgc1 > 2 ) {
			ostr << ". ";
			if( indent != -1 )
				indent -= incr;
		}
	
		free( (char *) listArgv1 );
	
		i++;
		if( i < listArgc ) {
			ostr << ", ";
			CG_PrintSpaces( ostr, indent, 0 );//-incr );
		} else if( count != 1 ) {
			ostr << ". ";
		}
	}
	free( (char *) listArgv );
	return TCL_OK;
}



int CG_ConvertToCG( ClientData, Tcl_Interp *interp, int argc, char *argv[] )
{
	int prettyprint = -1;
	int incr;
	char *cg = argv[ 1 ];

	if( argc == 4 ) {
		if( CG_CompareArg( "-prettyprint", argv[ 1 ] ) ) {
			prettyprint = 0;
			if( Tcl_GetInt( interp, argv[ 2 ],
				       &incr ) == TCL_ERROR )
				return TCL_ERROR;
			if( incr < 0 || incr > 42 ) {
				interp->result = "CG: Increment must be "
					"between 0 and 42";
				return TCL_ERROR;
			}
			cg = argv[ 3 ];
		} else {
			interp->result = "CG: Wrong usage";
			return TCL_ERROR;
		}
	} else if( argc != 2 ) {
		interp->result = "CG: Expected 1 or 2 args";
		return TCL_ERROR;
	}

	ostrstream Result;

	if( CG_DumpCG( interp, Result, cg, prettyprint, incr ) != TCL_OK ) {
		interp->result = "CG: Not a CG";
		Result.rdbuf()->freeze( 0 );
		return TCL_ERROR;
	}
	Result << ends;
	Tcl_SetResult( interp, Result.str(), TCL_VOLATILE );
	Result.rdbuf()->freeze( 0 );
	return TCL_OK;
}



int CG_ConvertToTcl (ClientData, Tcl_Interp *interp, int argc, char *argv[])
{
	if( *argv[ 1 ] == '\0' )
		return TCL_ERROR;

	ConceptualGraph *cg = ConvertStringToCG (interp, argv[1]);
	if (!cg)
		return TCL_ERROR;

	ostrstream ostr;
	LocalTclClient client;
	client.ConvertToTcl (ostr, cg);
	ostr << ends;
	
	Tcl_SetResult (interp, ostr.str(), TCL_VOLATILE);
	ostr.rdbuf()->freeze (0);
	delete cg;
	return TCL_OK;
}


int CG_GetMatchingRuleNames( ClientData, Tcl_Interp *interp,
			    int argc, char *argv[] )
{
	CG_CHECK( interp, argc, 1 );
	string names;
	TclClient->GetMatchingRuleNames (&names);
	int last_i = 0;
	for( int i = 0; i < names.length(); i++ ) {
		if( names[ i ] == '\n' ) {
			string aName = names.substr( last_i, i - last_i );
			Tcl_AppendElement( interp, STR2CHAR (aName) );
			last_i = i + 1;
		}
	}
	return TCL_OK;
}


int CG_GetMatchingRule( ClientData, Tcl_Interp *interp,
		       int argc, char *argv[] )
{
	CG_CHECK( interp, argc, 2 );
	string name( argv[ 1 ] );
	string rule;
	TclClient->GetMatchingRule( &name, &rule );
	Tcl_SetResult( interp, STR2CHAR (rule), TCL_VOLATILE );
	return TCL_OK;
}



int CG_DefineMatchingRule( ClientData, Tcl_Interp *interp,
			  int argc, char *argv[] )
{
	CG_CHECK( interp, argc, 3 );
	string name( argv[ 1 ] );
	string rule( argv[ 2 ] );
	TclClient->DefineMatchingRule( &name, &rule );
	return TCL_OK;
}



int CG_ConvertIDLtoCG( ClientData, Tcl_Interp *interp, int argc, char *argv[] )
{
    if (argc != 2) {
        Tcl_SetResult (interp, "cg_convert_idl_to_cg: wrong number of args",
                       TCL_STATIC);
        return TCL_ERROR;
    }

#ifdef HAVE_IDL
    ostrstream ostr;
    ostr << argv[ 1 ] << ends;
    istrstream istr( ostr.str() );
    ostr.rdbuf()->freeze( 0 );
    ostrstream output;
    IDLParser conv( istr, "stdin", output );
    conv.Parse();
    if( conv.ErrorOccured() ) {
        string err;
        err = string ("CG: ") + conv.GetErrorMsg() + " in line "
            + dec(conv.GetErrorLine());
        Tcl_SetResult( interp, STR2CHAR (err), TCL_VOLATILE );
        return TCL_ERROR;
    }
    output << ends;
    Tcl_SetResult( interp, output.str(), TCL_VOLATILE );
    output.rdbuf()->freeze( 0 );
    return TCL_OK;
#else
    Tcl_SetResult (interp, "cg_convert_idl_to_cg: not implemented",
                   TCL_STATIC);
    return TCL_ERROR;
#endif
}



int CG_ConvertCGtoIDL( ClientData, Tcl_Interp *interp, int argc, char *argv[] )
{
    if (argc != 3) {
        Tcl_SetResult (interp, "cg_convert_cg_to_idl: wrong number of args",
                       TCL_STATIC);
        return TCL_ERROR;
    }

#ifdef HAVE_IDL
    string type( argv[ 1 ] );
    ConceptualGraph *cg = ConvertStringToCG (interp, argv[ 2 ]);
    if (!cg)
        return TCL_ERROR;
    
    ostrstream ostr;
    
    if( type == "DCE" ) {
        Template2DCE conv( cg, ostr );
        conv.CreateSpecification();
        ostr << ends;
        Tcl_SetResult( interp, ostr.str(), TCL_VOLATILE );
        ostr.rdbuf()->freeze( 0 );
        delete cg;
        return TCL_OK;
    } else if( type == "CORBA" ) {
        Template2CORBA conv( cg, ostr );
        conv.CreateSpecification();
        ostr << ends;
        Tcl_SetResult( interp, ostr.str(), TCL_VOLATILE );
        ostr.rdbuf()->freeze( 0 );
        delete cg;
        return TCL_OK;
    }
    interp->result = "CG: Wrong IDL type";
    return TCL_ERROR;
#else
    Tcl_SetResult (interp, "cg_convert_cg_to_idl: not implemented",
                   TCL_STATIC);
    return TCL_ERROR;
#endif
}


int
CG_GetTemplates (ClientData, Tcl_Interp *interp, int argc, char *argv[])
{
    ConceptualGraph *cg = ConvertStringToCG (interp, STR2CHAR(templateGraph));
    if (!cg)
        return TCL_ERROR;

    LocalTclClient client;
    ConceptualGraph *cg2;
    for (cg2 = cg->GetWidth(); cg2; cg2 = cg2->GetDepth()) {
        string s = cg2->GetNode()->GetName();

	ostrstream ostr;
	client.ConvertToTcl (ostr, cg2);
	ostr << ends;
        s += string(" {") + ostr.str() + "}";
        ostr.rdbuf()->freeze (0);

	Tcl_AppendElement (interp, STR2CHAR(s));
    }
    return TCL_OK;
}

int
CG_CreateProfile (ClientData, Tcl_Interp *interp, int argc, char *argv[])
{
    CG_CHECK (interp, argc, 1);
    KeyCG key;
    ErrorString err = TclClient->CreateProfile (key);
    if (err.GetErrNum())
        return TCL_ERROR;
    Tcl_SetResult (interp, key2string (key), TCL_VOLATILE);
    return TCL_OK;
}

int
CG_RemoveProfile (ClientData, Tcl_Interp *interp, int argc, char *argv[])
{
    CG_CHECK (interp, argc, 2);
    ErrorString err = TclClient->RemoveProfile (string2key (argv[1]));
    if (err.GetErrNum())
        return TCL_ERROR;
    return TCL_OK;
}

int
CG_SetMinMatchval (ClientData, Tcl_Interp *interp, int argc, char *argv[])
{
    CG_CHECK (interp, argc, 3);
    ErrorString err = TclClient->SetMinmatchval (string2key (argv[1]),
                                                 atol (argv[2]));
    if (err.GetErrNum())
        return TCL_ERROR;
    return TCL_OK;
}

int
CG_SetTimeout (ClientData, Tcl_Interp *interp, int argc, char *argv[])
{
    CG_CHECK (interp, argc, 3);
    ErrorString err = TclClient->SetTimeout (string2key (argv[1]),
                                             atol (argv[2]));
    if (err.GetErrNum())
        return TCL_ERROR;
    return TCL_OK;
}

int
CG_GetMinMatchval (ClientData, Tcl_Interp *interp, int argc, char *argv[])
{
    CG_CHECK (interp, argc, 2);
    long mmv;
    ErrorString err = TclClient->GetMinmatchval (string2key (argv[1]), mmv);
    if (err.GetErrNum())
        return TCL_ERROR;
    Tcl_SetResult (interp, dec (mmv), TCL_VOLATILE);
    return TCL_OK;
}

int
CG_GetTimeout (ClientData, Tcl_Interp *interp, int argc, char *argv[])
{
    CG_CHECK (interp, argc, 2);
    long tmo;
    ErrorString err = TclClient->GetTimeout (string2key (argv[1]), tmo);
    if (err.GetErrNum())
        return TCL_ERROR;
    Tcl_SetResult (interp, dec (tmo), TCL_VOLATILE);
    return TCL_OK;
}

int CG_Init( Tcl_Interp *interp )
{
	static int count = 0;
    
	if (count++) {
		cerr << "CG: This package is not intended for "
			"multiple interpreters" << endl;
		exit (-1);
	}
	CG_Interp = interp;
	TclClient = NULL;
    
	Tcl_CreateCommand (interp, "cg_connect",
			   CG_Connect, NULL, (Tcl_CmdDeleteProc *)NULL);

	Tcl_CreateCommand (interp, "cg_disconnect",
			   CG_Disconnect, NULL, (Tcl_CmdDeleteProc *)NULL);

	Tcl_CreateCommand (interp, "cg_save_database",
			   CG_SaveDataBase, NULL, (Tcl_CmdDeleteProc *)NULL);
#if 0
	Tcl_CreateCommand (interp, "cg_shutdown",
			   CG_Shutdown, NULL, (Tcl_CmdDeleteProc *)NULL);
#endif

	Tcl_CreateCommand (interp, "cg_exists",
			   CG_ElemExists, NULL, (Tcl_CmdDeleteProc *)NULL);

	Tcl_CreateCommand (interp, "cg_get_version",
			   CG_GetVersion, NULL, (Tcl_CmdDeleteProc *)NULL);

	Tcl_CreateCommand (interp, "cg_contains_word",
			   CG_ContainsWord, NULL, (Tcl_CmdDeleteProc *)NULL);

	Tcl_CreateCommand (interp, "cg_add_word",
			   CG_AddWord, NULL, (Tcl_CmdDeleteProc *)NULL);

	Tcl_CreateCommand (interp, "cg_get_relation_pred",
			   CG_GetRelationPred, NULL,
			   (Tcl_CmdDeleteProc *)NULL);

	Tcl_CreateCommand (interp, "cg_get_relation_succ",
			   CG_GetRelationSucc, NULL,
			   (Tcl_CmdDeleteProc *)NULL);

	Tcl_CreateCommand (interp, "cg_get_all_relations",
			   CG_GetAllRelations, NULL,
			   (Tcl_CmdDeleteProc *)NULL);

	Tcl_CreateCommand (interp, "cg_add_relation",
			   CG_AddRelation, NULL, (Tcl_CmdDeleteProc *)NULL);

	Tcl_CreateCommand (interp, "cg_add_graph",
			   CG_AddGraph, NULL,
			   (Tcl_CmdDeleteProc *)NULL);

	Tcl_CreateCommand (interp, "cg_replace_graph",
			   CG_ReplaceGraph, NULL,
			   (Tcl_CmdDeleteProc *)NULL);

	Tcl_CreateCommand (interp, "cg_join_graph",
			   CG_JoinGraph, NULL,
			   (Tcl_CmdDeleteProc *)NULL);

	Tcl_CreateCommand (interp, "cg_get_graph",
			   CG_GetGraph, NULL, (Tcl_CmdDeleteProc *)NULL);

	Tcl_CreateCommand (interp, "cg_get_root_concept",
			   CG_GetRootConcept, NULL, (Tcl_CmdDeleteProc *)NULL);

	Tcl_CreateCommand (interp, "cg_get_providers",
			   CG_GetProviders, NULL,
			   (Tcl_CmdDeleteProc *) NULL);

	Tcl_CreateCommand (interp, "cg_add_provider",
			   CG_AddProvider, NULL,
			   (Tcl_CmdDeleteProc *) NULL);

	Tcl_CreateCommand (interp, "cg_get_instance",
			   CG_GetInstance, NULL, (Tcl_CmdDeleteProc *) NULL);

	Tcl_CreateCommand (interp, "cg_get_number_of_instances",
			   CG_GetNumberOfInstances, NULL,
			   (Tcl_CmdDeleteProc *) NULL );

	Tcl_CreateCommand (interp, "cg_get_variant",
			   CG_GetVariant, NULL, (Tcl_CmdDeleteProc *) NULL);

	Tcl_CreateCommand (interp, "cg_get_number_of_variants",
			   CG_GetNumberOfVariants, NULL,
			   (Tcl_CmdDeleteProc *) NULL);

	Tcl_CreateCommand (interp, "cg_free_command",
			   CG_FreeCommand, NULL, (Tcl_CmdDeleteProc *) NULL);

	Tcl_CreateCommand (interp, "cg_query_graph",
			   CG_QueryGraph, NULL, (Tcl_CmdDeleteProc *) NULL);

	Tcl_CreateCommand (interp, "cg_extract_type_from_graph",
			   CG_ExtractTypeFromGraph, NULL,
			   (Tcl_CmdDeleteProc *) NULL);

	Tcl_CreateCommand (interp, "cg_extract_instance_from_graph",
			   CG_ExtractInstanceFromGraph, NULL,
			   (Tcl_CmdDeleteProc *) NULL);
#if 0
	Tcl_CreateCommand (interp, "cg_get_verbose", CG_GetVerbose,
			   NULL, (Tcl_CmdDeleteProc *) NULL);

	Tcl_CreateCommand (interp, "cg_get_command_log_file",
			   CG_GetCommandLogFile, NULL,
			   (Tcl_CmdDeleteProc *) NULL);

	Tcl_CreateCommand (interp, "cg_get_output_log_file",
			   CG_GetOutputLogFile, NULL,
			   (Tcl_CmdDeleteProc *) NULL);

	Tcl_CreateCommand (interp, "cg_get_last_error_text",
			  CG_GetLastErrorText, NULL,
			   (Tcl_CmdDeleteProc *) NULL);
#endif
	Tcl_CreateCommand (interp, "cg_convert_to_cg",
			   CG_ConvertToCG, NULL, (Tcl_CmdDeleteProc *) NULL);

	Tcl_CreateCommand (interp, "cg_convert_to_tcl",
			   CG_ConvertToTcl, NULL, (Tcl_CmdDeleteProc *) NULL);

	Tcl_CreateCommand (interp, "cg_get_matching_rule_names",
			   CG_GetMatchingRuleNames, NULL,
			   (Tcl_CmdDeleteProc *) NULL);

	Tcl_CreateCommand (interp, "cg_get_matching_rule",
			   CG_GetMatchingRule, NULL,
			   (Tcl_CmdDeleteProc *) NULL);

	Tcl_CreateCommand (interp, "cg_define_matching_rule",
			   CG_DefineMatchingRule, NULL,
			   (Tcl_CmdDeleteProc *) NULL);

	Tcl_CreateCommand (interp, "cg_convert_idl_to_cg",
			   CG_ConvertIDLtoCG, NULL,
			   (Tcl_CmdDeleteProc *) NULL);

	Tcl_CreateCommand (interp, "cg_convert_cg_to_idl",
			   CG_ConvertCGtoIDL, NULL,
			   (Tcl_CmdDeleteProc *) NULL);

        Tcl_CreateCommand (interp, "cg_get_templates",
                           CG_GetTemplates, NULL,
			   (Tcl_CmdDeleteProc *) NULL);

        Tcl_CreateCommand (interp, "cg_create_profile",
                           CG_CreateProfile, NULL,
			   (Tcl_CmdDeleteProc *) NULL);

        Tcl_CreateCommand (interp, "cg_remove_profile",
                           CG_RemoveProfile, NULL,
			   (Tcl_CmdDeleteProc *) NULL);

        Tcl_CreateCommand (interp, "cg_set_minmatchval",
                           CG_SetMinMatchval, NULL,
			   (Tcl_CmdDeleteProc *) NULL);

        Tcl_CreateCommand (interp, "cg_set_timeout",
                           CG_SetTimeout, NULL,
			   (Tcl_CmdDeleteProc *) NULL);

        Tcl_CreateCommand (interp, "cg_get_minmatchval",
                           CG_GetMinMatchval, NULL,
			   (Tcl_CmdDeleteProc *) NULL);

        Tcl_CreateCommand (interp, "cg_get_timeout",
                           CG_GetTimeout, NULL,
			   (Tcl_CmdDeleteProc *) NULL);

        ReadTemplateFile ();
	return TCL_OK;
}

/*
 * Local variables:
 * c-indent-level: 8
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -8
 * c-argdecl-indent: 8
 * c-label-offset: -8
 * c-continued-statement-offset: 8
 * c-continued-brace-offset: 0
 * End:
 */
