
/*
 *  AI-Trader --- Service Trading using Conceptual Graphs
 *  Copyright (C) 1995 Arno Puder et al.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 aitrader@informatik.uni-frankfurt.de
 */


// $Id: cg.cc,v 1.10 1997/04/09 14:37:10 roemer Exp $


#include <assert.h>
#include "cg.h"
#include "util.h"

SetOfWord emptyWordSet;


Node::Node()
        : name ("")
{
}

Node::~Node()
{
}


Word &Node::GetName()
{
    return name;
}


RelationNode::RelationNode( const Word &aName )
{
    name = aName;
}


RelationNode::RelationNode( RelationNode *node )
{
    name = node->name;
}


RelationNode::~RelationNode()
{
}


int RelationNode::IsA()
{
    return RELATION_NODE;
}

int RelationNode::operator==( Node &node )
{
    return ( node.IsA() == RELATION_NODE ) && 
		( name == ((RelationNode&) node).name );
}


int RelationNode::operator<=( Node &node )
{
    return ( node.IsA() == RELATION_NODE ) && 
		( name <= ((RelationNode&) node).name );
}


void RelationNode::Print( ostream &ostr )
{
    ostr << "(" << name << ")";
}


ConceptNode::ConceptNode( const Word &aName, SetOfWord &theInst )
{
    name = aName;
    inst = theInst;
	value = NIL_VALUE;
}


ConceptNode::ConceptNode( ConceptNode *node )
{
    name = node->name;
    inst = node->inst;
	value = node->GetValue();
}


ConceptNode::~ConceptNode()
{
}


SetOfWord &ConceptNode::GetInstances()
{
	return inst;
}


void ConceptNode::AddInstances( SetOfWord &new_inst )
{
	inst += new_inst;
}


void ConceptNode::SetInstances( SetOfWord &new_inst )
{
	inst = new_inst;
}


void ConceptNode::SetValue( long aValue )
{
	value = aValue;
}


long ConceptNode::GetValue()
{
	return value;
}


void ConceptNode::Print( ostream &ostr )
{
    ostr << "[" << name;
    if( inst.size() != 0 ) {
        ostr << ":{";
        int first = 1;
        SetOfWord::const_iterator i;
        for (i = inst.begin(); i != inst.end(); ++i) {
            if( first )
                first = 0;
            else
                ostr << ", ";
            ostr << *i;
        }
        ostr << "}";
    } else {
        if( value != NIL_VALUE )
            ostr << ":" << value;
    }
    ostr << "]";
}


int ConceptNode::IsA()
{
    return CONCEPT_NODE;
}


int ConceptNode::operator==( Node &node )
{
    if (node.IsA () != CONCEPT_NODE)
        return false;

    ConceptNode &cnode = (ConceptNode&)node;
    return (name == cnode.name &&
            (inst == cnode.inst || inst < cnode.inst) &&
            value == cnode.value);
}


int ConceptNode::operator<=( Node &node )
{
    return ( node.IsA() == CONCEPT_NODE ) &&
		( name <= ((ConceptNode&) node).name );
}


ConceptualGraph::ConceptualGraph( Node *aNode, ConceptualGraph *theWidth, 
                                  ConceptualGraph *theDepth )
{
    validate ();
    node = aNode;
    width = theWidth;
    depth = theDepth;
    tag = NODE_UNKNOWN;
}

ConceptualGraph::~ConceptualGraph()
{
    invalidate ();
    assert( node != NULL );
    delete node;
    if( width )
        delete width;
    if( depth )
        delete depth;
}

void
ConceptualGraph::sanity ()
{
#if 1
    assert (magic == CGMAGIC);
#endif
}

void ConceptualGraph::SetNode( Node *aNode )
{
    sanity ();
    if( node )
        delete node;
    node = aNode;
}


Node *ConceptualGraph::GetNode()
{
    sanity ();
    return node;
}


void ConceptualGraph::SetWidth( ConceptualGraph *theWidth )
{
    sanity ();
    if( width )
        delete width;
    width = theWidth;
}

void ConceptualGraph::SetDepth( ConceptualGraph *theDepth )
{
    sanity ();
    if( depth )
        delete depth;
    depth = theDepth;
}


ConceptualGraph *ConceptualGraph::GetWidth()
{
    sanity ();
    return width;
}


ConceptualGraph *ConceptualGraph::GetDepth()
{
    sanity ();
    return depth;
}


void ConceptualGraph::SetTag( int aTag )
{
    sanity ();
    tag = aTag;
}


int ConceptualGraph::GetTag()
{
    sanity ();
    return tag;
}



void ConceptualGraph::ResetTags( int aTag )
{
    sanity ();
    tag = aTag;
    if( width )
        width->ResetTags( aTag );
    if( depth )
        depth->ResetTags( aTag );
}



void ConceptualGraph::Print( ostream &ostr, bool withTags )
{
    sanity ();
    node->Print( ostr );
    if( withTags && node->IsA() == CONCEPT_NODE ) {
        switch( tag ) {
        case NODE_CANTTELL:
        case NODE_UNKNOWN:
            ostr << "/*U*/";
            break;
        case NODE_ACCEPTED:
            ostr << "/*A*/";
            break;
        case NODE_REJECTED:
            ostr << "/*R*/";
            break;
        default:
            assert( 0 );
        }
    }
    if( width ) {
        if( width->GetDepth() )
            ostr << "-";
        ostr << "->";
        width->Print( ostr, withTags );
    }
    if( width && width->GetDepth() )
        ostr << ".";
    if( depth ) {
        ostr << ",->";
        depth->Print( ostr, withTags );
    }
}



void ConceptualGraph::PrettyPrint( ostream &ostr, int incr )
{
	DoPrettyPrint( ostr, 0, incr );
}



void ConceptualGraph::DoPrettyPrint( ostream &ostr, int indent, int incr )
{
	node->Print( ostr );
	if( width != NULL ) {
		if( width->GetDepth() != NULL ) {
			ostr << "-";
			indent = PrintSpaces( ostr, indent, incr );
		}
		ostr << "->";
		width->DoPrettyPrint( ostr, indent, incr );
	}
	int new_incr = 0;
	if( width != NULL && width->GetDepth() != NULL ) {
		ostr << ".";
		new_incr = -incr;
	}
	if( depth != NULL ) {
		ostr << ",";
		indent = PrintSpaces( ostr, indent, new_incr );
		ostr << "->";
		depth->DoPrettyPrint( ostr, indent, incr );
	}
}



int ConceptualGraph::PrintSpaces( ostream &ostr, int indent, int incr )
{
	if( indent == -1 )
		return indent;
	indent += incr;
	int num = indent;
	ostr << "\n";
	while( num-- != 0 )
		ostr << " ";
	return indent;
}


bool
ConceptualGraph::operator== (ConceptualGraph &cg2)
{
    if (!(*GetNode() == *cg2.GetNode()))
        return false;

    vector<ConceptualGraph *> childs;
    for (ConceptualGraph *p1 = cg2.GetWidth(); p1; p1 = p1->GetDepth())
        childs.push_back (p1);

    for (ConceptualGraph *p = GetWidth(); p; p = p->GetDepth()) {
        bool found = false;
        vector<ConceptualGraph *>::iterator i;
        for (i = childs.begin(); i != childs.end(); ++i) {
            if (**i == *p) {
                childs.erase (i);
                found = true;
                break;
            }
        }
        if (!found)
            return false;
    }
    return childs.size() == 0;
}


void
ConceptualGraph::Subtract (ConceptualGraph &cg2)
{
    for (ConceptualGraph *p2 = cg2.GetWidth(); p2; p2 = p2->GetDepth()) {
        ConceptualGraph **parent, *p1;
        for (parent = &width, p1 = *parent; p1; p1 = *parent) {
            if (*p1 == *p2) {
                *parent = p1->depth;
                p1->depth = NULL;
                delete p1;
            } else {
                parent = &p1->depth;
            }
        }
    }
}

		
Node *CopyNode( Node *node )
{
    assert( node->IsA() == CONCEPT_NODE || node->IsA() == RELATION_NODE );
    if( node->IsA() == CONCEPT_NODE ) 
        return new ConceptNode( (ConceptNode *) node );
    else
        return new RelationNode( (RelationNode *) node );
}


ConceptualGraph *CopyGraph( ConceptualGraph *node )
{
    if( node == NULL )
        return NULL;
    ConceptualGraph *width = CopyGraph( node->GetWidth() );
    ConceptualGraph *depth = CopyGraph( node->GetDepth() );
    Node *graph_node = CopyNode( node->GetNode() );
    ConceptualGraph *new_node = new ConceptualGraph( graph_node,
                                                     width, depth );
    new_node->SetTag( node->GetTag() );
    return new_node;
}
