
/*
 *  AI-Trader --- Service Trading using Conceptual Graphs
 *  Copyright (C) 1995 Arno Puder et al.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 aitrader@informatik.uni-frankfurt.de
 */

// $Id: dict.cc,v 1.14 1998/03/06 16:14:46 farsi Exp $


#include "valid.h"
#include "dict.h"

#ifdef NOT_USED
template<class keytype, class valtype, class comp, class OutputIterator>
void getmapkeys(const map<keytype, valtype, comp> &m, OutputIterator &result)
{
    typedef const pair<const keytype, valtype> &entry;
    struct extract : public unary_function<entry, string>
    {
        extract() {}
        string operator() (entry e)
        {
            return e.first;
        }
    };
    transform(m.begin(), m.end(), result, extract());
}
#endif

#ifdef NOT_USED
template<class A, class InputIterator>
static inline void diff(A &a, InputIterator first, InputIterator last)
{
    for (InputIterator i=first; i!=last; ++i)
        a.erase(*i);
}


template<class a, class b>
static inline void diff(a &aa, const b &bb)
{
    diff(aa, bb.begin(), bb.end());
}
#endif

Dict::veptr Dict::vepempty;


Dict::Dict()
        : isquickmode(false),
          isverbose(false)
{
}


Dict::Dict(const Dict &d)
        : entries(d.entries),
          relations(d.relations),
          isquickmode(d.isquickmode),
          isverbose(d.isverbose)
{
}


Dict::~Dict()
{
    meptr::iterator first=entries.begin();
    const meptr::iterator last=entries.end();
    for (; first != last; ++first) {
        delete (Word *)(const Word *)(*first).first;
        delete (Entry *)(const Entry *)(*first).second;
    }
}


Dict::vcwp
Dict::GetRelImmPred(const Word &w, const string &r)
#ifdef __GNUC__
    return words;
#endif
{
#ifndef __GNUC__
    vcwp words;
#endif
    const veptr &ents = GetRelImm (Entry::pred, w, r);
    for (veptr::const_iterator i = ents.begin(); i != ents.end(); ++i)
        words.push_back ((**i).GetWordp ());
#ifndef __GNUC__
    return words;
#endif
}


Dict::vcwp
Dict::GetRelImmSucc(const Word &w, const string &r)
#ifdef __GNUC__
    return words;
#endif
{
#ifndef __GNUC__
    vcwp words;
#endif
    const veptr &ents = GetRelImm (Entry::succ, w, r);
    for (veptr::const_iterator i = ents.begin(); i != ents.end(); ++i)
        words.push_back ((**i).GetWordp ());
#ifndef __GNUC__
    return words;
#endif
}


Dict::vcwp
Dict::GetRelAllSucc (const Word &w, const string &r)
#ifdef __GNUC__
    return words;
#endif
{
#ifndef __GNUC__
    vcwp words;
#endif
    if (ContainsWord(w) && ContainsRelation (r)) {
        pair<int,int> &rel = relations[r];
        eptr ep = entries[GetCWP(w)];

        septr wset = GetRelAll (Entry::succ, ep, rel);
        for (septr::const_iterator i = wset.begin(); i != wset.end(); ++i) {
            words.push_back ((**i).GetWordp());
        }
    } else {
        cerr << "Dict::GetRelAllSucc(" << w << ", " << r
             << "): word or relation not registered" << endl; 
    }
#ifndef __GNUC__
    return words;
#endif
}


Dict::vcwp
Dict::GetRelAllPred (const Word &w, const string &r)
#ifdef __GNUC__
    return words;
#endif
{
#ifndef __GNUC__
    vcwp words;
#endif
    if (ContainsWord(w) && ContainsRelation (r)) {
        pair<int,int> &rel = relations[r];
        eptr ep = entries[GetCWP(w)];

        septr wset = GetRelAll (Entry::pred, ep, rel);
        for (septr::const_iterator i = wset.begin(); i != wset.end(); ++i) {
            words.push_back ((**i).GetWordp());
        }
    } else {
        cerr << "Dict::GetRelAllSucc(" << w << ", " << r
             << "): word or relation not registered" << endl; 
    }
#ifndef __GNUC__
    return words;
#endif
}


bool Dict::RegisterRelation(const string &rel, int type)
{
    // some consistency checks:
    ASSERT(0==(type&~(RelSymmetric|RelReflexive|RelTransitive|RelAcyclic)));
    ASSERT(!(type&RelSymmetric && type&RelAcyclic));
    if (ContainsRelation(rel))
        return false;
    else {
        int number=relations.size();
        relations[rel]=make_pair(type, number);
        return true;
    }
}

vector<string> Dict::GetRelationNames() const
{
    vector<string> v(relations.size());
    for (msp::const_iterator i = relations.begin(); i != relations.end(); ++i)
        v[(*i).second.second] = (*i).first;
    return v;
}


string
Dict::GetRelationName (const pair<int, int> &rel)
{
    for (msp::const_iterator i = relations.begin(); i != relations.end(); ++i)
        if ((*i).second == rel)
            return (*i).first;
    return "NONE";
}


void Dict::DeleteWord(const Word &)
{
    NOT_IMPLEMENTED;
}


static inline void addrelation_errmsg(const Word &w1, const string &rel,
                                      const Word &w2, const char *errmsg)
{
    dprintf("Dict::AddRelation(\"%s\", \"%s\", \"%s\") : %s\n",
            (const char*)w1, rel.c_str(), (const char*)w2, errmsg);
}


static inline void addrelation_remrel(const Word &w1, const string &rel,
                                      const Word &w2,
                                      const Word &w3, const Word &w4)
{
    dprintf("Dict::AddRelation(\"%s\", \"%s\", \"%s\") : "
            "removed \"%s\"->\"%s\"\n",
            (const char*)w1, rel.c_str(), (const char*)w2, (const char*)w3,
            (const char*)w4);
}


bool Dict::AddRelation(const Word &w1, const string &srel, const Word &w2)
{
    if (!ContainsWord(w1)) {
        addrelation_errmsg(w1, srel, w2, "first word not in dictionary");
        return false;
    }
    if (!ContainsWord(w2)) {
        addrelation_errmsg(w1, srel, w2, "second word not in dictionary");
        return false;
    }
    if (!ContainsRelation(srel)) {
        addrelation_errmsg(w1, srel, w2, "relation not registered");
        return false;
    }
    if (isquickmode) {
        pair<int,int> &rel=relations[srel];
        const int type=rel.first;
        const int relno=rel.second;

        eptr ep1 = entries[GetCWP(w1)];
        eptr ep2 = entries[GetCWP(w2)];

        (*ep1).AddRel(Entry::succ, relno, ep2);
        (*ep2).AddRel(Entry::pred, relno, ep1);
        
        // if the relation is symmetric, add the relation in the
        // inverse sense, too
        if (type&RelSymmetric) {
            (*ep2).AddRel(Entry::succ, relno, ep1);
            (*ep1).AddRel(Entry::pred, relno, ep2);
        }
    } else {
        if (AreRelated(w1, srel, w2)) {
            addrelation_errmsg(w1, srel, w2, "exists");
            return true;
        }

        pair<int,int> &rel=relations[srel];
        const int type=rel.first;
        const int relno=rel.second;

        eptr ep1 = entries[GetCWP(w1)];
        eptr ep2 = entries[GetCWP(w2)];

        // verify that the acyclic characteristic does not lead to
        // contradiction. note that this check even works if the
        // relation is transitive
        if (type&RelAcyclic && AreRelated(w2, srel, w1)) {
            addrelation_errmsg(w1, srel, w2,
                               "attempt to contradict acyclic property");
            return false;
        }

        // now take care of the transitive characteristic. since we
        // only store a 'generating system' of the transitive
        // relation, it may be necessary to remove superfluous edges
        if (type&RelTransitive) {
            // build a list of all successors of w2 including w2
            septr allsucc = GetRelAll (Entry::succ, ep2, rel);
            allsucc.insert (ep2);
    
            // build a list of all predecessors of w1 including w1
            septr allpred = GetRelAll (Entry::pred, ep1, rel);
            allpred.insert (ep1);

            // go through all the predecessors
            for (septr::iterator i=allpred.begin(); i!=allpred.end(); ++i) {
                eptr ep3 = *i;

                // and get their immediate successors
                const veptr &immsucc=GetRelImm(Entry::succ, ep3, relno);

                // find words that are in the cut of immsucc and
                // allsucc.
                for (veptr::const_iterator j=immsucc.begin(); j!=immsucc.end(); ) {
                    eptr ep4 = *j;
                    if (allsucc.find(ep4) != allsucc.end()) {
                        // words that are in the cut are the
                        // destinations of relations that have to be
                        // removed
                        addrelation_remrel(w1, srel, w2,
                                           *((*ep3).GetWordp ()),
                                           *((*ep4).GetWordp ())),
                        (*ep3).DelRel (Entry::succ, relno, ep4);
                        (*ep4).DelRel (Entry::pred, relno, ep3);
                        if (type&RelSymmetric) {
                            (*ep3).DelRel (Entry::pred, relno, ep4);
                            (*ep4).DelRel (Entry::succ, relno, ep3);
                        }
                        // now immsucc has collapsed, so that we don't
                        // need to increment j
                    } else
                        ++j;
                }
            }
        }

        // Finally add the relation; if w1==w2 only add if the
        // relation is not reflexive per definiton
        if (w1 != w2 || !(type&RelReflexive)) {
            (*ep1).AddRel (Entry::succ, relno, ep2);
            (*ep2).AddRel (Entry::pred, relno, ep1);

            // if the relation is symmetric, add the relation in the
            // inverse sense, too
            if (type&RelSymmetric) {
                (*ep2).AddRel (Entry::succ, relno, ep1);
                (*ep1).AddRel (Entry::pred, relno, ep2);
            }
        }
    }

    return true;
}


void Dict::DeleteRelation(const Word &, const string &, const Word &)
{
    NOT_IMPLEMENTED;
}


Dict::septr
Dict::GetRelAll(Entry::direct d, eptr ep, const pair<int, int> &rel)
#ifdef __GNUC__
return done;
#endif
{
    const int type=rel.first;
    const int relno=rel.second;
#ifndef __GNUC__
    septr done;
#endif

    if (type&RelTransitive) {
        Entry::NewMark ();

        septr todo;
        todo.insert (ep);

        septr::iterator i;
        while ((i = todo.begin()) != todo.end()) {
            eptr ep2 = *i;
            todo.erase (i);
            if (!(*ep2).IsMarked ()) {
                (*ep2).Mark ();
                done.insert (ep2);
                const veptr &succ = GetRelImm (d, ep2, relno);
                todo.insert (succ.begin(), succ.end());
            }
        }
        done.erase (ep);
    } else {
        const veptr &imm = GetRelImm(d, ep, relno);
        done.insert(imm.begin(), imm.end());
    }
    // WARNING: even if rel is reflexive, pw is not included in the
    // returned set. the caller has to check this himself.
#ifndef __GNUC__
    return done;
#endif
}


// based on the assumption, that the number of successors in a given
// relation is much smaller than the number of predecessors, this
// implements the AreRelated function using a different algorithm:
//
// 1. if we only have transitive relations, we perform a BFS on the
//    successors of w1 using any of the given relations. if we find
//    w2, we return true
//
// 2. if we have non-transitive relations, we perform a BFS on the
// successors of w1 using the transitive relations. on each found
// node we follow non-transitive relations and start a new BFS for
// w2. all words in the new BFS are globally marked as seen.

int Dict::DFS (eptr ep, eptr end,
               const vcpii &trels,
               const vcpii &orels,
               const vcpii &orels_used)
{
    bool unmark = false;
    int r;

    /*
     * MUST use all non-transitive relations
     */
    if (ep == end && orels.empty()) {
        if (isverbose)
            cout << (string &)*((*ep).GetWordp());
        return 0;
    }
    (*ep).Mark ();

    vcpii::const_iterator rel;
    for (rel = trels.begin(); rel != trels.end(); ++rel) {
        const veptr &succ = GetRelImm (Entry::succ, ep, (*rel)->second);

        for (veptr::const_iterator i = succ.begin(); i != succ.end(); ++i) {
            if (!(**i).IsMarked ()) {
                r = DFS (*i, end, trels, orels, orels_used);
                if (!r) {
                    if (isverbose) {
                        cout << '<' << GetRelationName (**rel) << '>'
                             << (string &)*((*ep).GetWordp());
                    }
                    return 0;
                } else if (r < 0)
                    unmark = true;
            }
        }
    }

    for (rel = orels.begin(); rel != orels.end(); ++rel) {
        const veptr &succ = GetRelImm (Entry::succ, ep, (*rel)->second);

        vcpii neworels = orels;
        neworels.erase (find (neworels.begin(), neworels.end(), *rel));

        vcpii neworels_used = orels_used;
        neworels_used.push_back (*rel);

        for (veptr::const_iterator i = succ.begin(); i != succ.end(); ++i) {
            if (!(**i).IsMarked ()) {
                r = DFS (*i, end, trels, neworels, neworels_used);
                if (!r) {
                    if (isverbose) {
                        cout << '<' << GetRelationName (**rel) << '>'
                             << (string &)*((*ep).GetWordp());
                    }
                    return 0;
                } else if (r < 0)
                    unmark = true;
            }
        }
    }

    if (!unmark) {
        for (rel = orels_used.begin(); rel != orels_used.end(); ++rel) {
            const veptr &succ = GetRelImm (Entry::succ, ep, (*rel)->second);
            if (!succ.empty()) {
                unmark = true;
                break;
            }
        }
    }

    if (unmark) {
        (*ep).UnMark ();
        return -1;
    }
    return 1;
}

bool Dict::AreRelated(const Word &w1, const vector<string> &rels,
                      const Word &w2)
{
    // find out whether we have only reflexive relations and partition
    // the relations in transitive and non-transitive ones. we store
    // not the relation names, only their number and type
    bool only_reflexive=true;

    vcpii transrels;
    transrels.reserve(4);
    vcpii otherrels;
    otherrels.reserve(4);

    for (vector<string>::const_iterator i=rels.begin(); i!=rels.end(); ++i) {
        if (!ContainsRelation(*i))
            return false;
        const pair<int,int> &rel=relations[*i];
        const int type=rel.first;
        if (!(type&RelReflexive))
            only_reflexive=false;
        if (type&RelTransitive)
            transrels.push_back(&rel);
        else
            otherrels.push_back(&rel);
    }
    if (w1==w2)
        return only_reflexive;
    
    if (!ContainsWord(w1)) {
        dprintf ("Dict::AreRelated: (%s,_) not in dict.\n", (const char *)w1);
        return false;
    }
    if (!ContainsWord(w2)) {
        dprintf ("Dict::AreRelated: (_,%s) not in dict.\n", (const char *)w2);
        return false;
    }
    eptr ep1 = entries[GetCWP(w1)];
    eptr ep2 = entries[GetCWP(w2)];
    vcpii otherrels_used;

    Entry::NewMark ();
    int r = DFS (ep1, ep2, transrels, otherrels, otherrels_used);
    if (r == 0 && isverbose)
        cout << endl;
    return (r == 0);
}


ostream &operator<< (ostream &out, const Dict &d)
{
    out << "\n@dict-quick-mode-on\n";
    {
    // output the names and types of the relations
    for (Dict::msp::const_iterator i = d.relations.begin();
	 i != d.relations.end(); ++i) {
        out << "@register-relation '" << (*i).first << "'";
	const int reltype=(*i).second.first;
	if (reltype!=Dict::RelNone) {
	    out << " : ";
	    int apo=0;
	    if (reltype&Dict::RelSymmetric) {
	        if (apo++)
		    out << ", ";
		out << "@symmetric";
	    }
	    if (reltype&Dict::RelReflexive) {
	        if (apo++)
		    out << ", ";
		out << "@reflexive";
	    }
	    if (reltype&Dict::RelTransitive) {
	        if (apo++)
		    out << ", ";
		out << "@transitive";
	    }
	    if (reltype&Dict::RelAcyclic) {
	        if (apo++)
		    out << ", ";
		out << "@acyclic";
	    }
	    ASSERT(apo>0);
	}
	out << "\n";
    }
    }

    {
    // output the words
    for (Dict::meptr::const_iterator i = d.entries.begin();
	 i != d.entries.end(); ++i) {
        out << "@add-word " << *(*i).first << "\n";
    }
    }

    {
    // output the relations
    const vector<string> &relnames = d.GetRelationNames();
    for (Dict::meptr::const_iterator i = d.entries.begin();
	 i != d.entries.end(); ++i) {

        typedef vector<Entry::reldest> vrd;
	vrd::const_iterator begin
	    = (*(*i).second).GetRel(Entry::succ).begin();
	vrd::const_iterator end
	    = (*(*i).second).GetRel(Entry::succ).end();

	int relno=0;
	for (; begin!=end; ++relno, ++begin) {
	    // we do this instead of
	    // const int type=d.relations[relnames[relno]].second.first;
	    // because [] requires a non-const object
	    const int type=(*d.relations.find(relnames[relno])).second.first;
            const Dict::cwp &w1 = (*i).first;

	    for (Entry::reldest::const_iterator j = begin->begin();
		 j != begin->end(); ++j) {
	        const Dict::cwp w2 = (*(*j)).GetWordp ();
                // do not print symmetric relations twice
                if (!(type&Dict::RelSymmetric && *w1>*w2))
		    out << "@add-relation " << *w1 << " : '"
                        << relnames[relno] << "' : " << *w2 << "\n";
	    }
	}
    }
    }

    out << "@dict-quick-mode-off\n";

    return out;
}


// eof dict.cc

