
/*
 *  AI-Trader --- Service Trading using Conceptual Graphs
 *  Copyright (C) 1995 Arno Puder et al.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 aitrader@informatik.uni-frankfurt.de
 */

// $Id: dict.h,v 1.11 1998/02/06 11:05:49 roemer Exp $


#ifndef __DICT_H__
#define __DICT_H__


#include "entry.h"
#include "stl.h"


class Dict
{

public:

    enum eRelType
    {
        RelNone=0,
        RelSymmetric=1,  // aRb => bRa
        RelReflexive=2,  // aRa
        RelTransitive=4, // aRb && bRc => aRc
        RelAcyclic=8     // a!=b && aRb => !bRa
    };
    typedef Word::const_pointer cwp;
    typedef vector<cwp> vcwp;

    typedef Entry::pointer eptr;
    typedef vector<eptr> veptr;
    typedef set<eptr, less<eptr> > septr;
    typedef map<cwp, eptr, less<cwp> > meptr;

    typedef map<string, pair<int,int>, less<string> > msp;

    // constructor/destructor
    Dict();
    Dict(const Dict &);
    ~Dict();

    // set quick mode
    void SetQuickMode(bool);

    void SetVerboseMode (bool);
    
    // register relation. return value 0 means that this relation
    // already existed.
    bool RegisterRelation(const string &, int type);

    // get properties of named relation. return false if no such relation
    bool GetRelationProperties (const string &, int &prop);

    // retrieve all names of relations
    vector<string> GetRelationNames() const;
    
    // delete a given word.
    void DeleteWord(const Word &);

    // add a word to the dictionary.
    void AddWord(const Word &);

    // find out whether a given word exists in the dictionary
    bool ContainsWord(const Word &) const;

    // add a relation between two words return 1 on success, 0
    // otherwise. failure may be due to non registered word or relation
    // or to already existing relations that contradict the requested
    // one
    bool AddRelation(const Word &, const string &, const Word &);

    // delete a relation between two words
    void DeleteRelation(const Word &, const string &, const Word &);

    // get a list of _immediate_ predecessors for a given word and relation.
    // that is, disregard the possible transitivity of the relation
    // returns an empty list, if the relation was not defined for this
    // word or if the word does not exist
    vcwp GetRelImmPred(const Word &, const string &);
    vcwp GetRelAllPred(const Word &, const string &);

    // same for successors
    vcwp GetRelImmSucc(const Word &, const string &);
    vcwp GetRelAllSucc(const Word &, const string &);

    // returns 1 if the two words are related. transitive relations are
    // taken into account
    bool AreRelated(const Word &, const string &, const Word &);

    // return 1 if the two words are related by one of the relations in
    // the set, taking into account transitive relations among them
    bool AreRelated(const Word &, const vector<string> &, const Word &);

    // serialize
    friend ostream &operator<< (ostream &, const Dict &d);

private:
    typedef vector<const pair<int, int> *> vcpii;
    int DFS (eptr pw, eptr end,
             const vcpii &trels,
             const vcpii &orels,
             const vcpii &orels_deleted);

    bool ContainsRelation(const string &) const;
    cwp GetCWP(const Word &w);

    const veptr &GetRelImm(Entry::direct, eptr, int);
    const veptr &GetRelImm(Entry::direct d, const Word &w, const string &r);

    septr GetRelAll(Entry::direct, eptr, const pair<int, int> &);

    string GetRelationName(const pair<int, int> &rel);


    static veptr vepempty;
    bool isquickmode;
    bool isverbose;

#ifndef __GNUG__
    // needed by operator<<()
public:
#endif
    // map word pointers to entries pointers. Entries contain lists of
    // relations for a given word.
    meptr entries;
    
    // map relation names to (type,number). number is a unique int for
    // each relation
    msp relations;
};


template<class keytype, class valtype, class comp>
inline bool
dict_mapcontainskey(const map<keytype, valtype, comp> &m, const keytype &k)
{
    return m.find(k)!=m.end();
}

inline bool
Dict::GetRelationProperties (const string &s, int &prop)
{
    msp::const_iterator i = relations.find (s);
    if (i == relations.end())
        return false;
    prop = (*i).second.first;
    return true;
}

inline
void Dict::SetQuickMode(bool b)
{
    isquickmode=b;
}

inline void
Dict::SetVerboseMode (bool b)
{
    isverbose = b;
}

inline
bool Dict::ContainsWord(const Word &w) const
{
    return dict_mapcontainskey(entries, cwp(&w));
}


inline
bool Dict::ContainsRelation(const string &s) const
{
    return dict_mapcontainskey(relations, s);
}


inline
Dict::cwp Dict::GetCWP(const Word &w)
{
    ASSERT(ContainsWord(w));
    return (*entries.find(cwp(&w))).first;
}


inline
void Dict::AddWord(const Word &w)
{
    if (!ContainsWord(w)) {
        cwp nwp = new Word(w, false);
        entries[nwp] = new Entry (nwp);
    }
}


inline const Dict::veptr &
Dict::GetRelImm(Entry::direct d, eptr ep, int relno)
{
    return (*ep).GetRel(d, relno);
}


inline const Dict::veptr &
Dict::GetRelImm(Entry::direct d, const Word &w, const string &r)
{
    if (ContainsWord(w) && ContainsRelation(r))
        return GetRelImm (d, entries[GetCWP(w)], relations[r].second);
    else
        return vepempty;
}


inline
bool Dict::AreRelated(const Word &w1, const string &rel, const Word &w2)
{
    vector<string> rels(1);
    rels[0]=rel;
    return AreRelated(w1, rels, w2);
}


#ifdef NOT_USED
inline
bool Dict::AreSimplyRelated(cwp pw1, const pair<int, int> &rel, cwp pw2)
{
    const int type=rel.first;
    const int relno=rel.second;
    if (pw1==pw2 && type&RelReflexive)
        return true;
    else {
        const vcwp &rels=entries[pw1].GetRel(Entry::succ, relno);
        if (find(rels.begin(), rels.end(), pw2)!=rels.end())
            return true;
        else
            return false;
    }
}
#endif

#endif // __DICT_H__

