
/*
 *  AI-Trader --- Service Trading using Conceptual Graphs
 *  Copyright (C) 1995 Arno Puder et al.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 aitrader@informatik.uni-frankfurt.de
 */

// $Id: entry.h,v 1.4 1998/02/06 11:05:49 roemer Exp $


#ifndef __ENTRY_H__
#define __ENTRY_H__


#include "valid.h"
#include "word.h"
#include "stl.h"


class Entry
{
    
public:
    typedef Ptr<Entry> pointer, eptr;
    typedef Ptr<const Entry> const_pointer, cep;

    typedef vector<eptr> reldest;
    enum direct { pred=0, succ=1 };
    
    // constructors, destructors
    
    Entry(Word::const_pointer w)
    {
        rel[pred].reserve(4);
        rel[succ].reserve(4);
        wordp = w;
        mark = themark-1;
    }
    Entry(const Entry &e) 
    {
        rel[pred]=e.rel[pred];
        rel[succ]=e.rel[succ];
        wordp = e.wordp;
        mark = e.mark;
    }
    ~Entry() {}

    // get all words preceding/succeeding the current word in respect
    // to a certain relation. WARNING: after an add/delete operation
    // the return value may have changed
    const reldest &GetRel(direct d, int i) const
    {
        if (i<(signed)rel[d].size())
            return rel[d][i];
        else
            return empty;
    }
    // use with care: vector may be shorter than number of defined
    // relations
    const vector<reldest> &GetRel(direct d) const
    {
        return rel[d];
    }

    // (un)register a relation that originates in/points to the
    // current word
    void AddRel(direct d, int r, eptr c)
    {
        vector<reldest> &v=rel[d];
        const int incsize=r+1-v.size();
        if (incsize>0)
            v.insert(v.end(), incsize, reldest());
        reldest &vr=v[r];
        reldest::iterator oldpos=find(vr.begin(), vr.end(), c);
        if (oldpos==vr.end()) {
            if (oldpos==vr.begin())
                vr.reserve(2);
            vr.push_back(c);
        }
    }
    void DelRel(direct d, int r, eptr c)
    {
        reldest &vr=rel[d][r];
        reldest::iterator oldpos=find(vr.begin(), vr.end(), c);
        if (oldpos!=vr.end())
            vr.erase(oldpos);
        // currently we do not shrink the rel[d] vector
    }

    bool operator==(const Entry &e) const
    {
        return wordp == e.wordp;
    }
    bool operator!=(const Entry &e) const
    {
        return wordp != e.wordp;
    }
    bool operator<(const Entry &e) const
    {
        return wordp < e.wordp;
    }

    void Mark ()
    {
        mark = themark;
    }
    void UnMark ()
    {
        mark = themark-1;
    }
    bool IsMarked () const
    {
        return mark == themark;
    }
    static void NewMark ()
    {
        ++themark;
    }

    Word::const_pointer GetWordp () const
    {
        return wordp;
    }

private:
#if 0
    int refcount;
#endif
    vector<reldest> rel[2];
    Word::const_pointer wordp;
    long mark;

    static reldest empty;
    static long themark;
};


#endif
