
/*
 *  AI-Trader --- Service Trading using Conceptual Graphs
 *  Copyright (C) 1995 Arno Puder et al.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 aitrader@informatik.uni-frankfurt.de
 */


// $Id: interface.h,v 1.7 1997/06/27 16:19:13 roemer Exp $


#ifndef __INTERFACE_H__
#define __INTERFACE_H__

#include "key.h"
#include "match.h"
#include "cg.h"
#include "services.h"
#include "kb.h"
#include "error.h"


class AITraderInterface 
{
public:
    // Return version number
    virtual ErrorString GetVersion( string &version ) = 0;
	
    // Set the verbose flag. This determines if output statements in
    // matching rules have any effect.
    virtual ErrorString SetVerbose( bool flag ) = 0;
	
    // Save the complete database. "name" denotes the filename.
    virtual ErrorString SaveTraderDB( string *name ) = 0;

    // Define a new matching rule called "name".
    virtual ErrorString DefineMatchingRule( string *name, string *rule ) = 0;

    // Delete the matching rule called "name".
    virtual ErrorString DeleteMatchingRule( string *name ) = 0;

    // Get the names of all matching rules
    virtual ErrorString GetMatchingRuleNames( string *names ) = 0;
	
    // Get the matching rule
    virtual ErrorString GetMatchingRule( string *name, string *rule ) = 0;
	
// Dictionary commands

    // Set dictionary quick mode
    virtual ErrorString DictQuickMode( bool flag ) = 0;
	
    // Register a new kind of relation. The "prop" may contain one
    // or more of the following properties: RelNone, RelSymmetric,
    // RelReflexive, RelTransitive, RelAcyclic
    virtual ErrorString RegisterRelation( string &rel, int prop ) = 0;

    // Get all registered relations
    virtual ErrorString GetAllRelations( SetOfString &rels ) = 0;

    // Retrieve the properties (RelReflexive, ...) of the named relation
    virtual ErrorString GetRelationProperties (string &rel, int &prop) = 0;
	
    // Removes a word from the dictionary.
    virtual ErrorString DeleteWord( string &word ) = 0;

    // Adds a new word to the dictionary.
    virtual ErrorString AddWord( string &word ) = 0;

    // Checks whether word has previously been defined.
    virtual ErrorString ContainsWord( string &word, bool &yesno ) = 0;

    // Adds a new relation between words "word1" and "word2". The relation
    // "rel" has to be defined with RegisterRelation.
    virtual ErrorString AddRelation( string &word1, string &rel,
                                     string &word2 ) = 0;

    // Retrieve all predecessors for a given word and relation.
    virtual ErrorString GetRelationPred( string &word, string &rel,
                                         SetOfWord &pred ) = 0;
    virtual ErrorString GetRelationAllPred( string &word, string &rel,
                                            SetOfWord &pred ) = 0;

    // Retrieve all successors for a given word and relation.
    virtual ErrorString GetRelationSucc( string &word, string &rel,
                                         SetOfWord &succ ) = 0;
    virtual ErrorString GetRelationAllSucc( string &word, string &rel,
                                            SetOfWord &succ ) = 0;

    // Checks whether "word1" and "word2" are related via relation "rel".
    virtual ErrorString AreRelated( string &word1, SetOfString &rels,
                                    string &word2, bool &yesno ) = 0;

// Knowledgebase commands

    // Add a new conceptual graph.
    virtual ErrorString Add( ConceptualGraph *cg, KeyCG &newKey ) = 0;

    // Replace an existing graph
    virtual ErrorString Replace( KeyCG aKey, ConceptualGraph *cg ) = 0;

    // Join a new example to an existing conceptual graph.
    virtual ErrorString Join( KeyCG aKey, ConceptualGraph *cg ) = 0;

    // Match a given query graph with all stored type graphs.
    virtual ErrorString Match( ConceptualGraph *query,
                               SetOfString &selecRules,
                               SetOfMatchResult &matches,
                               KeyCG profile = -1) = 0;

    // Retrieve a type graph via its database key.
    virtual ErrorString GetGraph( KeyCG aKey,
                                  ConceptualGraph **cg ) = 0;

    // Get number of variants for a given concept type
    virtual ErrorString GetNumberOfVariants( string &conType, int &num ) = 0;
	
    // Retrieve a type graph via a concept type and a numbered instance
    virtual ErrorString GetNthVariant( string &conType, int num,
                                       KeyCG &theKey ) = 0;

    // Get number of type graphs
    virtual ErrorString GetNumberOfGraphs( int &num ) = 0;
	
    // Get n-th type graph
    virtual ErrorString GetNthGraph( int num, KeyCG &theKey ) = 0;

// Service offer management

    // Add a new provider for a previously exported service.
    virtual ErrorString AddProvider( KeyCG aKey,
                                     string &theMaintainer, string &anOID ) = 0;

    // Remove a provider from the list of exporters.
    virtual ErrorString RemoveProvider( KeyCG aKey,
                                        string &anOID ) = 0;

    // Get all providers for a given service.
    virtual ErrorString GetProviders( KeyCG aKey,
                                      SetOfServiceEntry &result ) = 0;

// Profile commands

    // Create new profile
    virtual ErrorString CreateProfile( KeyCG &aKey ) = 0;

    // Remove profile
    virtual ErrorString RemoveProfile( KeyCG aKey ) = 0;

    // Set minimum matchvalue of profile
    virtual ErrorString SetMinmatchval( KeyCG aKey, long minmatchval ) = 0;

    // Set timeout of profile
    virtual ErrorString SetTimeout( KeyCG aKey, long timeout ) = 0;

    // Get minimum matchvalue of profile
    virtual ErrorString GetMinmatchval( KeyCG aKey, long &minmatchval ) = 0;

    // Get timeout of profile
    virtual ErrorString GetTimeout( KeyCG aKey, long &timeout ) = 0;
    
// Answering mechanism
    virtual bool CheckForError() = 0;

    // Set an error message
    virtual void SetError( int err, string *msg ) = 0;
	
    // Set NULL answer
    virtual void SetAnswer() = 0;
    
    // Set conceptual graph as an answer
    virtual void SetAnswer( ConceptualGraph *cg ) = 0;
	
    // Set bool answer
    virtual void SetAnswer( bool flag ) = 0;
	
    // Set simple string (from dictionary) answer
    virtual void SetAnswer( string &str ) = 0;
	
    // Set match result answer
    virtual void SetAnswer( SetOfMatchResult &result ) = 0;
	
    // Set service entries answer
    virtual void SetAnswer( SetOfServiceEntry &entries ) = 0;
	
    // Set a set of words from the dictionary as the answer
    virtual void SetAnswer( SetOfWord &word_set ) = 0;

    // Set a database key as the answer
    virtual void SetAnswer( KeyCG aKey ) = 0;

    // Set an interger as the answer
    virtual void SetAnswer( long number ) = 0;
	
    // Get no answer except for error message
    virtual ErrorString GetAnswer() = 0;
	
    // Get conceptual graph
    virtual ErrorString GetAnswer( ConceptualGraph **cg ) = 0;

    // Get bool answer
    virtual ErrorString GetAnswer( bool &flag ) = 0;
	
    // Get simple string (from dictionary) answer
    virtual ErrorString GetAnswer( string &str ) = 0;
	
    // Get match result answer
    virtual ErrorString GetAnswer( SetOfMatchResult &result ) = 0;
	
    // Get service entries answer
    virtual ErrorString GetAnswer( SetOfServiceEntry &entries ) = 0;
	
    // Get a set of words from the dictionary as the answer
    virtual ErrorString GetAnswer( SetOfWord &word_set ) = 0;

    // Get database key
    virtual ErrorString GetAnswer( KeyCG &key ) = 0;

    // Get an integer
    virtual ErrorString GetAnswer( long &number ) = 0;
	
    // Get answer string
    virtual string GetAnswerString() = 0;
};

#endif
