
/*
 *  AI-Trader --- Service Trading using Conceptual Graphs
 *  Copyright (C) 1995 Arno Puder et al.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 aitrader@informatik.uni-frankfurt.de
 */


// $Id: kb.cc,v 1.13 1997/06/27 16:19:14 roemer Exp $

 
#include <assert.h>
#include <time.h>
#include "kb.h"
#include "eval.h"


KeyCG nilKeyCG;



KB::KB()
{
	uuid = 0;
}


KB::~KB()
{
	// !!! DB mu noch gelscht werden !!!
}


KeyCG KB::GetUniqueKeyCG()
{
	KeyCG newKey;
	
	do {
		newKey = uuid++;
	} while( db.count( newKey ) != 0 );
	
	return newKey;
}


/*
 * MUST NOT delete cg, make a copy instead and leave cg alone.
 * cgtcl.cc and yacc.y rely on it.
 */
void KB::Add( ConceptualGraph *cg, KeyCG &newKey )
{
	newKey = GetUniqueKeyCG();
	db[ newKey ] = CopyGraph (cg);
}


/*
 * MUST NOT delete cg, make a copy instead and leave cg alone.
 * cgtcl.cc and yacc.y rely on it.
 */
void KB::Replace( KeyCG aKey, ConceptualGraph *cg )
{
	if( db.count( aKey ) != 0 )
		// Delete old graph
		delete db[ aKey ];
	if( cg == NULL && db.count( aKey ) == 0 )
		// No new graph. Simply return
		return;
	if( cg == NULL ) {
		// Erase graph
		DB::iterator i = db.find( aKey );
		db.erase( i );
		return;
	}
	// Otherwise replace graph
	db[ aKey ] = CopyGraph (cg);
}



/*
 * MUST NOT delete cg, make a copy instead and leave cg alone.
 * cgtcl.cc and yacc.y rely on it.
 */
void KB::Join( KeyCG aKey, ConceptualGraph *cg )
{
	if( db.count( aKey ) == 0 ) {
		// Wrong key
		return;
	}

	ConceptualGraph *old_cg = db[aKey];

	assert (old_cg->GetNode()->IsA() == CONCEPT_NODE);
	if (*old_cg->GetNode()->GetName() == *cg->GetNode()->GetName()) {
		/*
		 * join instance lists
		 */
		ConceptNode *old_node = (ConceptNode *)old_cg->GetNode();
		ConceptNode *node = (ConceptNode *)cg->GetNode();
		old_node->AddInstances (node->GetInstances());
	}
	if (cg->GetWidth() == NULL) {
		// CG to be joined just consists of one concept node.
		// There is nothing to be done for this join
		return;
	}

        /*
         * make a copy of cg, remove all branches that are already in
         * old_cg and store a list of remaining childs into new_features
         */
        ConceptualGraph *cgcopy = CopyGraph (cg);
        cgcopy->Subtract (*old_cg);
        ConceptualGraph *new_features = CopyGraph (cgcopy->GetWidth());
        delete cgcopy;

	ConceptualGraph *rel = old_cg->GetWidth();
	if( rel == NULL ) {
		// Old Graph just consisted of one concept node
		old_cg->SetWidth( new_features );
		return;
	}
	while( rel->GetDepth() != NULL )
		rel = rel->GetDepth();
	// Add new features
	rel->SetDepth( new_features );
}



/*
 * MUST NOT delete cg, make a copy instead and leave cg alone.
 * cgtcl.cc and yacc.y rely on it.
 */
SetOfMatchResult KB::Match( Dict &dict, MatchingRule &rules,
                            SetOfString &selecRules,
                            ConceptualGraph *theQuery, bool verbose,
                            const ProfileEntry &prof)
{
    SetOfMatchResult result;

    time_t starttm = time (0);
	
    // Loop over all concepts in the knowledgebase
    DB::iterator i;
    for( i = db.begin(); i != db.end(); i++ ) {
        ConceptualGraph *type_graph = (*i).second;
        theQuery->ResetTags( NODE_UNKNOWN );
        Evaluator eval( rules, selecRules, dict, type_graph, theQuery );
        int matchval = eval.StartMatch( verbose );
        if( theQuery->GetTag() == NODE_ACCEPTED &&
            matchval >= prof.GetMinMatchval()) {
            ConceptualGraph *query = CopyGraph( theQuery );
            result.insert( MatchResult( (*i).first, query, matchval ) );
        }
        time_t tm = time (0);
        if (prof.GetTimeout() > 0 && tm - starttm > prof.GetTimeout())
            // timeout exceeded
            break;
    }
    return result;
}


/*
 * MUST make a copy
 */
ConceptualGraph *KB::GetGraphByKey( KeyCG key )
{
	if( db.count( key ) == 0 )
		// No such key in DB
		return NULL;
	return CopyGraph( (*(db.find( key ))).second );
}



int KB::GetNumberOfVariants( const string &conType )
{
	int count = 0;

	for( DB::iterator i = db.begin(); i != db.end(); i++ ) {
		if (((*i).second)->GetNode()->GetName() ==
		    Word (conType, false))
			count++;
	}
	return count;
}



KeyCG KB::GetVariant( const string &conType, int num )
{
	int count = 0;
	DB::iterator i;
	
	for( i = db.begin(); i != db.end(); i++ ) {
		if (((*i).second)->GetNode()->GetName() ==
		    Word (conType, false)) {
			if (num == count++)
				break;
		}
	}

	if( i == db.end() )
		return nilKeyCG;

	return (*i).first;
}

	

int KB::GetNumberOfGraphs()
{
	return db.size();
}


KeyCG KB::GetNthGraph( int num )
{
	DB::iterator i;

	assert( num < GetNumberOfGraphs() );
	for( i = db.begin(); i != db.end(); i++ ) {
		if( num-- == 0 )
			break;
	}
	if( i == db.end() )
		return nilKeyCG;
	return (*i).first;
}

			
void KB::Print( ostream &ostr )
{
	for( DB::iterator i = db.begin(); i != db.end(); i++ ) {
		ostr << "@add $";
		ostr << (*i).first;
		ostr << " : ";
		(*i).second->Print( ostr );
		ostr << endl;
	}
}



MatchResult::MatchResult( KeyCG theKey, ConceptualGraph *theQuery, int mval )
{
	key = theKey;
	query = theQuery;
        matchval = mval;
}

MatchResult::~MatchResult()
{
	if( query )
		delete query;
}


KeyCG MatchResult::GetKeyCG() const
{
	return key;
}


ConceptualGraph* MatchResult::GetQueryGraph() const
{
	return query;
}

int MatchResult::GetMatchVal() const
{
	return matchval;
}

void MatchResult::Print( ostream &ostr )
{
	ostr << "Query" << endl;
	if( query )
		query->Print( ostr, true );
	else
		ostr << "(unknown)";
	ostr << endl << "matched with concept ";
	ostr << key;
	ostr << endl << endl;
}


