
/*
 *  AI-Trader --- Service Trading using Conceptual Graphs
 *  Copyright (C) 1995 Arno Puder et al.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 aitrader@informatik.uni-frankfurt.de
 */

/*
 * this file implements class PrologInterp, a conceptual graph matcher that
 * uses SWI Prolog.
 *
 * (w) 1996, Kay Roemer.
 *
 * $Id: prolog.cc,v 1.15 1997/10/13 15:48:53 roemer Exp $
 */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <assert.h>

#include <stream.h>
#include <iostream.h>
#include <fstream.h>
#include <strstream.h>

#include "prolog.h"
#include "util.h"


PL_extension PL_extensions[] = {
  { NULL, 0, NULL, 0 }
};

static functor_t match_functor;
static atom_t accept_atom, reject_atom;

static MatchInfo *minfo;

/***********************************************************************
 * make/decode a prolog representation of a concept graph node, which
 * looks like:
 *	cgnode(<pointer to MatchInfo>, <pointer to ConceptualGraph>)
 *
 * the pointers are stored as floats, because PL_{get|put}_pointer()
 * does only work with pointers to internal prolog structures and
 * PL_{get|put}_integer() limits the range of the integers to 0..2^30-1.
 ***********************************************************************/

static term_t
cgnode_make (ConceptualGraph *np)
{
    term_t a1 = PL_new_term_ref ();
    PL_put_float (a1, (long)np);
    return a1;
}

static ConceptualGraph *
cgnode_decode (term_t cgnode)
{
    double d1;
    if (!PL_get_float (cgnode, &d1)) {
        PL_warning ("cgnode_decode: invalid cgnode\n");
        return NULL;
    }
    return (ConceptualGraph *)(long)d1;
}

/***********************************************************************
 * make/decode a prolog representation of a concept graph node's name,
 * which looks like:
 *	wnode(<pointer to MatchInfo>, <name as prolog string>)
 *
 * the pointers are stored as floats, because PL_{get|put}_pointer()
 * does only work with pointers to internal prolog structures and
 * PL_{get|put}_integer() limits the range of the integers to 0..2^30-1.
 ***********************************************************************/

static term_t
wnode_make (const string &word)
{
  term_t a1 = PL_new_term_ref ();
  PL_put_string_chars (a1, word.c_str());
  return a1;
}

static bool
wnode_decode (term_t wnode, string &s)
{
    int len;
    char *cp;

    switch (PL_term_type (wnode)) {
    case PL_ATOM:
        assert (PL_get_atom_chars (wnode, &cp));
        break;

    case PL_STRING:
        assert (PL_get_string (wnode, &cp, &len));
        break;

    default:
        PL_warning ("cgnode_decode: invalid wnode\n");
        return false;
    }
    s = cp;
    return true;
}

/***********************************************************************
 * foreign prolog predicates.
 ***********************************************************************/

/*
 * match two concep-graph nodes. current strategy is:
 *
 * (a) if there is a rule that matches with REJECT, then REJECT.
 * (b) else if there is a rule that matches with ACCEPT, then ACCEPT.
 * (c) else fail (meaning UNKNOWN).
 */
static foreign_t
pl_cg_match (term_t qgraph, term_t tgraph, term_t result)
{
    ConceptualGraph *cgp = cgnode_decode (qgraph);
    if (!cgp)
        PL_fail;

    int maxval = 0;
    SetOfString::const_iterator i;
    for (i = minfo->GetRules().begin(); i != minfo->GetRules().end(); ++i) {
        module_t module = PL_new_module (PL_new_atom ((*i).c_str()));

        term_t goal = PL_new_term_ref ();
        term_t res  = PL_new_term_ref ();
        PL_put_variable (res);
        PL_cons_functor (goal, match_functor, qgraph, tgraph, res);

        if (PL_call (goal, module)) {
            int resint=0;
            if (!PL_get_integer (res, &resint)) {
                PL_warning ("cg_match: cannot get result int\n");
            } else if (resint < 0) {
                /*
                 * reject for one rule -> reject node
                 */
                cgp->SetTag (NODE_REJECTED);
                return PL_unify_integer (result, -1);
            } else if (resint > maxval) {
                maxval = resint;
            }
        }
    }
    if (maxval > 0) {
        /*
         * at least one accept, but no rejects -> accept node
         */
        cgp->SetTag (NODE_ACCEPTED);
        return PL_unify_integer (result, maxval);
    }
    /*
     * no rejects or accepts -> mark node as visited with unknown result
     */
    cgp->SetTag (NODE_CANTTELL);
    PL_fail;
}

/*
 * match two words. `mode' is a list of relations that is passed to the
 * dictionary. If modelist is empty than just perform a case-insensitive
 * comparision.
 * the two words can be given as wnode's (as returned by cg_getname()) or
 * as strings, but at least one must be a wnode.
 */
static foreign_t
pl_cg_matchword (term_t qwnode, term_t mode, term_t twnode)
{
    vector<string> rels;

    /*
     * decode mode list
     */
    term_t l = PL_copy_term_ref (mode);
    term_t a = PL_new_term_ref ();
    while (PL_get_list (l, a, l)) {
        char *s;
        if (!PL_get_atom_chars (a, &s)) {
            PL_warning ("cg_matchword: bad mode list element\n");
            PL_fail;
        }
        string str (s);
        rels.insert (rels.end(), s);
    }
    if (!PL_get_nil (l)) {
        PL_warning ("cg_matchword: bad mode list\n");
        PL_fail;
    }

    /*
     * decode words to compare
     */
    string w1, w2;
    if (!wnode_decode (qwnode, w1) || !wnode_decode (twnode, w2))
        PL_fail;

    int r;
    if (rels.empty ()) {
        /*
         * empty mode list means: compare
         */
        r = !fcompare (w1, w2);
    } else {
        r = minfo->GetDict().AreRelated (w1, rels, w2);
    }
    if (r)
        PL_succeed;
    else
        PL_fail;
}

/*
 * get the name of a cg-node as a wnode.
 */
static foreign_t
pl_cg_getname (term_t cgnode, term_t wnode)
{
    ConceptualGraph *cgp = cgnode_decode (cgnode);

    if (!cgp)
        PL_fail;

    return PL_unify (wnode, wnode_make (cgp->GetNode()->GetName()));
}

/*
 * get the list of successors of a cg-node
 */
static foreign_t
pl_cg_getsucc (term_t cgnode, term_t succ)
{
    ConceptualGraph *cgp = cgnode_decode (cgnode);

    if (!cgp)
        PL_fail;

    term_t l = PL_copy_term_ref (succ);
    term_t h = PL_new_term_ref ();
    for (ConceptualGraph *cg = cgp->GetWidth (); cg; cg = cg->GetDepth ()) {
        /*
         * PL_unify_list (l, cgnode_make (matchp, child), l) does NOT work
         * for some reason.
         */
        if (!PL_unify_list (l, h, l) ||
            !PL_unify (h, cgnode_make (cg)))
            PL_fail;
    }
    return PL_unify_nil (l);
}

/*
 * get the node-type of a cg-node.
 */
static foreign_t
pl_cg_gettype (term_t cgnode, term_t type)
{
    ConceptualGraph *cgp = cgnode_decode (cgnode);

    if (!cgp)
        PL_fail;

    bool isconcept = (cgp->GetNode()->IsA() == CONCEPT_NODE);
    return PL_unify_atom_chars (type, isconcept ? "concept" : "relation");
}

/*
 * get the value of a concept-node if it has one.
 */
static foreign_t
pl_cg_getvalue (term_t cgnode, term_t value)
{
    ConceptualGraph *cgp = cgnode_decode (cgnode);

    if (!cgp)
        PL_fail;

    if (cgp->GetNode()->IsA() != CONCEPT_NODE) {
        cerr << "cg_getvalue: called on relation node." << endl;
        PL_fail;
    }
    ConceptNode *cnode = (ConceptNode *)cgp->GetNode ();
    if (cnode->GetValue () != NIL_VALUE) {
        return PL_unify_integer (value, cnode->GetValue ());
    }
    PL_fail;
}

/*
 * get a list of instances of a concept-node if it has one
 */
static foreign_t
pl_cg_getinst (term_t cgnode, term_t inst)
{
    ConceptualGraph *cgp = cgnode_decode (cgnode);

    if (!cgp)
        PL_fail;

    if (cgp->GetNode()->IsA() != CONCEPT_NODE) {
        cerr << "cg_getinst: called on relation node." << endl;
        PL_fail;
    }
    ConceptNode *cnode = (ConceptNode *)cgp->GetNode ();

    if (cnode->GetValue () != NIL_VALUE)
        PL_fail;

    term_t l = PL_copy_term_ref (inst);
    term_t h = PL_new_term_ref ();

    SetOfWord wset = cnode->GetInstances ();
    SetOfWord::const_iterator i;
    for (i = wset.begin (); i != wset.end (); ++i) {
        /*
         * PL_unify_list (l, ... , l) does NOT work for some reason.
         */
        if (!PL_unify_list (l, h, l) ||
            !PL_unify_string_chars (h, (*i).c_str()))
            PL_fail;
    }
    return PL_unify_nil (l);
}

/*
 * output a trace message, arguments are the same as for the prolog predicate
 * format/2.
 */
static foreign_t
pl_cg_trace (term_t format, term_t args)
{
    if (minfo->IsVerbose()) {
        term_t goal = PL_new_term_ref ();
        PL_cons_functor (goal, PL_new_functor (PL_new_atom ("format"), 2),
                         format, args);
        return PL_call (goal, NULL);
    }
    PL_succeed;
}

static foreign_t
pl_cg_cg2string (term_t cg, term_t str)
{
    ConceptualGraph *cgp = cgnode_decode (cg);
    if (!cg)
        PL_fail;
    ostrstream ostr;
    cgp->Print (ostr);
    int r = PL_unify_string_chars (str, ostr.str());
    ostr.rdbuf()->freeze (0);
    return r;
}

/***********************************************************************
 * class PrologInterp implementation.
 ***********************************************************************/

int PrologInterp::inited = 0;

PrologInterp::PrologInterp ()
{
    if (inited) {
        cerr << "class PrologInterp instanciated more than once!" << endl;
        return;
    }

    /*
     * setup prolog home directory if not already set
     */
    if (!getenv ("SWI_HOME_DIR")) {
        /*
         * 'buf' can be accessed after the putenv()!
         */
        static char buf[512];
        sprintf (buf, "SWI_HOME_DIR=%s", SWIHOME);
        putenv (buf);
    }

    char *argv[] = { "pl", "-O", "-g", "true", "-L40000", "-A10000", NULL };
    int argc = sizeof (argv) / sizeof (*argv) - 1;
    extern char **environ;

    if (!PL_initialise (argc, argv
#if PLVERSION < 20900
                        , environ
#endif
		       )) {
        cerr << "Cannot initialize SWI Prolog" << endl;
        exit (1);
    }
    /*
     * override SWI's own signal handlers. But take care to NOT override
     * SIGSEGV (swi uses it for stack expansion).
     */
    signal (SIGINT, SIG_DFL);

    /*
     * register foreign predicates
     */
    PL_register_foreign ("cg_match",     3, (pl_function_t)pl_cg_match,     0);
    PL_register_foreign ("cg_matchword", 3, (pl_function_t)pl_cg_matchword, 0);
    PL_register_foreign ("cg_getname",   2, (pl_function_t)pl_cg_getname,   0);
    PL_register_foreign ("cg_getsucc",   2, (pl_function_t)pl_cg_getsucc,   0);
    PL_register_foreign ("cg_gettype",   2, (pl_function_t)pl_cg_gettype,   0);
    PL_register_foreign ("cg_getvalue",  2, (pl_function_t)pl_cg_getvalue,  0);
    PL_register_foreign ("cg_getinst",   2, (pl_function_t)pl_cg_getinst,   0);
    PL_register_foreign ("cg_trace",     2, (pl_function_t)pl_cg_trace,     0);
    PL_register_foreign ("cg_cg2string", 2, (pl_function_t)pl_cg_cg2string, 0);

    /*
     * generate heavily used functors and atoms only once...
     */
    match_functor  = PL_new_functor (PL_new_atom ("match"), 3);

    accept_atom  = PL_new_atom ("accept");
    reject_atom  = PL_new_atom ("reject");

    old_stdout = dup (2);
    old_stderr = dup (3);
    new_stdout = open ("/dev/null", O_RDWR);

    inited = 1;
}

PrologInterp::~PrologInterp ()
{
    SetOfString::const_iterator i;
    for (i = files.begin(); i != files.end(); ++i) {
        unlink ((*i).c_str());
    }
    close (new_stdout);
}

void
PrologInterp::Silent ()
{
    if (!minfo->IsVerbose()) {
        dup2 (new_stdout, 2);
        dup2 (new_stdout, 3);
    }
}

void
PrologInterp::Verbose ()
{
    if (!minfo->IsVerbose()) {
        dup2 (old_stdout, 2);
        dup2 (old_stderr, 3);
    }
}

bool
PrologInterp::UpdateRule (const string &key, const string &value)
{
    string fname;
    fname += "/tmp/AIT-" + key + "." + dec (getpid ()) + ".pl";

    gsub (fname, " ", "-");

    ofstream module (fname.c_str(), ios::out|ios::trunc);
    if (!module) {
        cerr << "PrologInterp::UpdateRule: cannot open " << fname << endl;
        return false;
    }
    module << form (":- module('%s', [match/3]).\n\n", key.c_str());
    module << value << endl;
    module.close ();

    term_t a = PL_new_term_ref ();
    term_t f = PL_new_term_ref ();

    if (files.count (fname)) {
        /*
         * rule has been changed
         */
        PL_cons_functor (f, PL_new_functor (PL_new_atom ("make"), 0));
    } else {
        /*
         * new rule
         */
        PL_put_atom_chars (a, fname.c_str());
        PL_cons_functor (f, PL_new_functor (PL_new_atom ("consult"), 1), a);
        files.insert (fname);
    }

    Silent ();
    if (!PL_call (f, NULL)) {
        Verbose ();
        cerr << "PrologInterp: cannot load module " << key << endl;
        return false;
    }
    Verbose ();

    return true;
}

bool
PrologInterp::Match (MatchingRule &theRules, SetOfString &theSelRules,
                     Dict &theDict, ConceptualGraph *t, ConceptualGraph *q,
                     bool verbose, int *matchresult)
{
    /*
     * not that clean, but there is no better way to pass "user data" to
     * foreign predicates.
     */
    MatchInfo m (theDict, theSelRules, verbose);
    minfo = &m;

    if (theRules.IsModified ()) {
        MatchingRule::const_iterator i;
        for (i = theRules.begin (); i != theRules.end (); ++i) {
            if (!UpdateRule ((*i).first, (*i).second)) {
                minfo = NULL;
                return false;
            }
        }
    }

    fid_t frame = PL_open_foreign_frame ();

    term_t qterm  = cgnode_make (q);
    term_t tterm  = cgnode_make (t);
    term_t result = PL_new_term_ref ();
    PL_put_variable (result);

    /*
     * match them
     */
    pl_cg_match (qterm, tterm, result);
    if (matchresult) {
        int resint;
        if (!PL_get_integer (result, &resint)) {
            *matchresult = 0;
        } else {
            *matchresult = resint >= 0 ? resint : 0;
        }
        *matchresult = (*matchresult + 9) / 10;
    }
    PL_discard_foreign_frame (frame);

    minfo = NULL;
    return true;
}
