
/*
 *  AI-Trader --- Service Trading using Conceptual Graphs
 *  Copyright (C) 1995 Arno Puder et al.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 aitrader@informatik.uni-frankfurt.de
 */

#include <strstream.h>
#include <assert.h>
#include "parser.h"
#include "proxy.h"


ClientProxy::ClientProxy( TransportLayer *theConnection )
{
	transportLayer = theConnection;
}



void ClientProxy::ExecuteCommand( string command )
{
	string result;

	transportLayer->TransmitString( command, result );
	
	ostrstream ostr;
	ostr << result << ends;
	istrstream istr( ostr.str() );
	ostr.rdbuf()->freeze(0);


	Parser parser( &localServer, istr, NULL );
	parser.Parse();
        if (parser.ErrorOccured()) {
            cerr << parser.GetErrorFile() << ':'
                 << parser.GetErrorLine() << ':'
                 << parser.GetErrorMsg() << endl;
        }
}


ErrorString ClientProxy::GetVersion( string &version )
{
	ExecuteCommand( "@version" );
	return localServer.GetAnswer( version );
}


ErrorString ClientProxy::SetVerbose( bool flag )
{
	if( flag )
		ExecuteCommand( "@verbose-on" );
	else
		ExecuteCommand( "@verbose-off" );
	return localServer.GetAnswer();
}


ErrorString ClientProxy::DictQuickMode(bool)
{
    assert( 0 );
    return ErrorString();
}


ErrorString ClientProxy::SaveTraderDB( string *name )
{
	ExecuteCommand( "@save '" + (*name) + "'" );
	return localServer.GetAnswer();
}


ErrorString ClientProxy::DefineMatchingRule( string *name, string *rule )
{
	ExecuteCommand( "@define-matching-rule '" + (*name) + "' " + 
					(*rule) + '' );
	return localServer.GetAnswer();
}


ErrorString ClientProxy::DeleteMatchingRule( string *name )
{
	ExecuteCommand( "@delete-matching-rule '" + (*name) + "'" );
	return localServer.GetAnswer();
}


ErrorString ClientProxy::GetMatchingRuleNames( string *names )
{
	ExecuteCommand( "@show-matching-rules" );
	return localServer.GetAnswer( *names );
}


ErrorString ClientProxy::GetMatchingRule( string *name, string *rule )
{
	ExecuteCommand( "@show-matching-rule '" + (*name) + "'" );
	return localServer.GetAnswer( *rule );
}


ErrorString ClientProxy::RegisterRelation( string &rel, int prop )
{
	string command;
	
	command = "@register-relation '" + rel + "'";
	if( prop != Dict::RelNone ) {
		command += " : ";
	}
	bool flag = false;
	if( prop & Dict::RelSymmetric ) {
		command += "@symmetric";
		flag = true;
	}
	if( prop & Dict::RelReflexive ) {
		if( flag )
			command += ", ";
		command += "@reflexive";
		flag = true;
	}
	if( prop & Dict::RelTransitive ) {
		if( flag )
			command += ", ";
		command += "@transitive";
		flag = true;
	}
	if( prop & Dict::RelAcyclic ) {
		if( flag )
			command += ", ";
		command += "@acyclic";
	}
	ExecuteCommand( command );
	return localServer.GetAnswer();
}

ErrorString ClientProxy::GetRelationProperties (string &rel, int &prop)
{
    string command;

    command = "@get-relation-prop '" + rel + "'";
    ExecuteCommand (command);
    prop = 0;
    // XXX
    return ErrorString ();
}

ErrorString ClientProxy::GetAllRelations( SetOfString &rels )
{
	string result;
	
	ExecuteCommand( "@show-all-relations" );
	ErrorString err = localServer.GetAnswer( result );
	rels.erase (rels.begin(), rels.end());
	for( int last_i = 0, i = 0; i < result.length(); i++ ) {
		if( result[ i ] == '\n' ) {
			rels.insert (result.substr (last_i, i - last_i));
			last_i = i + 1;
		}
	}
	return err;
}


ErrorString ClientProxy::DeleteWord( string &word )
{
	ExecuteCommand( "@delete-word '" + word + "'" );
	return localServer.GetAnswer();
}


ErrorString ClientProxy::AddWord( string &word )
{
	ExecuteCommand( "@add-word '" + word + "'" );
	return localServer.GetAnswer();
}


ErrorString ClientProxy::ContainsWord( string &word, bool &yesno )
{
	ExecuteCommand( "@contains-word '" + word + "'" );
	return localServer.GetAnswer( yesno );
}



ErrorString ClientProxy::AddRelation( string &word1, string &rel,
                                      string &word2 )
{
	ExecuteCommand( "@add-relation '" + word1 + "' : '" +
                        rel + "' : '" + word2 + "'" );
	return localServer.GetAnswer();
}



ErrorString ClientProxy::GetRelationPred( string &word, string &rel,
					 SetOfWord &pred )
{
	ExecuteCommand( "@get-relation-pred '" + word + "' : '" +
                        rel + "'" );

	return localServer.GetAnswer( pred );
}


ErrorString ClientProxy::GetRelationAllPred( string &word, string &rel,
                                             SetOfWord &pred )
{
	ExecuteCommand( "@get-relation-all-pred '" + word + "' : '" +
                        rel + "'" );

	return localServer.GetAnswer( pred );
}


ErrorString ClientProxy::GetRelationSucc( string &word, string &rel,
					 SetOfWord &succ )
{
	ExecuteCommand( "@get-relation-succ '" + word + "' : '" +
                        rel + "'" );
	return localServer.GetAnswer( succ );
}


ErrorString ClientProxy::GetRelationAllSucc( string &word, string &rel,
                                             SetOfWord &succ )
{
	ExecuteCommand( "@get-relation-all-succ '" + word + "' : '" +
                        rel + "'" );
	return localServer.GetAnswer( succ );
}


ErrorString ClientProxy::AreRelated( string &word1, SetOfString &rels,
                                     string &word2, bool &yesno )
{
	string command;
	bool flag = false;
	
	command = "@are-related '" + word1 + "' : ";
    SetOfString::const_iterator i;
	for (i = rels.begin(); i != rels.end(); ++i) {
		if( flag )
			command += ", ";
		command += "'" + *i + "'";
		flag = true;
	}
	command += " : '" + word2 + "'";
	ExecuteCommand( command );
	return localServer.GetAnswer( yesno );
}


ErrorString ClientProxy::Add( ConceptualGraph *cg, KeyCG &newKey )
{
	ostrstream ostr;
	
	ostr << "@add ";
	cg->Print( ostr );
	ostr << ends;
	string command( ostr.str() );
	ostr.rdbuf()->freeze(0);
	ExecuteCommand( command );
	return localServer.GetAnswer( newKey );
}


ErrorString ClientProxy::Replace( KeyCG aKey, ConceptualGraph *cg )
{
    ostrstream ostr;

    if (cg) {
	ostr << "@add $" << aKey << " : ";
	cg->Print( ostr );
    } else {
        ostr << "@remove $" << aKey;
    }
    ostr << ends;
    string command( ostr.str() );
    ostr.rdbuf()->freeze(0);
    ExecuteCommand( command );
    return localServer.GetAnswer();
}


ErrorString ClientProxy::Join( KeyCG aKey, ConceptualGraph *cg )
{
	ostrstream ostr;
	
	ostr << "@join $" << aKey << " : ";
	cg->Print( ostr );
	ostr << ends;
	string command( ostr.str() );
	ostr.rdbuf()->freeze(0);
	ExecuteCommand( command );
	return localServer.GetAnswer();
}


ErrorString ClientProxy::Match( ConceptualGraph *query,
                                SetOfString &selecRules,
                                SetOfMatchResult &matches,
                                KeyCG profile)
{
    ostrstream ostr;
	
    ostr << "@match ";
    if (profile >= 0)
        ostr << "$" << profile << " : ";
    bool first = true;
    SetOfString::const_iterator i;
    for (i = selecRules.begin(); i != selecRules.end(); ++i) {
        if( !first )
            ostr << ", ";
        first = false;
        ostr << "'" << *i << "'";
    }
    ostr << " : ";
    query->Print( ostr );
    ostr << ends;
    string command( ostr.str() );
    ostr.rdbuf()->freeze(0);
    ExecuteCommand( command );
    return localServer.GetAnswer( matches );
}


ErrorString ClientProxy::GetGraph( KeyCG aKey, ConceptualGraph **cg )
{
	ostrstream ostr;
	
	ostr << "@get-graph $" << aKey;
	ostr << ends;
	string command( ostr.str() );
	ostr.rdbuf()->freeze(0);
	ExecuteCommand( command );
	return localServer.GetAnswer( cg );
}


ErrorString ClientProxy::GetNumberOfVariants( string &conType, int &num )
{
	string result;
	
	ExecuteCommand( "@number-of-variants \"" + conType + '"' );
	ErrorString err = localServer.GetAnswer( result );
	num = atoi( result.c_str() );
	return err;
}


ErrorString ClientProxy::GetNthVariant( string &conType, int num,
				       KeyCG &theKey )
{
	char buf[ 20 ];
	
	sprintf( buf, "%d", num );
	ExecuteCommand( "@get-graph '" + conType + "' : " + buf );
	return localServer.GetAnswer( theKey );
}



ErrorString ClientProxy::GetNumberOfGraphs( int &num )
{
    assert( 0 );
    return ErrorString();
}


ErrorString ClientProxy::GetNthGraph( int num, KeyCG &theKey )
{
    assert( 0 );
    return ErrorString();
}


ErrorString ClientProxy::AddProvider( KeyCG aKey,
				     string &theMaintainer, string &anOID )
{
	ostrstream ostr;
	
	ostr << "@add-provider $";
	ostr << aKey;
	ostr << ": '";
	ostr << theMaintainer;
	ostr << "', '";
	ostr << anOID;
	ostr << "'" << ends;
	string command( ostr.str() );
	ostr.rdbuf()->freeze(0);
	ExecuteCommand( command );
	return localServer.GetAnswer();
}


ErrorString ClientProxy::RemoveProvider( KeyCG aKey, string &anOID )
{
    ostrstream ostr;

    ostr << "@remove-provider $" << aKey << ", '" << anOID << "'" << ends;
    string command (ostr.str());
    ostr.rdbuf()->freeze (0);
    ExecuteCommand (command);
    return localServer.GetAnswer();
}


ErrorString ClientProxy::GetProviders( KeyCG aKey, SetOfServiceEntry &result )
{
	ostrstream ostr;
	
	ostr << "@get-provider $" << aKey;
	ostr << ends;
	string command( ostr.str() );
	ostr.rdbuf()->freeze(0);
	ExecuteCommand( command );
	return localServer.GetAnswer( result );
}

ErrorString ClientProxy::CreateProfile( KeyCG &aKey )
{
    ExecuteCommand ("@create-profile");
    return localServer.GetAnswer (aKey);
}

ErrorString ClientProxy::RemoveProfile( KeyCG aKey )
{
    ostrstream ostr;
    ostr << "@remove-profile $" << aKey << ends;

    string command (ostr.str());
    ostr.rdbuf()->freeze (0);
    ExecuteCommand (command);
    bool result;
    return localServer.GetAnswer (result);
}

ErrorString ClientProxy::SetMinmatchval( KeyCG aKey, long minmatchval )
{
    ostrstream ostr;
    ostr << "@set-minmatchval $" << aKey << " : " << minmatchval << ends;

    string command (ostr.str());
    ostr.rdbuf()->freeze (0);
    ExecuteCommand (command);
    bool result;
    return localServer.GetAnswer (result);
}

ErrorString ClientProxy::SetTimeout( KeyCG aKey, long timeout )
{
    ostrstream ostr;
    ostr << "@set-timeout $" << aKey << " : " << timeout << ends;

    string command (ostr.str());
    ostr.rdbuf()->freeze (0);
    ExecuteCommand (command);
    bool result;
    return localServer.GetAnswer (result);
}

ErrorString ClientProxy::GetMinmatchval( KeyCG aKey, long &minmatchval )
{
    ostrstream ostr;
    ostr << "@get-minmatchval $" << aKey << ends;

    string command (ostr.str());
    ostr.rdbuf()->freeze (0);
    ExecuteCommand (command);
    return localServer.GetAnswer (minmatchval);
}

ErrorString ClientProxy::GetTimeout( KeyCG aKey, long &timeout )
{
    ostrstream ostr;
    ostr << "@get-timeout $" << aKey << ends;

    string command (ostr.str());
    ostr.rdbuf()->freeze (0);
    ExecuteCommand (command);
    return localServer.GetAnswer (timeout);
}


bool ClientProxy::CheckForError()
{
    // This method should never be called
#if 0
    // ... but is called in Parser::Parse(), even on proxies
    assert( 0 );
#else
    return true;
#endif
}


void ClientProxy::SetError( int err, string *msg )
{
    // This method should never be called
    assert( 0 );
}


void ClientProxy::SetAnswer()
{
    // This method should never be called
    assert( 0 );
}


void ClientProxy::SetAnswer( ConceptualGraph *cg )
{
    // This method should never be called
    assert( 0 );
}


void ClientProxy::SetAnswer( bool flag )
{
    // This method should never be called
    assert( 0 );
}


void ClientProxy::SetAnswer( string &str )
{
    // This method should never be called
    assert( 0 );
}


void ClientProxy::SetAnswer( SetOfMatchResult &result )
{
    // This method should never be called
    assert( 0 );
}


void ClientProxy::SetAnswer( SetOfServiceEntry &entries )
{
    // This method should never be called
    assert( 0 );
}


void ClientProxy::SetAnswer( SetOfWord &word_set )
{
    // This method should never be called
    assert( 0 );
}


void ClientProxy::SetAnswer( KeyCG aKey )
{
    // This method should never be called
    assert( 0 );
}


void ClientProxy::SetAnswer( long number )
{
    // This method should never be called
    assert( 0 );
}


ErrorString ClientProxy::GetAnswer()
{
    // This method should never be called
    assert( 0 );
    return ErrorString();
}


ErrorString ClientProxy::GetAnswer( ConceptualGraph **cg )
{
    // This method should never be called
    assert( 0 );
    return ErrorString();
}


ErrorString ClientProxy::GetAnswer( bool &flag )
{
    // This method should never be called
    assert( 0 );
    return ErrorString();
}


ErrorString ClientProxy::GetAnswer( string &str )
{
    // This method should never be called
    assert( 0 );
    return ErrorString();
}


ErrorString ClientProxy::GetAnswer( SetOfMatchResult &result )
{
    // This method should never be called
    assert( 0 );
    return ErrorString();
}


ErrorString ClientProxy::GetAnswer( SetOfServiceEntry &entries )
{
    // This method should never be called
    assert( 0 );
    return ErrorString();
}


ErrorString ClientProxy::GetAnswer( SetOfWord &word_set )
{
    // This method should never be called
    assert( 0 );
    return ErrorString();
}


ErrorString ClientProxy::GetAnswer( KeyCG &key )
{
    // This method should never be called
    assert( 0 );
    return ErrorString();
}


ErrorString ClientProxy::GetAnswer( long &number )
{
    // This method should never be called
    assert( 0 );
    return ErrorString();
}


string ClientProxy::GetAnswerString()
{
    return localServer.GetAnswerString();
    return ErrorString();
}
