
/*
 *  AI-Trader --- Service Trading using Conceptual Graphs
 *  Copyright (C) 1995 Arno Puder et al.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 aitrader@informatik.uni-frankfurt.de
 */

// $Id: server.cc,v 1.24 1997/06/27 16:19:18 roemer Exp $

#include <stream.h>
#include <strstream.h>
#include "server.h"
#include "version.h"
#include "util.h"


#if 0
// The following function copies a WordAVLSet. This is required
// because this class doesn't define it's own overloaded assignment
// operator!!!

void CopyWordAVLSet( WordAVLSet &to, WordAVLSet &from )
{
	for( Pix i = from.first(); i != 0; from.next( i ) )
		to.add( from( i ) );
}
#endif

// Convert a vector of pointers to Words to a set of words
static SetOfWord &
vcwp2set (SetOfWord &set, const Dict::vcwp &vec)
{
    set.erase (set.begin(), set.end());
    Dict::vcwp::const_iterator i;
    for (i = vec.begin(); i != vec.end(); ++i)
        set.insert (**i);
    return set;
}



Server::Server()
{
	verbose = false;
	cgAnswer = NULL;
}


Server::~Server()
{
	if( cgAnswer )
		delete cgAnswer;
}


ErrorString Server::GetVersion( string &version )
{
	version = "AI-Trader ";
	version += TRADER_VERSION;
	return ErrorString();
}


ErrorString Server::SetVerbose( bool flag )
{
    verbose = flag;
    dict.SetVerboseMode (flag);
    return ErrorString();
}


ErrorString Server::SaveTraderDB( string *name )
{
    ofstream file( name->c_str() );
    file << "/*" << endl;
    file << " * Dictionary" << endl;
    file << " */" << endl;
    file << dict;
    file << endl << endl;
    file << "/*" << endl;
    file << " * Matching rules" << endl;
    file << " */" << endl << endl;
    rules.Print( file );
    file << endl << endl;
    file << "/*" << endl;
    file << " * Knowledgebase" << endl;
    file << " */" << endl << endl;
    kb.Print( file );
    file << endl << endl;
    file << "/*" << endl;
    file << " * Service Providers" << endl;
    file << " */" << endl << endl;
    services.Print( file );
    file << endl << endl;
    file << "/*" << endl;
    file << " * User Profiles" << endl;
    file << " */" << endl << endl;
    profiles.Print( file );
    return ErrorString();
}
    
    
ErrorString Server::DefineMatchingRule( string *name, string *rule )
{
    rules.DefineMatchingRule( *name, *rule );
    return ErrorString();
}


ErrorString Server::DeleteMatchingRule( string *name )
{
    rules.DeleteMatchingRule( *name );
    return ErrorString();
}


ErrorString Server::GetMatchingRuleNames( string *names )
{
	*names = "";
    MatchingRule::const_iterator i;
	for (i = rules.begin(); i != rules.end(); ++i) {
		*names += (*i).first;
		*names += '\n';
	}
	return ErrorString();
}



ErrorString Server::GetMatchingRule( string *name, string *rule )
{
	*rule = rules.GetMatchingRule(*name);
	return ErrorString();
}



ErrorString Server::DictQuickMode( bool flag )
{
	dict.SetQuickMode( flag );
	return  ErrorString();
}



ErrorString Server::RegisterRelation( string &rel, int prop )
{
    if( dict.RegisterRelation( rel, prop ) )
        return ErrorString();
    else
        return ErrorString( 2, rel.c_str() );
}

ErrorString Server::GetRelationProperties (string &rel, int &prop)
{
    if (dict.GetRelationProperties (rel, prop))
        return ErrorString ();
    else
        return ErrorString (2, rel.c_str());
}

ErrorString Server::GetAllRelations( SetOfString &rels )
{
    rels = vector2set (dict.GetRelationNames ());
	return ErrorString();
}


ErrorString Server::DeleteWord( string &word )
{
    //dict.DeleteWord (Word (word, false));
    return ErrorString(9);
}


ErrorString Server::AddWord( string &word )
{
    dict.AddWord (Word (word, false));
    return ErrorString();
}


ErrorString Server::ContainsWord( string &word, bool &yesno )
{
    yesno = (dict.ContainsWord ( Word (word, false)) == 1);
    return ErrorString();
}



ErrorString Server::AddRelation( string &word1, string &rel, string &word2 )
{
    if( dict.AddRelation( Word (word1, false), rel, Word (word2, false)))
        return ErrorString();
    else
        return ErrorString( 6, word1.c_str(), word2.c_str(), rel.c_str() );
}



ErrorString Server::GetRelationPred( string &word, string &rel,
                                     SetOfWord &pred )
{
    vcwp2set (pred, dict.GetRelImmPred( Word (word, false), rel ));
    return ErrorString();
}


ErrorString Server::GetRelationAllPred( string &word, string &rel,
                                        SetOfWord &pred )
{
    vcwp2set (pred, dict.GetRelAllPred( Word (word, false), rel ));
    return ErrorString();
}


ErrorString Server::GetRelationSucc( string &word, string &rel,
                                     SetOfWord &succ )
{
    vcwp2set (succ, dict.GetRelImmSucc( Word (word, false), rel ));
    return ErrorString();
}


ErrorString Server::GetRelationAllSucc( string &word, string &rel,
                                        SetOfWord &succ )
{
    vcwp2set (succ, dict.GetRelAllSucc( Word (word, false), rel ));
    return ErrorString();
}


ErrorString Server::AreRelated( string &word1, SetOfString &rels, string &word2,
                                bool &yesno )
{
    yesno = (dict.AreRelated (Word (word1, false), set2vector (rels),
                              Word (word2, false)) == 1);
    return ErrorString();
}


ErrorString Server::Add( ConceptualGraph *cg, KeyCG &newKey )
{
    kb.Add( cg, newKey );
    return ErrorString();
}


ErrorString Server::Replace( KeyCG aKey, ConceptualGraph *cg )
{
    kb.Replace( aKey, cg );
    if (!cg)
        services.DeleteAllProviders (aKey);
    return ErrorString();
}


ErrorString Server::Join( KeyCG aKey, ConceptualGraph *cg )
{
	kb.Join( aKey, cg );
	return ErrorString();
}


ErrorString Server::Match( ConceptualGraph *cg,
                           SetOfString &selecRules,
                           SetOfMatchResult &matches,
                           KeyCG pkey)
{
    ProfileEntry def;
    matches = kb.Match( dict, rules, selecRules, cg, verbose,
                        pkey >= 0 ? profiles.GetProfile (pkey) : def);
    return ErrorString();
}


ErrorString Server::GetGraph( KeyCG aKey, ConceptualGraph **cg )
{
    *cg = kb.GetGraphByKey( aKey );
    return ErrorString();
}


ErrorString Server::GetNumberOfVariants( string &conType, int &num )
{
    num = kb.GetNumberOfVariants( conType );
    return ErrorString();
}


ErrorString Server::GetNthVariant( string &conType, int num, KeyCG &theKey )
{
    theKey = kb.GetVariant( conType, num );
    return ErrorString();
}


ErrorString Server::GetNumberOfGraphs( int &num )
{
    num = kb.GetNumberOfGraphs();
    return ErrorString();
}


ErrorString Server::GetNthGraph( int num, KeyCG &theKey )
{
    theKey = kb.GetNthGraph( num );
    return ErrorString();
}


ErrorString Server::AddProvider( KeyCG aKey,
				string &theMaintainer, string &anOID )
{
    services.RegisterNewProvider( aKey, theMaintainer, anOID );
    return ErrorString();
}


ErrorString Server::RemoveProvider( KeyCG aKey, string &anOID )
{
    services.DeleteProvider( aKey, anOID );
    return ErrorString();
}


ErrorString Server::GetProviders( KeyCG aKey, SetOfServiceEntry &result )
{
    result = services.GetAllProviders( aKey );
    return ErrorString();
}


ErrorString Server::CreateProfile( KeyCG &aKey )
{
    aKey = profiles.RegisterProfile (ProfileEntry());
    return ErrorString();
}


ErrorString Server::RemoveProfile( KeyCG aKey )
{
    profiles.DeleteProfile (aKey);
    return ErrorString();
}


ErrorString Server::SetMinmatchval( KeyCG aKey, long minmatchval )
{
    if (!profiles.HaveProfile (aKey))
        profiles.RegisterProfile (aKey, ProfileEntry());
    ProfileEntry &prof = profiles.GetProfile (aKey);
    prof.SetMinMatchval (minmatchval);
    return ErrorString();
}


ErrorString Server::SetTimeout( KeyCG aKey, long timeout )
{
    if (!profiles.HaveProfile (aKey))
        profiles.RegisterProfile (aKey, ProfileEntry());
    ProfileEntry &prof = profiles.GetProfile (aKey);
    prof.SetTimeout (timeout);
    return ErrorString();
}


ErrorString Server::GetMinmatchval( KeyCG aKey, long &minmatchval )
{
    if (profiles.HaveProfile (aKey)) {
        ProfileEntry &prof = profiles.GetProfile (aKey);
        minmatchval = prof.GetMinMatchval ();
    }
    return ErrorString();
}


ErrorString Server::GetTimeout( KeyCG aKey, long &timeout )
{
    if (profiles.HaveProfile (aKey)) {
        ProfileEntry &prof = profiles.GetProfile (aKey);
        timeout = prof.GetTimeout ();
    }
    return ErrorString();
}


void Server::SetError( int err, string *msg )
{
	errorString = ErrorString( *msg, err );
}


bool Server::CheckForError()
{
	errorString.ResetError();
	if( !LastErrorString ) {
		errorString = LastErrorString;
		answerString = "@error ";
		char buf[ 10 ];
		sprintf( buf, "%d", LastErrorString.GetErrNum() );
		answerString += buf;
		answerString += " : '";
		answerString += LastErrorString;
		answerString += "'";
		return true;
	}
	return false;
}


void Server::SetAnswer()
{
	answerString = "@ok";
}


void Server::SetAnswer( ConceptualGraph *cg )
{
	if( cgAnswer )
		delete cgAnswer;
	cgAnswer = cg;
	answerString = "@answer-graph=";
	if( cg == NULL )
		return;
	ostrstream ostr;
	cg->Print( ostr );
	ostr << ends;
	answerString += ostr.str();
	ostr.rdbuf()->freeze( 0 );
}


void Server::SetAnswer( bool flag )
{
	boolAnswer = flag;
	answerString = "@answer-flag=";
	answerString += flag ? "@yes" : "@no";
}


void Server::SetAnswer( string &str )
{
	strAnswer = str;
	answerString = "@answer-string=" + str + '';
}


void Server::SetAnswer( SetOfMatchResult &result )
{
    matchesAnswer = result;
    answerString = "@answer-matches=";
    ostrstream ostr;
    SetOfMatchResult::iterator i = result.begin();
    while( i != result.end() ) {
        ostr << "$" << (*i).GetKeyCG() << ", " << (*i).GetMatchVal();
        i++;
        if( i != result.end() )
            ostr << " : ";
    }
    ostr << ends;
    answerString += ostr.str();
    ostr.rdbuf()->freeze( 0 );
}


void Server::SetAnswer( SetOfServiceEntry &entries )
{
	serviceEntriesAnswer = entries;
	answerString = "@answer-providers=";
	ostrstream ostr;
	SetOfServiceEntry::iterator i = entries.begin();
	while( i != entries.end() ) {
		ostr << "'";
		ostr << (*i).GetMaintainer();
		ostr << "', '";
		ostr << (*i).GetOID() << "'";
		i++;
		if( i != entries.end() )
			ostr << " : ";
	}
	ostr << ends;
	answerString += ostr.str();
	ostr.rdbuf()->freeze( 0 );
}



void Server::SetAnswer( SetOfWord &word_set )
{
    wordSetAnswer = word_set;
	
    answerString = "@answer-words=";
    ostrstream ostr;
    SetOfWord::const_iterator i = word_set.begin();
    while (i != word_set.end()) {
        ostr << *i;
        ++i;
        if( i != word_set.end() )
            ostr << ", ";
    }
    ostr << ends;
    answerString += ostr.str();
    ostr.rdbuf()->freeze( 0 );
}


void Server::SetAnswer( KeyCG aKey )
{
    keyAnswer = aKey;
    answerString = "@answer-key=$";
    answerString += dec (aKey);
}


void Server::SetAnswer( long number )
{
    intAnswer = number;
    answerString = "@answer-int=";
    answerString += dec (number);
}

ErrorString Server::GetAnswer()
{
    return errorString;
}


/*
 * when called this method MUST arrange to NOT delete cgAnswer
 * in Server::~Server(). cgtcl.cc relies on it.
 */
ErrorString Server::GetAnswer( ConceptualGraph **cg )
{
    *cg = cgAnswer;
    cgAnswer = NULL;
    return errorString;
}


ErrorString Server::GetAnswer( bool &flag )
{
	flag = boolAnswer;
	return errorString;
}


ErrorString Server::GetAnswer( string &str )
{
	str = strAnswer;
	return errorString;
}


ErrorString Server::GetAnswer( SetOfMatchResult &result )
{
	result = matchesAnswer;
	return errorString;
}


ErrorString Server::GetAnswer( SetOfServiceEntry &entries )
{
	entries = serviceEntriesAnswer;
	return errorString;
}


ErrorString Server::GetAnswer( SetOfWord &word_set )
{
    word_set = wordSetAnswer;
    return errorString;
}


ErrorString Server::GetAnswer( KeyCG &aKey )
{
	aKey = keyAnswer;
	return errorString;
}


ErrorString Server::GetAnswer( long &number )
{
    number = intAnswer;
    return errorString;
}


string Server::GetAnswerString()
{
	return answerString;
}
