/*

    $Id: valid.h,v 1.2 1997/04/09 14:37:13 roemer Exp $
  
    (c) 1994,95 Normed Verlag GmbH
    (c) 1994 Hein Roehrig and Holger Hehl
    
    Portions of this file have been -- with permission by Hein
    Roehrig -- extracted from
    
    i - The Interface Software to the Virtual Hardware Environment
    Copyright (C) 1994 Hein Roehrig, Holger Hehl

    File:       valid.h
    
    Purpose:    Validation macros. Replaces assert.h.
                ASSERT(expr) interrupts the program with an error message,
                if expr turns out to be false. If NDEBUG is defined,
                the expression will not be evaluated.
                VERIFY(expr) interrupts the program with an error message,
                if expr turns out to be false. If NDEBUG is defined,
                the expression will be evaluated, but will not be checked.
                If you want to get core-dumps but don't get them or vice-versa
                use the ulimit -c command.
                INTERNALFATALERROR(what) interrupts always the program
                with an error message starting with what
                INVALID_OPERATION always interrupts the program saying that
                an invalid operation has been executed
                NOT_IMPLEMENTED always interrupts the program saying that
                a not implemented operation has been executed

                dprintf() is like fprintf(stderr,...) if NDEBUG is not
                defined, otherwise it does nothing. Under Windows, it
                uses ::OutputDebugString.

                ExceptBase and UserExcept are exception handling
                classes for the use in conjunction with C++ exception
                handling.
                
                Share and enjoy!

*/


#ifndef __VALID_H__
#define __VALID_H__


#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>


// set to 1 if you want to override pre-defined ASSERT and VERIFY
#if 1
#ifdef ASSERT
#undef ASSERT
#endif
#ifdef VERIFY
#undef VERIFY
#endif
#endif


#ifndef INTERNALFATALERROR
#if defined(_MSC_VER) && defined(_WINDOWS)
static inline void
internal_fatal_error(const char *error, const char *file, const unsigned line)
{
static char t[200];
    sprintf(t, "%s in application %s file %s at line %u", error,
            afxCurrentAppName?afxCurrentAppName:"<unknown>", file, line);
    ::OutputDebugString(t);
    ::MessageBox(NULL, t, "INTERNAL FATAL ERROR", MB_SYSTEMMODAL|MB_ICONHAND|
                 MB_OK);
    AfxDebugBreak();
    abort();
}
#else // _MSC_VER
#include <stdlib.h>
#include <stdio.h>
#ifdef __GNUC__
static void internal_fatal_error(const char *, const char *, const unsigned)
__attribute__ ((noreturn));
#endif
static inline void
internal_fatal_error(const char *error, const char *file, const unsigned line)
{
    fprintf(stderr, "\n%s in file %s at line %u\n", error, file, line);
    abort();
}
#endif
#define INTERNALFATALERROR(x) (internal_fatal_error(x, __FILE__, __LINE__))
#endif // INTERNALFATALERROR



#ifndef ASSERT
#ifdef NDEBUG
#define ASSERT(x) ((void)(0))
#else // NDEBUG
#define ASSERT(x) ( (void) ((x)?(void)0:INTERNALFATALERROR("ASSERT error: '" #x "'")) )
#endif // NDEBUG
#endif // ASSERT



#ifndef VERIFY
#ifdef NDEBUG
#define VERIFY(x) (x)
#else // NDEBUG
#define VERIFY(x) ( (void) ((x)?0:INTERNALFATALERROR("VERIFY error: '" #x "'")) )
#endif // NDEBUG
#endif // VERIFY



#ifndef NOT_IMPLEMENTED
#define NOT_IMPLEMENTED INTERNALFATALERROR("NOT_IMPLEMENTED")
#endif // NOT_IMPLEMENTED



#ifndef INVALID_OPERATION
#define INVALID_OPERATION INTERNALFATALERROR("INVALID_OPERATION")
#endif // INVALID_OPERATION


#ifndef DPRINTF
#ifdef NDEBUG
#if !defined(__GNUC__)
inline
#else // __GNUC__
static
#endif // __GNUC__
int
dprintf(const char *format, ...)
{
    return(0);
}
#else // NDEBUG
#if !defined(__GNUC__)
inline
#else // __GNUC__
static
#endif // __GNUC__
int
dprintf(const char *format, ...)
{
    static char t[200];
    va_list arglist;
    va_start(arglist, format);
    const int retval=vsprintf(t, format, arglist);
    va_end(arglist);
#if !defined(_MSC_VER) || defined(_CONSOLE)
    fprintf(stderr, t);
#else // _MSC_VER
    ::OutputDebugString(t);
#endif // _MSC_VER
    return(retval);
}
#endif // NDEBUG
#endif // DPRINTF


#if 0
#ifdef __cplusplus


class ExceptBase
{

public:

    ExceptBase();
    ExceptBase(const ExceptBase &);
    virtual ~ExceptBase();

};


inline
ExceptBase::ExceptBase()
{
}


inline
ExceptBase::ExceptBase(const ExceptBase &)
{
}


inline
ExceptBase::~ExceptBase()
{
}


class UserExcept : public ExceptBase
{

public:
    
    UserExcept(const char *format, ...);
    UserExcept(const UserExcept &);
    virtual ~UserExcept();

    const char *GetDescription();

private:
    
    void Fill(const char*, va_list);
    char *m_Description;

};


inline void
UserExcept::Fill(const char *format, va_list arglist)
{
    if (format!=0)
    {
        char buffer[1000];
        const int retval=vsprintf(buffer, format, arglist);
        if (retval>0)
        {
            m_Description=new char [retval+1];
            strcpy(m_Description, buffer);
        }
        else
            INVALID_OPERATION;
    }
    else
        m_Description=0;

    dprintf("Generated User Exception '%s'\n", m_Description);
}


inline
UserExcept::UserExcept(const char *format, ...)
{
    va_list arglist;
    va_start(arglist, format);
    Fill(format, arglist);
    va_end(arglist);
}


inline
UserExcept::UserExcept(const UserExcept &o)
: ExceptBase(o)
{
    if (o.m_Description)
    {
        m_Description=new char [strlen(o.m_Description)+1];
        strcpy(m_Description, o.m_Description);
    }
    else
        m_Description=0;
}


inline
UserExcept::~UserExcept()
{
    if (m_Description)
        delete m_Description;
}


inline const char *
UserExcept::GetDescription()
{
    return(m_Description);
}


#endif // __cplusplus
#endif

#endif // __VALID_H__


